package edu.stanford.nlp.objectbank;

import edu.stanford.nlp.process.Function;
import edu.stanford.nlp.util.AbstractIterator;
import edu.stanford.nlp.util.XMLUtils;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.Reader;
import java.util.Iterator;

/**
 * A class which iterates over Strings occuring between the begin and end of a selected tag.
 * Ignores all other characters in the input Reader.
 *
 * @author Teg Grenager (grenager@stanford.edu)
 */
public class XMLBeginEndIterator extends AbstractIterator {

  private String tagName;
  private BufferedReader in;
  private Object nextToken;
  private Function op;
  private boolean keepInternalTags;
  private boolean keepDelimitingTags;

  public XMLBeginEndIterator(Reader in, String tagName) {
    this(in, tagName, new IdentityFunction(), false);
  }

  public XMLBeginEndIterator(Reader in, String tagName, boolean keepInternalTags) {
    this(in, tagName, new IdentityFunction(), keepInternalTags);
  }

  public XMLBeginEndIterator(Reader in, String tagName, Function op, boolean keepInternalTags) {
    this(in, tagName, op, keepInternalTags, false);
  }

  public XMLBeginEndIterator(Reader in, String tagName, boolean keepInternalTags, boolean keepDelimitingTags) {
    this(in, tagName, new IdentityFunction(), keepInternalTags, keepDelimitingTags);
  }

  public XMLBeginEndIterator(Reader in, String tagName, Function op, boolean keepInternalTags, boolean keepDelimitingTags) {
    this.tagName = tagName;
    this.op = op;
    this.keepInternalTags = keepInternalTags;
    this.keepDelimitingTags = keepDelimitingTags;
    this.in = new BufferedReader(in);
    setNext();
  }

  private void setNext() {
    String s = getNext();
    nextToken = parseString(s);
  }

  // returns null if there is no next object
  private String getNext() {
    StringBuffer result = new StringBuffer();
    try {
      XMLUtils.XMLTag tag;
      do {
        String text = XMLUtils.readUntilTag(in);
        // there may or may not be text before the next tag, but we discard it
        //        System.out.println("outside text: " + text );
        tag = XMLUtils.readAndParseTag(in);
        //        System.out.println("outside tag: " + tag);
        if (tag == null) {
          return null; // couldn't find any more tags, so no more elements
        }
      } while (!tag.name.equals(tagName) || tag.isEndTag);
      if (keepDelimitingTags) {
        result.append(tag.toString());
      }
      while (true) {
        String text = XMLUtils.readUntilTag(in);
        if (text != null) {
          // if the text isn't null, we append it
          //        System.out.println("inside text: " + text );
          result.append(text);
        }
        String tagString = XMLUtils.readTag(in);
        tag = XMLUtils.parseTag(tagString);
        if (tag == null) {
          return null; // unexpected end of this element, so no more elements
        }
        if (tag.name.equals(tagName) && tag.isEndTag) {
          if (keepDelimitingTags) {
            result.append(tagString);
          }
          // this is our end tag so we stop
          break;
        } else {
          // not our end tag, so we optionally append it and keep going
          if (keepInternalTags) {
            result.append(tagString);
          }
        }
      }
    } catch (Exception e) {
      e.printStackTrace();
    }
    return result.toString();
  }

  protected Object parseString(String s) {
    return op.apply(s);
  }

  public boolean hasNext() {
    return nextToken != null;
  }

  public Object next() {
    Object token = nextToken;
    setNext();
    return token;
  }

  public Object peek() {
    return nextToken;
  }

  /**
   * Returns a factory that vends BeginEndIterators that reads the contents of the
   * given Reader, extracts text between the specified Strings, then returns the result.
   */
  public IteratorFromReaderFactory getFactory(String tag) {
    return new XMLBeginEndIterator.XMLBeginEndIteratorFactory(tag);
  }

  class XMLBeginEndIteratorFactory implements IteratorFromReaderFactory {

    String tag;
    Function op;

    public XMLBeginEndIteratorFactory(String tag) {
      this(tag, new IdentityFunction());
    }

    public XMLBeginEndIteratorFactory(String tag, Function op) {
      this.tag = tag;
      this.op = op;
    }

    public Iterator getIterator(java.io.Reader r) {
      return new XMLBeginEndIterator(r, tag, op, false);
    }
  }

  public static void main(String[] args) throws Exception {
    Reader in = new FileReader(args[0]);
    Iterator iter = new XMLBeginEndIterator(in, args[1], args[2].equalsIgnoreCase("true"));
    int i = 0;
    while (iter.hasNext()) {
      String s = (String) iter.next();
      i++;
      System.out.println("*************************************************");
      System.out.println(s);
    }
  }
}

