package edu.stanford.nlp.parser.lexparser;

import edu.stanford.nlp.trees.Tree;
import edu.stanford.nlp.trees.TreebankLanguagePack;
import edu.stanford.nlp.util.Numberer;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.ObjectInputStream;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

/**
 * An abstract base class for dependency grammars.  The only thing you have
 * to implement in a subclass is scoreTB.  A subclass also has to either call
 * super() in its constructor, or otherwise initialize the tagBin array.
 * The call to initTagBins() (in the constructor) must be made after all
 * keys have been entered into Numberer.getGlobalNumberer("tags").
 *
 * @author Galen Andrew
 */
public abstract class AbstractDependencyGrammar implements DependencyGrammar {

  protected TagProjection tagProjection;
  private static Numberer tagNumberer;
  private static Numberer wordNumberer;
  protected int numTagBins;
  protected int[] tagBin;
  protected TreebankLanguagePack tlp;
  protected boolean directional;
  protected boolean useDistance;
  protected boolean useCoarseDistance;

  protected static final IntTaggedWord stopTW = new IntTaggedWord(-2, -2);
  protected static final IntTaggedWord wildTW = new IntTaggedWord(-1, -1);

  /** A temp holder for efficiency. Its value is restored in the readObject
   *  method, so it works after deserialization.
   */ 
  private transient Dependency tempDependency
    = new Dependency(-2, -2, -2, -2, false, 0);
  private transient Dependency internTempDependency = null;
  protected transient Map expandDependencyMap = new HashMap();

  private static final boolean DEBUG = false;


  public AbstractDependencyGrammar(TreebankLanguagePack tlp, TagProjection tagProjection, boolean directional, boolean useDistance, boolean useCoarseDistance) {
    this.tlp = tlp;
    this.tagProjection = tagProjection;
    this.directional = directional;
    this.useDistance = useDistance;
    this.useCoarseDistance = useCoarseDistance;
    initTagBins();
  }

  protected static Numberer tagNumberer() {
    if (tagNumberer == null) {
      tagNumberer = Numberer.getGlobalNumberer("tags");
    }
    return tagNumberer;
  }

  protected static Numberer wordNumberer() {
    if (wordNumberer == null) {
      wordNumberer = Numberer.getGlobalNumberer("words");
    }
    return wordNumberer;
  }

  /**
   * Default is no-op.
   * @param trees
   */
  public void tune(Collection<Tree> trees) {
    return;
  }

  public int numTagBins() {
    return numTagBins;
  }

  public int tagBin(int tag) {
    if (tag < 0) {
      return tag;
    } else {
      return tagBin[tag];
    }
  }

  public static boolean rootTW(IntTaggedWord rTW) {
    // System.out.println("rootTW: checking if " + rTW.toString("verbose") +
    // " == " + Lexicon.BOUNDARY_TAG + "[" +
    // tagNumberer().number(Lexicon.BOUNDARY_TAG) + "]" + ": " + 
    // (rTW.tag == tagNumberer().number(Lexicon.BOUNDARY_TAG)));
    return rTW.tag == tagNumberer().number(Lexicon.BOUNDARY_TAG);
  }

  protected short valenceBin(int distance) {
    if (!useDistance) {
      return 0;
    }
    if (distance < 0) {
      return -1;
    }
    if (distance == 0) {
      return 0;
    }
    return 1;
  }

  public int numDistBins() {
    return useCoarseDistance ? 4 : 5;
  }

  public short distanceBin(int distance) {
    if (!useDistance) {
      return 0;
    } else if (useCoarseDistance) {
      return courseDistanceBin(distance);
    } else {
      return regDistanceBin(distance);
    }
  }

  public static short regDistanceBin(int distance) {
    if (distance <= 0) {
      return 0;
    } else if (distance <= 1) {
      return 1;
    } else if (distance <= 5) {
      return 2;
    } else if (distance <= 10) {
      return 3;
    } else {
      return 4;
    }
  }

  public static short courseDistanceBin(int distance) {
    if (distance <= 0) {
      return 0;
    } else if (distance <= 2) {
      return 1;
    } else if (distance <= 5) {
      return 2;
    } else {
      return 3;
    }
  }

  private void initTagBins() {
    Numberer tagBinNumberer = new Numberer();
    if (DEBUG) System.err.println();
    tagBin = new int[tagNumberer().total()];
    for (int t = 0; t < tagBin.length; t++) {
      String tagStr = (String) tagNumberer().object(t);
      String binStr;
      if (tagProjection == null) {
        binStr = tagStr;
      } else {
        binStr = tagProjection.project(tagStr);
      }
      tagBin[t] = tagBinNumberer.number(binStr);
      if (DEBUG) {
        System.err.println("initTagBins: Mapped " + tagStr + " (" + t + 
                           ") to " + binStr + " (" + tagBin[t] + ")");
      }
    }
    numTagBins = tagBinNumberer.total();
    if (DEBUG) {
      System.err.println("initTagBins: tags " + tagBin.length + " bins " + 
                         numTagBins);
    }
  }

  public double score(Dependency dependency) {
    short hTBackup = dependency.head.tag;
    short aTBackup = dependency.arg.tag;

    dependency.head.tag = (short) tagBin(dependency.head.tag);
    dependency.arg.tag = (short) tagBin(dependency.arg.tag);

    double s = scoreTB(dependency);

    dependency.head.tag = hTBackup;
    dependency.arg.tag = aTBackup;

    return s;
  }

  /** cdm: this seems to fail to do the tagbin'ing you'd expect ... */
  public double score(int headWord, int headTag, int argWord, int argTag, boolean leftHeaded, int dist) {
    tempDependency.head.word = headWord;
    tempDependency.head.tag = (short) headTag;
    tempDependency.arg.word = argWord;
    tempDependency.arg.tag = (short) argTag;
    tempDependency.leftHeaded = leftHeaded;
    tempDependency.distance = (short) dist;
    return score(tempDependency);
  }

  public double scoreTB(int headWord, int headTag, int argWord, int argTag, boolean leftHeaded, int dist) {
    tempDependency.head.word = headWord;
    tempDependency.head.tag = (short) headTag;
    tempDependency.arg.word = argWord;
    tempDependency.arg.tag = (short) argTag;
    tempDependency.leftHeaded = leftHeaded;
    tempDependency.distance = (short) dist;
    return scoreTB(tempDependency);
  }

  private void readObject(ObjectInputStream ois) 
    throws IOException, ClassNotFoundException {
    ois.defaultReadObject();
    // reinitialize the transient tempDependency object
    tempDependency = new Dependency(-2, -2, -2, -2, false, 0);
  }

  /**
   * Default is to throw exception.
   * @param in
   * @throws IOException
   */
  public void readData(BufferedReader in) throws IOException {
    throw new UnsupportedOperationException();
  }

  /**
   * Default is to throw exception.
   * @param out
   * @throws IOException
   */
  public void writeData(PrintWriter out) throws IOException {
    throw new UnsupportedOperationException();
  }

  protected Dependency intern(IntTaggedWord headTW, IntTaggedWord argTW, boolean leftHeaded, int dist) {
    Map map = expandDependencyMap;
    if (internTempDependency == null) {
      internTempDependency = new Dependency();
    }
    internTempDependency.head = (IntTaggedWord) edu.stanford.nlp.util.Interner.globalIntern(headTW);
    internTempDependency.arg = (IntTaggedWord) edu.stanford.nlp.util.Interner.globalIntern(argTW);
    internTempDependency.leftHeaded = leftHeaded;
    internTempDependency.distance = (short) dist;
    Dependency returnDependency = internTempDependency;
    if (map != null) {
      returnDependency = (Dependency) map.get(internTempDependency);
      if (returnDependency == null) {
        map.put(internTempDependency, internTempDependency);
      }
      returnDependency = internTempDependency;
    }
    if (returnDependency == internTempDependency) {
      internTempDependency = null;
    }
    return returnDependency;
  }

  private static final long serialVersionUID = 21;

}
