package edu.stanford.nlp.parser.lexparser;

import edu.stanford.nlp.ling.TaggedWord;
import edu.stanford.nlp.trees.Tree;
import edu.stanford.nlp.stats.Counter;
import edu.stanford.nlp.util.Numberer;
import edu.stanford.nlp.util.StringUtils;
import edu.stanford.nlp.stats.Counter;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.Writer;
import java.text.NumberFormat;
import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * This is the default concrete instantiation of the Lexicon interface.
 * It was originally built for Penn Treebank English.
 *
 * @author Dan Klein
 * @author Galen Andrew
 * @author Christopher Manning
 */
public class BaseLexicon implements Lexicon {

  private static final boolean DEBUG_LEXICON = false;
  /** What type of equivalence classing is done in getSignature */
  protected int unknownLevel;
  protected int smoothInUnknownsThreshold;
  protected boolean smartMutation;
  // An array of Lists of rules (IntTaggedWord), indexed by word
  protected transient List[] rulesWithWord;
  // protected transient Set<IntTaggedWord> rules = new HashSet<IntTaggedWord>();
  // When it existed, rules somehow held a few less things than rulesWithWord
  // I never figured out why [cdm, Dec 2004]
  protected transient Set<IntTaggedWord> tags = new HashSet<IntTaggedWord>();
  protected transient Set<IntTaggedWord> words = new HashSet<IntTaggedWord>();
  // protected transient Set<IntTaggedWord> sigs=new HashSet<IntTaggedWord>();
  protected Counter<IntTaggedWord> seenCounter = new Counter<IntTaggedWord>();
  /** Has counts for taggings in terms of unseen signatures.
   *  The IntTagWords are for (tag,sig), (tag,null), (null,sig), (null,null).
   *  (None for basic UNK if there are signatures.)
   */
  protected Counter<IntTaggedWord> unSeenCounter = new Counter<IntTaggedWord>();
  protected static final int nullWord = -1;
  protected static final short nullTag = -1;
  /**
   * We cache the last signature looked up, because it asks for the same one
   * many times when an unknown word is encountered!  (Note that under the
   * current scheme, one unknown word, if seen sentence-initially and
   * non-initially, will be parsed with two different signatures....)
   */
  protected transient int lastSignatureIndex = -1;
  protected transient int lastSentencePosition = -1;
  protected transient int lastWordToSignaturize = -1;
  double[] smooth = {1.0, 1.0};
  transient double[][] m_TT = null;
  transient double[] m_T = null;

  public BaseLexicon() {
    this(new Options.LexOptions());
  }

  public BaseLexicon(Options.LexOptions op) {
    this.unknownLevel = op.useUnknownWordSignatures;
    this.smoothInUnknownsThreshold = op.smoothInUnknownsThreshold;
    this.smartMutation = op.smartMutation;
  }

  /**
   * Checks whether a word is in the lexicon.
   * This version will compile the lexicon into the rulesWithWord array, if 
   * that hasn't already happened
   *
   * @param word The word as an int index to a Numberer
   * @return Whether the word is in the lexicon
   */
  public boolean isKnown(int word) {
    if (rulesWithWord == null) {
      initRulesWithWord();
    }
    return (word < rulesWithWord.length && rulesWithWord[word].size() > 0);
  }

  /**
   * Checks whether a word is in the lexicon.
   * This version works even while compiling lexicon with current counters
   * (rather than using the compiled rulesWithWord array).
   *
   * @param word The word as a String
   * @return Whether the word is in the lexicon
   */
  public boolean isKnown(String word) {
    IntTaggedWord iW = new IntTaggedWord(Numberer.number("words", word), nullTag);
    return seenCounter.getCount(iW) > 0.0;
  }

  /** Returns the possible POS taggings for a word.
   *
   *  @param word The word, represented as an integer in Numberer
   *  @param loc  The position of the word in the sentence (counting from 0). 
   *                <i>Implementation note: The BaseLexicon class doesn't
   *                actually make use of this position information.</i>
   *  @return An Iterator over a List ofIntTaggedWords, which pair the word 
   *                with possible taggings as integer pairs.  (Each can be
   *                thought of as a <code>tag -&gt; word<code> rule.)
   */
  public Iterator ruleIteratorByWord(int word, int loc) {
    if (rulesWithWord == null) {
      initRulesWithWord();
    }
    if (isKnown(word)) {
      Iterator i = rulesWithWord[word].iterator();
      if (i.hasNext()) {
	if (DEBUG_LEXICON) {
	  System.err.println("Lexicon: " + Numberer.object("words", word) +
			     " (known, loc=" + loc + ", n="+ word + 
			     ") taggings: " + rulesWithWord[word]);
	}
        return i;
      }
    }
    List newTaggings = new ArrayList(40);
    for (Iterator i = rulesWithWord[Numberer.number("words", UNKNOWN_WORD)].iterator(); i.hasNext();) {
      IntTaggedWord iTW = (IntTaggedWord) i.next();
      newTaggings.add(new IntTaggedWord(word, iTW.tag));
    }
    if (DEBUG_LEXICON) {
      System.err.println("Lexicon: " + UNKNOWN_WORD + " (unknown, loc=" + loc + ", n="+ Numberer.number("words", UNKNOWN_WORD) + ") taggings: " + newTaggings);
    }
    return newTaggings.iterator();
  }

  protected void initRulesWithWord() {
    // int numWords = words.size()+sigs.size()+1;
    Numberer.number("words", UNKNOWN_WORD);
    int numWords = Numberer.getGlobalNumberer("words").total();
    rulesWithWord = new List[numWords];
    for (int w = 0; w < numWords; w++) {
      rulesWithWord[w] = new ArrayList(1);   // most have 1 or 2 items in them
    }
    //for (Iterator ruleI = rules.iterator(); ruleI.hasNext();) {
    Set<IntTaggedWord> tags = new HashSet<IntTaggedWord>();
    for (Iterator<IntTaggedWord> ruleI = seenCounter.keySet().iterator(); ruleI.hasNext();) {
      IntTaggedWord iTW = ruleI.next();
      if (iTW.tag() == nullTag) {
        continue;
      }
      if (iTW.word() == nullWord) {
        tags.add(iTW);
        continue;
      }
      //IntTaggedWord iW = new IntTaggedWord(iTW.word, nullTag);
      //if (seenCounter.getCount(iTW) > 2 || seenCounter.getCount(iW) < 1000) {
      rulesWithWord[iTW.word].add(iTW);
      //}
    }
    // tags for unknown words
    //System.out.println("Considering UNKNOWN WORD");
    for (Iterator<IntTaggedWord> tagI = tags.iterator(); tagI.hasNext();) {
      IntTaggedWord iT = tagI.next();
      double types = unSeenCounter.getCount(iT);
      if (types > Train.openClassTypesThreshold) {
        // Number of types before it's treated as open class
        IntTaggedWord iTW = new IntTaggedWord(Numberer.number("words", UNKNOWN_WORD), iT.tag);
        //System.out.println("Approved UNKNOWN WORD ("+Numberer.number("words",Lexicon.UNKNOWN_WORD)+") tag "+iTW+" (tag "+iTW.tag()+") length is "+rulesWithWord[iTW.word]);
        rulesWithWord[iTW.word].add(iTW);
      }
    }
  }

  protected List<IntTaggedWord> treeToEvents(Tree tree) {
    List taggedWords = tree.taggedYield();
    List<IntTaggedWord> itwList = new ArrayList<IntTaggedWord>();
    for (int i = 0; i < taggedWords.size(); i++) {
      TaggedWord tw = (TaggedWord) taggedWords.get(i);
      IntTaggedWord iTW = new IntTaggedWord(Numberer.number("words", tw.word()), Numberer.number("tags", tw.tag()));
      itwList.add(iTW);
    }
    return itwList;
  }

  /**
   * Trains this lexicon on the Collection of trees.
   */
  public void train(Collection trees) {
    train(trees, 1.0);
  }

  /**
   * Trains this lexicon on the Collection of trees.
   */
  public void train(Collection trees, double weight) {
    // scan data
    int tNum = 0;
    int tSize = trees.size();
    for (Iterator<Tree> tI = trees.iterator(); tI.hasNext();) {
      Tree tree = tI.next();
      tNum++;
      List<IntTaggedWord> taggedWords = treeToEvents(tree);
      for (int w = 0; w < taggedWords.size(); w++) {
        IntTaggedWord iTW = taggedWords.get(w);
        seenCounter.incrementCount(iTW, weight);
        IntTaggedWord iT = new IntTaggedWord(nullWord, iTW.tag);
        seenCounter.incrementCount(iT, weight);
        IntTaggedWord iW = new IntTaggedWord(iTW.word, nullTag);
        seenCounter.incrementCount(iW, weight);
        IntTaggedWord i = new IntTaggedWord(nullWord, nullTag);
        seenCounter.incrementCount(i, weight);
        // rules.add(iTW);
        tags.add(iT);
        words.add(iW);
        if (tNum > (int) (tSize * Train.fractionBeforeUnseenCounting)) {
          // start doing this once we're halfway through the trees
          if (seenCounter.getCount(iW) < 2) {
            // it's an entirely unknown word
            int s = getSignatureIndex(iTW.word, w);
            Numberer.number("words", UNKNOWN_WORD); //cdm: outside loop?
            IntTaggedWord iTS = new IntTaggedWord(s, iTW.tag);
            IntTaggedWord iS = new IntTaggedWord(s, nullTag);
            unSeenCounter.incrementCount(iTS, weight);
            unSeenCounter.incrementCount(iT, weight);
            unSeenCounter.incrementCount(iS, weight);
            unSeenCounter.incrementCount(i, weight);
            // rules.add(iTS);
            // sigs.add(iS);
          } else {
            if (seenCounter.getCount(iTW) < 2) {
              // it's a new tag for a known word
              // do nothing for now
            }
          }
        }
      }
    }
    // index the possible tags for each word
    //numWords = wordNumberer.total();
    //unknownWordIndex = Numberer.number("words",Lexicon.UNKNOWN_WORD);
    //initRulesWithWord();
    tune(trees);
    if (DEBUG_LEXICON) {
      printLexStats();
    }
  }

  /**
   * Adds the tagging with count to the data structures in this Lexicon.
   */
  protected void addTagging(boolean seen, IntTaggedWord itw, double count) {
    if (seen) {
      seenCounter.incrementCount(itw, count);
      if (itw.tag() == nullTag) {
        words.add(itw);
      } else if (itw.word() == nullWord) {
        tags.add(itw);
      } else {
        // rules.add(itw);
      }
    } else {
      unSeenCounter.incrementCount(itw, count);
      // if (itw.tag() == nullTag) {
      //   sigs.add(itw);
      // }
    }
  }

  /**
   * Returns the index of the signature of the word numbered wordIndex,
   * where the signature is the String representation of unknown word
   * features.  Caches the last signature index returned.
   */
  protected int getSignatureIndex(int wordIndex, int sentencePosition) {
    if (wordIndex == lastWordToSignaturize && sentencePosition == lastSentencePosition) {
      // System.err.println("Signature: cache mapped " + wordIndex + " to " + lastSignatureIndex);
      return lastSignatureIndex;
    } else {
      String uwSig = getSignature((String) Numberer.object("words", wordIndex), sentencePosition);
      int sig = Numberer.number("words", uwSig);
      lastSignatureIndex = sig;
      lastSentencePosition = sentencePosition;
      lastWordToSignaturize = wordIndex;
      return sig;
    }
  }

  /**
   * This routine returns a String that is the "signature" of the class of a
   * word.
   * For, example, it might represent whether it is a number of ends in -s.
   * The strings returned by convention match the pattern UNK-.* , which
   * is just assumed to not match any real word.
   * Behavior depends on the unknownLevel (-uwm flag) passed in to the class.
   * The recognized numbers are 1-5: 5 is fairly English-specific; 4, 3, and 2
   * look for various word features (digits, dashes, etc.) which are only
   * vaguely English-specific; 1 uses the last two characters combined with 
   * a simple classification by capitalization.
   *
   * @param word The word to make a signature for
   * @param loc  Its position in the sentence (mainly so sentence-initial
   *             capitalized words can be treated differently)
   * @return A String that is its signature (equivalence class)
   */
  protected String getSignature(String word, int loc) {
    //    int unknownLevel = Options.get().useUnknownWordSignatures;
    StringBuffer sb = new StringBuffer("UNK");
    switch (unknownLevel) {

      case 5:
        {
          // Reformed Mar 2004 (cdm); hopefully much better now.
          // { -CAPS, -INITC ap, -LC lowercase, 0 } +
          // { -KNOWNLC, 0 } +          [only for INITC]
          // { -NUM, 0 } +
          // { -DASH, 0 } +
          // { -last lowered char(s) if known discriminating suffix, 0}
          int wlen = word.length();
          int numCaps = 0;
          boolean hasDigit = false;
          boolean hasDash = false;
          boolean hasLower = false;
          for (int i = 0; i < wlen; i++) {
            char ch = word.charAt(i);
            if (Character.isDigit(ch)) {
              hasDigit = true;
            } else if (ch == '-') {
              hasDash = true;
            } else if (Character.isLetter(ch)) {
              if (Character.isLowerCase(ch)) {
                hasLower = true;
              } else if (Character.isTitleCase(ch)) {
                hasLower = true;
                numCaps++;
              } else {
                numCaps++;
              }
            }
          }
          char ch0 = word.charAt(0);
          String lowered = word.toLowerCase();
          if (Character.isUpperCase(ch0) || Character.isTitleCase(ch0)) {
            if (loc == 0 && numCaps == 1) {
              sb.append("-INITC");
              if (isKnown(lowered)) {
                sb.append("-KNOWNLC");
              }
            } else {
              sb.append("-CAPS");
            }
          } else if (!Character.isLetter(ch0) && numCaps > 0) {
            sb.append("-CAPS");
          } else if (hasLower) { // (Character.isLowerCase(ch0)) {
            sb.append("-LC");
          }
          if (hasDigit) {
            sb.append("-NUM");
          }
          if (hasDash) {
            sb.append("-DASH");
          }
          if (lowered.endsWith("s") && wlen >= 3) {
            // here length 3, so you don't miss out on ones like 80s
            char ch2 = lowered.charAt(wlen - 2);
            // not -ess suffixes or greek/latin -us, -is
            if (ch2 != 's' && ch2 != 'i' && ch2 != 'u') {
              sb.append("-s");
            }
          } else if (word.length() >= 5 && !hasDash && !(hasDigit && numCaps > 0)) {
            // don't do for very short words;
            // Implement common discriminating suffixes
            if (lowered.endsWith("ed")) {
              sb.append("-ed");
            } else if (lowered.endsWith("ing")) {
              sb.append("-ing");
            } else if (lowered.endsWith("ion")) {
              sb.append("-ion");
            } else if (lowered.endsWith("er")) {
              sb.append("-er");
            } else if (lowered.endsWith("est")) {
              sb.append("-est");
            } else if (lowered.endsWith("ly")) {
              sb.append("-ly");
            } else if (lowered.endsWith("ity")) {
              sb.append("-ity");
            } else if (lowered.endsWith("y")) {
              sb.append("-y");
            } else if (lowered.endsWith("al")) {
              sb.append("-al");
              // } else if (lowered.endsWith("ble")) {
              // sb.append("-ble");
              // } else if (lowered.endsWith("e")) {
              // sb.append("-e");
            }
          }
          break;
        }

      case 4:
        {
          boolean hasDigit = false;
          boolean hasNonDigit = false;
          boolean hasLetter = false;
          boolean hasLower = false;
          boolean hasDash = false;
          boolean hasPeriod = false;
          boolean hasComma = false;
          for (int i = 0; i < word.length(); i++) {
            char ch = word.charAt(i);
            if (Character.isDigit(ch)) {
              hasDigit = true;
            } else {
              hasNonDigit = true;
              if (Character.isLetter(ch)) {
                hasLetter = true;
                if (Character.isLowerCase(ch) || Character.isTitleCase(ch)) {
                  hasLower = true;
                }
              } else {
                if (ch == '-') {
                  hasDash = true;
                } else if (ch == '.') {
                  hasPeriod = true;
                } else if (ch == ',') {
                  hasComma = true;
                }
              }
            }
          }
          // 6 way on letters
          if (Character.isUpperCase(word.charAt(0)) || Character.isTitleCase(word.charAt(0))) {
            if (!hasLower) {
              sb.append("-AC");
            } else if (loc == 0) {
              sb.append("-SC");
            } else {
              sb.append("-C");
            }
          } else if (hasLower) {
            sb.append("-L");
          } else if (hasLetter) {
            sb.append("-U");
          } else {
            // no letter
            sb.append("-S");
          }
          // 3 way on number
          if (hasDigit && !hasNonDigit) {
            sb.append("-N");
          } else if (hasDigit) {
            sb.append("-n");
          }
          // binary on period, dash, comma
          if (hasDash) {
            sb.append("-H");
          }
          if (hasPeriod) {
            sb.append("-P");
          }
          if (hasComma) {
            sb.append("-C");
          }
          if (word.length() > 3) {
            // don't do for very short words: "yes" isn't an "-es" word
            // try doing to lower for further densening and skipping digits
            char ch = word.charAt(word.length() - 1);
            if (Character.isLetter(ch)) {
              sb.append("-");
              sb.append(Character.toLowerCase(ch));
            }
          }
          break;
        }

      case 3:
        {
          // This basically works right, except note that 'S' is applied to all
          // capitalized letters in first word of sentence, not just first....
          sb.append("-");
          char lastClass = '-';  // i.e., nothing
          char newClass;
          int num = 0;
          for (int i = 0; i < word.length(); i++) {
            char ch = word.charAt(i);
            if (Character.isUpperCase(ch) || Character.isTitleCase(ch)) {
              if (loc == 0) {
                newClass = 'S';
              } else {
                newClass = 'L';
              }
            } else if (Character.isLetter(ch)) {
              newClass = 'l';
            } else if (Character.isDigit(ch)) {
              newClass = 'd';
            } else if (ch == '-') {
              newClass = 'h';
            } else if (ch == '.') {
              newClass = 'p';
            } else {
              newClass = 's';
            }
            if (newClass != lastClass) {
              lastClass = newClass;
              sb.append(lastClass);
              num = 1;
            } else {
              if (num < 2) {
                sb.append('+');
              }
              num++;
            }
          }
          if (word.length() > 3) {
            // don't do for very short words: "yes" isn't an "-es" word
            // try doing to lower for further densening and skipping digits
            char ch = Character.toLowerCase(word.charAt(word.length() - 1));
            sb.append('-');
            sb.append(ch);
          }
          break;
        }

      case 2:
        {
          // {-ALLC, -INIT, -UC, -LC, zero} +
          // {-DASH, zero} +
          // {-NUM, -DIG, zero} +
          // {lowerLastChar, zeroIfShort}
          boolean hasDigit = false;
          boolean hasNonDigit = false;
          boolean hasLower = false;
          for (int i = 0; i < word.length(); i++) {
            char ch = word.charAt(i);
            if (Character.isDigit(ch)) {
              hasDigit = true;
            } else {
              hasNonDigit = true;
              if (Character.isLetter(ch)) {
                if (Character.isLowerCase(ch) || Character.isTitleCase(ch)) {
                  hasLower = true;
                }
              }
            }
          }
          if (Character.isUpperCase(word.charAt(0)) || Character.isTitleCase(word.charAt(0))) {
            if (!hasLower) {
              sb.append("-ALLC");
            } else if (loc == 0) {
              sb.append("-INIT");
            } else {
              sb.append("-UC");
            }
          } else if (hasLower) {   // if (Character.isLowerCase(word.charAt(0))) {
            sb.append("-LC");
          }
          // no suffix = no (lowercase) letters
          if (word.indexOf('-') >= 0) {
            sb.append("-DASH");
          }
          if (hasDigit) {
            if (!hasNonDigit) {
              sb.append("-NUM");
            } else {
              sb.append("-DIG");
            }
          } else if (word.length() > 3) {
            // don't do for very short words: "yes" isn't an "-es" word
            // try doing to lower for further densening and skipping digits
            char ch = word.charAt(word.length() - 1);
            sb.append(Character.toLowerCase(ch));
          }
          // no suffix = short non-number, non-alphabetic
          break;
        }

      default:
        sb.append("-");
        sb.append(word.substring(Math.max(word.length() - 2, 0), word.length()));
        sb.append("-");
        if (Character.isLowerCase(word.charAt(0))) {
          sb.append("LOWER");
        } else {
          if (Character.isUpperCase(word.charAt(0))) {
            if (loc == 0) {
              sb.append("INIT");
            } else {
              sb.append("UPPER");
            }
          } else {
            sb.append("OTHER");
          }
        }
    } // end switch (unknownLevel)
    // System.err.println("Summarized " + word + " to " + sb.toString());
    return sb.toString();
  } // end getSignature()

  void buildPT_T() {
    Numberer tagNumberer = Numberer.getGlobalNumberer("tags");
    int numTags = tagNumberer.total();
    m_TT = new double[numTags][numTags];
    m_T = new double[numTags];
    double[] tmp = new double[numTags];
    for (Iterator<IntTaggedWord> i = words.iterator(); i.hasNext();) {
      IntTaggedWord word = i.next();
      IntTaggedWord iTW = new IntTaggedWord(word.word, nullTag);
      double tot = 0.0;
      for (int t = 0; t < numTags; t++) {
        iTW.tag = (short) t;
        tmp[t] = seenCounter.getCount(iTW);
        tot += tmp[t];
      }
      if (tot < 10) {
        continue;
      }
      for (int t = 0; t < numTags; t++) {
        for (int t2 = 0; t2 < numTags; t2++) {
          if (tmp[t2] > 0.0) {
            double c = tmp[t] / tot;
            m_T[t] += c;
            m_TT[t2][t] += c;
          }
        }
      }
    }
  }

  /**
   * Get the score of this word with this tag (as an IntTaggedWord) at this 
   * loc.
   * (Presumably an estimate of P(word | tag).)
   * <p>
   * <i>Implementation documentation:</i>  Seen:
   * c_W = count(W)    c_TW = count(T,W)
   * c_T = count(T)    c_Tunseen = count(T) among new words in 2nd half
   * total = count(seen words)    totalUnseen = count("unseen" words)
   * p_T_U = Pmle(T|"unseen") 
   * pb_T_W = P(T|W).  If (c_W > smoothInUnknownsThreshold) = c_TW/c_W
   * Else (if not smart mutation) pb_T_W = bayes prior smooth[1] with p_T_U
   * p_T= Pmle(T)    p_W = Pmle(W)
   * pb_W_T = log(pb_T_W * p_W / p_T)    [Bayes rule]
   * Note that this doesn't really properly reserve mass to unknowns.
   *
   * Unseen: 
   * c_TS = count(T,Sig|Unseen)    c_S = count(Sig)    c_T = count(T|Unseen)
   * c_U = totalUnseen above
   * p_T_U = Pmle(T|Unseen)
   * pb_T_S = Bayes smooth of Pmle(T|S) with P(T|Unseen)   [smooth[0]]
   * pb_W_T = log(P(W|T)) inverted
   *
   * @param iTW An IntTaggedWord pairing a word and POS tag
   * @param loc The position in the sentence.  <i>In the default implementation
   *               this is used only for unknown words to change their
   *               probability distribution when sentence initial
   * @return A double valued score, usually - log P(word|tag)
   */
  public double score(IntTaggedWord iTW, int loc) {
    int word = iTW.word;
    short tag = iTW.tag;

    iTW.tag = nullTag;
    double c_W = seenCounter.getCount(iTW);
    //double x_W = xferCounter.getCount(iTW);
    iTW.tag = tag;

    double pb_W_T; // always set below

    boolean seen = (c_W > 0.0);
    if (seen) {

      // known word model for P(T|W)

      double c_TW = seenCounter.getCount(iTW);
      //double x_TW = xferCounter.getCount(iTW);
      iTW.word = nullWord;
      double c_T = seenCounter.getCount(iTW);
      double c_Tunseen = unSeenCounter.getCount(iTW);
      iTW.tag = nullTag;
      double total = seenCounter.getCount(iTW);
      double totalUnseen = unSeenCounter.getCount(iTW);
      iTW.tag = tag;
      iTW.word = word;

      //c_TW = Math.sqrt(c_TW);
      //c_TW += 0.5;

      double p_T_U = c_Tunseen / totalUnseen;
      double pb_T_W;  // always set below

      // System.err.println("c_W is " + c_W + " THRESH is " + Options.get().smoothInUnknownsThreshold + " mle = " + (c_TW/c_W));
      if (c_W > smoothInUnknownsThreshold) {
        // we've seen the word enough times to have confidence in its tagging
        pb_T_W = c_TW / c_W;
      } else {
        // we haven't seen the word enough times to have confidence in its tagging
        if (smartMutation) {
          Numberer tagNumberer = Numberer.getGlobalNumberer("tags");
          int numTags = tagNumberer.total();
          if (m_TT == null || numTags != m_T.length) {
            buildPT_T();
          }
          p_T_U *= 0.1;
          //System.out.println("Checking "+iTW);
          for (int t = 0; t < numTags; t++) {
            IntTaggedWord iTW2 = new IntTaggedWord(word, t);
            double p_T_W2 = seenCounter.getCount(iTW2) / c_W;
            if (p_T_W2 > 0) {
              //System.out.println(" Observation of "+tagNumberer.object(t)+" ("+seenCounter.getCount(iTW2)+") mutated to "+tagNumberer.object(iTW.tag)+" at rate "+(m_TT[tag][t]/m_T[t]));
              p_T_U += p_T_W2 * m_TT[tag][t] / m_T[t] * 0.9;
            }
          }
        }
        pb_T_W = (c_TW + smooth[1] * p_T_U) / (c_W + smooth[1]);
      }
      //double pb_T_W = (c_TW+smooth[1]*x_TW)/(c_W+smooth[1]*x_W);

      double p_T = (c_T / total);
      double p_W = (c_W / total);
      pb_W_T = Math.log(pb_T_W * p_W / p_T);

      if (DEBUG_LEXICON) {
	if (iTW.word != debugLastWord) {
	  if (debugLastWord >= 0 && debugPrefix != null) {
	    // the 2nd conjunct in test above handles older serialized files
	    System.err.println(debugPrefix + debugProbs);
	  }
	  debugLastWord = iTW.word;
	  debugLoc = loc;
	  debugProbs = new StringBuffer();
	  debugNoProbs = new StringBuffer(" impossible: ");
	  debugPrefix = "Lexicon: " + Numberer.object("words", debugLastWord) + " (known): ";
	}
	if (pb_W_T > Double.NEGATIVE_INFINITY) {
	  NumberFormat nf = NumberFormat.getNumberInstance();
	  nf.setMaximumFractionDigits(3);
	  debugProbs.append(Numberer.object("tags", tag) + ": cTW=" + c_TW + " c_T=" + c_T + " pb_T_W=" + nf.format(pb_T_W) + " log pb_W_T=" + nf.format(pb_W_T) +", ");
          // debugProbs.append("\n" + "smartMutation=" + smartMutation + " smoothInUnknownsThreshold=" + smoothInUnknownsThreshold + " smooth0=" + smooth[0] + "smooth1=" + smooth[1] + " p_T_U=" + p_T_U + " c_W=" + c_W);
	} else {
	  debugNoProbs.append(Numberer.object("tags", tag)).append(" ");
	}
      } // end if (DEBUG_LEXICON)

    } else {

      // unknown word model for P(T|S)
      int sig = getSignatureIndex(iTW.word, loc);

      iTW.word = sig;
      double c_TS = unSeenCounter.getCount(iTW);
      iTW.tag = nullTag;
      double c_S = unSeenCounter.getCount(iTW);
      iTW.word = nullWord;
      double c_U = unSeenCounter.getCount(iTW);
      double total = seenCounter.getCount(iTW);
      iTW.tag = tag;
      double c_T = unSeenCounter.getCount(iTW);
      double c_Tseen = seenCounter.getCount(iTW);
      iTW.word = word;

      double p_T_U = c_T / c_U;
      if (unknownLevel == 0) {
        c_TS = 0;
        c_S = 0;
      }
      double pb_T_S = (c_TS + smooth[0] * p_T_U) / (c_S + smooth[0]);

      double p_T = (c_Tseen / total);
      double p_W = 1.0 / total;
      pb_W_T = Math.log(pb_T_S * p_W / p_T);
      if (DEBUG_LEXICON) {
	if (iTW.word != debugLastWord) {
	  if (debugLastWord >= 0) {
	    System.err.println(debugPrefix + debugProbs + debugNoProbs);
	  }
	  debugLastWord = iTW.word;
	  debugLoc = loc;
	  debugProbs = new StringBuffer();
	  debugNoProbs = new StringBuffer(" impossible: ");
	  int sigIdx = getSignatureIndex(debugLastWord, debugLoc);
	  debugPrefix = "Lexicon: " + Numberer.object("words", debugLastWord) + " (" + debugLastWord + ") idx " + debugLoc + " -> " + Numberer.getGlobalNumberer("words").object(sigIdx) + " (" + sigIdx + "): ";
	}
	if (pb_W_T > Double.NEGATIVE_INFINITY) {
	  NumberFormat nf = NumberFormat.getNumberInstance();
	  nf.setMaximumFractionDigits(3);
	  debugProbs.append(Numberer.getGlobalNumberer("tags").object(tag) + ": cTS=" + c_TS + " c_T=" + c_T + " pb_T_S=" + nf.format(pb_T_S) + " log pb_W_T=" + nf.format(pb_W_T) +", ");
	  // + " c_S=" + c_S + " c_Unk=" + c_U + " c_T=" + c_T 
	  // + " pb_W_T=" + nf.format(Math.exp(pb_W_T))
	} else {
	  debugNoProbs.append(Numberer.getGlobalNumberer("tags").object(tag)).append(" ");
	}
      } // end if (DEBUG_LEXICON)
    }

    if (pb_W_T > -100.0) {
      return pb_W_T;
    }
    return Double.NEGATIVE_INFINITY;
  } // end score()

  private transient int debugLastWord = -1;
  private transient int debugLoc = -1;
  private transient StringBuffer debugProbs;
  private transient StringBuffer debugNoProbs;
  private transient String debugPrefix;

  public void tune(Collection<Tree> trees) {
    double bestScore = Double.NEGATIVE_INFINITY;
    double[] bestSmooth = {0.0, 0.0};
    for (smooth[0] = 1; smooth[0] <= 1; smooth[0] *= 2.0) {//64
      for (smooth[1] = 0.2; smooth[1] <= 0.2; smooth[1] *= 2.0) {//3
        //for (smooth[0]=0.5; smooth[0]<=64; smooth[0] *= 2.0) {//64
        //for (smooth[1]=0.1; smooth[1]<=12.8; smooth[1] *= 2.0) {//3
        double score = 0.0;
        //score = scoreAll(trees);
        if (Test.verbose) {
          System.out.println("Tuning lexicon: s0 " + smooth[0] + " s1 " + smooth[1] + " is " + score + " " + trees.size() + " trees.");
        }
        if (score > bestScore) {
          System.arraycopy(smooth, 0, bestSmooth, 0, smooth.length);
          bestScore = score;
        }
      }
    }
    System.arraycopy(bestSmooth, 0, smooth, 0, bestSmooth.length);
    if (smartMutation) {
      smooth[0] = 8.0;
      //smooth[1] = 1.6;
      //smooth[0] = 0.5;
      smooth[1] = 0.1;
    }
    if (Test.unseenSmooth > 0.0) {
      smooth[0] = Test.unseenSmooth;
    }
    if (Test.verbose) {
      System.out.println("Tuning selected smoothUnseen " + smooth[0] + 
			 " smoothSeen " + smooth[1] + " at " + bestScore);
    }
  }

  /**
   * Populates data in this Lexicon from the character stream
   * given by the Reader r.
   */
  public void readData(BufferedReader in) throws IOException {
    final String SEEN = "SEEN";
    String line;
    int lineNum = 1;
    boolean seen;
    // all lines have one tagging with raw count per line
    line = in.readLine();
    Pattern p = Pattern.compile("^smooth\\[([0-9])\\] = (.*)$");
    while (line != null && line.length() > 0) {
      try {
        Matcher m = p.matcher(line);
        if (m.matches()) {
          int i = Integer.parseInt(m.group(1));
          double value = Double.parseDouble(m.group(2));
          smooth[i] = value;
        } else {
          String[] fields = StringUtils.splitOnCharWithQuoting(line, ' ', '\"', '\\'); // split on spaces, quote with doublequote, and escape with backslash
          //        System.out.println("fields:\n" + fields[0] + "\n" + fields[1] + "\n" + fields[2] + "\n" + fields[3] + "\n" + fields[4]);
          if (fields[3].equals(SEEN)) {
            seen = true;
          } else {
            seen = false;
          }
          addTagging(seen, new IntTaggedWord(fields[2], fields[0]), Double.parseDouble(fields[4]));
        }
      } catch (Exception e) {
        throw new IOException("Error on line " + lineNum + ": " + line);
      }
      lineNum++;
      line = in.readLine();
    }
  }

  /**
   * Writes out data from this Object to the Writer w. Rules are separated by
   * newline, and rule elements are delimited by \t.
   */
  public void writeData(Writer w) throws IOException {
    PrintWriter out = new PrintWriter(w);
    IntTaggedWord itw;
    for (Iterator<IntTaggedWord> sI = seenCounter.keySet().iterator(); sI.hasNext();) {
      itw = sI.next();
      out.println(itw.toLexicalEntry() + " SEEN " + seenCounter.getCount(itw));
    }
    for (Iterator<IntTaggedWord> uI = unSeenCounter.keySet().iterator(); uI.hasNext();) {
      itw = uI.next();
      out.println(itw.toLexicalEntry() + " UNSEEN " + unSeenCounter.getCount(itw));
    }
    for (int i = 0; i < smooth.length; i++) {
      out.println("smooth[" + i + "] = " + smooth[i]);
    }
    out.flush();
  }

  private static final int STATS_BINS = 15;

  /** Print some statistics about this lexicon. */
  public void printLexStats() {
    if (rulesWithWord == null) {
      initRulesWithWord();
    }
    System.out.println("BaseLexicon statistics");
    System.out.println("unknownLevel is " + unknownLevel);
    // System.out.println("Rules size: " + rules.size());
    int accumulated = 0;
    for (int k = 0; k < rulesWithWord.length; k++) {
      accumulated += rulesWithWord[k].size();
    }
    System.out.println("Sum of rulesWithWord: " + accumulated);
    System.out.println("Tags size: " + tags.size());
    System.out.println("Words size: " + words.size());
    // System.out.println("Unseen Sigs size: " + sigs.size() + 
    //		       " [number of unknown equivalence classes]");
    System.out.println("rulesWithWord length: " + rulesWithWord.length + 
		       " [should be sum of words + sigs]");
    int[] lengths = new int[STATS_BINS];
    ArrayList[] words = new ArrayList[STATS_BINS];
    for (int j = 0; j < STATS_BINS; j++) {
      words[j] = new ArrayList();
    }
    for (int i = 0; i < rulesWithWord.length; i++) {
      int num = rulesWithWord[i].size();
      if (num > STATS_BINS - 1) {
	num = STATS_BINS - 1;
      }
      lengths[num]++;
      if (num >= STATS_BINS / 2) {
	words[num].add(Numberer.object("words", i));
      }
    }
    System.out.println("Stats on how many taggings for how many words");
    for (int j = 0; j < STATS_BINS; j++) {
      System.out.print(j + " taggings: " + lengths[j] + " words ");
      if (j >= STATS_BINS / 2) {
	System.out.print(words[j]);
      }
      System.out.println();
    }
    NumberFormat nf = NumberFormat.getNumberInstance();
    nf.setMaximumFractionDigits(0);
    System.out.println("Unseen counter: " + unSeenCounter.toString(nf));
  }

  /**
   * Evaluates how many words (= terminals) in a collection of trees
   * are covered by the lexicon. First arg is the collection of
   * trees; second through fourth args get the results.
   */
  public double evaluateCoverage(Collection<Tree> trees, Set missingWords, Set missingTags, Set<IntTaggedWord> missingTW) {

    List<IntTaggedWord> iTW1 = new ArrayList<IntTaggedWord>();
    for (Iterator<Tree> i1 = trees.iterator(); i1.hasNext();) {
      iTW1.addAll(treeToEvents(i1.next()));
    }
    List<IntTaggedWord> iTW = iTW1;
    int total = 0;
    int unseen = 0;

    for (Iterator<IntTaggedWord> i = iTW.iterator(); i.hasNext();) {
      total++;
      IntTaggedWord itw = i.next();
      if (!words.contains(new IntTaggedWord(itw.word(), nullTag))) {
        missingWords.add(Numberer.object("word", itw.word()));
      }
      if (!tags.contains(new IntTaggedWord(nullWord, itw.tag()))) {
        missingTags.add(Numberer.object("tag", itw.tag()));
      }
      //      if (!rules.contains(itw)) {
      if (seenCounter.getCount(itw) == 0.0) {
        unseen++;
        missingTW.add(itw);
      }
    }
    return (double) unseen / total;
  }

  private static final long serialVersionUID = 40L;

}
