package edu.stanford.nlp.parser.lexparser;

import edu.stanford.nlp.ling.Label;
import edu.stanford.nlp.ling.Sentence;
import edu.stanford.nlp.ling.TaggedWord;
import edu.stanford.nlp.ling.Word;
import edu.stanford.nlp.math.ArrayMath;
import edu.stanford.nlp.trees.Tree;
import edu.stanford.nlp.trees.Treebank;
import edu.stanford.nlp.trees.international.pennchinese.CharacterLevelTagExtender;
import edu.stanford.nlp.stats.GeneralizedCounter;
import edu.stanford.nlp.util.Numberer;
import edu.stanford.nlp.stats.Counter;
import edu.stanford.nlp.stats.Distribution;
import edu.stanford.nlp.stats.*;

import java.util.*;

/**
 * Performs word segmentation with a hierarchical markov model over POS
 * and over characters given POS.
 *
 * @author Galen Andrew
 */
public class ChineseMarkovWordSegmenter implements WordSegmenter {
  Distribution initialPOSDist;
  Map markovPOSDists;
  ChineseCharacterBasedLexicon lex;
  Set POSes;

  public ChineseMarkovWordSegmenter(ChineseCharacterBasedLexicon lex) {
    this.lex = lex;
  }

  public ChineseMarkovWordSegmenter() {
    lex = new ChineseCharacterBasedLexicon();
  }

  public void train(Collection trees) {
    Numberer tagNumberer = Numberer.getGlobalNumberer("tags");
    lex.train(trees);
    Counter initial = new Counter();
    GeneralizedCounter ruleCounter = new GeneralizedCounter(2);
    for (Iterator treeIter = trees.iterator(); treeIter.hasNext();) {
      Tree tree = (Tree) treeIter.next();
      List tags = tree.preTerminalYield();
      String last = null;
      for (Iterator iter = tags.iterator(); iter.hasNext();) {
        String tag = ((Label) iter.next()).value();
        tagNumberer.number(tag);
        if (last == null) {
          initial.incrementCount(tag);
        } else {
          ruleCounter.incrementCount2D(last, tag);
        }
        last = tag;
      }
    }
    int numTags = tagNumberer.total();
    POSes = new HashSet(tagNumberer.objects());
    initialPOSDist = Distribution.laplaceSmoothedDistribution(initial, numTags, 0.5);
    markovPOSDists = new HashMap();
    //    markovPOSDists = new HashMap<String, Distribution>();  // 1.5
    Set entries = ruleCounter.lowestLevelCounterEntrySet();
    for (Iterator iter = entries.iterator(); iter.hasNext();) {
      Map.Entry entry = (Map.Entry) iter.next();
      //      Map.Entry<List<String>, Counter> entry = (Map.Entry<List<String>, Counter>) iter.next();
      Distribution d = Distribution.laplaceSmoothedDistribution((Counter) entry.getValue(), numTags, 0.5);
      markovPOSDists.put(((List) entry.getKey()).get(0), d);
    }
  }

  public Sentence segmentWords(String s) {
    return segmentWordsWithMarkov(s);
  }

  public Sentence basicSegmentWords(String s) {
    Numberer tagNumberer = Numberer.getGlobalNumberer("tags");
    int length = s.length();
    //    Set<String> POSes = (Set<String>) POSDistribution.keySet();  // 1.5
    // best score of span
    double[][] scores = new double[length][length + 1];
    // best (last index of) first word for this span
    int[][] splitBacktrace = new int[length][length + 1];
    // best tag for word over this span
    int[][] POSbacktrace = new int[length][length + 1];
    for (int i = 0; i < length; i++) {
      Arrays.fill(scores[i], Double.NEGATIVE_INFINITY);
    }
    // first fill in word probabilities
    for (int diff = 1; diff <= 10; diff++) {
      for (int start = 0; start + diff <= length; start++) {
        int end = start + diff;
        StringBuffer wordBuf = new StringBuffer();
        for (int pos = start; pos < end; pos++) {
          wordBuf.append(s.charAt(pos));
        }
        String word = wordBuf.toString();
        //        for (String tag : POSes) {  // 1.5
        for (Iterator iter = POSes.iterator(); iter.hasNext();) {
          String tag = (String) iter.next();
          IntTaggedWord itw = new IntTaggedWord(word, tag);
          double newScore = lex.score(itw, 0) + Math.log(lex.getPOSDistribution().probabilityOf(tag));
          if (newScore > scores[start][end]) {
            scores[start][end] = newScore;
            splitBacktrace[start][end] = end;
            POSbacktrace[start][end] = itw.tag();
          }
        }
      }
    }
    // now fill in word combination probabilities
    for (int diff = 2; diff <= length; diff++) {
      for (int start = 0; start + diff <= length; start++) {
        int end = start + diff;
        for (int split = start + 1; split < end && split - start <= 10; split++) {
          if (splitBacktrace[start][split] != split) {
            continue; // only consider words on left
          }
          double newScore = scores[start][split] + scores[split][end];
          if (newScore > scores[start][end]) {
            scores[start][end] = newScore;
            splitBacktrace[start][end] = split;
          }
        }
      }
    }

    List words = new ArrayList();
    int start = 0;
    while (start < length) {
      int end = splitBacktrace[start][length];
      StringBuffer wordBuf = new StringBuffer();
      for (int pos = start; pos < end; pos++) {
        wordBuf.append(s.charAt(pos));
      }
      String word = wordBuf.toString();
      String tag = (String) tagNumberer.object(POSbacktrace[start][end]);

      words.add(new TaggedWord(word, tag));
      start = end;
    }

    return new Sentence(words);
  }

  public Sentence segmentWordsWithMarkov(String s) {
    Numberer tagNumberer = Numberer.getGlobalNumberer("tags");
    int length = s.length();
    //    Set<String> POSes = (Set<String>) POSDistribution.keySet();  // 1.5
    int numTags = POSes.size();
    // score of span with initial word of this tag
    double[][][] scores = new double[length][length + 1][numTags];
    // best (length of) first word for this span with this tag
    int[][][] splitBacktrace = new int[length][length + 1][numTags];
    // best tag for second word over this span, if first is this tag
    int[][][] POSbacktrace = new int[length][length + 1][numTags];
    for (int i = 0; i < length; i++) {
      for (int j = 0; j < length + 1; j++) {
        Arrays.fill(scores[i][j], Double.NEGATIVE_INFINITY);
      }
    }
    // first fill in word probabilities
    for (int diff = 1; diff <= 10; diff++) {
      for (int start = 0; start + diff <= length; start++) {
        int end = start + diff;
        StringBuffer wordBuf = new StringBuffer();
        for (int pos = start; pos < end; pos++) {
          wordBuf.append(s.charAt(pos));
        }
        String word = wordBuf.toString();
        //        for (String tag : POSes) {  // 1.5
        for (Iterator iter = POSes.iterator(); iter.hasNext();) {
          String tag = (String) iter.next();
          IntTaggedWord itw = new IntTaggedWord(word, tag);
          double score = lex.score(itw, 0);
          if (start == 0) {
            score += Math.log(initialPOSDist.probabilityOf(tag));
          }
          scores[start][end][itw.tag()] = score;
          splitBacktrace[start][end][itw.tag()] = end;
        }
      }
    }
    // now fill in word combination probabilities
    for (int diff = 2; diff <= length; diff++) {
      for (int start = 0; start + diff <= length; start++) {
        int end = start + diff;
        for (int split = start + 1; split < end && split - start <= 10; split++) {
          //        for (String tag : POSes) {  // 1.5
          for (Iterator iter = POSes.iterator(); iter.hasNext();) {
            String tag = (String) iter.next();
            int tagNum = tagNumberer.number(tag);
            if (splitBacktrace[start][split][tagNum] != split) {
              continue;
            }
            Distribution rTagDist = (Distribution) markovPOSDists.get(tag);
            if (rTagDist == null) {
              continue; // this happens with "*" POS
            }
            //        for (String rTag : POSes) {  // 1.5
            for (Iterator iter2 = POSes.iterator(); iter2.hasNext();) {
              String rTag = (String) iter2.next();
              int rTagNum = tagNumberer.number(rTag);
              double newScore = scores[start][split][tagNum] + scores[split][end][rTagNum] + Math.log(rTagDist.probabilityOf(rTag));
              if (newScore > scores[start][end][tagNum]) {
                scores[start][end][tagNum] = newScore;
                splitBacktrace[start][end][tagNum] = split;
                POSbacktrace[start][end][tagNum] = rTagNum;
              }
            }
          }
        }
      }
    }
    int nextPOS = ArrayMath.argmax(scores[0][length]);
    List words = new ArrayList();

    int start = 0;
    while (start < length) {
      int split = splitBacktrace[start][length][nextPOS];
      StringBuffer wordBuf = new StringBuffer();
      for (int i = start; i < split; i++) {
        wordBuf.append(s.charAt(i));
      }
      String word = wordBuf.toString();
      String tag = (String) tagNumberer.object(nextPOS);
      words.add(new TaggedWord(word, tag));
      if (split < length) {
        nextPOS = POSbacktrace[start][length][nextPOS];
      }
      start = split;
    }

    return new Sentence(words);
  }

  private Distribution getSegmentedWordLengthDistribution(Treebank tb) {
    CharacterLevelTagExtender ext = new CharacterLevelTagExtender();
    Counter c = new Counter();
    for (Iterator iterator = tb.iterator(); iterator.hasNext();) {
      Tree gold = (Tree) iterator.next();
      StringBuffer goldChars = new StringBuffer();
      Sentence goldYield = gold.yield();
      for (Iterator wordIter = goldYield.iterator(); wordIter.hasNext();) {
        Word word = (Word) wordIter.next();
        goldChars.append(word);
      }
      Sentence ourWords = segmentWords(goldChars.toString());
      for (int i = 0; i < ourWords.size(); i++) {
        c.incrementCount(new Integer(ourWords.get(i).toString().length()));
      }
    }
    Distribution wordLengthDist = Distribution.getDistribution(c);
    return wordLengthDist;
  }

  private static final long serialVersionUID = 1559606198270645508L;
}
