package edu.stanford.nlp.parser.lexparser;

import edu.stanford.nlp.util.Numberer;
import edu.stanford.nlp.util.Timing;
import edu.stanford.nlp.trees.TreeFactory;
import edu.stanford.nlp.trees.TreebankLanguagePack;
import edu.stanford.nlp.trees.Tree;
import edu.stanford.nlp.trees.LabeledScoredTreeFactory;
import edu.stanford.nlp.math.SloppyMath;
import edu.stanford.nlp.ling.HasWord;
import edu.stanford.nlp.ling.StringLabel;
import edu.stanford.nlp.ling.HasTag;
import edu.stanford.nlp.ling.StringLabelFactory;

import java.util.*;
import java.util.regex.Matcher;

/** An exhaustive generalized CKY PCFG parser.
 *  Fairly carefully optimized to be fast.
 *
 *  @author Dan Klein
 */
public class ExhaustivePCFGParser implements Scorer {

  // public static long insideTime = 0;  // for profiling
  // public static long outsideTime = 0;

  protected String goalStr;
  protected String stateSpace;
  protected Numberer stateNumberer;
  protected Numberer wordNumberer = Numberer.getGlobalNumberer("words");
  protected Numberer tagNumberer = Numberer.getGlobalNumberer("tags");

  protected TreeFactory tf;

  protected BinaryGrammar bg;
  protected UnaryGrammar ug;
  protected Lexicon lex;
  protected Options op;
  protected TreebankLanguagePack tlp;

  protected OutsideRuleFilter orf;

  // inside scores
  protected float[][][] iScore;  // start idx, end idx, state -> logProb
  // outside scores
  protected float[][][] oScore;  // start idx, end idx, state -> logProb
  protected float bestScore;

  protected int[][][] wordsInSpan; // number of words in span with this state

  protected boolean[][] oFilteredStart;
  protected boolean[][] oFilteredEnd;

  protected boolean[][] iPossibleByL;
  protected boolean[][] iPossibleByR;
  protected boolean[][] oPossibleByL;
  protected boolean[][] oPossibleByR;

  protected int[] words;
  protected int length;
  protected boolean[][] tags;
  private int myMaxLength = -0xDEADBEEF;

  protected int numStates;
  private int arraySize = 0;

  public void setGoalString(String goalStr) {
    this.goalStr = goalStr;
  }

  public double oScore(Edge edge) {
    double oS = oScore[edge.start][edge.end][edge.state];
    if (Test.pcfgThreshold) {
      double iS = iScore[edge.start][edge.end][edge.state];
      if (iS + oS - bestScore < Test.pcfgThresholdValue) {
        return Double.NEGATIVE_INFINITY;
      }
    }
    return oS;
  }

  public double iScore(Edge edge) {
    return iScore[edge.start][edge.end][edge.state];
  }

  public boolean oPossible(Hook hook) {
    return (hook.isPreHook() ? oPossibleByR[hook.end][hook.state] : oPossibleByL[hook.start][hook.state]);
  }

  public boolean iPossible(Hook hook) {
    return (hook.isPreHook() ? iPossibleByR[hook.start][hook.subState] : iPossibleByL[hook.end][hook.subState]);
  }


  public boolean oPossibleL(int state, int start) {
    return oPossibleByL[start][state];
  }

  public boolean oPossibleR(int state, int end) {
    return oPossibleByR[end][state];
  }

  public boolean iPossibleL(int state, int start) {
    return iPossibleByL[start][state];
  }

  public boolean iPossibleR(int state, int end) {
    return iPossibleByR[end][state];
  }

  protected void buildOFilter() {
    orf.init();
    for (int start = 0; start < length; start++) {
      orf.leftAccepting(oFilteredStart[start]);
      orf.advanceRight(tags[start]);
    }
    for (int end = length; end > 0; end--) {
      orf.rightAccepting(oFilteredEnd[end]);
      orf.advanceLeft(tags[end - 1]);
    }
  }


  public double validateBinarizedTree(Tree tree, int start) {
    if (tree.isLeaf()) {
      return 0.0;
    }
    double epsilon = 0.0001;
    if (tree.isPreTerminal()) {
      int tag = Numberer.number("tags", tree.label().value());
      int word = Numberer.number("words", tree.children()[0].label().value());
      IntTaggedWord iTW = new IntTaggedWord(word, tag);
      double score = lex.score(iTW, start);
      double bound = iScore[start][start + 1][Numberer.number(stateSpace, tree.label().value())];
      if (score > bound + epsilon) {
        System.out.println("Invalid tagging:");
        System.out.println("  Tag: " + tree.label().value());
        System.out.println("  Word: " + tree.children()[0].label().value());
        System.out.println("  Score: " + score);
        System.out.println("  Bound: " + bound);
      }
      return score;
    }
    int parent = Numberer.number(stateSpace, tree.label().value());
    int firstChild = Numberer.number(stateSpace, tree.children()[0].label().value());
    if (tree.children().length == 1) {
      UnaryRule ur = new UnaryRule();
      ur.parent = parent;
      ur.child = firstChild;
      double score = SloppyMath.max(ug.scoreRule(ur), -10000.0) + validateBinarizedTree(tree.children()[0], start);
      double bound = iScore[start][start + tree.yield().size()][parent];
      if (score > bound + epsilon) {
        System.out.println("Invalid unary:");
        System.out.println("  Parent: " + tree.label().value());
        System.out.println("  Child: " + tree.children()[0].label().value());
        System.out.println("  Start: " + start);
        System.out.println("  End: " + (start + tree.yield().size()));
        System.out.println("  Score: " + score);
        System.out.println("  Bound: " + bound);
      }
      return score;
    }
    BinaryRule br = new BinaryRule();
    br.parent = parent;
    br.leftChild = firstChild;
    br.rightChild = Numberer.number(stateSpace, tree.children()[1].label().value());
    double score = SloppyMath.max(bg.scoreRule(br), -10000.0) + validateBinarizedTree(tree.children()[0], start) + validateBinarizedTree(tree.children()[1], start + tree.children()[0].yield().size());
    double bound = iScore[start][start + tree.yield().size()][parent];
    if (score > bound + epsilon) {
      System.out.println("Invalid binary:");
      System.out.println("  Parent: " + tree.label().value());
      System.out.println("  LChild: " + tree.children()[0].label().value());
      System.out.println("  RChild: " + tree.children()[1].label().value());
      System.out.println("  Start: " + start);
      System.out.println("  End: " + (start + tree.yield().size()));
      System.out.println("  Score: " + score);
      System.out.println("  Bound: " + bound);
    }
    return score;
  }

  // needs to be set up so that uses same Train options...
  public Tree scoreNonBinarizedTree(Tree tree) {
    TreeAnnotatorAndBinarizer binarizer = new TreeAnnotatorAndBinarizer(op.tlpParams, op.forceCNF, !Train.outsideFactor(), true);
    tree = binarizer.transformTree(tree);
    scoreBinarizedTree(tree, 0);
    return op.tlpParams.subcategoryStripper().transformTree(new Debinarizer(tlp,op.forceCNF).transformTree(tree));
    //    return debinarizer.transformTree(t);
  }

  //
  public double scoreBinarizedTree(Tree tree, int start) {
    if (tree.isLeaf()) {
      return 0.0;
    }
    if (tree.isPreTerminal()) {
      int tag = Numberer.number("tags", tree.label().value());
      int word = Numberer.number("words", tree.children()[0].label().value());
      IntTaggedWord iTW = new IntTaggedWord(word, tag);
      // if (lex.score(iTW,(leftmost ? 0 : 1)) == Double.NEGATIVE_INFINITY) {
      //   System.out.println("NO SCORE FOR: "+iTW);
      // }
      double score = lex.score(iTW, start);
      tree.setScore(score);
      return score;
    }
    int parent = Numberer.number(stateSpace, tree.label().value());
    int firstChild = Numberer.number(stateSpace, tree.children()[0].label().value());
    if (tree.children().length == 1) {
      UnaryRule ur = new UnaryRule();
      ur.parent = parent;
      ur.child = firstChild;
      //+ DEBUG
      // if (ug.scoreRule(ur) < -10000) {
      //        System.out.println("Grammar doesn't have rule: " + ur);
      // }
      //      return SloppyMath.max(ug.scoreRule(ur), -10000.0) + scoreBinarizedTree(tree.children()[0], leftmost);
      double score = ug.scoreRule(ur) + scoreBinarizedTree(tree.children()[0], start);
      tree.setScore(score);
      return score;
    }
    BinaryRule br = new BinaryRule();
    br.parent = parent;
    br.leftChild = firstChild;
    br.rightChild = Numberer.number(stateSpace, tree.children()[1].label().value());
    //+ DEBUG
    // if (bg.scoreRule(br) < -10000) {
    //  System.out.println("Grammar doesn't have rule: " + br);
    // }
    //    return SloppyMath.max(bg.scoreRule(br), -10000.0) +
    //            scoreBinarizedTree(tree.children()[0], leftmost) +
    //            scoreBinarizedTree(tree.children()[1], false);
    double score = bg.scoreRule(br) + scoreBinarizedTree(tree.children()[0], start) + scoreBinarizedTree(tree.children()[1], start + tree.children()[0].yield().size());
    tree.setScore(score);
    return score;
  }


  protected static final boolean spillGuts = false;
  protected static boolean dumpTagging = false;
  protected static long time = System.currentTimeMillis();

  protected static void tick(String str) {
    long time2 = System.currentTimeMillis();
    long diff = time2 - time;
    time = time2;
    System.err.print("done.  " + diff + "\n" + str);
  }

  protected boolean floodTags = false;
  protected List sentence = null;
  protected LatticeReader lr = null;

  protected int[][] narrowLExtent = null; // the rightmost left extent of state s ending at position i
  protected int[][] wideLExtent = null; // the leftmost left extent of state s ending at position i
  protected int[][] narrowRExtent = null; // the leftmost right extent of state s starting at position i
  protected int[][] wideRExtent = null; // the rightmost right extent of state s starting at position i

  protected boolean[] isTag = null;

  public boolean parse(List sentence) { // ExhaustivePCFGParser
    lr = null; // better nullPointer exception than silent error
    //System.out.println("is it a taggedword?" + (sentence.get(0) instanceof TaggedWord)); //debugging
    if (sentence != this.sentence) {
      this.sentence = sentence;
      floodTags = false;
    }
    if (Test.verbose) {
      Timing.tick("Starting pcfg parse.");
    }
    if (spillGuts) {
      tick("Starting PCFG parse...");
    }
    length = sentence.size();
    if (length > arraySize) {
      if (length > Test.maxLength + 1 || length >= myMaxLength) {
        throw new OutOfMemoryError("Refusal to create such large arrays.");
      } else {
        try {
          createArrays(length + 1);
        } catch (OutOfMemoryError e) {
          myMaxLength = length;
          if (arraySize > 0) {
            try {
              createArrays(arraySize);
            } catch (OutOfMemoryError e2) {
              throw new RuntimeException("CANNOT EVEN CREATE ARRAYS OF ORIGINAL SIZE!!!");
            }
          }
          throw e;
        }
      }
      arraySize = length + 1;
      if (Test.verbose) {
        System.err.println("Created PCFG parser arrays of size " + arraySize);
      }
    }
    int goal = stateNumberer.number(goalStr);
    if (Test.verbose) {
      //      System.out.println(numStates + " states, " + goal + " is the goal state.");
      System.err.println(bg.numRules() + " binary rules, " + ug.numRules() + " unary rules.");
      // System.err.println(new ArrayList(ug.coreRules.keySet()));
      System.err.print("Initializing...");
    }
    // map to words
    words = new int[length];
    int unk = 0;
    StringBuffer unkWords = new StringBuffer("[");
    for (int i = 0; i < length; i++) {
      Object o = sentence.get(i);
      if (o instanceof HasWord) {
        o = ((HasWord) o).word();
      }
      if (Test.verbose && !lex.isKnown(wordNumberer.number(o.toString()))) {
        unk++;
        unkWords.append(" ");
        unkWords.append(o);
      }
      words[i] = wordNumberer.number(o.toString());
      //else
      //      words[i] = wordNumberer.number(Lexicon.UNKNOWN_WORD);

    }
    // initialize inside and outside score arrays
    if (spillGuts) {
      tick("Wiping arrays...");
    }
    for (int start = 0; start < length; start++) {
      for (int end = start + 1; end <= length; end++) {
        Arrays.fill(iScore[start][end], Float.NEGATIVE_INFINITY);
        if (op.doDep) {
          Arrays.fill(oScore[start][end], Float.NEGATIVE_INFINITY);
        }
        if (Test.lengthNormalization) {
          Arrays.fill(wordsInSpan[start][end], 1);
        }
      }
    }
    for (int loc = 0; loc <= length; loc++) {
      Arrays.fill(narrowLExtent[loc], -1); // the rightmost left with state s ending at i that we can get is the beginning
      Arrays.fill(wideLExtent[loc], length + 1); // the leftmost left with state s ending at i that we can get is the end
      Arrays.fill(narrowRExtent[loc], length + 1); // the leftmost right with state s starting at i that we can get is the end
      Arrays.fill(wideRExtent[loc], -1); // the rightmost right with state s starting at i that we can get is the beginning
    }
    // int puncTag = stateNumberer.number(".");
    // boolean lastIsPunc = false;
    if (Test.verbose) {
      Timing.tick("done.");
      unkWords.append(" ]");
      op.tlpParams.pw().println("Unknown words: " + unk + " " + unkWords);
      System.err.print("Starting filters...");
    }
    // do tags
    if (spillGuts) {
      tick("Tagging...");
    }
    initializeChart(sentence);
    //if (Test.outsideFilter)
    // buildOFilter();
    if (Test.verbose) {
      Timing.tick("done.");
      System.err.print("Starting insides...");
    }
    // do the inside probabilities
    doInsideScores();
    if (Test.verbose) {
      // insideTime += Timing.tick("done.");
      Timing.tick("done.");
      System.out.println("PCFG parsing " + length + " words (incl. stop): insideScore = " + iScore[0][length][goal]);
    }
    bestScore = iScore[0][length][goal];
    boolean succeeded = hasParse();
    if (Test.doRecovery && !succeeded && !floodTags) {
      floodTags = true; // sentence will try to reparse
      System.err.println("Trying recovery parse...");
      return parse(sentence);
    }
    if (!op.doDep) {
      return succeeded;
    }
    if (Test.verbose) {
      System.err.print("Starting outsides...");
    }
    // outside scores
    oScore[0][length][goal] = 0.0f;
    doOutsideScores();
    //System.out.println("State rate: "+((int)(1000*ohits/otries))/10.0);
    //System.out.println("Traversals: "+ohits);
    if (Test.verbose) {
      // outsideTime += Timing.tick("Done.");
      Timing.tick("done.");
      System.out.print("Starting half-filters...");
    }
    for (int loc = 0; loc <= length; loc++) {
      Arrays.fill(iPossibleByL[loc], false);
      Arrays.fill(iPossibleByR[loc], false);
      Arrays.fill(oPossibleByL[loc], false);
      Arrays.fill(oPossibleByR[loc], false);
    }
    for (int start = 0; start < length; start++) {
      for (int end = start + 1; end <= length; end++) {
        for (int state = 0; state < numStates; state++) {
          if (iScore[start][end][state] > Float.NEGATIVE_INFINITY && oScore[start][end][state] > Float.NEGATIVE_INFINITY) {
            iPossibleByL[start][state] = true;
            iPossibleByR[end][state] = true;
            oPossibleByL[start][state] = true;
            oPossibleByR[end][state] = true;
          }
        }
      }
    }
    if (Test.verbose) {
      Timing.tick("done.");
    }
    if (false) {
      long iNZ = 0;
      long oNZ = 0;
      long bNZ = 0;
      long tot = 0;
      for (int start = 0; start < length; start++) {
        for (int end = start + 1; end <= length; end++) {
          for (int s = 0; s < numStates; s++) {
            tot++;
            if (iScore[start][end][s] != Double.NEGATIVE_INFINITY) {
              iNZ++;
            }
            if (oScore[start][end][s] != Double.NEGATIVE_INFINITY) {
              oNZ++;
            }
            if (oScore[start][end][s] != Double.NEGATIVE_INFINITY && iScore[start][end][s] != Double.NEGATIVE_INFINITY) {
              bNZ++;
            }
          }
        }
      }
      System.out.println("Zero Saturation (i): " + ((int) (1000 * iNZ / tot)) / 10.0);
      System.out.println("Zero Saturation (o): " + ((int) (1000 * oNZ / tot)) / 10.0);
      System.out.println("Zero Saturation (b): " + ((int) (1000 * bNZ / tot)) / 10.0);
      Timing.tick("Done with zero scan.");
    }
    return succeeded;
  }

  public boolean parse(LatticeReader lr) { // ExhaustivePCFGParser
    sentence = null; // better nullPointer exception than silent error
    if (lr != this.lr) {
      this.lr = lr;
      floodTags = false;
    }

    if (Test.verbose) {
      Timing.tick("Starting pcfg parse.");
    }
    if (spillGuts) {
      tick("Starting PCFG parse...");
    }
    length = lr.getNumStates();
    if (length > arraySize) {
      if (length > Test.maxLength + 1 || length >= myMaxLength) {
        throw new OutOfMemoryError("Refusal to create such large arrays.");
      } else {
        try {
          createArrays(length + 1);
        } catch (OutOfMemoryError e) {
          myMaxLength = length;
          if (arraySize > 0) {
            try {
              createArrays(arraySize);
            } catch (OutOfMemoryError e2) {
              throw new RuntimeException("CANNOT EVEN CREATE ARRAYS OF ORIGINAL SIZE!!!");
            }
          }
          throw e;
        }
        arraySize = length + 1;
        if (Test.verbose) {
          System.out.println("Created PCFG parser arrays of size " + arraySize);
        }
      }
    }
    int goal = stateNumberer.number(goalStr);
    if (Test.verbose) {
      //      System.out.println(numStates + " states, " + goal + " is the goal state.");
      System.out.println(bg.numRules() + " binary rules, " + ug.numRules() + " unary rules.");
      //System.out.println(new ArrayList(ug.coreRules.keySet()));
      System.out.print("Initializing...");
    }
    // map to words
    List lWords = lr.getLatticeWords();
    words = new int[lWords.size()];
    int unk = 0;
    StringBuffer unkWords = new StringBuffer("[");
    for (int i = 0, size = lWords.size(); i < size; i++) {
      String s = ((LatticeReader.LatticeWord) lWords.get(i)).word;
      if (Test.verbose && !lex.isKnown(wordNumberer.number(s))) {
        unk++;
        unkWords.append(" ");
        unkWords.append(s);
      }
      words[i] = wordNumberer.number(s);
      //else
      //      words[i] = wordNumberer.number(Lexicon.UNKNOWN_WORD);

      /*
      if (o instanceof HasTag) { // collect the tags of the input sentence
        String tag = ((HasTag)o).tag();
        if(tagNumberer.hasSeen(tag))
          inputTags[i] = tagNumberer.number(tag);
      }
      */
    }
    // initialize inside and outside score arrays
    if (spillGuts) {
      tick("Wiping arrays...");
    }
    for (int start = 0; start < length; start++) {
      for (int end = start + 1; end <= length; end++) {
        Arrays.fill(iScore[start][end], Float.NEGATIVE_INFINITY);
        if (op.doDep) {
          Arrays.fill(oScore[start][end], Float.NEGATIVE_INFINITY);
        }
      }
    }
    for (int loc = 0; loc <= length; loc++) {
      Arrays.fill(narrowLExtent[loc], -1); // the rightmost left with state s ending at i that we can get is the beginning
      Arrays.fill(wideLExtent[loc], length + 1); // the leftmost left with state s ending at i that we can get is the end
      Arrays.fill(narrowRExtent[loc], length + 1); // the leftmost right with state s starting at i that we can get is the end
      Arrays.fill(wideRExtent[loc], -1); // the rightmost right with state s starting at i that we can get is the beginning
    }
    // int puncTag = stateNumberer.number(".");
    // boolean lastIsPunc = false;
    if (Test.verbose) {
      Timing.tick("done.");
      unkWords.append(" ]");
      op.tlpParams.pw().println("Unknown words: " + unk + " " + unkWords);
      System.out.print("Starting filters...");
    }
    // do tags
    if (spillGuts) {
      tick("Tagging...");
    }
    initializeChart(lr);
    //if (Test.outsideFilter)
    // buildOFilter();
    if (Test.verbose) {
      Timing.tick("done.");
      System.out.print("Starting insides...");
    }
    // do the inside probabilities
    doInsideScores();
    if (Test.verbose) {
      // insideTime += Timing.tick("done.");
      Timing.tick("done.");
      System.out.println("PCFG " + length + " words (incl. stop) iScore " + iScore[0][length][goal]);
    }
    bestScore = iScore[0][length][goal];
    boolean succeeded = hasParse();
    if (Test.doRecovery && bestScore == Float.NEGATIVE_INFINITY && !floodTags) {
      floodTags = true; // sentence will try to reparse
      System.err.println("Trying recovery parse...");
      return parse(lr);
    }
    if (!op.doDep) {
      return succeeded;
    }
    if (Test.verbose) {
      System.err.print("Starting outsides...");
    }
    // outside scores
    oScore[0][length][goal] = 0.0f;
    doOutsideScores();
    //System.out.println("State rate: "+((int)(1000*ohits/otries))/10.0);
    //System.out.println("Traversals: "+ohits);
    if (Test.verbose) {
      // outsideTime += Timing.tick("Done.");
      Timing.tick("Done.");
      System.out.print("Starting half-filters...");
    }
    for (int loc = 0; loc <= length; loc++) {
      Arrays.fill(iPossibleByL[loc], false);
      Arrays.fill(iPossibleByR[loc], false);
      Arrays.fill(oPossibleByL[loc], false);
      Arrays.fill(oPossibleByR[loc], false);
    }
    for (int start = 0; start < length; start++) {
      for (int end = start + 1; end <= length; end++) {
        for (int state = 0; state < numStates; state++) {
          if (iScore[start][end][state] > Float.NEGATIVE_INFINITY && oScore[start][end][state] > Float.NEGATIVE_INFINITY) {
            iPossibleByL[start][state] = true;
            iPossibleByR[end][state] = true;
            oPossibleByL[start][state] = true;
            oPossibleByR[end][state] = true;
          }
        }
      }
    }
    if (Test.verbose) {
      Timing.tick("done.");
    }
    if (false) {
      long iNZ = 0;
      long oNZ = 0;
      long bNZ = 0;
      long tot = 0;
      for (int start = 0; start < length; start++) {
        for (int end = start + 1; end <= length; end++) {
          for (int s = 0; s < numStates; s++) {
            tot++;
            if (iScore[start][end][s] != Double.NEGATIVE_INFINITY) {
              iNZ++;
            }
            if (oScore[start][end][s] != Double.NEGATIVE_INFINITY) {
              oNZ++;
            }
            if (oScore[start][end][s] != Double.NEGATIVE_INFINITY && iScore[start][end][s] != Double.NEGATIVE_INFINITY) {
              bNZ++;
            }
          }
        }
      }
      System.out.println("Zero Saturation (i): " + ((int) (1000 * iNZ / tot)) / 10.0);
      System.out.println("Zero Saturation (o): " + ((int) (1000 * oNZ / tot)) / 10.0);
      System.out.println("Zero Saturation (b): " + ((int) (1000 * bNZ / tot)) / 10.0);
      Timing.tick("Done with zero scan.");
    }
    return succeeded;
  }

  private void doOutsideScores() {
    for (int diff = length; diff >= 1; diff--) {
      for (int start = 0; start + diff <= length; start++) {
        int end = start + diff;
        // do unaries
        for (int s = 0; s < numStates; s++) {
          float oS = oScore[start][end][s];
          if (oS == Float.NEGATIVE_INFINITY) {
            continue;
          }
          UnaryRule[] rules = ug.closedRulesByParent(s);
          for (int r = 0; r < rules.length; r++) {
            UnaryRule ur = rules[r];
            float pS = (float) ur.score;
            float tot = oS + pS;
            if (tot > oScore[start][end][ur.child] && iScore[start][end][ur.child] > Float.NEGATIVE_INFINITY) {
              oScore[start][end][ur.child] = tot;
            }
          }
        }
        // do binaries
        for (int s = 0; s < numStates; s++) {
          int min1 = narrowRExtent[start][s];
          if (end < min1) {
            continue;
          }
          BinaryRule[] rules = bg.splitRulesWithLC(s);
          for (int r = 0; r < rules.length; r++) {
            BinaryRule br = rules[r];
            float oS = oScore[start][end][br.parent];
            if (oS == Float.NEGATIVE_INFINITY) {
              continue;
            }
            int max1 = narrowLExtent[end][br.rightChild];
            if (max1 < min1) {
              continue;
            }
            int min = min1;
            int max = max1;
            if (max - min > 2) {
              int min2 = wideLExtent[end][br.rightChild];
              min = (min1 > min2 ? min1 : min2);
              if (max1 < min) {
                continue;
              }
              int max2 = wideRExtent[start][br.leftChild];
              max = (max1 < max2 ? max1 : max2);
              if (max < min) {
                continue;
              }
            }
            float pS = (float) br.score;
            for (int split = min; split <= max; split++) {
              float lS = iScore[start][split][br.leftChild];
              if (lS == Float.NEGATIVE_INFINITY) {
                continue;
              }
              float rS = iScore[split][end][br.rightChild];
              if (rS == Float.NEGATIVE_INFINITY) {
                continue;
              }
              float totL = pS + rS + oS;
              if (totL > oScore[start][split][br.leftChild]) {
                oScore[start][split][br.leftChild] = totL;
              }
              float totR = pS + lS + oS;
              if (totR > oScore[split][end][br.rightChild]) {
                oScore[split][end][br.rightChild] = totR;
              }
            }
          }
        }
        for (int s = 0; s < numStates; s++) {
          int max1 = narrowLExtent[end][s];
          if (max1 < start) {
            continue;
          }
          BinaryRule[] rules = bg.splitRulesWithRC(s);
          for (int r = 0; r < rules.length; r++) {
            BinaryRule br = rules[r];
            float oS = oScore[start][end][br.parent];
            if (oS == Float.NEGATIVE_INFINITY) {
              continue;
            }
            int min1 = narrowRExtent[start][br.leftChild];
            if (max1 < min1) {
              continue;
            }
            int min = min1;
            int max = max1;
            if (max - min > 2) {
              int min2 = wideLExtent[end][br.rightChild];
              min = (min1 > min2 ? min1 : min2);
              if (max1 < min) {
                continue;
              }
              int max2 = wideRExtent[start][br.leftChild];
              max = (max1 < max2 ? max1 : max2);
              if (max < min) {
                continue;
              }
            }
            float pS = (float) br.score;
            for (int split = min; split <= max; split++) {
              float lS = iScore[start][split][br.leftChild];
              if (lS == Float.NEGATIVE_INFINITY) {
                continue;
              }
              float rS = iScore[split][end][br.rightChild];
              if (rS == Float.NEGATIVE_INFINITY) {
                continue;
              }
              float totL = pS + rS + oS;
              if (totL > oScore[start][split][br.leftChild]) {
                oScore[start][split][br.leftChild] = totL;
              }
              float totR = pS + lS + oS;
              if (totR > oScore[split][end][br.rightChild]) {
                oScore[split][end][br.rightChild] = totR;
              }
            }
          }
        }
        /*
for (int s = 0; s < numStates; s++) {
float oS = oScore[start][end][s];
          //if (iScore[start][end][s] == Float.NEGATIVE_INFINITY ||
          //             oS == Float.NEGATIVE_INFINITY)
          if (oS == Float.NEGATIVE_INFINITY)
continue;
          BinaryRule[] rules = bg.splitRulesWithParent(s);
for (int r=0; r<rules.length; r++) {
BinaryRule br = rules[r];
            int min1 = narrowRExtent[start][br.leftChild];
            if (end < min1)
              continue;
            int max1 = narrowLExtent[end][br.rightChild];
            if (max1 < min1)
              continue;
            int min2 = wideLExtent[end][br.rightChild];
            int min = (min1 > min2 ? min1 : min2);
            if (max1 < min)
              continue;
            int max2 = wideRExtent[start][br.leftChild];
            int max = (max1 < max2 ? max1 : max2);
            if (max < min)
              continue;
float pS = (float) br.score;
for (int split = min; split <= max; split++) {
float lS = iScore[start][split][br.leftChild];
if (lS == Float.NEGATIVE_INFINITY)
          continue;
float rS = iScore[split][end][br.rightChild];
              if (rS == Float.NEGATIVE_INFINITY)
continue;
float totL = pS+rS+oS;
if (totL > oScore[start][split][br.leftChild]) {
oScore[start][split][br.leftChild] = totL;
}
float totR = pS+lS+oS;
if (totR > oScore[split][end][br.rightChild]) {
oScore[split][end][br.rightChild] = totR;
}
}
}
}
        */
      }
    }
  }

  /** Fills in the iScore array of each category over each span
   *  of length 2 or more.
   */
  void doInsideScores() {

    for (int diff = 2; diff <= length; diff++) {
      // usually stop one short because boundary symbol only combines
      // with whole sentence span
      for (int start = 0; start < ((diff == length) ? 1: length - diff); start++) {

        if (spillGuts) {
          tick("Binaries for span " + diff + "...");
        }
        int end = start + diff;

        if (Test.constraints != null) {
          boolean skip = false;
          for (Test.Constraint c : Test.constraints) {
            if ((start > c.start && start < c.end && end > c.end) || (end > c.start && end < c.end && start < c.start)) {
              skip = true;
              break;
            }
          }
          if (skip) {
            continue;
          }
        }

        for (int leftState = 0; leftState < numStates; leftState++) {

          int narrowR = narrowRExtent[start][leftState];
          boolean iPossibleL = (narrowR < end); // can this left constituent leave space for a right constituent?
          if (!iPossibleL) {
            continue;
          }
          BinaryRule[] leftRules = bg.splitRulesWithLC(leftState);
          //      if (spillGuts) System.out.println("Found " + leftRules.length + " left rules for state " + stateNumberer.object(leftState));
          for (int i = 0; i < leftRules.length; i++) {
            //      if (spillGuts) System.out.println("Considering rule for " + start + " to " + end + ": " + leftRules[i]);
            BinaryRule r = leftRules[i];

            int narrowL = narrowLExtent[end][r.rightChild];
            boolean iPossibleR = (narrowL >= narrowR); // can this right constituent fit next to the left constituent?
            if (!iPossibleR) {
              continue;
            }
            int min1 = narrowR;
            int min2 = wideLExtent[end][r.rightChild];
            int min = (min1 > min2 ? min1 : min2);
            if (min > narrowL) { // can this right constituent stretch far enough to reach the left constituent?
              continue;
            }
            int max1 = wideRExtent[start][leftState];
            int max2 = narrowL;
            int max = (max1 < max2 ? max1 : max2);
            if (min > max) { // can this left constituent stretch far enough to reach the right constituent?
              continue;
            }
            float pS = (float) r.score;
            int parentState = r.parent;
            float oldIScore = iScore[start][end][parentState];
            float bestIScore = oldIScore;
            boolean foundBetter;  // always set below for this rule
            //System.out.println("Min "+min+" max "+max+" start "+start+" end "+end);

            if (!Test.lengthNormalization) {
              // find the split that can use this rule to make the max score
              for (int split = min; split <= max; split++) {

                if (Test.constraints != null) {
                  boolean skip = false;
                  for (Test.Constraint c : Test.constraints) {
                    if (((start < c.start && end >= c.end) || (start <= c.start && end > c.end)) && split > c.start && split < c.end) {
                      skip = true;
                      break;
                    }
                    if ((start == c.start && split == c.end)) {
                      String tag = (String) stateNumberer.object(leftState);
                      Matcher m = c.state.matcher(tag);
                      if (!m.matches()) {
                        skip = true;
                        break;
                      }
                    }
                    if ((split == c.start && end == c.end)) {
                      String tag = (String) stateNumberer.object(r.rightChild);
                      Matcher m = c.state.matcher(tag);
                      if (!m.matches()) {
                        skip = true;
                        break;
                      }
                    }
                  }
                  if (skip) {
                    continue;
                  }
                }

                float lS = iScore[start][split][leftState];
                if (lS == Float.NEGATIVE_INFINITY) {
                  continue;
                }
                float rS = iScore[split][end][r.rightChild];
                if (rS == Float.NEGATIVE_INFINITY) {
                  continue;
                }
                float tot = pS + lS + rS;
                if (tot > bestIScore) {
                  bestIScore = tot;
                }
              } // for split point
              foundBetter = bestIScore > oldIScore;
            } else {
              // find split that uses this rule to make the max *length normalized* score
              int bestWordsInSpan = wordsInSpan[start][end][parentState];
              float oldNormIScore = oldIScore / bestWordsInSpan;
              float bestNormIScore = oldNormIScore;

              for (int split = min; split <= max; split++) {
                float lS = iScore[start][split][leftState];
                if (lS == Float.NEGATIVE_INFINITY) {

                  continue;
                }
                float rS = iScore[split][end][r.rightChild];
                if (rS == Float.NEGATIVE_INFINITY) {
                  continue;
                }
                float tot = pS + lS + rS;
                int newWordsInSpan = wordsInSpan[start][split][leftState] + wordsInSpan[split][end][r.rightChild];
                float normTot = tot / newWordsInSpan;
                if (normTot > bestNormIScore) {
                  bestIScore = tot;
                  bestNormIScore = normTot;
                  bestWordsInSpan = newWordsInSpan;
                }
              } // for split point
              foundBetter = bestNormIScore > oldNormIScore;
              if (foundBetter) {
                wordsInSpan[start][end][parentState] = bestWordsInSpan;
              }
            } // fi Test.lengthNormalization
            if (foundBetter) { // this way of making "parentState" is better than previous
              iScore[start][end][parentState] = bestIScore;

              //              if (spillGuts) System.out.println("Could build " + stateNumberer.object(parentState) + " from " + start + " to " + end);
              if (oldIScore == Float.NEGATIVE_INFINITY) {
                if (start > narrowLExtent[end][parentState]) {
                  narrowLExtent[end][parentState] = start;
                  wideLExtent[end][parentState] = start;
                } else {
                  if (start < wideLExtent[end][parentState]) {
                    wideLExtent[end][parentState] = start;
                  }
                }
                if (end < narrowRExtent[start][parentState]) {
                  narrowRExtent[start][parentState] = end;
                  wideRExtent[start][parentState] = end;
                } else {
                  if (end > wideRExtent[start][parentState]) {
                    wideRExtent[start][parentState] = end;
                  }
                }
              }
            } // end if foundBetter
          } // end for leftRules
        } // end for leftState
        // do right restricted rules
        for (int rightState = 0; rightState < numStates; rightState++) {
          int narrowL = narrowLExtent[end][rightState];
          boolean iPossibleR = (narrowL > start);
          if (!iPossibleR) {
            continue;
          }
          BinaryRule[] rightRules = bg.splitRulesWithRC(rightState);
          //      if (spillGuts) System.out.println("Found " + rightRules.length + " right rules for state " + stateNumberer.object(rightState));
          for (int i = 0; i < rightRules.length; i++) {
            //      if (spillGuts) System.out.println("Considering rule for " + start + " to " + end + ": " + rightRules[i]);
            BinaryRule r = rightRules[i];

            int narrowR = narrowRExtent[start][r.leftChild];
            boolean iPossibleL = (narrowR <= narrowL);
            if (!iPossibleL) {
              continue;
            }
            int min1 = narrowR;
            int min2 = wideLExtent[end][rightState];
            int min = (min1 > min2 ? min1 : min2);
            if (min > narrowL) {
              continue;
            }
            int max1 = wideRExtent[start][r.leftChild];
            int max2 = narrowL;
            int max = (max1 < max2 ? max1 : max2);
            if (min > max) {
              continue;
            }
            float pS = (float) r.score;
            int parentState = r.parent;
            float oldIScore = iScore[start][end][parentState];
            float bestIScore = oldIScore;
            boolean foundBetter; // always initialized below
            //System.out.println("Start "+start+" end "+end+" min "+min+" max "+max);
            if (!Test.lengthNormalization) {
              // find the split that can use this rule to make the max score
              for (int split = min; split <= max; split++) {

                if (Test.constraints != null) {
                  boolean skip = false;
                  for (Test.Constraint c : Test.constraints) {
                    if (((start < c.start && end >= c.end) || (start <= c.start && end > c.end)) && split > c.start && split < c.end) {
                      skip = true;
                      break;
                    }
                    if ((start == c.start && split == c.end)) {
                      String tag = (String) stateNumberer.object(r.leftChild);
                      Matcher m = c.state.matcher(tag);
                      if (!m.matches()) {
                        //if (!tag.startsWith(c.state+"^")) {
                        skip = true;
                        break;
                      }
                    }
                    if ((split == c.start && end == c.end)) {
                      String tag = (String) stateNumberer.object(rightState);
                      Matcher m = c.state.matcher(tag);
                      if (!m.matches()) {
                        //if (!tag.startsWith(c.state+"^")) {
                        skip = true;
                        break;
                      }
                    }
                  }
                  if (skip) {
                    continue;
                  }
                }

                float lS = iScore[start][split][r.leftChild];
                if (lS == Float.NEGATIVE_INFINITY) {
                  continue;
                }
                float rS = iScore[split][end][rightState];
                if (rS == Float.NEGATIVE_INFINITY) {
                  continue;
                }
                float tot = pS + lS + rS;
                if (tot > bestIScore) {
                  bestIScore = tot;
                }
              } // end for split
              foundBetter = bestIScore > oldIScore;
            } else {
              // find split that uses this rule to make the max *length normalized* score
              int bestWordsInSpan = wordsInSpan[start][end][parentState];
              float oldNormIScore = oldIScore / bestWordsInSpan;
              float bestNormIScore = oldNormIScore;
              for (int split = min; split <= max; split++) {
                float lS = iScore[start][split][r.leftChild];
                if (lS == Float.NEGATIVE_INFINITY) {
                  continue;
                }
                float rS = iScore[split][end][rightState];
                if (rS == Float.NEGATIVE_INFINITY) {
                  continue;
                }
                float tot = pS + lS + rS;
                int newWordsInSpan = wordsInSpan[start][split][r.leftChild] + wordsInSpan[split][end][rightState];
                float normTot = tot / newWordsInSpan;
                if (normTot > bestNormIScore) {
                  bestIScore = tot;
                  bestNormIScore = normTot;
                  bestWordsInSpan = newWordsInSpan;
                }
              } // end for split
              foundBetter = bestNormIScore > oldNormIScore;
              if (foundBetter) {
                wordsInSpan[start][end][parentState] = bestWordsInSpan;
              }
            } // end if lengthNormalization
            if (foundBetter) { // this way of making "parentState" is better than previous
              iScore[start][end][parentState] = bestIScore;
              //              if (spillGuts) System.out.println("Could build " + stateNumberer.object(parentState) + " from " + start + " to " + end);
              if (oldIScore == Float.NEGATIVE_INFINITY) {
                if (start > narrowLExtent[end][parentState]) {
                  narrowLExtent[end][parentState] = start;
                  wideLExtent[end][parentState] = start;
                } else {
                  if (start < wideLExtent[end][parentState]) {
                    wideLExtent[end][parentState] = start;
                  }
                }
                if (end < narrowRExtent[start][parentState]) {
                  narrowRExtent[start][parentState] = end;
                  wideRExtent[start][parentState] = end;
                } else {
                  if (end > wideRExtent[start][parentState]) {
                    wideRExtent[start][parentState] = end;
                  }
                }
              }
            } // end if foundBetter
          } // for rightRules
        } // for rightState
        if (spillGuts) {
          tick("Unaries for span " + diff + "...");
        }
        // do unary rules -- one could promote this loop and put start inside
        for (int state = 0; state < numStates; state++) {
          float iS = iScore[start][end][state];
          if (iS == Float.NEGATIVE_INFINITY) {
            continue;
          }
          UnaryRule[] unaries = ug.closedRulesByChild(state);
          for (int r = 0; r < unaries.length; r++) {

            UnaryRule ur = unaries[r];

            if (Test.constraints != null) {
              boolean skip = false;
              for (Test.Constraint c : Test.constraints) {
                if ((start == c.start && end == c.end)) {
                  String tag = (String) stateNumberer.object(ur.parent);
                  Matcher m = c.state.matcher(tag);
                  if (!m.matches()) {
                    //if (!tag.startsWith(c.state+"^")) {
                    skip = true;
                    break;
                  }
                }
              }
              if (skip) {
                continue;
              }
            }

            int parentState = ur.parent;
            float pS = (float) ur.score;
            float tot = iS + pS;
            float cur = iScore[start][end][parentState];
            boolean foundBetter;  // always set below
            if (Test.lengthNormalization) {
              int totWordsInSpan = wordsInSpan[start][end][state];
              float normTot = tot / totWordsInSpan;
              int curWordsInSpan = wordsInSpan[start][end][parentState];
              float normCur = cur / curWordsInSpan;
              foundBetter = normTot > normCur;
              if (foundBetter) {
                wordsInSpan[start][end][parentState] = wordsInSpan[start][end][state];
              }
            } else {
              foundBetter = (tot > cur);
            }
            if (foundBetter) {
              //              if (spillGuts) System.out.println("Could build " + stateNumberer.object(parentState) + " from " + start + " to " + end);
              iScore[start][end][parentState] = tot;
              if (cur == Float.NEGATIVE_INFINITY) {
                if (start > narrowLExtent[end][parentState]) {
                  narrowLExtent[end][parentState] = start;
                  wideLExtent[end][parentState] = start;
                } else {
                  if (start < wideLExtent[end][parentState]) {
                    wideLExtent[end][parentState] = start;
                  }
                }
                if (end < narrowRExtent[start][parentState]) {
                  narrowRExtent[start][parentState] = end;
                  wideRExtent[start][parentState] = end;
                } else {
                  if (end > wideRExtent[start][parentState]) {
                    wideRExtent[start][parentState] = end;
                  }
                }
              }
            } // end if foundBetter
          } // for UnaryRule r
        } // for unary rules
      } // for start
    } // for diff (i.e., span)
  } // end doInsideScores()


  private void initializeChart(LatticeReader lr) {
    List latticeWords = lr.getLatticeWords();
    for (Iterator wordIter = latticeWords.iterator(); wordIter.hasNext();) {
      LatticeReader.LatticeWord lWord = (LatticeReader.LatticeWord) wordIter.next();
      int start = lWord.startNode, end = lWord.endNode;
      String word = lWord.word;
      for (int state = 0; state < numStates; state++) {
        if (isTag[state]) {
          IntTaggedWord itw = new IntTaggedWord(wordNumberer.number(word), Numberer.translate(stateSpace, "tags", state));
          // ignoring LatticeReader's language model score
          //        float lmS = (float) lWord.lm;
          // the position passed into lex shouldn't really be start here, but
          // lets leave it for now
          float newScore = (float) lex.score(itw, start) + (float) lWord.am;
          if (newScore > iScore[start][end][state]) {
            iScore[start][end][state] = newScore;
            narrowRExtent[start][state] = start + 1;
            narrowLExtent[end][state] = end - 1;
            wideRExtent[start][state] = start + 1;
            wideLExtent[end][state] = end - 1;
          }
        }
      }
      // if parse failed because of tag coverage
      if (floodTags && (!Test.noRecoveryTagging)) { // sends to recovery parse with more flexible tagging
        for (int state = 0; state < numStates; state++) {
          float iS = iScore[start][end][state];
          if (iS == Float.NEGATIVE_INFINITY && isTag[state]) {
            iScore[start][end][state] = -1000.0f;
            narrowRExtent[start][state] = end;
            narrowLExtent[end][state] = start;
            wideRExtent[start][state] = end;
            wideLExtent[end][state] = start;
          }
        }
      }
      if (spillGuts) {
        tick("Terminal Unary...");
      }
      for (int state = 0; state < numStates; state++) {
        float iS = iScore[start][end][state];
        if (iS == Float.NEGATIVE_INFINITY) {
          continue;
        }
        UnaryRule[] unaries = ug.closedRulesByChild(state);
        for (int r = 0; r < unaries.length; r++) {
          UnaryRule ur = unaries[r];
          int parentState = ur.parent;
          float pS = (float) ur.score;
          float tot = iS + pS;
          if (tot > iScore[start][end][parentState]) {
            iScore[start][end][parentState] = tot;
            narrowRExtent[start][parentState] = end;
            narrowLExtent[end][parentState] = start;
            wideRExtent[start][parentState] = end;
            wideLExtent[end][parentState] = start;
          }
        }
      }
    }
  }

  private void initializeChart(List sentence) {
    for (int start = 0; start + 1 <= length; start++) {
      if (Test.maxSpanForTags > 1) {
        // note we don't look for "words" including the end symbol!
        for (int end = start + 1; (end < length - 1 && end - start <= Test.maxSpanForTags) || (start + 1 == end); end++) {
          StringBuffer word = new StringBuffer();
          // this is ugly and should be fixed...
          for (int i = start; i < end; i++) {
            if (sentence.get(i) instanceof StringLabel) {
              word.append(((StringLabel) sentence.get(i)).value());
            } else {
              word.append((String) sentence.get(i));
            }
          }
          for (int state = 0; state < numStates; state++) {
            float iS = iScore[start][end][state];
            if (iS == Float.NEGATIVE_INFINITY && isTag[state]) {
              IntTaggedWord itw = new IntTaggedWord(wordNumberer.number(word.toString()), Numberer.translate(stateSpace, "tags", state));
              iScore[start][end][state] = (float) lex.score(itw, start);
              if (iScore[start][end][state] > Float.NEGATIVE_INFINITY) {
                narrowRExtent[start][state] = start + 1;
                narrowLExtent[end][state] = end - 1;
                wideRExtent[start][state] = start + 1;
                wideLExtent[end][state] = end - 1;
              }
            }
          }
        }
      } else {
        int word = words[start];
        int end = start + 1;
        Arrays.fill(tags[start], false);
        String trueTagStr = null;
        if (sentence.get(start) instanceof HasTag) {
          trueTagStr = ((HasTag) sentence.get(start)).tag();
        }
        // setting the below line to false generates ALL tags in lexicon for
        // every word (though some may still be scored at -Inf).  Beware! XXXX
        if (!op.flexiTag) {
          if (dumpTagging) {
            System.out.println("Normal tagging " + Numberer.getGlobalNumberer("words").object(word));
          }
          for (Iterator taggingI = lex.ruleIteratorByWord(word, start); taggingI.hasNext();) {
            IntTaggedWord tagging = (IntTaggedWord) taggingI.next();
            int state = Numberer.translate("tags", stateSpace, tagging.tag);
            if (trueTagStr != null) {
              if (!tlp.basicCategory(tagging.tagString()).equals(trueTagStr)) {
                continue;
              }
            }
            // try {
            iScore[start][end][state] = (float) lex.score(tagging, start);
            if (iScore[start][end][state] > Float.NEGATIVE_INFINITY) {
              narrowRExtent[start][state] = end;
              narrowLExtent[end][state] = start;
              wideRExtent[start][state] = end;
              wideLExtent[end][state] = start;
            }
            // } catch (Exception e) {
            // e.printStackTrace();
            // System.out.println("State: " + state + " tags " + Numberer.getGlobalNumberer("tags").object(tagging.tag));
            // }
            int tag = tagging.tag;
            tags[start][tag] = true;
            if (dumpTagging) {
              System.out.println("Word " + start + " tagging " + tagging + " score " + iScore[start][start + 1][state] + " -- " + state + " = " + Numberer.object("states", state));
            }
            //if (start == length-2 && tagging.parent == puncTag)
            //  lastIsPunc = true;
          }
        } else {
          if (dumpTagging) {
            System.out.println("Flexitagging " + Numberer.translate(stateSpace, "words", word));
          }
          for (int state = 0; state < numStates; state++) {
            if (trueTagStr != null) {
              String tagString = (String) stateNumberer.object(state);
              if (!tlp.basicCategory(tagString).equals(trueTagStr)) {
                continue;
              }
            }
            float iS = iScore[start][end][state];
            if (iS == Float.NEGATIVE_INFINITY && isTag[state]) {
              iScore[start][end][state] = (float) lex.score(new IntTaggedWord(word, Numberer.translate(stateSpace, "tags", state)), start);
              if (iScore[start][end][state] > Float.NEGATIVE_INFINITY) {
                narrowRExtent[start][state] = end;
                narrowLExtent[end][state] = start;
                wideRExtent[start][state] = end;
                wideLExtent[end][state] = start;
              }
              if (dumpTagging) {
                System.out.println("Word " + start + " tagging " + (new IntTaggedWord(word, Numberer.translate(stateSpace, "tags", state))) + " score " + iScore[start][start + 1][state] + " -- " + state + " = " + Numberer.object("states", state));
              }
            }
          }
        }
        // tag multi-counting
        if (op.dcTags) {
          for (int state = 0; state < numStates; state++) {
            if (isTag[state]) {
              iScore[start][end][state] *= (1.0 + Test.depWeight);
            }
          }
        }
        // if parse failed because of tag coverage
        if (floodTags && (!Test.noRecoveryTagging)) { // sends to recovery parse with more flexible tagging
          for (int state = 0; state < numStates; state++) {
            float iS = iScore[start][end][state];
            if (iS == Float.NEGATIVE_INFINITY && isTag[state]) {
              iScore[start][end][state] = -1000.0f;
              narrowRExtent[start][state] = end;
              narrowLExtent[end][state] = start;
              wideRExtent[start][state] = end;
              wideLExtent[end][state] = start;
            }
          }
        }
        if (spillGuts) {
          tick("Terminal Unary...");
        }
        for (int state = 0; state < numStates; state++) {
          float iS = iScore[start][end][state];
          if (iS == Float.NEGATIVE_INFINITY) {
            continue;
          }
          UnaryRule[] unaries = ug.closedRulesByChild(state);
          for (int r = 0; r < unaries.length; r++) {
            UnaryRule ur = unaries[r];
            int parentState = ur.parent;
            float pS = (float) ur.score;
            float tot = iS + pS;
            if (tot > iScore[start][end][parentState]) {
              iScore[start][end][parentState] = tot;
              narrowRExtent[start][parentState] = end;
              narrowLExtent[end][parentState] = start;
              wideRExtent[start][parentState] = end;
              wideLExtent[end][parentState] = start;
            }
          }
        }
        if (spillGuts) {
          tick("Next word...");
        }
      }
    }
  }

  public boolean hasParse() {
    if (length > arraySize) {
      return false;
    }
    int goal = stateNumberer.number(goalStr);
    return (iScore[0][length][goal] > Float.NEGATIVE_INFINITY);
  }

  public double getLogScore() {
    int goal = stateNumberer.number(goalStr);
    return iScore[0][length][goal];
  }


  private static final double TOL = 1e-5;

  protected static boolean matches(double x, double y) {
    return (Math.abs(x - y) / (Math.abs(x) + Math.abs(y) + 1e-10) < TOL);
  }

  protected Tree extractBestParse(int goal, int start, int end) {
    // note that the returned tree may be missing intermediate nodes in
    // a unary chain because it parses with a unary-closed grammar.

    // find source of inside score
    // no backtraces so we can speed up the parsing for its primary use
    double bestScore = iScore[start][end][goal];
    double normBestScore = Test.lengthNormalization ? (bestScore / wordsInSpan[start][end][goal]) : bestScore;
    String goalStr = (String) stateNumberer.object(goal);
    //System.out.println("Searching for "+goalStr+" from "+start+" to "+end+" scored "+bestScore);
    // check tags
    if (end - start <= Test.maxSpanForTags && tagNumberer.hasSeen(goalStr)) {
      if (Test.maxSpanForTags > 1) {
        Tree wordNode = null;
        if (sentence != null) {
          StringBuffer word = new StringBuffer();
          for (int i = start; i < end; i++) {
            if (sentence.get(i) instanceof StringLabel) {
              word.append(((StringLabel) sentence.get(i)).value());
            } else {
              word.append((String) sentence.get(i));
            }
          }
          wordNode = tf.newLeaf(new StringLabel(word.toString()));
        } else if (lr != null) {
          List<LatticeReader.LatticeWord> latticeWords = lr.getWordsOverSpan(start, end);
          for (LatticeReader.LatticeWord lWord : latticeWords) {
            IntTaggedWord itw = new IntTaggedWord(wordNumberer.number(lWord.word), Numberer.translate(stateSpace, "tags", goal));
            double tagScore = lex.score(itw, start);
            if (matches(bestScore, tagScore + (float) lWord.am)) {
              wordNode = tf.newLeaf(new StringLabel(lWord.word));
              break;
            }
          }
          if (wordNode == null) {
            throw new RuntimeException("could not find matching word from lattice in parse reconstruction");
          }
        } else {
          throw new RuntimeException("attempt to get word when sentence and lattice are null!");
        }
        Tree tagNode = tf.newTreeNode(new StringLabel(goalStr), Collections.singletonList(wordNode));
        tagNode.setScore(bestScore);
        return tagNode;
      } else {
        IntTaggedWord tagging = new IntTaggedWord(words[start], tagNumberer.number(goalStr));
        double tagScore = lex.score(tagging, start);
        if (tagScore > Double.NEGATIVE_INFINITY || floodTags) {
          // return a pre-terminal tree
          String wordStr = (String) wordNumberer.object(words[start]);
          Tree wordNode = tf.newLeaf(new StringLabel(wordStr));
          List childList = new ArrayList();
          childList.add(wordNode);
          Tree tagNode = tf.newTreeNode(new StringLabel(goalStr), childList);
          tagNode.setScore(bestScore);
          //System.out.println("Tag node: "+tagNode);
          return tagNode;
        }
      }
    }
    // check binaries first
    for (int split = start + 1; split < end; split++) {
      for (Iterator binaryI = bg.ruleIteratorByParent(goal); binaryI.hasNext();) {
        BinaryRule br = (BinaryRule) binaryI.next();
        double score = br.score + iScore[start][split][br.leftChild] + iScore[split][end][br.rightChild];
        boolean matches;
        if (Test.lengthNormalization) {
          double normScore = score / (wordsInSpan[start][split][br.leftChild] + wordsInSpan[split][end][br.rightChild]);
          matches = matches(normScore, normBestScore);
        } else {
          matches = matches(score, bestScore);
        }
        if (matches) {
          // build binary split
          Tree leftChildTree = extractBestParse(br.leftChild, start, split);
          Tree rightChildTree = extractBestParse(br.rightChild, split, end);
          List children = new ArrayList();
          children.add(leftChildTree);
          children.add(rightChildTree);
          Tree result = tf.newTreeNode(new StringLabel(goalStr), children);
          result.setScore(score);
          //System.out.println("Binary node: "+result);
          return result;
        }
      }
    }
    // check unaries
    for (Iterator unaryI = ug.ruleIteratorByParent(goal); unaryI.hasNext();) {
      UnaryRule ur = (UnaryRule) unaryI.next();
      double score = ur.score + iScore[start][end][ur.child];
      boolean matches;
      if (Test.lengthNormalization) {
        double normScore = score / wordsInSpan[start][end][ur.child];
        matches = matches(normScore, normBestScore);
      } else {
        matches = matches(score, bestScore);
      }
      if (ur.child != ur.parent && matches) {
        // build unary
        Tree childTree = extractBestParse(ur.child, start, end);
        List children = new ArrayList();
        children.add(childTree);
        Tree result = tf.newTreeNode(new StringLabel(goalStr), children);
        //System.out.println("Unary node: "+result);
        result.setScore(score);
        return result;
      }
    }
    System.err.println("Warning: no parse found in ExhaustivePCFGParser.extractBestParse: failing on: [" + start + ", " + end + "] looking for " + goalStr);
    return null;
  }

  protected void restoreUnaries(Tree t) {
    //System.out.println("In restoreUnaries...");
    for (Iterator nodeI = t.subTreeList().iterator(); nodeI.hasNext();) {
      Tree node = (Tree) nodeI.next();
      //System.out.println("Doing node: "+node.label());
      if (node.isLeaf() || node.isPreTerminal() || node.children().length != 1) {
        //System.out.println("Skipping node: "+node.label());
        continue;
      }
      //System.out.println("Not skipping node: "+node.label());
      Tree parent = node;
      Tree child = node.children()[0];
      List path = ug.getBestPath(stateNumberer.number(parent.label().value()), stateNumberer.number(child.label().value().toString()));
      //System.out.println("Got path: "+path);
      int pos = 1;
      while (pos < path.size() - 1) {
        int interState = ((Integer) path.get(pos)).intValue();
        Tree intermediate = tf.newTreeNode(new StringLabel((String) stateNumberer.object(interState)), parent.getChildrenAsList());
        List children = new ArrayList();
        children.add(intermediate);
        parent.setChildren(children);
        pos++;
      }
      //System.out.println("Done with node: "+node.label());
    }
  }

  public Tree getBestParse() {//String goalStr, int start, int end) {
    int start = 0;
    int end = words.length;
    int goal = stateNumberer.number(goalStr);
    Tree internalTree = extractBestParse(goal, start, end);
    //System.out.println("Got internal best parse...");
    if (internalTree == null) {
      System.err.println("Warning: no parse found in ExhaustivePCFGParser.extractBestParse");
    } else {
      restoreUnaries(internalTree);
    }
    //System.out.println("Restored unaries...");
    return internalTree;
    //TreeTransformer debinarizer = BinarizerFactory.getDebinarizer();
    //return debinarizer.transformTree(internalTree);
  }


  /**
   * Return all best parses (except no ties allowed on POS tags?).
   * Note that the returned tree may be missing intermediate nodes in
   * a unary chain because it parses with a unary-closed grammar.
   */
  protected List extractBestParses(int goal, int start, int end) {
    // find sources of inside score
    // no backtraces so we can speed up the parsing for its primary use
    double bestScore = iScore[start][end][goal];
    String goalStr = (String) stateNumberer.object(goal);
    //System.out.println("Searching for "+goalStr+" from "+start+" to "+end+" scored "+bestScore);
    // check tags
    if (end - start == 1 && tagNumberer.hasSeen(goalStr)) {
      IntTaggedWord tagging = new IntTaggedWord(words[start], tagNumberer.number(goalStr));
      double tagScore = lex.score(tagging, start);
      if (tagScore > Double.NEGATIVE_INFINITY || floodTags) {
        // return a pre-terminal tree
        String wordStr = (String) wordNumberer.object(words[start]);
        Tree wordNode = tf.newLeaf(new StringLabel(wordStr));
        List childList = new ArrayList();
        childList.add(wordNode);
        Tree tagNode = tf.newTreeNode(new StringLabel(goalStr), childList);
        //System.out.println("Tag node: "+tagNode);
        return Collections.singletonList(tagNode);
      }
    }
    // check binaries first
    List bestTrees = new ArrayList();
    for (int split = start + 1; split < end; split++) {
      for (Iterator binaryI = bg.ruleIteratorByParent(goal); binaryI.hasNext();) {
        BinaryRule br = (BinaryRule) binaryI.next();
        double score = br.score + iScore[start][split][br.leftChild] + iScore[split][end][br.rightChild];
        if (matches(score, bestScore)) {
          // build binary split
          List leftChildTrees = extractBestParses(br.leftChild, start, split);
          List rightChildTrees = extractBestParses(br.rightChild, split, end);
          // System.out.println("Found a best way to build " + goalStr + "(" +
          //                 start + "," + end + ") with " +
          //                 leftChildTrees.size() + "x" +
          //                 rightChildTrees.size() + " ways to build.");
          for (Iterator it1 = leftChildTrees.iterator(); it1.hasNext();) {
            Tree leftChildTree = (Tree) it1.next();
            for (Iterator it2 = rightChildTrees.iterator(); it2.hasNext();) {
              Tree rightChildTree = (Tree) it2.next();
              List children = new ArrayList();
              children.add(leftChildTree);
              children.add(rightChildTree);
              Tree result = tf.newTreeNode(new StringLabel(goalStr), children);
              //System.out.println("Binary node: "+result);
              bestTrees.add(result);
            }
          }
        }
      }
    }
    // check unaries
    for (Iterator unaryI = ug.ruleIteratorByParent(goal); unaryI.hasNext();) {
      UnaryRule ur = (UnaryRule) unaryI.next();
      double score = ur.score + iScore[start][end][ur.child];
      if (ur.child != ur.parent && matches(score, bestScore)) {
        // build unary
        List childTrees = extractBestParses(ur.child, start, end);
        for (Iterator it = childTrees.iterator(); it.hasNext();) {
          Tree childTree = (Tree) it.next();
          List children = new ArrayList();
          children.add(childTree);
          Tree result = tf.newTreeNode(new StringLabel(goalStr), children);
          //System.out.println("Unary node: "+result);
          bestTrees.add(result);
        }
      }
    }
    if (bestTrees.size() == 0) {
      System.err.println("Warning: no parse found in ExhaustivePCFGParser.extractBestParse: failing on: [" + start + ", " + end + "] looking for " + goalStr);
    }
    return bestTrees;
  }

  /**
   * Return the set of trees that are tied best parses for a sentences.
   */
  public List getBestParses() {
    int start = 0;
    int end = words.length;
    int goal = stateNumberer.number(goalStr);
    List internalTrees = extractBestParses(goal, start, end);
    //System.out.println("Got internal best parse...");
    for (int i = 0; i < internalTrees.size(); i++) {
      Tree internalTree = (Tree) internalTrees.get(i);
      restoreUnaries(internalTree);
    }
    //System.out.println("Restored unaries...");
    return internalTrees;
    //TreeTransformer debinarizer = BinarizerFactory.getDebinarizer();
    //return debinarizer.transformTree(internalTree);
  }


  public ExhaustivePCFGParser(BinaryGrammar bg, UnaryGrammar ug, Lexicon lex, Options op) {
    //    System.out.println("ExhaustivePCFGParser constructor called.");
    this.op = op;
    this.tlp = op.langpack();
    goalStr = tlp.startSymbol();
    this.stateSpace = bg.stateSpace();
    stateNumberer = Numberer.getGlobalNumberer(stateSpace);
    this.bg = bg;
    this.ug = ug;
    this.lex = lex;
    tf = new LabeledScoredTreeFactory(new StringLabelFactory());

    numStates = stateNumberer.total();
    isTag = new boolean[numStates];
    for (int state = 0; state < numStates; state++) {
      isTag[state] = tagNumberer.hasSeen(stateNumberer.object(state));
    }
  }

  private void createArrays(int length) {

    // zero out some stuff first in case we recently ran out of memory and are reallocating
    clearArrays();

    int numTags = tagNumberer.total();
    // allocate just the parts of iScore and oScore used (end > start, etc.)
    //    System.out.println("initializing iScore arrays with length " + length + " and numStates " + numStates);
    iScore = new float[length + 1][length + 1][];
    for (int start = 0; start <= length; start++) {
      for (int end = start + 1; end <= length; end++) {
        iScore[start][end] = new float[numStates];
      }
    }
    //    System.out.println("finished initializing iScore arrays");
    if (op.doDep) {
      //      System.out.println("initializing oScore arrays with length " + length + " and numStates " + numStates);
      oScore = new float[length + 1][length + 1][];
      for (int start = 0; start <= length; start++) {
        for (int end = start + 1; end <= length; end++) {
          oScore[start][end] = new float[numStates];
        }
      }
      //      System.out.println("finished initializing oScore arrays");
    }
    iPossibleByL = new boolean[length + 1][numStates];
    iPossibleByR = new boolean[length + 1][numStates];
    narrowRExtent = new int[length + 1][numStates];
    wideRExtent = new int[length + 1][numStates];
    narrowLExtent = new int[length + 1][numStates];
    wideLExtent = new int[length + 1][numStates];
    if (op.doDep) {
      oPossibleByL = new boolean[length + 1][numStates];
      oPossibleByR = new boolean[length + 1][numStates];

      oFilteredStart = new boolean[length + 1][numStates];
      oFilteredEnd = new boolean[length + 1][numStates];
    }
    tags = new boolean[length + 1][numTags];

    if (Test.lengthNormalization) {
      wordsInSpan = new int[length + 1][length + 1][];
      for (int start = 0; start <= length; start++) {
        for (int end = start + 1; end <= length; end++) {
          wordsInSpan[start][end] = new int[numStates];
        }
      }
    }
    //    System.out.println("ExhaustivePCFGParser constructor finished.");
  }

  private void clearArrays() {
    iScore = oScore = null;
    iPossibleByL = iPossibleByR = oFilteredEnd = oFilteredStart = oPossibleByL = oPossibleByR = tags = null;
    narrowRExtent = wideRExtent = narrowLExtent = wideLExtent = null;
  }

} // end class ExhaustivePCFGParser

