// StanfordLexicalizedParser -- a probabilistic lexicalized NL CFG parser
// Copyright (c) 2002, 2003, 2004, 2005 The Board of Trustees of 
// The Leland Stanford Junior University. All Rights Reserved.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
// For more information, bug reports, fixes, contact:
//    Christopher Manning
//    Dept of Computer Science, Gates 4A
//    Stanford CA 94305-9040
//    USA
//    parser-support@lists.stanford.edu
//    http://nlp.stanford.edu/downloads/lex-parser.shtml

package edu.stanford.nlp.parser.lexparser;

import edu.stanford.nlp.io.NumberRangeFileFilter;
import edu.stanford.nlp.ling.*;
import edu.stanford.nlp.math.SloppyMath;
import edu.stanford.nlp.parser.ViterbiParser;
import edu.stanford.nlp.process.Function;
import edu.stanford.nlp.trees.*;
import edu.stanford.nlp.util.*;
import edu.stanford.nlp.stats.Counter;

import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.Serializable;
import java.util.*;
import java.util.regex.Matcher;

/**
 * @author Dan Klein (original version)
 * @author Christopher Manning (better features, ParserParams, serialization)
 * @author Roger Levy (internationalization)
 * @author Teg Grenager (grammar compaction, etc., tokenization, etc.)
 * @author Galen Andrew (lattice parsing)
 */


class Train {
  /* THESE OPTIONS AFFECT ONLY TRAIN TIME */

  /**
   * if true, leave all PTB (functional tag) annotations (bad)
   */
  public static boolean leaveItAll = false;
  public static boolean cheatPCFG = false;

  public static boolean markovFactor = false;
  public static int markovOrder = 1;
  public static boolean hSelSplit = false; // good with true;
  public static int HSEL_CUT = 10;

  public static boolean markFinalStates = true; // whether or not to mark final states in binarized grammar
  // must be off to get most value out of grammar compaction

  /**
   * A POS tag has to have been attributed to more than this number of word
   * types before it is regarded as an open-class tag.  Unknown words will
   * only possibly be tagged as open-class tags (unless flexiTag is on).
   */
  public static int openClassTypesThreshold = 50;

  /**
   * Start to aggregate signature-tag pairs only for words unseen in the first
   * this fraction of the data.
   */
  public static double fractionBeforeUnseenCounting = 0.5;

  /**
   * If true, declare early -- leave this on except maybe with markov on.
   */
  public static boolean outsideFactor() {
    return !markovFactor;
  }

  /**
   * This variable controls doing parent annotation of phrasal nodes.  Good.
   */
  public static boolean PA = true;
  /**
   * This variable controls doing 2 levels of parent annotation.  Bad.
   */
  public static boolean gPA = false;

  public static boolean postPA = false;
  public static boolean postGPA = false;

  /**
   * Only split the "common high KL divergence" parent categories.... Good.
   */
  public static boolean selectiveSplit = false; //true;

  public static double selectiveSplitCutOff = 0.0;

  public static boolean selectivePostSplit = false;

  public static double selectivePostSplitCutOff = 0.0;

  /** Whether, in post-splitting of categories, nodes are annotated with the 
   *  (grand)parent's base category or with its complete subcategorized
   *  category. 
   */
  public static boolean postSplitWithBaseCategory = false;

  /**
   * Selective Sister annotation.
   */
  public static boolean sisterAnnotate = false;

  public static Set sisterSplitters;

  /**
   * Mark all unary nodes specially.  Good for just PCFG. Bad for factored.
   * (1 better than 2 in combos)
   */
  public static boolean markUnary = false;//true;
  public static boolean markUnary2 = false;
  public static boolean markUnaryTags = false;


  /**
   * Mark all pre-preterminals (also does splitBaseNP: don't need both)
   */
  public static boolean splitPrePreT = false;


  /**
   * Parent annotation on tags.  Good (for PCFG?)
   */
  public static boolean tagPA = false;//true;

  /**
   * Do parent annotation on tags selectively.  Neutral, but less splits.
   */
  public static boolean tagSelectiveSplit = false;

  public static double tagSelectiveSplitCutOff = 0.0;

  public static boolean tagSelectivePostSplit = false;

  public static double tagSelectivePostSplitCutOff = 0.0;

  /**
   * Right edge is right-recursive (X << X) Bad. (NP only is good)
   */
  public static boolean rightRec = false;//true;

  /**
   * Left edge is right-recursive (X << X)  Bad.
   */
  public static boolean leftRec = false;

  /**
   * X over X is marked (subsumes baseNP marking) Bad.
   */
  public static boolean xOverX = false;

  /**
   * Promote/delete punctuation like Collins.  Bad (!)
   */
  public static boolean collinsPunc = false;

  /**
   * Set the splitter strings.  These are a set of parent and/or grandparent
   * annotated categories which should be split off.
   */
  public static Set splitters;

  public static Set postSplitters;

  public static Set<String> deleteSplitters;

  /**
   * Just for debugging: check that your tree transforms work right or states
   */
  public static boolean printTreeTransformations = false;

  public static PrintWriter printAnnotatedPW;
  public static PrintWriter printBinarizedPW;

  public static boolean printStates = false;

  public static int compactGrammar = 3; // exact compaction on by default

  public static boolean leftToRight = false; // whether to binarize left to right or head out

  public static int compactGrammar() {
    if (markovFactor) {
      return compactGrammar;
    }
    return 0;
  }

  public static boolean noTagSplit = false;

  /**
   * CHANGE ANYTHING BELOW HERE AT YOUR OWN RISK
   */

  public static boolean smoothing = false;
  public static boolean smoothedBound = false;

  /*  public static boolean factorOut = false;
  public static boolean rightBonus = false;
  public static boolean brokenDep = false;*/

  public static double ruleDiscount = 0.0;

  //public static boolean outsideFilter = false;


  public static void display() {
    System.err.println("Train parameters: smooth=" + smoothing + " PA=" + PA + " GPA=" + gPA + " selSplit=" + selectiveSplit + " (" + selectiveSplitCutOff + ((deleteSplitters != null) ? ("; deleting " + deleteSplitters): "") + ")" + " mUnary=" + markUnary + " mUnary2=" + markUnary2 + " mUnaryTags=" + markUnaryTags + " sPPT=" + splitPrePreT + " tagPA=" + tagPA + " tagSelSplit=" + tagSelectiveSplit + " (" + tagSelectiveSplitCutOff + ")" + " rightRec=" + rightRec + " leftRec=" + leftRec + " xOverX=" + xOverX + " collinsPunc=" + collinsPunc + " markov=" + markovFactor + " mOrd=" + markovOrder + " hSelSplit=" + hSelSplit + "(" + HSEL_CUT + ")" + " compactGrammar=" + compactGrammar + " leaveItAll=" + leaveItAll + " postPA=" + postPA + " postGPA=" + postGPA + " selPSplit=" + selectivePostSplit + " (" + selectivePostSplitCutOff + ")" + " tagSelPSplit=" + tagSelectivePostSplit + " (" + tagSelectivePostSplitCutOff + ")" + " postSplitWithBase=" + postSplitWithBaseCategory + " fractionBeforeUnseenCounting=" + fractionBeforeUnseenCounting);
  }

  public static void printTrainTree(PrintWriter pw, String message, Tree t) {
    PrintWriter myPW;
    if (pw == null) {
      myPW = new PrintWriter(System.out, true);
    } else {
      myPW = pw;
    }
    if (message != null && pw == null) {
      // hard coded to not print message if using file output!
      myPW.println(message);
    }
    boolean previousState = CategoryWordTag.printWordTag;
    CategoryWordTag.printWordTag = false;
    t.pennPrint(myPW);
    CategoryWordTag.printWordTag = previousState;
  }

} // end class Train


interface Extractor {
  public Object extract(Collection<Tree> trees);

  public Object extract(Iterator<Tree> iterator, Function<Tree, Tree> f);
}

abstract class AbstractTreeExtractor implements Extractor {

  protected void tallyLeaf(Tree lt) {
  }

  protected void tallyPreTerminal(Tree lt) {
  }

  protected void tallyInternalNode(Tree lt) {
  }

  protected void tallyRoot(Tree lt) {
  }

  protected Object formResult() {
    return null;
  }

  protected void tallyLocalTree(Tree lt) {
    // printTrainTree(null, "Tallying local tree:", lt);

    if (lt.isLeaf()) {
      //      System.out.println("it's a leaf");
      tallyLeaf(lt);
    } else if (lt.isPreTerminal()) {
      //      System.out.println("it's a preterminal");
      tallyPreTerminal(lt);
    } else {
      //      System.out.println("it's a internal node");
      tallyInternalNode(lt);
    }
  }

  protected void tallyTree(Tree t) {
    tallyRoot(t);
    for (Tree localTree : t.subTreeList()) {
      tallyLocalTree(localTree);
    }
  }

  protected void tallyTrees(Collection<Tree> trees) {
    for (Tree tree : trees) {
      tallyTree(tree);
    }
  }

  protected void tallyTreeIterator(Iterator<Tree> treeIterator, Function<Tree, Tree> f) {
    while (treeIterator.hasNext()) {
      Tree tree = treeIterator.next();
      try {
        tree = f.apply(tree);
      } catch (Exception e) {
        if (Test.verbose) {
          e.printStackTrace();
        }
        continue;  // UGH!
      }
      tallyTree(tree);
    }
  }

  public Object extract() {
    return formResult();
  }

  public Object extract(Collection<Tree> treeList) {
    tallyTrees(treeList);
    return formResult();
  }

  double weight = 1.0;

  public Object extract(Collection<Tree> trees1, Collection<Tree> trees2, double weight) {
    tallyTrees(trees1);
    this.weight = weight;
    tallyTrees(trees2);
    return formResult();
  }

  public Object extract(Iterator<Tree> treeIterator, Function<Tree, Tree> f) {
    tallyTreeIterator(treeIterator, f);
    return formResult();
  }
}


class FA {
  protected boolean[] inStatePrev;
  protected boolean[] inStateNext;
  protected boolean[] loopState;
  protected int acceptingState;
  protected int initialState;
  protected int numStates;
  protected int numSymbols;
  protected int[][] transition; // state x tag

  public void init() {
    Arrays.fill(inStatePrev, false);
    Arrays.fill(inStateNext, false);
    inStatePrev[initialState] = true;
  }

  public void input(int symbol) {
    for (int prevState = 0; prevState < numStates; prevState++) {
      if (inStatePrev[prevState]) {
        inStateNext[transition[prevState][symbol]] = true;
      }
    }
  }

  public void advance() {
    boolean[] temp = inStatePrev;
    inStatePrev = inStateNext;
    inStateNext = temp;
    Arrays.fill(inStateNext, false);
    for (int state = 0; state < numStates; state++) {
      if (inStatePrev[state] && loopState[state]) {
        inStateNext[state] = true;
      }
    }
  }

  public boolean isAccepting() {
    return inStatePrev[acceptingState];
  }

  public void setTransition(int state, int symbol, int result) {
    transition[state][symbol] = result;
  }

  public void setLoopState(int state, boolean loops) {
    loopState[state] = loops;
  }

  public FA(int numStates, int numSymbols) {
    this.numStates = numStates;
    this.numSymbols = numSymbols;
    acceptingState = numStates - 1;
    inStatePrev = new boolean[numStates];
    inStateNext = new boolean[numStates];
    loopState = new boolean[numStates];
    transition = new int[numStates][numSymbols];
  }

} // end class FA


class OutsideRuleFilter {

  private Numberer tagNumberer = Numberer.getGlobalNumberer("tags");
  private Numberer stateNumberer = Numberer.getGlobalNumberer("states");
  private int numTags;
  private int numFAs;

  protected FA[] leftFA;
  protected FA[] rightFA;

  protected List reverse(List list) {
    List reverse = new ArrayList(list.size());
    for (int i = list.size() - 1; i >= 0; i--) {
      reverse.add(list.get(i));
    }
    return reverse;
  }

  protected FA buildFA(List tags) {
    FA fa = new FA(tags.size() + 1, numTags);
    fa.setLoopState(0, true);
    for (int state = 1; state <= tags.size(); state++) {
      Object tagO = tags.get(state - 1);
      if (tagO == null) {
        fa.setLoopState(state, true);
        for (int symbol = 0; symbol < numTags; symbol++) {
          fa.setTransition(state - 1, symbol, state);
        }
      } else {
        int tag = tagNumberer.number(tagO);
        fa.setTransition(state - 1, tag, state);
      }
    }
    return fa;
  }

  protected void registerRule(List leftTags, List rightTags, int state) {
    leftFA[state] = buildFA(leftTags);
    rightFA[state] = buildFA(reverse(rightTags));
  }

  public void init() {
    for (int rule = 0; rule < numFAs; rule++) {
      leftFA[rule].init();
      rightFA[rule].init();
    }
  }

  public void advanceRight(boolean[] tags) {
    for (int tag = 0; tag < numTags; tag++) {
      if (!tags[tag]) {
        continue;
      }
      for (int rule = 0; rule < numFAs; rule++) {
        leftFA[rule].input(tag);
      }
    }
    for (int rule = 0; rule < numFAs; rule++) {
      leftFA[rule].advance();
    }
  }

  public void leftAccepting(boolean[] result) {
    for (int rule = 0; rule < numFAs; rule++) {
      result[rule] = leftFA[rule].isAccepting();
    }
  }

  public void advanceLeft(boolean[] tags) {
    for (int tag = 0; tag < numTags; tag++) {
      if (!tags[tag]) {
        continue;
      }
      for (int rule = 0; rule < numFAs; rule++) {
        rightFA[rule].input(tag);
      }
    }
    for (int rule = 0; rule < numFAs; rule++) {
      rightFA[rule].advance();
    }
  }

  public void rightAccepting(boolean[] result) {
    for (int rule = 0; rule < numFAs; rule++) {
      result[rule] = rightFA[rule].isAccepting();
    }
  }

  private void allocate(int numFAs) {
    this.numFAs = numFAs;
    leftFA = new FA[numFAs];
    rightFA = new FA[numFAs];
  }

  public OutsideRuleFilter(BinaryGrammar bg) {
    int numStates = stateNumberer.total();
    numTags = tagNumberer.total();
    allocate(numStates);
    for (int state = 0; state < numStates; state++) {
      String stateStr = (String) stateNumberer.object(state);
      List left = new ArrayList();
      List right = new ArrayList();
      if (!bg.isSynthetic(state)) {
        registerRule(left, right, state);
        continue;
      }
      boolean foundSemi = false;
      boolean foundDots = false;
      List array = left;
      StringBuffer sb = new StringBuffer();
      for (int c = 0; c < stateStr.length(); c++) {
        if (stateStr.charAt(c) == ':') {
          foundSemi = true;
          continue;
        }
        if (!foundSemi) {
          continue;
        }
        if (stateStr.charAt(c) == ' ') {
          if (sb.length() > 0) {
            String str = sb.toString();
            if (!tagNumberer.hasSeen(str)) {
              str = null;
            }
            array.add(str);
            sb = new StringBuffer();
          }
          continue;
        }
        if (!foundDots && stateStr.charAt(c) == '.') {
          c += 3;
          foundDots = true;
          array = right;
          continue;
        }
        sb.append(stateStr.charAt(c));
      }
      registerRule(left, right, state);
    }
  }

} // end class OutsideRuleFilter


// Binarization, Debinarization

/** This does nothing but a function to change the tree nodes into 
 *  CategoryWordTag, while the leaves are StringLabels. That's what the
 *  rest of the code assumes.
 */
class TreeNullAnnotator implements TreeTransformer {
  
  private TreeFactory tf = 
    new LabeledScoredTreeFactory(new CategoryWordTagFactory());
  private HeadFinder hf;

  public Tree transformTree(Tree t) {
    // make a defensive copy which the helper method can then mangle
    Tree copy = t.deepCopy(tf);
    return transformTreeHelper(copy);
  }

  private Tree transformTreeHelper(Tree t) {
    if (t == null) {
    } else {
      String cat = t.label().value();
      if (t.isLeaf()) {
        Label label = new StringLabel(cat); //new CategoryWordTag(cat,cat,"");
        t.setLabel(label);
      } else {
        Tree[] kids = t.children();
        for (int childNum = 0; childNum < kids.length; childNum++) {
          Tree child = kids[childNum];
          transformTreeHelper(child); // recursive call
        }
        Tree headChild = hf.determineHead(t);
        String tag;
        String word;
        if (headChild == null) {
          System.err.println("ERROR: null head for tree\n" + t.toString());
          word = null;
          tag = null;
        } else if (headChild.isLeaf()) {
          tag = cat;
          word = headChild.label().value();
        } else {
          CategoryWordTag headLabel = (CategoryWordTag) headChild.label();
          word = headLabel.word();
          tag = headLabel.tag();
        }
        Label label = new CategoryWordTag(cat, word, tag);
        t.setLabel(label);
      }
    }
    return t;
  }

  public TreeNullAnnotator(HeadFinder hf) {
    this.hf = hf;
  }

}


class TreeAnnotator implements TreeTransformer {

  private TreeFactory tf;
  private TreebankLangParserParams tlpParams;
  private HeadFinder hf;

  /** Do the category splitting of the tree passed in.
   *  This method defensively copies its argument, which is not changed.
   *
   *  @param t The tree to be annotated.  This can be any tree with a 
   *     <code>value() stored in Labels.  The tree is assumed to have 
   *     preterminals that are parts of speech.
   *  @return The annotated version of the Tree (which is a completely 
   *     separate Tree with new tree structure and new labels).  The
   *     non-leaf nodes of the tree will be CategoryWordTag objects.
   */
  public Tree transformTree(Tree t) {
    // make a defensive copy which the helper method can then mangle
    Tree copy = t.deeperCopy(tf);
    return transformTreeHelper(copy, copy);
  }

  /* Do the category splitting of the tree passed in.
   * This is initially called on the root node of a tree, and it recursively 
   * calls itself on children.  A depth first left-to-right traversal is
   * done whereby a tree node's children are first transformed and then 
   * the parent is transformed.  At the time of calling, the original root 
   * always sits above the current node.  This routine can be assumed to,
   * and does, change the tree passed in: it destructively modifies tree nodes,
   * and makes new tree structure when it needs to.
   * 
   * @param t The tree node to subcategorize.
   * @param root The root of the tree.  It must contain <code>t</code> or 
   *     this code will throw a NullPointerException.
   * @return The annotated tree.
   */
  private Tree transformTreeHelper(Tree t, Tree root) {
    if (t == null) {
      // handle null
      return null;
    }

    String cat = t.label().value();
    if (t.isLeaf()) {
      // handle words
      Label label = new StringLabel(cat); //new CategoryWordTag(cat,cat,"");
      t.setLabel(label);
      return t;
    }

    Tree parent;
    Tree grandParent;
    String parentStr;
    String grandParentStr;
    if (root == null || t.equals(root)) {
      parent = null;
      parentStr = "";
    } else {
      parent = t.parent(root);
      parentStr = parent.label().value();
    }
    if (parent == null || parent.equals(root)) {
      grandParent = null;
      grandParentStr = "";
    } else {
      grandParent = parent.parent(root);
      grandParentStr = grandParent.label().value();
    }
    String baseParentStr = tlpParams.treebankLanguagePack().basicCategory(parentStr);
    String baseGrandParentStr = tlpParams.treebankLanguagePack().basicCategory(grandParentStr);
    //System.out.println(t.label().value() + " " + parentStr + " " + grandParentStr);

    if (t.isPreTerminal()) {
      // handle tags
      Tree childResult = transformTreeHelper(t.children()[0], null); // recurse
      String word = childResult.value();  // would be nicer if Word/CWT ??

      if ( ! Train.noTagSplit) {
        if (Train.tagPA) {
          String test = cat + "^" + baseParentStr;
          if (!Train.tagSelectiveSplit || Train.splitters.contains(test)) {
            cat = test;
          }
        }
        if (Train.markUnaryTags && parent.children().length == 1) {
          cat = cat + "^U";
        }
      } // otherwise, leave the tags alone!

      Label label = new CategoryWordTag(cat, word, cat);
      t.setLabel(label);
      t.setChild(0, childResult);  // just in case word is changed
      // language-specific transforms
      return tlpParams.transformTree(t, root);
    } // end isPreTerminal()
    // handle phrasal categories
    Tree[] kids = t.children();
    for (int childNum = 0; childNum < kids.length; childNum++) {
      Tree child = kids[childNum];
      Tree childResult = transformTreeHelper(child, root); // recursive call
      t.setChild(childNum, childResult);
    }
    CategoryWordTag headLabel;
    Tree headChild = hf.determineHead(t);
    try {
      headLabel = (CategoryWordTag) headChild.label();
    } catch (ClassCastException cce) {
      tlpParams.pw().println("TreeAnnotator: Ill-formed tree:\n" + t);
      throw cce;
    } catch (NullPointerException npe) {
      tlpParams.pw().println("TreeAnnotator: null head found for tree:\n" + t);
      throw npe;
    }

    String word = headLabel.word();
    String tag = headLabel.tag();
    String baseTag = tlpParams.treebankLanguagePack().basicCategory(tag);
    String baseCat = tlpParams.treebankLanguagePack().basicCategory(cat);

    /* Sister annotation. Potential problem: if multiple sisters are
     * strong indicators for a single category's expansions.  This
     * happens concretely in the Chinese Treebank when NP (object)
     * has left sisters VV and AS.  Could lead to too much
     * sparseness.  The ideal solution would be to give the
     * splitting list an ordering, and take only the highest (~most
     * informative/reliable) sister annotation.
     */
    if (Train.sisterAnnotate && !Train.smoothing && baseParentStr.length() > 0) {
      List leftSis = listBasicCategories(SisterAnnotationStats.leftSisterLabels(t, parent));
      List rightSis = listBasicCategories(SisterAnnotationStats.rightSisterLabels(t, parent));

      List leftAnn = new ArrayList();
      List rightAnn = new ArrayList();

      for (Iterator i = leftSis.iterator(); i.hasNext();) {
        String s = (String) i.next();
        //s = baseCat+"=l="+tlpParams.treebankLanguagePack().basicCategory(s);
        leftAnn.add(baseCat + "=l=" + tlpParams.treebankLanguagePack().basicCategory(s));
        //System.out.println("left-annotated test string " + s);
      }
      for (Iterator i = rightSis.iterator(); i.hasNext();) {
        String s = (String) i.next();
        //s = baseCat+"=r="+tlpParams.treebankLanguagePack().basicCategory(s);
        rightAnn.add(baseCat + "=r=" + tlpParams.treebankLanguagePack().basicCategory(s));
      }
      for (Iterator j = rightAnn.iterator(); j.hasNext();) {
        //System.out.println("new rightsis " + (String)j.next()); //debugging
      }
      for (Iterator i = Train.sisterSplitters.iterator(); i.hasNext();) {
        String annCat = (String) i.next();
        //System.out.println("annotated test string " + annCat);
        if (leftAnn.contains(annCat) || rightAnn.contains(annCat)) {
          cat = cat + annCat.replaceAll("^" + baseCat, "");
          break;
        }
      }
    }

    if (Train.PA && !Train.smoothing && baseParentStr.length() > 0) {
      String cat2 = baseCat + "^" + baseParentStr;
      if (!Train.selectiveSplit || Train.splitters.contains(cat2)) {
        cat = cat + "^" + baseParentStr;
      }
    }
    if (Train.gPA && !Train.smoothing && grandParentStr.length() > 0) {
      if (Train.selectiveSplit) {
        String cat2 = baseCat + "^" + baseParentStr + "~" + baseGrandParentStr;
        if (cat.indexOf("^") >= 0 && Train.splitters.contains(cat2)) {
          cat = cat + "~" + baseGrandParentStr;
        }
      } else {
        cat = cat + "~" + baseGrandParentStr;
      }
    }
    if (Train.markUnary && kids.length == 1 && kids[0].depth() >= 2) {
      cat = cat + "-U";
    }
    if (Train.markUnary2 && parent != null && parent.children().length == 1 && t.depth() >= 2) {
      cat = cat + "-u";
    }
    if (Train.xOverX && xOverX(t, baseCat)) {
      cat = cat + "-X";
    }
    if (Train.rightRec && rightRec(t, baseCat)) {
      cat = cat + "-R";
    }
    if (Train.leftRec && leftRec(t, baseCat)) {
      cat = cat + "-L";
    }
    if (Train.splitPrePreT && t.isPrePreTerminal()) {
      cat = cat + "-PPT";
    }
    Label label = new CategoryWordTag(cat, word, tag);
    t.setLabel(label);
    // System.out.println("internal transformTree: " +
    //                         baseCat + "^" + baseParentStr + "~" +
    //                         baseGrandParentStr + " to " +
    //                         label.value());
    return tlpParams.transformTree(t, root);
  }


  private List listBasicCategories(List l) {
    List l1 = new ArrayList();
    for (Iterator i = l.iterator(); i.hasNext();) {
      l1.add(tlpParams.treebankLanguagePack().basicCategory((String) i.next()));
    }
    return l1;
  }


  private boolean rightRec(Tree t, String baseCat) {
    if (//! baseCat.equals("S") &&
            !baseCat.equals("NP")) {
      return false;
    }
    while (!t.isLeaf()) {
      t = t.lastChild();
      String str = t.label().value();
      if (str.startsWith(baseCat)) {
        return true;
      }
    }
    return false;
  }

  private boolean leftRec(Tree t, String baseCat) {
    while (!t.isLeaf()) {
      t = t.firstChild();
      String str = t.label().value();
      if (str.startsWith(baseCat)) {
        return true;
      }
    }
    return false;
  }

  private boolean xOverX(Tree t, String baseCat) {
    for (Iterator i = t.subTreeList().iterator(); i.hasNext();) {
      Tree s = (Tree) i.next();
      if (s == t) {
        continue;
      }
      if (s.label().value().startsWith(baseCat)) {
        return true;
      }
    }
    return false;
  }

  public TreeAnnotator(HeadFinder hf, TreebankLangParserParams tlpp) {
    this.tlpParams = tlpp;
    this.hf = hf;
    this.tf = new LabeledScoredTreeFactory(new CategoryWordTagFactory());
  }

} // end class TreeAnnotator


class TreeAnnotatorAndBinarizer implements TreeTransformer {

  TreeFactory tf;
  private TreebankLanguagePack tlp;
  private TreeTransformer annotator;
  private TreeTransformer binarizer;
  private TreeTransformer postSplitter;
  private boolean forceCNF;

  public TreeAnnotatorAndBinarizer(TreebankLangParserParams tlpParams, boolean forceCNF, boolean insideFactor, boolean doSubcategorization) {
    this(tlpParams.headFinder(), tlpParams.headFinder(), tlpParams, forceCNF, insideFactor, doSubcategorization);
  }

  public TreeAnnotatorAndBinarizer(HeadFinder annotationHF, HeadFinder binarizationHF, TreebankLangParserParams tlpParams, boolean forceCNF, boolean insideFactor, boolean doSubcategorization) {
    if (doSubcategorization) {
      annotator = new TreeAnnotator(annotationHF, tlpParams);
    } else {
      annotator = new TreeNullAnnotator(annotationHF);
    }
    binarizer = new TreeBinarizer(binarizationHF, tlpParams.treebankLanguagePack(), insideFactor, Train.markovFactor, Train.markovOrder, Train.compactGrammar() > 0, Train.compactGrammar() > 1, Train.HSEL_CUT, Train.markFinalStates);
    if (Train.selectivePostSplit) {
      postSplitter = new PostSplitter(tlpParams);
    }
    this.tf = new LabeledScoredTreeFactory(new CategoryWordTagFactory());
    this.tlp = tlpParams.treebankLanguagePack();
    this.forceCNF = forceCNF;
  }

  public void dumpStats() {
    if (Train.selectivePostSplit) {
      ((PostSplitter) postSplitter).dumpStats();
    }
  }

  public void setDoSelectiveSplit(boolean doSelectiveSplit) {
    ((TreeBinarizer) binarizer).setDoSelectiveSplit(doSelectiveSplit);
  }

  /**
   * Changes the ROOT label, and adds a Lexicon.BOUNDARY daughter to it.
   * This is needed for the dependency parser.
   */
  public void addRoot(Tree t) {
    if (t.isLeaf()) {
      System.err.println("Warning: tree is leaf: " + t);
      t = tf.newTreeNode(tlp.startSymbol(), Collections.singletonList(t));
    }
    t.setLabel(new CategoryWordTag(tlp.startSymbol(), Lexicon.BOUNDARY, Lexicon.BOUNDARY_TAG));
    List preTermChildList = new ArrayList();
    Tree boundaryTerm = tf.newLeaf(new StringLabel(Lexicon.BOUNDARY));//CategoryWordTag(Lexicon.BOUNDARY,Lexicon.BOUNDARY,""));
    preTermChildList.add(boundaryTerm);
    Tree boundaryPreTerm = tf.newTreeNode(new CategoryWordTag(Lexicon.BOUNDARY_TAG, Lexicon.BOUNDARY, Lexicon.BOUNDARY_TAG), preTermChildList);
    List childList = t.getChildrenAsList();
    childList.add(boundaryPreTerm);
    t.setChildren(childList);
  }

  /** The tree t is normally expected to be a Penn-Treebank-style tree
   *  in which the top node is an extra node that rewrites unarily.
   *  If this isn't the case, an extra node is added and the user is
   *  warned.
   */

  public Tree transformTree(Tree t) {
    if (Train.printTreeTransformations) {
      Train.printTrainTree(null, "ORIGINAL TREE:", t);
    }
    Tree trTree = annotator.transformTree(t);
    if (Train.selectivePostSplit) {
      trTree = postSplitter.transformTree(trTree);
    }
    if (Train.printTreeTransformations) {
      Train.printTrainTree(Train.printAnnotatedPW, "ANNOTATED TREE:", trTree);
    }

    // if we add the ROOT first, then we don't know how to percolate the heads at the top
    addRoot(trTree); // this creates a few non-binarized rules at the top

    Tree binarizedTree = binarizer.transformTree(trTree);
    if (Train.printTreeTransformations) {
      Train.printTrainTree(Train.printBinarizedPW, "BINARIZED TREE:", binarizedTree);
    }
    //      addRoot(binarizedTree); // this creates a few non-binarized rules at the top
    if (forceCNF) {
      binarizedTree = new CNFTransformers.ToCNFTransformer().transformTree(binarizedTree);
      //        System.out.println("BinarizedCNF:\n");
      //        binarizedTree.pennPrint();
    }
    return binarizedTree;
  }

} // end class TreeAnnotatorAndBinarizer


/**
 * This splits on parents using the same algorithm as the earlier
 * parent (selective) annotation algorithms, but applied AFTER the tree
 * has been annotated.
 */
class PostSplitter implements TreeTransformer {

  Counter nonTerms = new Counter();
  TreebankLangParserParams tlpParams;
  TreeFactory tf;
  HeadFinder hf;

  public Tree transformTree(Tree t) {
    tf = t.treeFactory();
    return transformTreeHelper(t, t);
  }

  public Tree transformTreeHelper(Tree t, Tree root) {
    Tree result;
    Tree parent;
    Tree grandParent;
    String parentStr;
    String grandParentStr;
    if (root == null || t.equals(root)) {
      parent = null;
      parentStr = "";
    } else {
      parent = t.parent(root);
      parentStr = parent.label().value();
    }
    if (parent == null || parent.equals(root)) {
      grandParent = null;
      grandParentStr = "";
    } else {
      grandParent = parent.parent(root);
      grandParentStr = grandParent.label().value();
    }
    String cat = t.label().value();
    String baseParentStr = tlpParams.treebankLanguagePack().basicCategory(parentStr);
    String baseGrandParentStr = tlpParams.treebankLanguagePack().basicCategory(grandParentStr);
    if (t.isLeaf()) {
      return tf.newLeaf(new StringLabel(t.label().value()));
    }
    String tag = t.headPreTerminal(hf).value();
    String word = t.headTerminal(hf).value();
    if (t.isPreTerminal()) {
      nonTerms.incrementCount(t.label().value());
    } else {
      nonTerms.incrementCount(t.label().value());
      if (Train.postPA && !Train.smoothing && baseParentStr.length() > 0) {
        String cat2;
        if (Train.postSplitWithBaseCategory) {
          cat2 = cat + "^" + baseParentStr;
        } else {
          cat2 = cat + "^" + parentStr;
        }
        if (!Train.selectivePostSplit || Train.postSplitters.contains(cat2)) {
          cat = cat2;
        }
      }
      if (Train.postGPA && !Train.smoothing && grandParentStr.length() > 0) {
        String cat2;
        if (Train.postSplitWithBaseCategory) {
          cat2 = cat + "~" + baseGrandParentStr;
        } else {
          cat2 = cat + "~" + grandParentStr;
        }
        if (Train.selectivePostSplit) {
          if (cat.indexOf("^") >= 0 && Train.postSplitters.contains(cat2)) {
            cat = cat2;
          }
        } else {
          cat = cat2;
        }
      }
    }
    result = tf.newTreeNode(new CategoryWordTag(cat, word, cat), Collections.EMPTY_LIST);
    ArrayList newKids = new ArrayList();
    Tree[] kids = t.children();
    for (int i = 0; i < kids.length; i++) {
      newKids.add(transformTreeHelper(kids[i], root));
    }
    result.setChildren(newKids);
    return result;
  }

  public void dumpStats() {
    System.out.println("%% Counts of nonterminals:");
    List biggestCounts = new ArrayList(nonTerms.keySet());
    Collections.sort(biggestCounts, nonTerms.comparator(false));
    for (Iterator it = biggestCounts.iterator(); it.hasNext();) {
      Object ob = it.next();
      System.out.println(ob + ": " + nonTerms.getCount(ob));
    }
  }

  public PostSplitter(TreebankLangParserParams tlpParams) {
    this.tlpParams = tlpParams;
    this.hf = tlpParams.headFinder();
  }

} // end class PostSplitter


class Debinarizer implements TreeTransformer {

  private TreeFactory tf;
  private TreebankLanguagePack tlp;
  private boolean forceCNF;

  /**
   * Remove the last daughter of a ROOT node.  This is a boundary symbol.
   * This is needed because the dependency parser uses such symbols.
   */
  private void removeRoot(Tree t) {
    if (!tlp.isStartSymbol(t.label().value())) {
      return;
    }
    Tree[] origChildren = t.children();
    Tree[] children = new Tree[origChildren.length - 1];
    System.arraycopy(origChildren, 0, children, 0, origChildren.length - 1);
    t.setChildren(children);
  }

  protected Tree transformTreeHelper(Tree t) {
    if (t.isLeaf()) {
      Tree leaf = tf.newLeaf(new StringLabel(t.label().value()));
      leaf.setScore(t.score());
      return leaf;
    }
    List newChildren = new ArrayList(20);
    Label label = new StringLabel(t.label().value());
    for (int childNum = 0; childNum < t.children().length; childNum++) {
      Tree child = t.children()[childNum];
      Tree newChild = transformTreeHelper(child);
      if ((!newChild.isLeaf()) && newChild.label().value().indexOf('@') >= 0) {
        newChildren.addAll(newChild.getChildrenAsList());
      } else {
        newChildren.add(newChild);
      }
    }
    Tree node = tf.newTreeNode(label, newChildren);
    node.setScore(t.score());
    return node;
  }

  public Tree transformTree(Tree t) {
    Tree result = transformTreeHelper(t);
    if (forceCNF) {
      result = new CNFTransformers.FromCNFTransformer().transformTree(result);
    }
    removeRoot(result);
    return result;
  }

  public Debinarizer(TreebankLanguagePack tlp, boolean forceCNF) {
    this.tlp = tlp;
    this.forceCNF = forceCNF;
    tf = new LabeledScoredTreeFactory(new StringLabelFactory());
  }

} // end class Debinarizer


class BinaryGrammarExtractor extends AbstractTreeExtractor {
  protected Numberer stateNumberer = Numberer.getGlobalNumberer("states");
  private Counter ruleCounter = new Counter();
  private Counter symbolCounter = new Counter();
  private Set binaryRules = new HashSet();
  private Set unaryRules = new HashSet();

  //  protected void tallyTree(Tree t) {
  //    super.tallyTree(t);
  //    System.out.println("Tree:");
  //    t.pennPrint();
  //  }

  protected void tallyRule(Rule r, double score) {
    symbolCounter.incrementCount(stateNumberer.object(r.parent), score);
    ruleCounter.incrementCount(r, score);
    if (r.isUnary()) {
      unaryRules.add(r);
    } else {
      binaryRules.add(r);
    }
  }

  protected void tallyInternalNode(Tree lt) {
    Rule r = null;
    if (lt.children().length == 1) {
      UnaryRule ur = new UnaryRule();
      ur.parent = stateNumberer.number(lt.label().value());
      ur.child = stateNumberer.number(lt.children()[0].label().value());
      r = ur;
    } else {
      BinaryRule br = new BinaryRule();
      br.parent = stateNumberer.number(lt.label().value());
      br.leftChild = stateNumberer.number(lt.children()[0].label().value());
      br.rightChild = stateNumberer.number(lt.children()[1].label().value());
      r = br;
    }
    tallyRule(r, weight);
  }

  protected Object formResult() {
    stateNumberer.number(Lexicon.BOUNDARY_TAG);
    BinaryGrammar bg = new BinaryGrammar(stateNumberer.total());
    UnaryGrammar ug = new UnaryGrammar(stateNumberer.total());
    // add unaries
    for (Iterator uI = unaryRules.iterator(); uI.hasNext();) {
      UnaryRule ur = (UnaryRule) uI.next();
      ur.score = Math.log((ruleCounter.getCount(ur)) / symbolCounter.getCount(stateNumberer.object(ur.parent)));
      if (Train.compactGrammar() >= 4) {
        ur.score = ruleCounter.getCount(ur);
      }
      ug.addRule(ur);
    }
    // add binaries
    for (Iterator bI = binaryRules.iterator(); bI.hasNext();) {
      BinaryRule br = (BinaryRule) bI.next();
      br.score = Math.log((ruleCounter.getCount(br) - Train.ruleDiscount) / symbolCounter.getCount(stateNumberer.object(br.parent)));
      if (Train.compactGrammar() >= 4) {
        br.score = ruleCounter.getCount(br);
      }
      bg.addRule(br);
    }
    return new Pair(ug, bg);
  }


} // end class BinaryGrammarExtractor


class SmoothedBinaryGrammarExtractor extends AbstractTreeExtractor {

  protected int HISTORY_DEPTH() {
    return (Train.gPA ? 3 : (Train.PA ? 2 : 1));
  }

  protected Numberer stateNumberer = Numberer.getGlobalNumberer("states");
  protected Numberer tagNumberer = Numberer.getGlobalNumberer("tags");

  protected Map ruleToLabel = new HashMap();
  protected Counter rulePairs = new Counter();
  protected Counter labelPairs = new Counter();

  public void tallyTree(Tree t) {
    LinkedList ll = new LinkedList();
    //ll.add("ROOT");
    //ll.add("ROOT");
    tallyTree(t, ll);
  }

  protected void tallyTree(Tree t, LinkedList parents) {
    // traverse tree, building parent list
    String str = t.label().value();
    boolean strIsPassive = (str.indexOf('@') == -1);
    if (strIsPassive) {
      parents.addFirst(str);
    }
    if (!t.isLeaf()) {
      if (!t.children()[0].isLeaf()) {
        tallyInternalNode(t, parents);
        for (int c = 0; c < t.children().length; c++) {
          Tree child = t.children()[c];
          tallyTree(child, parents);
        }
      } else {
        tagNumberer.number(t.label().value());
      }
    }
    if (strIsPassive) {
      parents.removeFirst();
    }
  }

  protected Rule ltToRule(Tree lt) {
    if (lt.children().length == 1) {
      UnaryRule ur = new UnaryRule();
      ur.parent = stateNumberer.number(lt.label().value());
      ur.child = stateNumberer.number(lt.children()[0].label().value());
      return ur;
    } else {
      BinaryRule br = new BinaryRule();
      br.parent = stateNumberer.number(lt.label().value());
      br.leftChild = stateNumberer.number(lt.children()[0].label().value());
      br.rightChild = stateNumberer.number(lt.children()[1].label().value());
      return br;
    }
  }

  protected boolean isSynthetic(int state) {
    return ((String) stateNumberer.object(state)).indexOf('@') > -1;
  }

  protected boolean isTag(int state) {
    return tagNumberer.hasSeen(stateNumberer.object(state));
  }

  protected Rule specifyRule(Rule rule, List history, int childDepth) {
    Rule r = null;
    String topHistoryStr = historyToString(history.subList(1, history.size()));
    String bottomHistoryStr = historyToString(history.subList(0, childDepth));
    if (rule instanceof UnaryRule) {
      UnaryRule ur = new UnaryRule();
      UnaryRule urule = (UnaryRule) rule;
      ur.parent = stateNumberer.number(stateNumberer.object(urule.parent) + topHistoryStr);
      if (isSynthetic(urule.child)) {
        ur.child = stateNumberer.number(stateNumberer.object(urule.child) + topHistoryStr);
      } else if (isTag(urule.child)) {
        ur.child = urule.child;
      } else {
        ur.child = stateNumberer.number(stateNumberer.object(urule.child) + bottomHistoryStr);
      }
      r = ur;
    } else {
      BinaryRule br = new BinaryRule();
      BinaryRule brule = (BinaryRule) rule;
      br.parent = stateNumberer.number(stateNumberer.object(brule.parent) + topHistoryStr);
      if (isSynthetic(brule.leftChild)) {
        br.leftChild = stateNumberer.number(stateNumberer.object(brule.leftChild) + topHistoryStr);
      } else if (isTag(brule.leftChild)) {
        br.leftChild = brule.leftChild;
      } else {
        br.leftChild = stateNumberer.number(stateNumberer.object(brule.leftChild) + bottomHistoryStr);
      }
      if (isSynthetic(brule.rightChild)) {
        br.rightChild = stateNumberer.number(stateNumberer.object(brule.rightChild) + topHistoryStr);
      } else if (isTag(brule.rightChild)) {
        br.rightChild = brule.rightChild;
      } else {
        br.rightChild = stateNumberer.number(stateNumberer.object(brule.rightChild) + bottomHistoryStr);
      }
      r = br;
    }
    return r;
  }

  protected void tallyInternalNode(Tree lt, List parents) {
    // form base rule
    String label = lt.label().value();
    Rule baseR = ltToRule(lt);
    ruleToLabel.put(baseR, label);
    // act on each history depth
    for (int depth = 0, maxDepth = Math.min(HISTORY_DEPTH(), parents.size()); depth <= maxDepth; depth++) {
      List history = new ArrayList(parents.subList(0, depth));
      // tally each history level / rewrite pair
      rulePairs.incrementCount(new Pair(baseR, history), 1);
      labelPairs.incrementCount(new Pair(label, history), 1);
    }
  }

  protected Map historyToString = new HashMap();

  protected String historyToString(List history) {
    String str = (String) historyToString.get(history);
    if (str == null) {
      StringBuffer sb = new StringBuffer();
      for (int i = 0; i < history.size(); i++) {
        sb.append('^');
        sb.append(history.get(i));
      }
      str = sb.toString();
      historyToString.put(history, str);
    }
    return str;
  }

  protected Object formResult() {
    Set brs = new HashSet();
    Set urs = new HashSet();
    // scan each rule / history pair
    int ruleCount = 0;
    for (Iterator pairI = rulePairs.keySet().iterator(); pairI.hasNext();) {
      if (ruleCount % 100 == 0) {
        System.err.println("Rules multiplied: " + ruleCount);
      }
      ruleCount++;
      Pair rulePair = (Pair) pairI.next();
      Rule baseRule = (Rule) rulePair.first;
      String baseLabel = (String) ruleToLabel.get(baseRule);
      List history = (List) rulePair.second;
      double totalProb = 0;
      for (int depth = 1; depth <= HISTORY_DEPTH() && depth <= history.size(); depth++) {
        List subHistory = history.subList(0, depth);
        double c_label = labelPairs.getCount(new Pair(baseLabel, subHistory));
        double c_rule = rulePairs.getCount(new Pair(baseRule, subHistory));
        //System.out.println("Multiplying out "+baseRule+" with history "+subHistory);
        //System.out.println("Count of "+baseLabel+" with "+subHistory+" is "+c_label);
        //System.out.println("Count of "+baseRule+" with "+subHistory+" is "+c_rule );

        double prob = (1.0 / HISTORY_DEPTH()) * (c_rule) / (c_label);
        totalProb += prob;
        for (int childDepth = 0; childDepth <= Math.min(HISTORY_DEPTH() - 1, depth); childDepth++) {
          Rule rule = specifyRule(baseRule, subHistory, childDepth);
          rule.score = Math.log(totalProb);
          //System.out.println("Created  "+rule+" with score "+rule.score);
          if (rule instanceof UnaryRule) {
            urs.add(rule);
          } else {
            brs.add(rule);
          }
        }
      }
    }
    System.out.println("Total states: " + stateNumberer.total());
    BinaryGrammar bg = new BinaryGrammar(stateNumberer.total());
    UnaryGrammar ug = new UnaryGrammar(stateNumberer.total());
    for (Iterator brI = brs.iterator(); brI.hasNext();) {
      BinaryRule br = (BinaryRule) brI.next();
      bg.addRule(br);
    }
    for (Iterator urI = urs.iterator(); urI.hasNext();) {
      UnaryRule ur = (UnaryRule) urI.next();
      ug.addRule(ur);
    }
    return new Pair(ug, bg);
  }

} // end class SmoothedBinaryGrammarExtractor


class XBarGrammarProjection implements GrammarProjection {
  UnaryGrammar sourceUG;
  BinaryGrammar sourceBG;
  Numberer sourceNumberer;

  UnaryGrammar targetUG;
  BinaryGrammar targetBG;
  Numberer targetNumberer;

  int[] projection;

  public int project(int state) {
    return projection[state];
  }

  public UnaryGrammar sourceUG() {
    return sourceUG;
  }

  public BinaryGrammar sourceBG() {
    return sourceBG;
  }

  public UnaryGrammar targetUG() {
    return targetUG;
  }

  public BinaryGrammar targetBG() {
    return targetBG;
  }


  protected String projectString(String str) {
    if (str.indexOf('@') == -1) {
      if (str.indexOf('^') == -1) {
        return str;
      }
      return str.substring(0, str.indexOf('^'));
    }
    StringBuffer sb = new StringBuffer();
    sb.append(str.substring(0, str.indexOf(' ')));
    if (str.indexOf('^') > -1) {
      //sb.append(str.substring(str.indexOf('^'),str.length()));
    }
    int num = -2;
    for (int i = 0; i < str.length(); i++) {
      if (str.charAt(i) == ' ') {
        num++;
      }
    }
    sb.append(" w " + num);
    return sb.toString();
  }

  protected void scanStates(Numberer source, Numberer target) {
    for (int i = 0; i < source.total(); i++) {
      String stateStr = (String) source.object(i);
      String projStr = projectString(stateStr);
      projection[i] = target.number(projStr);
    }
  }

  protected BinaryRule projectBinaryRule(BinaryRule br) {
    BinaryRule br2 = new BinaryRule();
    br2.parent = projection[br.parent];
    br2.leftChild = projection[br.leftChild];
    br2.rightChild = projection[br.rightChild];
    br2.score = br.score;
    return br2;
  }

  protected UnaryRule projectUnaryRule(UnaryRule ur) {
    UnaryRule ur2 = new UnaryRule();
    ur2.parent = projection[ur.parent];
    ur2.child = projection[ur.child];
    ur2.score = ur.score;
    return ur2;
  }

  public XBarGrammarProjection(BinaryGrammar bg, UnaryGrammar ug) {
    Map binaryRules = new HashMap();
    Map unaryRules = new HashMap();
    sourceUG = ug;
    sourceBG = bg;
    sourceNumberer = Numberer.getGlobalNumberer(bg.stateSpace());
    targetNumberer = Numberer.getGlobalNumberer(bg.stateSpace() + "-xbar");
    projection = new int[sourceNumberer.total()];
    scanStates(sourceNumberer, targetNumberer);
    targetBG = new BinaryGrammar(targetNumberer.total(), bg.stateSpace() + "-xbar");
    targetUG = new UnaryGrammar(targetNumberer.total());
    for (Iterator brI = bg.iterator(); brI.hasNext();) {
      Rule rule = projectBinaryRule((BinaryRule) brI.next());
      Rule old = (BinaryRule) binaryRules.get(rule);
      if (old == null || rule.score > old.score) {
        binaryRules.put(rule, rule);
      }
    }
    for (Iterator brI = binaryRules.keySet().iterator(); brI.hasNext();) {
      BinaryRule br = (BinaryRule) binaryRules.get(brI.next());
      targetBG.addRule(br);
      //System.out.println("BR: "+targetNumberer.object(br.parent)+" -> "+targetNumberer.object(br.leftChild)+" "+targetNumberer.object(br.rightChild)+" %% "+br.score);
    }
    targetBG.splitRules();
    for (int parent = 0; parent < sourceNumberer.total(); parent++) {
      for (Iterator urI = ug.ruleIteratorByParent(parent); urI.hasNext();) {
        UnaryRule sourceRule = (UnaryRule) urI.next();
        Rule rule = projectUnaryRule(sourceRule);
        Rule old = (UnaryRule) unaryRules.get(rule);
        if (old == null || rule.score > old.score) {
          unaryRules.put(rule, rule);
        }
        /*
          if (((UnaryRule)rule).child == targetNumberer.number("PRP") &&
            ((String)sourceNumberer.object(rule.parent)).charAt(0) == 'N') {
          System.out.println("Source UR: "+sourceRule+" %% "+sourceRule.score);
          System.out.println("Score of "+rule+"is now: "+((UnaryRule)unaryRules.get(rule)).score);
        }
        */
      }
    }
    for (Iterator urI = unaryRules.keySet().iterator(); urI.hasNext();) {
      UnaryRule ur = (UnaryRule) unaryRules.get(urI.next());
      targetUG.addRule(ur);
      //System.out.println("UR: "+targetNumberer.object(ur.parent)+" -> "+targetNumberer.object(ur.child)+" %% "+ur.score);
    }
    targetUG.purgeRules();
    System.out.println("Projected " + sourceNumberer.total() + " states to " + targetNumberer.total() + " states.");
  }
}


interface GrammarProjection {
  int project(int state);

  UnaryGrammar sourceUG();

  BinaryGrammar sourceBG();

  UnaryGrammar targetUG();

  BinaryGrammar targetBG();
}

class NullGrammarProjection implements GrammarProjection {
  UnaryGrammar ug;
  BinaryGrammar bg;

  public int project(int state) {
    return state;
  }

  public UnaryGrammar sourceUG() {
    return ug;
  }

  public BinaryGrammar sourceBG() {
    return bg;
  }

  public UnaryGrammar targetUG() {
    return ug;
  }

  public BinaryGrammar targetBG() {
    return bg;
  }

  NullGrammarProjection(BinaryGrammar bg, UnaryGrammar ug) {
    this.ug = ug;
    this.bg = bg;
  }
}


class ProjectionScorer implements Scorer {
  protected GrammarProjection gp;
  protected Scorer scorer;

  protected Edge tempEdge = new Edge();

  protected Edge project(Edge edge) {
    tempEdge.start = edge.start;
    tempEdge.end = edge.end;
    tempEdge.state = gp.project(edge.state);
    tempEdge.head = edge.head;
    tempEdge.tag = edge.tag;
    return tempEdge;
  }

  protected Hook tempHook = new Hook();

  protected Hook project(Hook hook) {
    tempHook.start = hook.start;
    tempHook.end = hook.end;
    tempHook.state = gp.project(hook.state);
    tempHook.head = hook.head;
    tempHook.tag = hook.tag;
    tempHook.subState = gp.project(hook.subState);
    return tempHook;
  }

  public double oScore(Edge edge) {
    return scorer.oScore(project(edge));
  }

  public double iScore(Edge edge) {
    return scorer.iScore(project(edge));
  }

  public boolean oPossible(Hook hook) {
    return scorer.oPossible(project(hook));
  }

  public boolean iPossible(Hook hook) {
    return scorer.iPossible(project(hook));
  }

  public boolean parse(List words) {
    return scorer.parse(words);
  }

  public ProjectionScorer(Scorer scorer, GrammarProjection gp) {
    this.scorer = scorer;
    this.gp = gp;
  }
} // end ProjectionScorer


interface TagProjection {
  String project(String tagStr);
}

class BasicCategoryTagProjection implements TagProjection, Serializable {
  TreebankLanguagePack tlp;

  public BasicCategoryTagProjection(TreebankLanguagePack tlp) {
    this.tlp = tlp;
  }

  public String project(String tagStr) {
    //return tagStr;
    return tlp.basicCategory(tagStr);
  }
}


// Looks like the intended behavior of TestTagProjection is:
// 1) Include the basic category (everything before a '-' or '^' annotation)
// 2) Include any annotation introduced with '-'
// 3) Exclude any annotation introduced with '^'
// 4) Annotations introduced with other characters will be included or excluded
//    as determined by the previous annotation or basic category.
//
// This seems awfully haphazard :(
//
//  Roger

class TestTagProjection implements TagProjection, Serializable {
  public String project(String tagStr) {
    StringBuffer sb = new StringBuffer();
    boolean good = true;
    for (int pos = 0; pos < tagStr.length(); pos++) {
      char c = tagStr.charAt(pos);
      if (c == '-') {
        good = true;
      }
      if (c == '^') {
        good = false;
      }
      if (good) {
        sb.append(c);
      }
    }
    return sb.toString();
  }
}


class Dependency implements Serializable {

  public static final String LEFT = "left";
  public static final String RIGHT = "right";

  public IntTaggedWord head;
  public IntTaggedWord arg;
  public boolean leftHeaded;
  public short distance;

  public int hashCode() {
    return head.hashCode() ^ (arg.hashCode() << 8) ^ ((leftHeaded ? 1 : 0) << 15) ^ (distance << 16);
  }

  public boolean equals(Object o) {
    if (this == o) {
      return true;
    }
    if (o instanceof Dependency) {
      Dependency d = (Dependency) o;
      return (head.equals(d.head) && arg.equals(d.arg) && distance == d.distance && leftHeaded == d.leftHeaded);
    } else {
      return false;
    }
  }

  private static final char[] charsToEscape = new char[]{'\"'};

  public String toString() {
    return "\"" + StringUtils.escapeString(head.toString(), charsToEscape, '\\') + "\" -> \"" + StringUtils.escapeString(arg.toString(), charsToEscape, '\\') + "\" " + (leftHeaded ? LEFT : RIGHT) + " " + distance;
  }

  public Dependency() {
  };

  public Dependency(IntTaggedWord head, IntTaggedWord arg, boolean leftHeaded, int distance) {
    this.head = head;
    this.arg = arg;
    this.distance = (short) distance;
    this.leftHeaded = leftHeaded;
  }

  public Dependency(int headWord, int headTag, int argWord, int argTag, boolean leftHeaded, int distance) {
    this.head = new IntTaggedWord(headWord, headTag);
    this.arg = new IntTaggedWord(argWord, argTag);
    this.distance = (short) distance;
    this.leftHeaded = leftHeaded;
  }

  private static final long serialVersionUID = 1L;

} // end class Dependency

class MLEDependencyGrammarExtractor extends AbstractTreeExtractor {
  protected Numberer wordNumberer = Numberer.getGlobalNumberer("words");
  protected Numberer tagNumberer = Numberer.getGlobalNumberer("tags");
  protected Counter dependencyCounter = new Counter();
  //private Set dependencies = new HashSet();

  protected TreebankLangParserParams tlpParams;
  protected boolean directional, useDistance, useCoarseDistance;

  public MLEDependencyGrammarExtractor(Options op) {
    tlpParams = op.tlpParams;
    directional = op.directional;
    useDistance = op.distance;
    useCoarseDistance = op.coarseDistance;
  }

  protected void tallyRoot(Tree lt) {
    List deps = MLEDependencyGrammar.treeToDependencyList(lt);
    for (Iterator i = deps.iterator(); i.hasNext();) {
      Dependency dependency = (Dependency) i.next();
      dependencyCounter.incrementCount(dependency, 1.0);
    }
  }

  protected Object formResult() {
    wordNumberer.number(Lexicon.UNKNOWN_WORD);
    MLEDependencyGrammar dg = new MLEDependencyGrammar(tlpParams, directional, useDistance, useCoarseDistance);
    for (Iterator dependencyI = dependencyCounter.keySet().iterator(); dependencyI.hasNext();) {
      Dependency dependency = (Dependency) dependencyI.next();
      dg.addRule(dependency, dependencyCounter.getCount(dependency));
    }
    return dg;
  }

} // end class MLEDependencyGrammarExtractor


class ExhaustiveDependencyParser implements Scorer {

  private static final boolean debug = false;

  private Numberer tagNumberer = Numberer.getGlobalNumberer("tags");
  private Numberer wordNumberer = Numberer.getGlobalNumberer("words");

  private TreeFactory tf;

  private DependencyGrammar dg;
  private Lexicon lex;
  private Options op;
  private TreebankLanguagePack tlp;

  private List sentence;
  private int[] words;

  private float[][][] iScoreH; // headPos, headTag, cornerPosition (non-head)
  private float[][][] oScoreH; // headPos, headTag, cornerPosition (non-head)

  private int[][] rawDistance;
  int[][] binDistance;       // reused in other class, so can't be private
  float[][][][][] headScore;
  float[][][] headStop; // headPos, headTag, split
  private boolean[][][] oPossibleByL;
  private boolean[][][] oPossibleByR;
  private boolean[][][] iPossibleByL;
  private boolean[][][] iPossibleByR;
  private int arraySize = 0;
  private int myMaxLength = -0xDEADBEEF;

  float oScore(int start, int end, int head, int tag) {
    return oScoreH[head][dg.tagBin(tag)][start] + oScoreH[head][dg.tagBin(tag)][end];
  }

  float iScore(int start, int end, int head, int tag) {
    return iScoreH[head][dg.tagBin(tag)][start] + iScoreH[head][dg.tagBin(tag)][end];
  }

  public double oScore(Edge edge) {
    return oScore(edge.start, edge.end, edge.head, edge.tag);
  }

  public double iScore(Edge edge) {
    return iScore(edge.start, edge.end, edge.head, edge.tag);
  }

  public boolean oPossible(Hook hook) {
    return (hook.isPreHook() ? oPossibleByR[hook.end][hook.head][dg.tagBin(hook.tag)] : oPossibleByL[hook.start][hook.head][dg.tagBin(hook.tag)]);
  }

  public boolean iPossible(Hook hook) {
    return (hook.isPreHook() ? iPossibleByR[hook.start][hook.head][dg.tagBin(hook.tag)] : iPossibleByL[hook.end][hook.head][dg.tagBin(hook.tag)]);
  }

  public boolean parse(List sentence) {
    if (Test.verbose) {
      Timing.tick("Starting dependency parse.");
    }
    this.sentence = sentence;
    int length = sentence.size();
    if (length > arraySize) {
      if (length > Test.maxLength + 1 || length >= myMaxLength) {
        throw new OutOfMemoryError("Refusal to create such large arrays.");
      } else {
        try {
          createArrays(length + 1);
        } catch (OutOfMemoryError e) {
          myMaxLength = length;
          if (arraySize > 0) {
            try {
              createArrays(arraySize);
            } catch (OutOfMemoryError e2) {
              throw new RuntimeException("CANNOT EVEN CREATE ARRAYS OF ORIGINAL SIZE!!! " + arraySize);
            }
          }
          throw e;
        }
        arraySize = length + 1;
        if (Test.verbose) {
          System.err.println("Created dparser arrays of size " + arraySize);
        }
      }
    }
    if (Test.verbose) {
      System.err.print("Initializing...");
    }

    // map to words
    words = new int[length];
    int numTags = dg.numTagBins();//tagNumberer.total();
    //System.out.println("\nNumTags: "+numTags);
    //System.out.println(tagNumberer);
    boolean[][] hasTag = new boolean[length][numTags];
    for (int i = 0; i < length; i++) {
      //if (wordNumberer.hasSeen(sentence.get(i).toString()))
      Object o = sentence.get(i);
      if (o instanceof HasWord) {
        o = ((HasWord) o).word();
      }
      words[i] = wordNumberer.number(o.toString());
      //else
      //words[i] = wordNumberer.number(Lexicon.UNKNOWN_WORD);
    }
    for (int head = 0; head < length; head++) {
      for (int tag = 0; tag < numTags; tag++) {
        Arrays.fill(iScoreH[head][tag], Float.NEGATIVE_INFINITY);
        Arrays.fill(oScoreH[head][tag], Float.NEGATIVE_INFINITY);
      }
    }
    for (int head = 0; head < length; head++) {
      for (int loc = 0; loc <= length; loc++) {
        rawDistance[head][loc] = (head >= loc ? head - loc : loc - head - 1);
        binDistance[head][loc] = dg.distanceBin(rawDistance[head][loc]);
      }
    }
    // do tags
    for (int start = 0; start + 1 <= length; start++) {
      String trueTagStr = null;
      if (sentence.get(start) instanceof HasTag) {
        trueTagStr = ((HasTag) sentence.get(start)).tag();
      }
      if (!op.flexiTag) {
        int word = words[start];
        for (Iterator taggingI = lex.ruleIteratorByWord(word, start); taggingI.hasNext();) {
          IntTaggedWord tagging = (IntTaggedWord) taggingI.next();
          if (trueTagStr != null) {
            if (!tlp.basicCategory(tagging.tagString()).equals(trueTagStr)) {
              continue;
            }
          }
          float score = (float) lex.score(tagging, start);
          //iScoreH[start][tag][start] = (op.dcTags ? (float)Test.depWeight*score : 0.0f);
          if (score > Float.NEGATIVE_INFINITY) {
            int tag = tagging.tag;
            iScoreH[start][dg.tagBin(tag)][start] = 0.0f;
            iScoreH[start][dg.tagBin(tag)][start + 1] = 0.0f;
            if (debug) System.err.println("Accepted tagging: " + wordNumberer.object(tagging.word)+"|"+tagNumberer.object(tagging.tag) + ", got score " + score);
          }
        }
      } else {
        // THIS CAN AND SHOULD BE SPED UP BY NULLING IMPOSSIBLE BINS
        // CDM: ALSO THIS BRANCH DOESN'T OBSERVE TRUETAGSTR VARIABLE
        int word = words[start];
        IntTaggedWord tagging = new IntTaggedWord(word, -1);
        for (int tag = 0; tag < tagNumberer.total(); tag++) {
          tagging.tag = (short) tag;
          if (trueTagStr != null) {
            if (!tlp.basicCategory(tagging.tagString()).equals(trueTagStr)) {
              continue;
            }
          }
          if (lex.score(tagging, start) == Float.NEGATIVE_INFINITY) {
            continue;
          }
          if (debug) System.err.println("Accepted tagging: " + wordNumberer.object(tagging.word)+"|"+tagNumberer.object(tagging.tag));
          //iScoreH[start][tag][start] = (float)(op.dcTags ? (float)Test.depWeight*lex.score(new IntTaggedWord(word, tag), start) : 0.0f);
          iScoreH[start][dg.tagBin(tag)][start] = 0.0f;
          iScoreH[start][dg.tagBin(tag)][start + 1] = 0.0f;
        }
      }
    }
    for (int hWord = 0; hWord < length; hWord++) {
      for (int hTag = 0; hTag < numTags; hTag++) {
        hasTag[hWord][hTag] = (iScoreH[hWord][hTag][hWord] + iScoreH[hWord][hTag][hWord + 1] > Float.NEGATIVE_INFINITY);
        Arrays.fill(headStop[hWord][hTag], Float.NEGATIVE_INFINITY);
        for (int aWord = 0; aWord < length; aWord++) {
          for (int dist = 0; dist < dg.numDistBins(); dist++) {
            Arrays.fill(headScore[dist][hWord][hTag][aWord], Float.NEGATIVE_INFINITY);
          }
        }
      }
    }
    // score and cache all pairs -- headScores and stops
    //int hit = 0;
    for (int hWord = 0; hWord < length; hWord++) {
      for (int hTag = 0; hTag < numTags; hTag++) {
        //Arrays.fill(headStopL[hWord][hTag], Float.NEGATIVE_INFINITY);
        //Arrays.fill(headStopR[hWord][hTag], Float.NEGATIVE_INFINITY);
        //Arrays.fill(headStop[hWord][hTag], Float.NEGATIVE_INFINITY);
        if (!hasTag[hWord][hTag]) {
          continue;
        }
        for (int split = 0; split <= length; split++) {
          if (split <= hWord) {
            headStop[hWord][hTag][split] = (float) dg.scoreTB(words[hWord], hTag, -2, -2, false, hWord - split);
            //System.out.println("headstopL " + hWord +" " + hTag + " " + split + " " + headStopL[hWord][hTag][split]); // debugging
          } else {
            headStop[hWord][hTag][split] = (float) dg.scoreTB(words[hWord], hTag, -2, -2, true, split - hWord - 1);
            //System.out.println("headstopR " + hWord +" " + hTag + " " + split + " " + headStopR[hWord][hTag][split]); // debugging
          }
          //hit++;
        }
        //Timing.tick("hWord: "+hWord+" hTag: "+hTag+" piddle count: "+hit);
        for (int aWord = 0; aWord < length; aWord++) {
          if (aWord == hWord) {
            continue;
          }
          boolean leftHeaded = hWord < aWord;
          int start;
          int end;
          if (leftHeaded) {
            start = hWord + 1;
            end = aWord + 1;
          } else {
            start = aWord + 1;
            end = hWord + 1;
          }
          for (int aTag = 0; aTag < numTags; aTag++) {
            if (!hasTag[aWord][aTag]) {
              continue;
            }
            for (int split = start; split < end; split++) {
              // Moved this stuff out two loops- GMA
              //              for (int split = 0; split <= length; split++) {
              // if leftHeaded, go from hWord+1 to aWord
              // else go from aWord+1 to hWord
              //              if ((leftHeaded && (split <= hWord || split > aWord)) ||
              //                      ((!leftHeaded) && (split <= aWord || split > hWord)))
              //                continue;
              int headDistance = rawDistance[hWord][split];
              int binDist = binDistance[hWord][split];
              headScore[binDist][hWord][hTag][aWord][aTag] = (float) dg.scoreTB(words[hWord], hTag, words[aWord], aTag, leftHeaded, headDistance);
              //hit++;
              while (split + 1 < end && binDistance[hWord][split + 1] == binDist) {
                split++;
              }
              if (debug) {
                System.out.println("Dependency headScore: " + wordNumberer.object(words[hWord]) + "/" + tagNumberer.object(hTag) + "[" + hWord + "] -> " + wordNumberer.object(words[aWord]) + "/" + tagNumberer.object(aTag) + "[" + aWord + "] = " + headScore[binDist][hWord][hTag][aWord][aTag]);
              }
            }
          }
        }
      }
    }
    if (Test.verbose) {
      Timing.tick("done.");
      // displayHeadScores();
      System.err.print("Starting insides...");
    }
    // do larger spans
    for (int diff = 2; diff <= length; diff++) {
      for (int start = 0; start + diff <= length; start++) {
        int end = start + diff;
        // left extension
        int endHead = end - 1;
        for (int endTag = 0; endTag < numTags; endTag++) {
          if (!hasTag[endHead][endTag]) {
            continue;
          }
          for (int argHead = start; argHead < endHead; argHead++) {
            for (int argTag = 0; argTag < numTags; argTag++) {
              if (!hasTag[argHead][argTag]) {
                continue;
              }
              float bestScore = iScoreH[endHead][endTag][start];
              for (int split = argHead + 1; split < end; split++) {
                float score = (iScoreH[endHead][endTag][split] + iScoreH[argHead][argTag][start] + iScoreH[argHead][argTag][split] + headScore[binDistance[endHead][split]][endHead][endTag][argHead][argTag] + headStop[argHead][argTag][start] + headStop[argHead][argTag][split]);
                if (score > bestScore) {
                  bestScore = score;
                }
              }
              iScoreH[endHead][endTag][start] = bestScore;
            }
          }
        }
        // right extension
        int startHead = start;
        for (int startTag = 0; startTag < numTags; startTag++) {
          if (!hasTag[startHead][startTag]) {
            continue;
          }
          for (int argHead = start + 1; argHead < end; argHead++) {
            for (int argTag = 0; argTag < numTags; argTag++) {
              if (!hasTag[argHead][argTag]) {
                continue;
              }
              float bestScore = iScoreH[startHead][startTag][end];
              for (int split = start + 1; split <= argHead; split++) {
                float score = (iScoreH[startHead][startTag][split] + iScoreH[argHead][argTag][split] + iScoreH[argHead][argTag][end] + headScore[binDistance[startHead][split]][startHead][startTag][argHead][argTag] + headStop[argHead][argTag][end] + headStop[argHead][argTag][split]);
                if (score > bestScore) {
                  bestScore = score;
                }
              }
              iScoreH[startHead][startTag][end] = bestScore;
            }
          }
        }
      }
    }
    int goalTag = dg.tagBin(tagNumberer.number(Lexicon.BOUNDARY_TAG));
    if (Test.verbose) {
      Timing.tick("done.");
      System.out.println("Dep  parsing " + length + " words (incl. stop): insideScore " + (iScoreH[length - 1][goalTag][0] + iScoreH[length - 1][goalTag][length]));
      System.err.print("Starting outsides...");
    }
    oScoreH[length - 1][goalTag][0] = 0.0f;
    oScoreH[length - 1][goalTag][length] = 0.0f;
    for (int diff = length; diff > 1; diff--) {
      for (int start = 0; start + diff <= length; start++) {
        int end = start + diff;
        // left half
        int endHead = end - 1;
        for (int endTag = 0; endTag < numTags; endTag++) {
          if (!hasTag[endHead][endTag]) {
            continue;
          }
          for (int argHead = start; argHead < endHead; argHead++) {
            for (int argTag = 0; argTag < numTags; argTag++) {
              if (!hasTag[argHead][argTag]) {
                continue;
              }
              for (int split = argHead; split <= endHead; split++) {
                float subScore = (oScoreH[endHead][endTag][start] + headScore[binDistance[endHead][split]][endHead][endTag][argHead][argTag] + headStop[argHead][argTag][start] + headStop[argHead][argTag][split]);
                float scoreRight = (subScore + iScoreH[argHead][argTag][start] + iScoreH[argHead][argTag][split]);
                float scoreMid = (subScore + iScoreH[argHead][argTag][start] + iScoreH[endHead][endTag][split]);
                float scoreLeft = (subScore + iScoreH[argHead][argTag][split] + iScoreH[endHead][endTag][split]);
                if (scoreRight > oScoreH[endHead][endTag][split]) {
                  oScoreH[endHead][endTag][split] = scoreRight;
                }
                if (scoreMid > oScoreH[argHead][argTag][split]) {
                  oScoreH[argHead][argTag][split] = scoreMid;
                }
                if (scoreLeft > oScoreH[argHead][argTag][start]) {
                  oScoreH[argHead][argTag][start] = scoreLeft;
                }
              }
            }
          }
        }
        // right half
        int startHead = start;
        for (int startTag = 0; startTag < numTags; startTag++) {
          if (!hasTag[startHead][startTag]) {
            continue;
          }
          for (int argHead = startHead + 1; argHead < end; argHead++) {
            for (int argTag = 0; argTag < numTags; argTag++) {
              if (!hasTag[argHead][argTag]) {
                continue;
              }
              for (int split = startHead + 1; split <= argHead; split++) {
                float subScore = (oScoreH[startHead][startTag][end] + headScore[binDistance[startHead][split]][startHead][startTag][argHead][argTag] + headStop[argHead][argTag][split] + headStop[argHead][argTag][end]);
                float scoreLeft = (subScore + iScoreH[argHead][argTag][split] + iScoreH[argHead][argTag][end]);
                float scoreMid = (subScore + iScoreH[startHead][startTag][split] + iScoreH[argHead][argTag][end]);
                float scoreRight = (subScore + iScoreH[startHead][startTag][split] + iScoreH[argHead][argTag][split]);
                if (scoreLeft > oScoreH[startHead][startTag][split]) {
                  oScoreH[startHead][startTag][split] = scoreLeft;
                }
                if (scoreMid > oScoreH[argHead][argTag][split]) {
                  oScoreH[argHead][argTag][split] = scoreMid;
                }
                if (scoreRight > oScoreH[argHead][argTag][end]) {
                  oScoreH[argHead][argTag][end] = scoreRight;
                }
              }
            }
          }
        }
      }
    }
    if (Test.verbose) {
      Timing.tick("done.");
      System.err.print("Starting half-filters...");
    }
    for (int loc = 0; loc <= length; loc++) {
      for (int head = 0; head < length; head++) {
        Arrays.fill(iPossibleByL[loc][head], false);
        Arrays.fill(iPossibleByR[loc][head], false);
        Arrays.fill(oPossibleByL[loc][head], false);
        Arrays.fill(oPossibleByR[loc][head], false);
      }
    }
    for (int head = 0; head < length; head++) {
      for (int tag = 0; tag < numTags; tag++) {
        if (!hasTag[head][tag]) {
          continue;
        }
        for (int start = 0; start <= head; start++) {
          for (int end = head + 1; end <= length; end++) {
            if (iScoreH[head][tag][start] + iScoreH[head][tag][end] > Float.NEGATIVE_INFINITY && oScoreH[head][tag][start] + oScoreH[head][tag][end] > Float.NEGATIVE_INFINITY) {
              iPossibleByR[end][head][tag] = true;
              iPossibleByL[start][head][tag] = true;
              oPossibleByR[end][head][tag] = true;
              oPossibleByL[start][head][tag] = true;
            }
          }
        }
      }
    }
    if (Test.verbose) {
      Timing.tick("done.");
    }
    return hasParse();
  }

  public boolean hasParse() {
    int length = sentence.size();
    if (length > arraySize) {
      return false;
    }
    int goalTag = tagNumberer.number(Lexicon.BOUNDARY_TAG);
    return (iScore(0, length, length - 1, goalTag) > Float.NEGATIVE_INFINITY);
  }

  /**
   * This displays a headScore matrix, which will be valid after parsing
   * a sentence.  Unclear yet whether this is valid/useful [cdm].
   */
  public void displayHeadScores() {
    int numTags = tagNumberer.total();
    System.out.println("---- headScore matrix (head x dep, best tags) ----");
    System.out.print(StringUtils.padOrTrim("", 6));
    for (int i = 0; i < words.length; i++) {
      System.out.print(" " + StringUtils.padOrTrim(wordNumberer.object(words[i]).toString(), 2));
    }
    System.out.println();
    for (int hWord = 0; hWord < words.length; hWord++) {
      System.out.print(StringUtils.padOrTrim(wordNumberer.object(words[hWord]).toString(), 6));
      int bigBD = -1, bigHTag = -1, bigATag = -1;
      for (int aWord = 0; aWord < words.length; aWord++) {
        // we basically just max of all the variables, but for distance > 0, we
        // include a factor for generating something at distance 0, or else
        // the result is too whacked out to be useful
        float biggest = Float.NEGATIVE_INFINITY;
        for (int bd = 0; bd < dg.numDistBins(); bd++) {
          for (int hTag = 0; hTag < numTags; hTag++) {
            /*
            float penalty = 0.0f;
            if (bd != 0) {
              penalty = (float) dg.score(words[hWord], hTag, -2, -2, aWord > hWord, 0);
              penalty = (float) Math.log(1.0 - Math.exp(penalty));
            }
            for (int aTag = 0; aTag < numTags; aTag++) {
              if (headScore[bd][hWord][hTag][aWord][aTag] + penalty > biggest) {
                biggest = headScore[bd][hWord][hTag][aWord][aTag] + penalty;
            */
            for (int aTag = 0; aTag < numTags; aTag++) {
              if (headScore[bd][hWord][dg.tagBin(hTag)][aWord][dg.tagBin(aTag)] > biggest) {
                biggest = headScore[bd][hWord][dg.tagBin(hTag)][aWord][dg.tagBin(aTag)];
                bigBD = bd;
                bigHTag = hTag;
                bigATag = aTag;
              }
            }
          }
        }
        if (Float.isInfinite(biggest)) {
          System.out.print(" " + StringUtils.padOrTrim("in", 2));
        } else {
          int score = Math.round(Math.abs(headScore[bigBD][hWord][dg.tagBin(bigHTag)][aWord][dg.tagBin(bigATag)]));
          System.out.print(" " + StringUtils.padOrTrim(Integer.toString(score), 2));
        }
      }
      System.out.println();
    }
  }

  private static final double TOL = 1e-5;

  private static boolean matches(double x, double y) {
    return (Math.abs(x - y) / (Math.abs(x) + Math.abs(y) + 1e-10) < TOL);
  }

  private Tree extractBestParse(int start, int end, int hWord, int hTag) {
    final boolean verbose = true;
    //System.out.println("Span "+start+" to "+end+" word "+wordNumberer.object(words[hWord])+"/"+hWord+" tag "+tagNumberer.object(hTag)+"/"+hTag+" score "+iScore[start][end][hWord][hTag]);
    String headWordStr = (String) wordNumberer.object(words[hWord]);
    String headTagStr = (String) tagNumberer.object(hTag);
    Label headLabel = new CategoryWordTag(headWordStr, headWordStr, headTagStr);
    List children = new ArrayList();
    // deal with span 1
    if (end - start == 1) {
      Tree leaf = tf.newLeaf(new Word(headWordStr));
      children.add(leaf);
      return tf.newTreeNode(headLabel, children);
    }
    // find backtrace
    double bestScore = iScore(start, end, hWord, hTag);
    for (int split = start + 1; split < end; split++) {
      int binD = binDistance[hWord][split];
      if (hWord < split) {
        for (int aWord = split; aWord < end; aWord++) {
          for (int aTag = 0; aTag < tagNumberer.total(); aTag++) {
            if (matches(iScore(start, split, hWord, hTag) + iScore(split, end, aWord, aTag) + headScore[binD][hWord][dg.tagBin(hTag)][aWord][dg.tagBin(aTag)] + headStop[aWord][dg.tagBin(aTag)][split] + headStop[aWord][dg.tagBin(aTag)][end], bestScore)) {
              if (verbose) {
                //String argWordStr=(String)wordNumberer.object(words[aWord]);
                //String argTagStr = (String)tagNumberer.object(aTag);
                //System.out.println(headWordStr+"|"+headTagStr+" -> "+argWordStr+"|"+argTagStr+" "+((int)(10.0*pairScore[hWord][hTag][aWord][aTag]))/10.0);
              }
              // build it
              children.add(extractBestParse(start, split, hWord, hTag));
              children.add(extractBestParse(split, end, aWord, aTag));
              return tf.newTreeNode(headLabel, children);
            }
          }
        }
      } else {
        for (int aWord = start; aWord < split; aWord++) {
          for (int aTag = 0; aTag < tagNumberer.total(); aTag++) {
            if (matches(iScore(start, split, aWord, aTag) + iScore(split, end, hWord, hTag) + headScore[binD][hWord][dg.tagBin(hTag)][aWord][dg.tagBin(aTag)] + headStop[aWord][dg.tagBin(aTag)][start] + headStop[aWord][dg.tagBin(aTag)][split], bestScore)) {
              if (verbose) {
                //String argWordStr=(String)wordNumberer.object(words[aWord]);
                //String argTagStr = (String)tagNumberer.object(aTag);
                //System.out.println(headWordStr+"|"+headTagStr+" -> "+argWordStr+"|"+argTagStr+" "+((int)(10.0*pairScore[hWord][hTag][aWord][aTag]))/10.0);
              }
              children.add(extractBestParse(start, split, aWord, aTag));
              children.add(extractBestParse(split, end, hWord, hTag));
              return tf.newTreeNode(headLabel, children);
              // build it
            }
          }
        }
      }
    }
    System.err.println("Problem in ExhaustiveDependencyParser::extractBestParse");
    return null;
  }

  private Tree flatten(Tree tree) {
    if (tree.isLeaf() || tree.isPreTerminal()) {
      return tree;
    }
    List newChildren = new ArrayList();
    Tree[] children = tree.children();
    for (int c = 0; c < children.length; c++) {
      Tree child = children[c];
      Tree newChild = flatten(child);
      if (!newChild.isPreTerminal() && newChild.label().toString().equals(tree.label().toString())) {
        newChildren.addAll(newChild.getChildrenAsList());
      } else {
        newChildren.add(newChild);
      }
    }
    return tf.newTreeNode(tree.label(), newChildren);
  }

  public Tree getBestParse() {
    if (!hasParse()) {
      return null;
    }
    return flatten(extractBestParse(0, words.length, words.length - 1, tagNumberer.number(Lexicon.BOUNDARY_TAG)));
  }

  public ExhaustiveDependencyParser(DependencyGrammar dg, Lexicon lex, Options op) {
    this.dg = dg;
    this.lex = lex;
    this.op = op;
    this.tlp = op.langpack();
    tf = new LabeledScoredTreeFactory(new StringLabelFactory());
  }

  private void createArrays(int length) {
    iScoreH = oScoreH = headStop = null;
    iPossibleByL = iPossibleByR = oPossibleByL = oPossibleByR = null;
    headScore = null;
    rawDistance = binDistance = null;

    int tagNum = dg.numTagBins();//tagNumberer.total();

    iScoreH = new float[length + 1][tagNum][length + 1];
    oScoreH = new float[length + 1][tagNum][length + 1];
    iPossibleByL = new boolean[length + 1][length + 1][tagNum];
    iPossibleByR = new boolean[length + 1][length + 1][tagNum];
    oPossibleByL = new boolean[length + 1][length + 1][tagNum];
    oPossibleByR = new boolean[length + 1][length + 1][tagNum];
    headScore = new float[dg.numDistBins()][length][tagNum][length][tagNum];
    headStop = new float[length + 1][tagNum][length + 1];
    rawDistance = new int[length + 1][length + 1];
    binDistance = new int[length + 1][length + 1];
  }
}

// A*

interface Scorer {
  public double oScore(Edge edge);

  public double iScore(Edge edge);

  public boolean oPossible(Hook hook);

  public boolean iPossible(Hook hook);

  public boolean parse(List words);
}

class TwinScorer implements Scorer {
  private Scorer scorer1;
  private Scorer scorer2;

  public double oScore(Edge edge) {
    return scorer1.oScore(edge) + scorer2.oScore(edge);
  }

  public double iScore(Edge edge) {
    return scorer1.iScore(edge) + scorer2.iScore(edge);
  }

  public boolean oPossible(Hook hook) {
    return scorer1.oPossible(hook) && scorer2.oPossible(hook);
  }

  public boolean iPossible(Hook hook) {
    return scorer1.iPossible(hook) && scorer2.iPossible(hook);
  }

  public boolean parse(List words) {
    boolean b1 = scorer1.parse(words);
    boolean b2 = scorer2.parse(words);
    return (b1 && b2);
  }

  public TwinScorer(Scorer scorer1, Scorer scorer2) {
    this.scorer1 = scorer1;
    this.scorer2 = scorer2;
  }
}

class MaxScorer implements Scorer {
  private Scorer scorer1;
  private Scorer scorer2;

  public double oScore(Edge edge) {
    return SloppyMath.min(scorer1.oScore(edge), scorer2.oScore(edge));
  }

  public double iScore(Edge edge) {
    return SloppyMath.min(scorer1.iScore(edge), scorer2.iScore(edge));
  }

  public boolean oPossible(Hook hook) {
    return scorer1.oPossible(hook) && scorer2.oPossible(hook);
  }

  public boolean iPossible(Hook hook) {
    return scorer1.iPossible(hook) && scorer2.iPossible(hook);
  }

  public boolean parse(List words) {
    boolean b1 = scorer1.parse(words);
    boolean b2 = scorer2.parse(words);
    return (b1 && b2);
  }

  public MaxScorer(Scorer scorer1, Scorer scorer2) {
    this.scorer1 = scorer1;
    this.scorer2 = scorer2;
  }
}


// A* Guts


class Interner {
  private Map oToO = new HashMap();

  public Object intern(Object o) {
    Object i = oToO.get(o);
    if (i == null) {
      i = o;
      oToO.put(o, o);
    }
    return i;
  }
}


abstract class Item implements Scored {
  public int start;
  public int end;
  public int state;
  public int head;
  public int tag;
  public Edge backEdge;
  public double iScore = Double.NEGATIVE_INFINITY;
  public double oScore = Double.NEGATIVE_INFINITY;

  public double score() {
    return iScore + (Test.exhaustiveTest ? 0.0 : oScore);
  }

  public boolean isEdge() {
    return false;
  }

  public boolean isPreHook() {
    return false;
  }

  public boolean isPostHook() {
    return false;
  }
}


class Edge extends Item {
  public Hook backHook;

  public boolean isEdge() {
    return true;
  }

  public String toString() {
    return "Edge(" + Numberer.getGlobalNumberer("states").object(state) + ":" + start + "-" + end + "," + head + "/" + Numberer.getGlobalNumberer("tags").object(tag) + ")";
  }

  public int hashCode() {
    return (state << 1) ^ (head << 8) ^ (tag << 16) ^ (start << 4) ^ (end << 24);
  }

  public boolean equals(Object o) {
    if (this == o) {
      return true;
    } else if (o instanceof Edge) {
      Edge e = (Edge) o;
      if (state == e.state && head == e.head && tag == e.tag && start == e.start && end == e.end) {
        return true;
      }
      return false;
    } else {
      return false;
    }
  }
}


class Hook extends Item {
  public int subState;

  public boolean isPreHook() {
    return head < start;
  }

  public boolean isPostHook() {
    return head >= end;
  }

  public String toString() {
    return (isPreHook() ? "Pre" : "Post") + "Hook(" + Numberer.getGlobalNumberer("states").object(state) + "|" + Numberer.getGlobalNumberer("states").object(subState) + ":" + start + "-" + end + "," + head + "/" + Numberer.getGlobalNumberer("tags").object(tag) + ")";
  }

  public int hashCode() {
    return 1 + (state << 14) ^ (subState << 16) ^ (head << 22) ^ (tag << 27) ^ (start << 1) ^ (end << 7);
  }

  public boolean equals(Object o) {
    //System.out.println("\nCHECKING HOOKS: "+this+" vs "+o);
    if (this == o) {
      return true;
    }
    try {
      if (((Item) o).isEdge()) {
        return false;
      }
      Hook e = (Hook) o;
      if (state == e.state && subState == e.subState && head == e.head && tag == e.tag && start == e.start && end == e.end) {
        return true;
      }
      return false;
    } catch (ClassCastException e) {
      return false;
    }
  }
}


class HookChart {
  private Map registeredPreHooks = new HashMap();
  private Map registeredPostHooks = new HashMap();
  private Map registeredEdgesByLeftIndex = new HashMap();
  private Map registeredEdgesByRightIndex = new HashMap();

  private Map realEdgesByL = new HashMap();
  private Map realEdgesByR = new HashMap();
  private Set builtLIndexes = new HashSet();
  private Set builtRIndexes = new HashSet();

  private Interner interner = new Interner();

  private static class ChartIndex {
    public int state;
    public int head;
    public int tag;
    public int loc;

    public int hashCode() {
      return state ^ (head << 8) ^ (tag << 16) ^ (loc << 24);
    }

    public boolean equals(Object o) {
      if (this == o) {
        return true;
      }
      try {
        ChartIndex ci = (ChartIndex) o;
        if (state == ci.state && head == ci.head && tag == ci.tag && loc == ci.loc) {
          return true;
        }
        return false;
      } catch (ClassCastException e) {
        return false;
      }
    }
  }

  private static class WeakChartIndex {
    public int state;
    public int loc;

    public int hashCode() {
      return state ^ (loc << 16);
    }

    public boolean equals(Object o) {
      if (this == o) {
        return true;
      }
      try {
        WeakChartIndex ci = (WeakChartIndex) o;
        if (state == ci.state && loc == ci.loc) {
          return true;
        }
        return false;
      } catch (ClassCastException e) {
        return false;
      }
    }
  }

  private static final Collection empty = new ArrayList(0);
  private WeakChartIndex tempWeakIndex = new WeakChartIndex();

  public void registerEdgeIndexes(Edge edge) {
    tempIndex.state = edge.state;
    tempIndex.head = edge.head;
    tempIndex.tag = edge.tag;
    tempIndex.loc = edge.start;
    ChartIndex index = (ChartIndex) interner.intern(tempIndex);
    if (index == tempIndex) {
      tempIndex = new ChartIndex();
    }
    builtLIndexes.add(index);
    //System.out.println("Edge registered: "+edge);
    tempIndex.state = edge.state;
    tempIndex.head = edge.head;
    tempIndex.tag = edge.tag;
    tempIndex.loc = edge.end;
    index = (ChartIndex) interner.intern(tempIndex);
    if (index == tempIndex) {
      tempIndex = new ChartIndex();
    }
    builtRIndexes.add(index);
  }

  public void registerRealEdge(Edge edge) {
    tempWeakIndex.state = edge.state;
    tempWeakIndex.loc = edge.start;
    WeakChartIndex index = (WeakChartIndex) interner.intern(tempWeakIndex);
    if (index == tempWeakIndex) {
      tempWeakIndex = new WeakChartIndex();
    }
    insert(realEdgesByL, index, edge);
    tempWeakIndex.state = edge.state;
    tempWeakIndex.loc = edge.end;
    index = (WeakChartIndex) interner.intern(tempWeakIndex);
    if (index == tempWeakIndex) {
      tempWeakIndex = new WeakChartIndex();
    }
    insert(realEdgesByR, index, edge);
  }

  public boolean isBuiltL(int state, int start, int head, int tag) {
    tempIndex.state = state;
    tempIndex.head = head;
    tempIndex.tag = tag;
    tempIndex.loc = start;
    return builtLIndexes.contains(tempIndex);
  }

  public boolean isBuiltR(int state, int end, int head, int tag) {
    tempIndex.state = state;
    tempIndex.head = head;
    tempIndex.tag = tag;
    tempIndex.loc = end;
    return builtRIndexes.contains(tempIndex);
  }

  public Collection getRealEdgesWithL(int state, int start) {
    tempWeakIndex.state = state;
    tempWeakIndex.loc = start;
    Collection edges = (Collection) realEdgesByL.get(tempWeakIndex);
    if (edges == null) {
      return empty;
    }
    return edges;
  }

  public Collection getRealEdgesWithR(int state, int end) {
    tempWeakIndex.state = state;
    tempWeakIndex.loc = end;
    Collection edges = (Collection) realEdgesByR.get(tempWeakIndex);
    if (edges == null) {
      return empty;
    }
    return edges;
  }

  public Collection getPreHooks(Edge edge) {
    tempIndex.state = edge.state;
    tempIndex.head = edge.head;
    tempIndex.tag = edge.tag;
    tempIndex.loc = edge.end;
    Collection result = (Collection) registeredPreHooks.get(tempIndex);
    if (result == null) {
      result = empty;
    }
    //System.out.println("For "+edge+" returning "+result.size()+" pre hooks");
    return result;
  }

  public Collection getPostHooks(Edge edge) {
    tempIndex.state = edge.state;
    tempIndex.head = edge.head;
    tempIndex.tag = edge.tag;
    tempIndex.loc = edge.start;
    Collection result = (Collection) registeredPostHooks.get(tempIndex);
    if (result == null) {
      result = empty;
    }
    //System.out.println("For "+edge+" returning "+result.size()+" post hooks");
    return result;
  }

  public Collection getEdges(Hook hook) {
    tempIndex.state = hook.subState;
    tempIndex.head = hook.head;
    tempIndex.tag = hook.tag;
    Collection result;
    if (hook.isPreHook()) {
      tempIndex.loc = hook.start;
      result = (Collection) registeredEdgesByRightIndex.get(tempIndex);
    } else {
      tempIndex.loc = hook.end;
      result = (Collection) registeredEdgesByLeftIndex.get(tempIndex);
    }
    if (result == null) {
      result = empty;
    }
    //System.out.println("For "+hook+" returning "+result.size()+" edges");
    return result;
  }

  private void insert(Map map, Object index, Object item) {
    List list = (List) map.get(index);
    if (list == null) {
      list = new ArrayList();
      map.put(index, list);
    }
    list.add(item);
  }

  ChartIndex tempIndex = new ChartIndex();

  public void addEdge(Edge edge) {
    tempIndex.state = edge.state;
    tempIndex.head = edge.head;
    tempIndex.tag = edge.tag;
    // left index
    tempIndex.loc = edge.start;
    ChartIndex index = (ChartIndex) interner.intern(tempIndex);
    if (index == tempIndex) {
      tempIndex = new ChartIndex();
      tempIndex.state = edge.state;
      tempIndex.head = edge.head;
      tempIndex.tag = edge.tag;
    }
    insert(registeredEdgesByLeftIndex, index, edge);
    tempIndex.loc = edge.end;
    index = (ChartIndex) interner.intern(tempIndex);
    if (index == tempIndex) {
      tempIndex = new ChartIndex();
    }
    insert(registeredEdgesByRightIndex, index, edge);
  }

  public void addHook(Hook hook) {
    Map map;
    tempIndex.state = hook.subState;
    tempIndex.head = hook.head;
    tempIndex.tag = hook.tag;
    if (hook.isPreHook()) {
      tempIndex.loc = hook.start;
      map = registeredPreHooks;
    } else {
      tempIndex.loc = hook.end;
      map = registeredPostHooks;
    }
    ChartIndex index = (ChartIndex) interner.intern(tempIndex);
    if (index == tempIndex) {
      tempIndex = new ChartIndex();
    }
    insert(map, index, hook);
  }
} // HookChart


class N5BiLexPCFGParser extends BiLexPCFGParser {
  protected void relaxTempHook() {
    relaxHook1++;
    if (verbose) {
      System.out.println("Considering: " + tempHook + " iP: " + scorer.iPossible(tempHook) + " oP: " + scorer.oPossible(tempHook));
    }
    if (!Test.exhaustiveTest) {
      if (!scorer.oPossible(tempHook) || !scorer.iPossible(tempHook)) {
        return;
      }
    }
    relaxHook2++;
    Hook resultHook = tempHook;
    //Hook resultHook = (Hook)interner.intern(tempHook);
    if (verbose) {
      System.out.println("Formed " + resultHook + " " + (resultHook == tempHook ? "new" : "old") + " " + tempHook.iScore + " was " + resultHook.iScore);
    }
    if (resultHook == tempHook) {
      relaxHook3++;
      tempHook = new Hook();
      processHook(resultHook);
      builtHooks++;
    }
  }

  N5BiLexPCFGParser(Scorer scorer, ExhaustivePCFGParser fscorer, ExhaustiveDependencyParser leach, BinaryGrammar bg, UnaryGrammar ug, DependencyGrammar dg, Lexicon lex, Options op) {
    super(scorer, fscorer, leach, bg, ug, dg, lex, op, new NullGrammarProjection(bg, ug));
  }

  N5BiLexPCFGParser(Scorer scorer, ExhaustivePCFGParser fscorer, ExhaustiveDependencyParser leach, BinaryGrammar bg, UnaryGrammar ug, DependencyGrammar dg, Lexicon lex, Options op, GrammarProjection proj) {
    super(scorer, fscorer, leach, bg, ug, dg, lex, op, proj);
  }
}


class BiLexPCFGParser implements ViterbiParser {

  protected static final boolean verbose = false;

  protected HookChart chart;
  protected Heap agenda;
  protected int length;
  protected int[] words;
  protected Edge goal;
  protected Interner interner;
  protected Scorer scorer;
  protected ExhaustivePCFGParser fscorer;
  protected ExhaustiveDependencyParser dparser;
  protected GrammarProjection projection;
  //pair dep scores

  protected BinaryGrammar bg;
  protected UnaryGrammar ug;
  protected DependencyGrammar dg;
  protected Lexicon lex;
  protected Options op;
  protected List[] taggedWordList;

  protected Numberer wordNumberer = Numberer.getGlobalNumberer("words");
  protected Numberer tagNumberer = Numberer.getGlobalNumberer("tags");
  protected Numberer stateNumberer = Numberer.getGlobalNumberer("states");

  protected TreeFactory tf = new LabeledScoredTreeFactory();

  // temp
  protected long relaxHook1 = 0;
  protected long relaxHook2 = 0;
  protected long relaxHook3 = 0;
  protected long relaxHook4 = 0;

  protected long builtHooks = 0;
  protected long builtEdges = 0;
  protected long extractedHooks = 0;
  protected long extractedEdges = 0;


  private static final double TOL = 1e-10;

  protected static boolean better(double x, double y) {
    return ((x - y) / (Math.abs(x) + Math.abs(y) + 1e-100) > TOL);
  }


  protected Tree extractParse(Edge edge) {
    String head = (String) wordNumberer.object(words[edge.head]);
    String tag = (String) tagNumberer.object(edge.tag);
    String state = (String) stateNumberer.object(edge.state);
    Label label = new CategoryWordTag(state, head, tag);
    if (edge.backEdge == null && edge.backHook == null) {
      // leaf, but needs word terminal
      List childList = new ArrayList(1);
      childList.add(tf.newLeaf(new StringLabel(head)));
      return tf.newTreeNode(label, childList);
    }
    List children = new ArrayList();
    if (edge.backHook == null) {
      // unary
      children.add(extractParse(edge.backEdge));
      return tf.newTreeNode(label, children);
    }
    // binary
    if (edge.backHook.isPreHook()) {
      children.add(extractParse(edge.backEdge));
      children.add(extractParse(edge.backHook.backEdge));
    } else {
      children.add(extractParse(edge.backHook.backEdge));
      children.add(extractParse(edge.backEdge));
    }
    return tf.newTreeNode(label, children);
  }

  public Tree getBestParse() {
    return extractParse(goal);
  }

  protected Edge tempEdge = new Edge();

  protected void combine(Edge edge, Hook hook) {
    if (verbose) {
      System.out.println("Combining: " + edge + " and " + hook);
    }
    // make result edge
    if (hook.isPreHook()) {
      tempEdge.start = edge.start;
      tempEdge.end = hook.end;
    } else {
      tempEdge.start = hook.start;
      tempEdge.end = edge.end;
    }
    tempEdge.state = hook.state;
    tempEdge.head = hook.head;
    tempEdge.tag = hook.tag;
    tempEdge.iScore = hook.iScore + edge.iScore;
    tempEdge.backEdge = edge;
    tempEdge.backHook = hook;
    relaxTempEdge();
  }

  protected void relaxTempEdge() {
    // if (tempEdge.iScore > scorer.iScore(tempEdge)+1e-4) {
    //   System.out.println(tempEdge+" has i "+tempEdge.iScore+" iE "+scorer.iScore(tempEdge));
    // }
    Edge resultEdge = (Edge) interner.intern(tempEdge);
    if (verbose) {
      System.out.println("Formed " + resultEdge + " " + (resultEdge == tempEdge ? "new" : "old") + " " + tempEdge.iScore + " was " + resultEdge.iScore + " better? " + better(tempEdge.iScore, resultEdge.iScore));
    }
    if (resultEdge == tempEdge) {
      //System.out.println(resultEdge+" was new.");
      tempEdge = new Edge();
      discoverEdge(resultEdge);
    } else {
      if (better(tempEdge.iScore, resultEdge.iScore) && resultEdge.oScore > Double.NEGATIVE_INFINITY) {
        double back = resultEdge.iScore;
        Edge backE = resultEdge.backEdge;
        Hook backH = resultEdge.backHook;
        resultEdge.iScore = tempEdge.iScore;
        resultEdge.backEdge = tempEdge.backEdge;
        resultEdge.backHook = tempEdge.backHook;
        try {
          agenda.decreaseKey(resultEdge);
        } catch (NullPointerException e) {
          if (false) {
            System.out.println("");
            System.out.println("Old backEdge: " + backE + " i " + backE.iScore + " o " + backE.oScore + " s " + backE.score());
            System.out.println("Old backEdge: " + backE + " iE " + scorer.iScore(backE));
            System.out.println("Old backHook: " + backH + " i " + backH.iScore + " o " + backH.oScore + " s " + backH.score());
            System.out.println("New backEdge: " + tempEdge.backEdge + " i " + tempEdge.backEdge.iScore + " o " + tempEdge.backEdge.oScore + " s " + tempEdge.backEdge.score());
            System.out.println("New backEdge: " + tempEdge.backEdge + " iE " + scorer.iScore(tempEdge.backEdge));
            System.out.println("New backHook: " + tempEdge.backHook + " i " + tempEdge.backHook.iScore + " o " + tempEdge.backHook.oScore + " s " + tempEdge.backHook.score());
            System.out.println("ERROR: Formed " + resultEdge + " i " + tempEdge.iScore + " o " + resultEdge.oScore + " s " + resultEdge.score());
            System.out.println("ERROR: Formed " + resultEdge + " " + (resultEdge == tempEdge ? "new" : "old") + " " + tempEdge.iScore + " was " + back + " better? " + better(tempEdge.iScore, back));
          }
        }
      }
    }
  }

  protected void discoverEdge(Edge edge) {
    // create new edge
    edge.oScore = scorer.oScore(edge);
    agenda.add(edge);
    builtEdges++;
  }

  protected void discoverHook(Hook hook) {
    hook.oScore = buildOScore(hook);
    if (hook.oScore == Double.NEGATIVE_INFINITY) {
      relaxHook4++;
    }
    builtHooks++;
    agenda.add(hook);
  }

  protected Edge iTemp = new Edge();
  protected Edge oTemp = new Edge();

  protected double buildOScore(Hook hook) {
    double bestOScore = Double.NEGATIVE_INFINITY;
    iTemp.head = hook.head;
    iTemp.tag = hook.tag;
    iTemp.state = hook.subState;
    oTemp.head = hook.head;
    oTemp.tag = hook.tag;
    oTemp.state = hook.state;
    if (hook.isPreHook()) {
      iTemp.end = hook.start;
      oTemp.end = hook.end;
      for (int start = 0; start <= hook.head; start++) {
        iTemp.start = start;
        oTemp.start = start;
        double oScore = scorer.oScore(oTemp) + scorer.iScore(iTemp);
        //System.out.println("Score for "+hook+" is i "+iTemp+" ("+scorer.iScore(iTemp)+") o "+oTemp+" ("+scorer.oScore(oTemp)+")");
        bestOScore = SloppyMath.max(bestOScore, oScore);
      }
    } else {
      iTemp.start = hook.end;
      oTemp.start = hook.start;
      for (int end = hook.head + 1; end <= length; end++) {
        iTemp.end = end;
        oTemp.end = end;
        double oScore = scorer.oScore(oTemp) + scorer.iScore(iTemp);
        bestOScore = SloppyMath.max(bestOScore, oScore);
      }
    }
    return bestOScore;
  }

  protected Hook tempHook = new Hook();

  protected void projectHooks(Edge edge) {
    // form hooks
    // POST HOOKS
    //for (Iterator rI = bg.ruleIteratorByLeftChild(edge.state);
    //      rI.hasNext(); ) {
    List ruleList = bg.ruleListByLeftChild(edge.state);
    for (int r = 0; r < ruleList.size(); r++) {
      //BinaryRule br = (BinaryRule)rI.next();
      BinaryRule br = (BinaryRule) ruleList.get(r);
      if (!fscorer.oPossibleL(project(br.parent), edge.start) || !fscorer.iPossibleL(project(br.rightChild), edge.end)) {
        if (!Test.exhaustiveTest) {
          continue;
        }
      }
      for (int head = edge.end; head < length; head++) {
        for (Iterator iTWI = taggedWordList[head].iterator(); iTWI.hasNext();) {
          IntTaggedWord iTW = (IntTaggedWord) iTWI.next();
          int tag = iTW.tag;
          tempHook.start = edge.start;
          tempHook.end = edge.end;
          tempHook.head = head;
          tempHook.tag = tag;
          tempHook.state = br.parent;
          tempHook.subState = br.rightChild;
          if (!chart.isBuiltL(tempHook.subState, tempHook.end, tempHook.head, tempHook.tag)) {
            continue;
          }
          tempHook.iScore = edge.iScore + br.score + dparser.headScore[dparser.binDistance[head][edge.end]][head][dg.tagBin(tag)][edge.head][dg.tagBin(edge.tag)] + dparser.headStop[edge.head][dg.tagBin(edge.tag)][edge.start] + dparser.headStop[edge.head][dg.tagBin(edge.tag)][edge.end];
          tempHook.backEdge = edge;
          relaxTempHook();
        }
      }
    }
    // PRE HOOKS
    //for (Iterator rI = bg.ruleIteratorByRightChild(edge.state);
    //     rI.hasNext(); ) {
    ruleList = bg.ruleListByRightChild(edge.state);
    for (int r = 0; r < ruleList.size(); r++) {
      //BinaryRule br = (BinaryRule)rI.next();
      BinaryRule br = (BinaryRule) ruleList.get(r);
      if (!fscorer.oPossibleR(project(br.parent), edge.end) || !fscorer.iPossibleR(project(br.leftChild), edge.start)) {
        if (!Test.exhaustiveTest) {
          continue;
        }
      }
      for (int head = 0; head < edge.start; head++) {
        for (Iterator iTWI = taggedWordList[head].iterator(); iTWI.hasNext();) {
          IntTaggedWord iTW = (IntTaggedWord) iTWI.next();
          int tag = iTW.tag;
          tempHook.start = edge.start;
          tempHook.end = edge.end;
          tempHook.head = head;
          tempHook.tag = tag;
          tempHook.state = br.parent;
          tempHook.subState = br.leftChild;
          if (!chart.isBuiltR(tempHook.subState, tempHook.start, tempHook.head, tempHook.tag)) {
            continue;
          }
          tempHook.iScore = edge.iScore + br.score + dparser.headScore[dparser.binDistance[head][edge.start]][head][dg.tagBin(tag)][edge.head][dg.tagBin(edge.tag)] + dparser.headStop[edge.head][dg.tagBin(edge.tag)][edge.start] + dparser.headStop[edge.head][dg.tagBin(edge.tag)][edge.end];
          tempHook.backEdge = edge;
          relaxTempHook();
        }
      }
    }
  }

  protected void registerReal(Edge real) {
    chart.registerRealEdge(real);
  }

  protected void triggerHooks(Edge edge) {
    // we might have built a synth edge, enabling some old real edges to project hooks
    boolean newL = !chart.isBuiltL(edge.state, edge.start, edge.head, edge.tag);
    boolean newR = !chart.isBuiltR(edge.state, edge.end, edge.head, edge.tag);
    //if (newL)
    //  System.out.println("Triggering on L: "+edge);
    //if (newR)
    //  System.out.println("Triggering on R: "+edge);
    chart.registerEdgeIndexes(edge);
    if (newR) {
      // PRE HOOKS
      BinaryRule[] rules = bg.splitRulesWithLC(edge.state);
      for (int i = 0; i < rules.length; i++) {
        BinaryRule br = rules[i];
        Collection realEdges = chart.getRealEdgesWithL(br.rightChild, edge.end);
        for (Iterator realI = realEdges.iterator(); realI.hasNext();) {
          Edge real = (Edge) realI.next();
          tempHook.start = real.start;
          tempHook.end = real.end;
          tempHook.state = br.parent;
          tempHook.subState = br.leftChild;
          tempHook.head = edge.head;
          tempHook.tag = edge.tag;
          tempHook.backEdge = real;
          tempHook.iScore = real.iScore + br.score + dparser.headScore[dparser.binDistance[edge.head][edge.end]][edge.head][dg.tagBin(edge.tag)][real.head][dg.tagBin(real.tag)] + dparser.headStop[real.head][dg.tagBin(real.tag)][real.start] + dparser.headStop[real.head][dg.tagBin(real.tag)][real.end];
          relaxTempHook();
        }
      }
    }
    if (newL) {
      // POST HOOKS
      BinaryRule[] rules = bg.splitRulesWithRC(edge.state);
      for (int i = 0; i < rules.length; i++) {
        BinaryRule br = rules[i];
        Collection realEdges = chart.getRealEdgesWithR(br.leftChild, edge.start);
        for (Iterator realI = realEdges.iterator(); realI.hasNext();) {
          Edge real = (Edge) realI.next();
          tempHook.start = real.start;
          tempHook.end = real.end;
          tempHook.state = br.parent;
          tempHook.subState = br.rightChild;
          tempHook.head = edge.head;
          tempHook.tag = edge.tag;
          tempHook.backEdge = real;
          tempHook.iScore = real.iScore + br.score + dparser.headScore[dparser.binDistance[edge.head][edge.start]][edge.head][dg.tagBin(edge.tag)][real.head][dg.tagBin(real.tag)] + dparser.headStop[real.head][dg.tagBin(real.tag)][real.start] + dparser.headStop[real.head][dg.tagBin(real.tag)][real.end];
          relaxTempHook();
        }
      }
    }
  }

  protected void triggerAllHooks(Edge edge) {
    // we might have built a new edge, enabling some old edges to project hooks
    boolean newL = !chart.isBuiltL(edge.state, edge.start, edge.head, edge.tag);
    boolean newR = !chart.isBuiltR(edge.state, edge.end, edge.head, edge.tag);
    if (verbose) {
      if (newL) {
        System.out.println("Triggering on L: " + edge);
      }
      if (newR) {
        System.out.println("Triggering on R: " + edge);
      }
    }
    chart.registerEdgeIndexes(edge);
    if (newR) {
      // PRE HOOKS
      for (Iterator rI = bg.ruleIteratorByLeftChild(edge.state); rI.hasNext();) {
        BinaryRule br = (BinaryRule) rI.next();
        Collection edges = chart.getRealEdgesWithL(br.rightChild, edge.end);
        for (Iterator realI = edges.iterator(); realI.hasNext();) {
          Edge real = (Edge) realI.next();
          tempHook.start = real.start;
          tempHook.end = real.end;
          tempHook.state = br.parent;
          tempHook.subState = br.leftChild;
          tempHook.head = edge.head;
          tempHook.tag = edge.tag;
          tempHook.backEdge = real;
          tempHook.iScore = real.iScore + br.score + dparser.headScore[dparser.binDistance[edge.head][edge.end]][edge.head][dg.tagBin(edge.tag)][real.head][dg.tagBin(real.tag)] + dparser.headStop[real.head][dg.tagBin(real.tag)][real.start] + dparser.headStop[real.head][dg.tagBin(real.tag)][real.end];
          relaxTempHook();
        }
      }
    }
    if (newL) {
      // POST HOOKS
      for (Iterator rI = bg.ruleIteratorByRightChild(edge.state); rI.hasNext();) {
        BinaryRule br = (BinaryRule) rI.next();
        Collection edges = chart.getRealEdgesWithR(br.leftChild, edge.start);
        if (false && verbose) {
          System.out.println("Looking for: " + Numberer.getGlobalNumberer("states").object(br.leftChild) + " ending at " + edge.start);
          System.out.println("Found: " + edges);
        }
        for (Iterator realI = edges.iterator(); realI.hasNext();) {
          Edge real = (Edge) realI.next();
          tempHook.start = real.start;
          tempHook.end = real.end;
          tempHook.state = br.parent;
          tempHook.subState = br.rightChild;
          tempHook.head = edge.head;
          tempHook.tag = edge.tag;
          tempHook.backEdge = real;
          tempHook.iScore = real.iScore + br.score + dparser.headScore[dparser.binDistance[edge.head][edge.start]][edge.head][dg.tagBin(edge.tag)][real.head][dg.tagBin(real.tag)] + dparser.headStop[real.head][dg.tagBin(real.tag)][real.start] + dparser.headStop[real.head][dg.tagBin(real.tag)][real.end];
          relaxTempHook();
        }
      }
    }
  }


  protected void relaxTempHook() {
    relaxHook1++;
    if (verbose) {
      System.out.println("Considering: " + tempHook + " iP: " + scorer.iPossible(tempHook) + " oP: " + scorer.oPossible(tempHook));
    }
    if (!Test.exhaustiveTest) {
      if (!scorer.oPossible(tempHook) || !scorer.iPossible(tempHook)) {
        return;
      }
    }
    relaxHook2++;
    Hook resultHook = (Hook) interner.intern(tempHook);
    if (verbose) {
      System.out.println("Formed " + resultHook + " " + (resultHook == tempHook ? "new" : "old") + " " + tempHook.iScore + " was " + resultHook.iScore);
    }
    if (resultHook == tempHook) {
      relaxHook3++;
      tempHook = new Hook();
      discoverHook(resultHook);
    }
    if (better(tempHook.iScore, resultHook.iScore)) {
      resultHook.iScore = tempHook.iScore;
      resultHook.backEdge = tempHook.backEdge;
      try {
        agenda.decreaseKey(resultHook);
      } catch (NullPointerException e) {
      }
    }
  }

  protected void projectUnaries(Edge edge) {
    for (Iterator rI = ug.ruleIteratorByChild(edge.state); rI.hasNext();) {
      UnaryRule ur = (UnaryRule) rI.next();
      if (ur.child == ur.parent) {
        continue;
      }
      tempEdge.start = edge.start;
      tempEdge.end = edge.end;
      tempEdge.head = edge.head;
      tempEdge.tag = edge.tag;
      tempEdge.state = ur.parent;
      tempEdge.backEdge = edge;
      tempEdge.backHook = null;
      tempEdge.iScore = edge.iScore + ur.score;
      relaxTempEdge();
    }
  }

  protected void processEdge(Edge edge) {
    // add to chart
    if (verbose) {
      System.out.println("Adding to chart: " + edge);
    }
    chart.addEdge(edge);
    // fetch existing hooks that can combine with it
    Collection preHooks = chart.getPreHooks(edge);
    Collection postHooks = chart.getPostHooks(edge);
    // combine them
    for (Iterator preI = preHooks.iterator(); preI.hasNext();) {
      Hook hook = (Hook) preI.next();
      combine(edge, hook);
    }
    for (Iterator postI = postHooks.iterator(); postI.hasNext();) {
      Hook hook = (Hook) postI.next();
      combine(edge, hook);
    }
    // do projections
    //if (verbose) System.out.println("Projecting: "+edge);
    projectUnaries(edge);
    if (!bg.isSynthetic(edge.state) && !op.freeDependencies) {
      projectHooks(edge);
      registerReal(edge);
    }
    if (op.freeDependencies) {
      projectHooks(edge);
      registerReal(edge);
      triggerAllHooks(edge);
    } else {
      triggerHooks(edge);
    }
  }

  protected void processHook(Hook hook) {
    // add to chart
    //if (verbose) System.out.println("Adding to chart: "+hook);
    chart.addHook(hook);
    Collection edges = chart.getEdges(hook);
    for (Iterator edgeI = edges.iterator(); edgeI.hasNext();) {
      Edge edge = (Edge) edgeI.next();
      combine(edge, hook);
    }
  }

  protected void processItem(Item item) {
    if (item.isEdge()) {
      processEdge((Edge) item);
    } else {
      processHook((Hook) item);
    }
  }

  protected void discoverItem(Item item) {
    if (item.isEdge()) {
      discoverEdge((Edge) item);
    } else {
      discoverHook((Hook) item);
    }
  }

  protected Item makeInitialItem(int pos, int tag, int state, double iScore) {
    Edge edge = new Edge();
    edge.start = pos;
    edge.end = pos + 1;
    edge.state = state;
    edge.head = pos;
    edge.tag = tag;
    edge.iScore = iScore;
    return edge;
  }

  protected List makeInitialItems(List wordList) {
    List itemList = new ArrayList();
    int length = wordList.size();
    int numTags = tagNumberer.total();
    words = new int[length];
    taggedWordList = new List[length];
    int terminalCount = 0;
    for (int i = 0; i < length; i++) {
      taggedWordList[i] = new ArrayList(numTags);
      String wordStr = "";
      String tagStr = "";
      Object wordObject = wordList.get(i);
      if (wordObject instanceof HasWord) {
        wordStr = ((HasWord) wordObject).word();
      } else {
        wordStr = wordObject.toString();
      }
      //if (wordObject instanceof HasTag) {
      //  tagStr = ((HasTag)wordObject).tag();
      //}
      if (!wordNumberer.hasSeen(wordStr)) {
        wordStr = Lexicon.UNKNOWN_WORD;
      }
      int word = wordNumberer.number(wordStr);
      words[i] = word;
      if (!op.flexiTag) {
        for (Iterator tagI = lex.ruleIteratorByWord(word, i); tagI.hasNext();) {
          IntTaggedWord tagging = (IntTaggedWord) tagI.next();
          int tag = tagging.tag;
          //String curTagStr = (String)tagNumberer.object(tag);
          //if (!tagStr.equals("") && !tagStr.equals(curTagStr))
          //  continue;
          int state = Numberer.translate("tags", "states", tag);
          //itemList.add(makeInitialItem(i,tag,state,1.0*tagging.score));
          // THIS WILL CAUSE BUGS!!!  Don't use with another A* scorer
          tempEdge.state = state;
          tempEdge.head = i;
          tempEdge.start = i;
          tempEdge.end = i + 1;
          tempEdge.tag = tag;
          itemList.add(makeInitialItem(i, tag, state, scorer.iScore(tempEdge)));
          terminalCount++;
          taggedWordList[i].add(new IntTaggedWord(word, tag));
        }
      } else {
        for (int state = 0; state < stateNumberer.total(); state++) {
          if (!tagNumberer.hasSeen(stateNumberer.object(state))) {
            continue;
          }
          int tag = Numberer.translate("states", "tags", state);
          //if (!tagStr.equals("") && !tagStr.equals(curTagStr))
          //  continue;
          // THIS WILL CAUSE BUGS!!!  Don't use with another A* scorer
          tempEdge.state = state;
          tempEdge.head = i;
          tempEdge.start = i;
          tempEdge.end = i + 1;
          tempEdge.tag = tag;
          if (scorer.iScore(tempEdge) == Double.NEGATIVE_INFINITY) {
            continue;
          }
          itemList.add(makeInitialItem(i, tag, state, scorer.iScore(tempEdge)));
          terminalCount++;
          taggedWordList[i].add(new IntTaggedWord(word, tag));
        }
      }
    }
    if (Test.verbose) {
      System.out.println("Terminals (# of tag edges in chart): " +
                         terminalCount);
    }
    return itemList;
  }

  protected void scoreDependencies() {
    // just leach it off the dparser for now...
    /*
    Dependency dependency = new Dependency();
    for (int head = 0; head < words.length; head++) {
      for (int hTag = 0; hTag < tagNumberer.total(); hTag++) {
        for (int arg = 0; arg < words.length; arg++) {
          for (int aTag = 0; aTag < tagNumberer.total(); aTag++) {
            Arrays.fill(depScore[head][hTag][arg][aTag],Float.NEGATIVE_INFINITY);
          }
        }
      }
    }
    for (int head = 0; head < words.length; head++) {
      for (int arg = 0; arg < words.length; arg++) {
        if (head == arg)
          continue;
        for (Iterator headTWI=taggedWordList[head].iterator(); headTWI.hasNext();) {
          IntTaggedWord headTW = (IntTaggedWord)headTWI.next();
          for (Iterator argTWI=taggedWordList[arg].iterator(); argTWI.hasNext();) {
            IntTaggedWord argTW = (IntTaggedWord)argTWI.next();
            dependency.head = headTW;
            dependency.arg = argTW;
            dependency.leftHeaded = (head < arg);
            dependency.distance = Math.abs(head-arg);
            depScore[head][headTW.tag][arg][argTW.tag] =
              dg.score(dependency);
            if (false && depScore[head][headTW.tag][arg][argTW.tag] > -100)
              System.out.println(wordNumberer.object(headTW.word)+"/"+tagNumberer.object(headTW.tag)+" -> "+wordNumberer.object(argTW.word)+"/"+tagNumberer.object(argTW.tag)+" score "+depScore[head][headTW.tag][arg][argTW.tag]);
          }
        }
      }
    }
    */
  }

  protected void setGoal(int length) {
    goal = new Edge();
    goal.start = 0;
    goal.end = length;
    goal.state = stateNumberer.number(op.langpack().startSymbol());
    goal.tag = tagNumberer.number(Lexicon.BOUNDARY_TAG);
    goal.head = length - 1;
    //goal = (Edge)interner.intern(goal);
  }

  protected void initialize(List words) {
    length = words.size();
    interner = new Interner();
    agenda = new ArrayHeap(new ScoredComparator(ScoredComparator.DESCENDING));
    chart = new HookChart();
    setGoal(length);
    List initialItems = makeInitialItems(words);
    scoreDependencies();
    for (int i = 0; i < initialItems.size(); i++) {
      Item item = (Item) initialItems.get(i);
      item = (Item) interner.intern(item);
      //if (verbose) System.out.println("Initial: "+item);
      discoverItem(item);
    }
  }


  /**
   * Parse a Sentence.
   * This hasn't yet been implemented.  At present the goal is ignored.
   *
   * @return true iff it could be parsed
   */
  public boolean parse(Sentence sentence, String goal) {
    return parse((List) sentence);
  }


  /**
   * Parse a Sentence.
   *
   * @return true iff it could be parsed
   */
  public boolean parse(Sentence sentence) {
    return parse((List) sentence);
  }


  /**
   * Parse a Sentence.
   *
   * @return true iff it could be parsed
   */
  public boolean parse(List words) {
    int spanFound = 0;
    long last = 0;
    int exHook = 0;
    relaxHook1 = 0;
    relaxHook2 = 0;
    relaxHook3 = 0;
    relaxHook4 = 0;
    builtHooks = 0;
    builtEdges = 0;
    extractedHooks = 0;
    extractedEdges = 0;
    if (Test.verbose) {
      Timing.tick("Starting combined parse.");
    }
    dparser.binDistance = dparser.binDistance; // THIS IS TERRIBLE, BUT SAVES MEMORY
    initialize(words);
    while (!agenda.isEmpty()) {
      Item item = (Item) agenda.extractMin();
      if (!item.isEdge()) {
        exHook++;
        extractedHooks++;
      } else {
        extractedEdges++;
      }
      if (relaxHook1 > last + 1000000) {
        last = relaxHook1;
        if (Test.verbose) {
          System.out.println("Proposed hooks:   " + relaxHook1);
          System.out.println("Unfiltered hooks: " + relaxHook2);
          System.out.println("Built hooks:      " + relaxHook3);
          System.out.println("Waste hooks:      " + relaxHook4);
          System.out.println("Extracted hooks:  " + exHook);
        }
      }
      if (item.end - item.start > spanFound) {
        spanFound = item.end - item.start;
        if (Test.verbose) {
          System.out.print(spanFound + " ");
        }
      }
      //if (item.end < 5) System.out.println("Extracted: "+item+" iScore "+item.iScore+" oScore "+item.oScore+" score "+item.score());
      if (item.equals(goal)) {
        if (Test.verbose) {
          System.out.println("Found goal!");
          System.out.println("Comb iScore " + item.iScore); // was goal.iScore
          Timing.tick("Done, parse found.");
          System.out.println("Built items:      " + (builtEdges + builtHooks));
          System.out.println("Built hooks:      " + builtHooks);
          System.out.println("Built edges:      " + builtEdges);
          System.out.println("Extracted items:  " + (extractedEdges + extractedHooks));
          System.out.println("Extracted hooks:  " + extractedHooks);
          System.out.println("Extracted edges:  " + extractedEdges);
          //postMortem();
        }
        goal = (Edge) item;
        interner = null;
        agenda = null;
        return true;
      }
      if (item.score() == Double.NEGATIVE_INFINITY) {
        System.err.println("FactoredParser: no consistent parse [hit A*-blocked edges, aborting].");
        if (Test.verbose) {
          Timing.tick("FactoredParser: no consistent parse [hit A*-blocked edges, aborting].");
        }
        return false;
      }
      if (Test.MAX_ITEMS > 0 && (builtEdges + builtHooks) >= Test.MAX_ITEMS) {
        System.err.println("FactoredParser: exceeded MAX_ITEMS work limit [" + 
                           Test.MAX_ITEMS + " items]; aborting.");
        if (Test.verbose) {
          Timing.tick("FactoredParser: exceeded MAX_ITEMS work limit [" + 
                      Test.MAX_ITEMS + " items]; aborting.");
        }
        return false;
      }
      if (verbose && item.score() != Double.NEGATIVE_INFINITY) {
        System.out.println("Removing from agenda: " + item + " score i " + item.iScore + " + o " + item.oScore + " = " + item.score());
        System.out.println("Backtrace: " + (item.backEdge != null ? item.backEdge.toString() : "null") + " " + (item.isEdge() ? (((Edge) item).backHook != null ? ((Edge) item).backHook.toString() : "") : ""));
      }
      processItem(item);
    } // end while agenda is not empty
    System.err.println("FactoredParser: emptied agenda, no parse found!");
    if (Test.verbose) {
      Timing.tick("FactoredParser: emptied agenda, no parse found!");
    }
    return false;
  }


  protected void postMortem() {
    int numHooks = 0;
    int numEdges = 0;
    int numUnmatchedHooks = 0;
    int total = agenda.size();
    int done = 0;
    while (!agenda.isEmpty()) {
      Item item = (Item) agenda.extractMin();
      done++;
      //if(done % (total/10) == 0)
      //        System.out.println("Scanning: "+100*done/total);
      if (item.isEdge()) {
        numEdges++;
      } else {
        numHooks++;
        Collection edges = chart.getEdges((Hook) item);
        if (edges.size() == 0) {
          numUnmatchedHooks++;
        }
      }
    }
    System.out.println("--- Agenda Post-Mortem ---");
    System.out.println("Edges:           " + numEdges);
    System.out.println("Hooks:           " + numHooks);
    System.out.println("Unmatched Hooks: " + numUnmatchedHooks);
  }

  protected int project(int state) {
    return projection.project(state);
  }

  BiLexPCFGParser(Scorer scorer, ExhaustivePCFGParser fscorer, ExhaustiveDependencyParser dparser, BinaryGrammar bg, UnaryGrammar ug, DependencyGrammar dg, Lexicon lex, Options op) {
    this(scorer, fscorer, dparser, bg, ug, dg, lex, op, new NullGrammarProjection(bg, ug));
  }

  BiLexPCFGParser(Scorer scorer, ExhaustivePCFGParser fscorer, ExhaustiveDependencyParser dparser, BinaryGrammar bg, UnaryGrammar ug, DependencyGrammar dg, Lexicon lex, Options op, GrammarProjection projection) {
    this.fscorer = fscorer;
    this.projection = projection;
    this.dparser = dparser;
    this.scorer = scorer;
    this.bg = bg;
    this.ug = ug;
    this.dg = dg;
    this.lex = lex;
    this.op = op;
  }
}


abstract class AbstractEval {

  protected String str = "";

  private double precision = 0.0;
  private double recall = 0.0;
  private double f1 = 0.0;
  private double num = 0.0;
  private double exact = 0.0;

  private double precision2 = 0.0;
  private double recall2 = 0.0;
  private double pnum2 = 0.0;
  private double rnum2 = 0.0;

  public double getSentAveF1() {
    return f1 / num;
  }

  public double getEvalbF1() {
    return 2.0 / (rnum2 / (double) recall2 + pnum2 / (double) precision2);
  }

  // should be able to pass in a comparator!
  protected double precision(Set s1, Set s2) {
    double n = 0.0;
    double p = 0.0;
    for (Iterator i1 = s1.iterator(); i1.hasNext();) {
      Object o1 = i1.next();
      if (s2.contains(o1)) {
        p += 1.0;
      }
      //if (s2.contains(o1))
      //        System.out.println("Found: "+o1);
      //else
      //        System.out.println("Failed to find: "+o1);
      n += 1.0;
    }
    //System.out.println("Matched " + p + " of " + n);
    return (n > 0.0 ? p / n : 0.0);
  }

  abstract Set makeObjects(Tree tree);

  public void evaluate(Tree guess, Tree gold) {
    evaluate(guess, gold, new PrintWriter(System.out, true));
  }

  /* evaluates precision and recall by calling makeObjects() to make a
   * set of structures for guess Tree and gold Tree, and compares them
   * with each other.  */
  public void evaluate(Tree guess, Tree gold, PrintWriter pw) {
    Set dep1 = makeObjects(guess);
    //List dep1s = new ArrayList(dep1);
    //Collections.sort(dep1s);
    //System.out.println("Guess set is " + dep1s);
    Set dep2 = makeObjects(gold);
    //List dep2s = new ArrayList(dep2);
    //Collections.sort(dep2s);
    //System.out.println("Gold set is " + dep2s);
    //System.out.println("Precision");
    double curPrecision = precision(dep1, dep2);
    //System.out.println("Recall");
    double curRecall = precision(dep2, dep1);
    double curF1 = (curPrecision > 0.0 && curRecall > 0.0 ? 2.0 / (1.0 / curPrecision + 1.0 / curRecall) : 0.0);
    precision += curPrecision;
    recall += curRecall;
    f1 += curF1;
    num += 1.0;

    precision2 += dep1.size() * curPrecision;
    pnum2 += dep1.size();

    recall2 += dep2.size() * curRecall;
    rnum2 += dep2.size();

    if (curF1 > 0.9999) {
      exact += 1.0;
    }

    pw.println(" P: " + ((int) (curPrecision * 10000)) / 100.0 + " (sent ave " + ((int) (precision * 10000 / num)) / 100.0 + ") (evalb " + ((int) (precision2 * 10000 / pnum2)) / 100.0 + ")");
    pw.println(" R: " + ((int) (curRecall * 10000)) / 100.0 + " (sent ave " + ((int) (recall * 10000 / num)) / 100.0 + ") (evalb " + ((int) (recall2 * 10000 / rnum2)) / 100.0 + ")");

    double cF1 = 2.0 / (rnum2 / (double) recall2 + pnum2 / (double) precision2);
    String emit = str + " F1: " + ((int) (curF1 * 10000)) / 100.0 + " (sent ave " + ((int) (10000 * f1 / num)) / 100.0 + ", evalb " + ((int) (10000 * cF1)) / 100.0 + ")   Exact: " + ((int) (10000 * exact / num)) / 100.0 + " N: " + (int) num;
    pw.println(emit);
    //System.err.println(emit);

    /*
      Sentence s = guess.yield();
      for (Iterator it2 = s.iterator(); it2.hasNext(); ) {
        Object obj = it2.next();
        if (curF1 < 0.7) {
          badwords.incrementCount(obj);
        } else {
          goodwords.incrementCount(obj);
        }
      }
    */
  }

  /*
  private Counter goodwords = new Counter();
  private Counter badwords = new Counter();

  public void printGoodBad() {
    System.out.println("Printing bad categories");
    for (Iterator it = badwords.keysAbove(5.0).iterator(); it.hasNext(); ) {
      Object key = it.next();
      System.out.println("In badwords 5 times: " + key);
      double numb = badwords.getCount(key);
      double numg = goodwords.getCount(key);
      if (numb / (numb + numg) > 0.1) {
        System.out.println("Bad word!  " + key + " (" +
                           (numb / (numb + numg)) + " bad)");
        // EncodingPrintWriter.out.println("Bad word!  " + key + " (" +
        //                 (numb / (numb + numg)) + " bad)",
        //                              "GB18030");
      }
    }
  }
  */

  public void display(boolean verbose) {
    display(verbose, new PrintWriter(System.out, true));
  }

  public void display(boolean verbose, PrintWriter pw) {
    double prec = precision2 / pnum2;//(num > 0.0 ? precision/num : 0.0);
    double rec = recall2 / rnum2;//(num > 0.0 ? recall/num : 0.0);
    double f = 2.0 / (1.0 / prec + 1.0 / rec);//(num > 0.0 ? f1/num : 0.0);
    //System.out.println(" Precision: "+((int)(10000.0*prec))/100.0);
    //System.out.println(" Recall:    "+((int)(10000.0*rec))/100.0);
    //System.out.println(" F1:        "+((int)(10000.0*f))/100.0);
    pw.println("LP: " + ((int) (10000.0 * prec)) / 100.0 + " LR: " + ((int) (10000.0 * rec)) / 100.0 + " F1: " + ((int) (10000.0 * f)) / 100.0 + " Exact: " + ((int) (10000.0 * exact / num)) / 100.0);
    /*
    double prec = (num > 0.0 ? precision/num : 0.0);
    double rec = (num > 0.0 ? recall/num : 0.0);
    double f = (num > 0.0 ? f1/num : 0.0);
    System.out.println(" Precision: "+prec);
    System.out.println(" Recall:    "+rec);
    System.out.println(" F1:        "+f);
    */
  }
}


class CBEval extends LabeledConstituentEval {
  private double cb = 0.0;
  private double num = 0.0;
  private double zeroCB = 0.0;

  protected void checkCrossing(Set s1, Set s2) {
    double c = 0.0;
    for (Iterator i1 = s1.iterator(); i1.hasNext();) {
      Constituent constit = (Constituent) i1.next();
      if (constit.crosses(s2)) {
        c += 1.0;
      }
    }
    if (c == 0) {
      zeroCB += 1.0;
    }
    cb += c;
    num += 1;
  }

  public void evaluate(Tree t1, Tree t2, PrintWriter pw) {
    Set b1 = makeObjects(t1);
    Set b2 = makeObjects(t2);
    checkCrossing(b1, b2);
  }

  public void display(boolean verbose, PrintWriter pw) {
    pw.println("AvgCB: " + ((int) (100.0 * cb / num)) / 100.0 + " ZeroCB: " + ((int) (10000.0 * zeroCB / num)) / 100.0);
  }

  public CBEval(String str, TreebankLanguagePack tlp) {
    super(str, tlp);
  }

}


/**
 * CDM Mar 2004: This should be rewritten so as to root a word at an
 * index position; otherwise it doesn't work correctly when you get two
 * identical dependents (like with "I went to Greece to see the ruins").
 */
class DependencyEval extends AbstractEval {

  private static final boolean DEBUG = false;
  Filter punctFilter;

  /**
   * Build the set of dependencies for evaluation.  This set excludes
   * all dependencies for which the argument is a punctuation tag.
   */
  Set makeObjects(Tree tree) {
    Set deps = new HashSet();
    for (Iterator nodeI = tree.subTreeList().iterator(); nodeI.hasNext();) {
      // every child with a different head is an argument, as are ones with the
      // same head after the first one found
      Tree node = (Tree) nodeI.next();
      if (node.isLeaf() || node.children().length < 2) {
        continue;
      }
      // System.err.println("XXX node is " + node + "; label type is " +
      //                         node.label().getClass().getName());
      String head = ((HasWord) node.label()).word();
      boolean seenHead = false;
      for (int cNum = 0; cNum < node.children().length; cNum++) {
        Tree child = node.children()[cNum];
        String arg = ((HasWord) child.label()).word();
        if (head.equals(arg) && !seenHead) {
          seenHead = true;
        } else if (!punctFilter.accept(arg)) {
          deps.add(new UnnamedDependency(head, arg));
        }
      }
    }
    if (DEBUG) {
      System.err.println("Deps: " + deps);
    }
    return deps;
  }

  public DependencyEval(String str, Filter punctFilter) {
    this.str = str;
    this.punctFilter = punctFilter; // = Options.get().langpack().punctuationWordAcceptFilter();
  }

} // end class DependencyEval


class LabeledConstituentEval extends AbstractEval {

  private static final boolean DEBUG = false;

  private ConstituentFactory cf = new LabeledScoredConstituentFactory();
  private TreeFactory tf = new LabeledScoredTreeFactory();
  private TreebankLanguagePack tlp;

  protected Tree stripLeaves(Tree tree) {
    if (tree.isLeaf()) {
      return null;
    }
    if (tree.isPreTerminal()) {
      return tf.newLeaf(tree.label());
    }
    List children = new ArrayList(tree.children().length);
    for (int cNum = 0; cNum < tree.children().length; cNum++) {
      children.add(stripLeaves(tree.children()[cNum]));
    }
    return tf.newTreeNode(tree.label(), children);
  }

  Set makeObjects(Tree tree) {
    // Strip off a root node. Allow it to have multiple daughters.
    // Assumes boundary symbol is already removed.
    Tree noLeafTree = stripLeaves(tree);
    Set set = new HashSet();
    if (tlp.isStartSymbol(noLeafTree.label().value())) {
      for (int i = 0; i < noLeafTree.children().length - 1; i++) {
        set.addAll(noLeafTree.children()[i].constituents(cf));
      }
    } else {
      set.addAll(noLeafTree.constituents(cf));
    }
    return set;
  }

  public void evaluate(Tree guess, Tree gold, PrintWriter pw) {
    if (guess.yield().size() != gold.yield().size()) {
      pw.println("Warning: yield differs:");
      pw.println("Guess: " + guess.yield());
      pw.println("Gold: " + gold.yield());
    }
    if (DEBUG) {
      Set sguess = makeObjects(guess);
      Set sgold = makeObjects(gold);
      System.err.println("LCEval Guess " + sguess.size() + " constits: " + makeObjects(guess));
      System.err.println("LCEval Gold  " + sgold.size() + " constits: " + makeObjects(gold));
    }
    super.evaluate(guess, gold, pw);
  }

  public LabeledConstituentEval(String str, TreebankLanguagePack tlp) {
    this.str = str;
    this.tlp = tlp;
  }

} // end class LabeledConstituentEval


class TaggingEval extends AbstractEval {

  private static final boolean DEBUG = false;
  private Lexicon lex;

  Set makeObjects(Tree tree) {
    List twList = tree.taggedYield();
    Set set = new HashSet();
    for (int i = 0; i < twList.size(); i++) {
      TaggedWord tw = (TaggedWord) twList.get(i);
      //IntTaggedWord iTW = new IntTaggedWord(Numberer.number("words",tw.word()), Numberer.number("tags",tw.tag()));
      Pair positionWT = new Pair(new Integer(i), new WordTag(tw.value(), tw.tag()));
      //WordTag positionWT = new WordTag(tw.value(),tw.tag());
      //System.out.println(iTW);
      //if (! tw.tag.equals("*"))
      set.add(positionWT);
    }
    // if (DEBUG) System.err.println("Tags: " + set);
    return set;
  }

  public TaggingEval(String str) {
    this.str = str;
  }

  public TaggingEval(String str, Lexicon lex) {
    this(str);
    this.lex = lex;
  }

  public void evaluate(Tree guess, Tree gold, PrintWriter pw) {
    Sentence sGold = gold.taggedYield();
    Sentence sGuess = guess.taggedYield();
    if (sGuess.size() != sGold.size()) {
      pw.println("Warning: yield differs:");
      pw.println("Guess: " + sGuess);
      pw.println("Gold: " + sGold);
    } else {
      if (DEBUG) {
        for (Iterator goldIt = sGold.iterator(), guessIt = sGuess.iterator(); goldIt.hasNext();) {
          TaggedWord goldNext = (TaggedWord) goldIt.next();
          TaggedWord guessNext = (TaggedWord) guessIt.next();
          if (!goldNext.tag().equals(guessNext.tag())) {
            pw.print("TaggingError ");
            if (lex.isKnown(goldNext.word())) {
              pw.print("seen ");
            } else {
              pw.print("unseen ");
            }
            pw.println(goldNext.word() + " correct " + goldNext.tag() + " chose " + guessNext.tag());
          }
        }
      }
    }
    super.evaluate(guess, gold, pw);
  }

} // end class TaggingEval


class RuleErrorEval extends AbstractEval {

  private boolean verbose = false;

  private Counter over = new Counter();
  private Counter under = new Counter();

  protected static String localize(Tree tree) {
    if (tree.isLeaf()) {
      return "";
    }
    StringBuffer sb = new StringBuffer();
    sb.append(tree.label());
    sb.append(" ->");
    for (int i = 0; i < tree.children().length; i++) {
      sb.append(" ");
      sb.append(tree.children()[i].label());
    }
    return sb.toString();
  }

  Set makeObjects(Tree tree) {
    Set localTrees = new HashSet();
    for (Iterator i = tree.subTreeList().iterator(); i.hasNext();) {
      Tree st = (Tree) i.next();
      localTrees.add(localize(st));
    }
    return localTrees;
  }

  public void evaluate(Tree t1, Tree t2, PrintWriter pw) {
    Set s1 = makeObjects(t1);
    Set s2 = makeObjects(t2);
    for (Iterator i1 = s1.iterator(); i1.hasNext();) {
      Object o1 = i1.next();
      if (!s2.contains(o1)) {
        over.incrementCount(o1);
      }
    }
    for (Iterator i2 = s2.iterator(); i2.hasNext();) {
      Object o2 = i2.next();
      if (!s1.contains(o2)) {
        under.incrementCount(o2);
      }
    }
  }

  protected void display(Counter c, int num, PrintWriter pw) {
    List rules = new ArrayList(c.keySet());
    Collections.sort(rules, c.comparator(false));
    int rSize = rules.size();
    if (num > rSize) {
      num = rSize;
    }
    for (int i = 0; i < num; i++) {
      pw.println(rules.get(i) + " " + c.getCount(rules.get(i)));
    }
  }

  public void display(boolean verbose, PrintWriter pw) {
    this.verbose = verbose;
    pw.println("Most freqently underproposed rules:");
    display(under, (verbose ? 100 : 10), pw);
    pw.println("Most freqently overproposed rules:");
    display(over, (verbose ? 100 : 10), pw);
  }

  public RuleErrorEval(String str) {
    this.str = str;
  }

} // end class RuleErrorEval


/** This class counts which categories are over and underproposed in trees. */
class CatErrorEval extends AbstractEval {

  private Counter over = new Counter();
  private Counter under = new Counter();

  /** Unused. Fake satisfying the abstract class. */
  Set makeObjects(Tree tree) {
    return null;
  }

  List myMakeObjects(Tree tree) {
    List cats = new LinkedList();
    for (Iterator i = tree.subTreeList().iterator(); i.hasNext(); ) {
      Tree st = (Tree) i.next();
      cats.add(st.value());
    }
    return cats;
  }

  public void evaluate(Tree t1, Tree t2, PrintWriter pw) {
    List s1 = myMakeObjects(t1);
    List s2 = myMakeObjects(t2);
    List del2 = new LinkedList(s2);
    // we delete out as we find them so we can score correctly a cat with
    // a certain cardinality in a tree.
    for (Iterator i1 = s1.iterator(); i1.hasNext(); ) {
      Object o1 = i1.next();
      if ( ! del2.remove(o1)) {
        over.incrementCount(o1);
      }
    }
    for (Iterator i2 = s2.iterator(); i2.hasNext(); ) {
      Object o2 = i2.next();
      if (! s1.remove(o2)) {
        under.incrementCount(o2);
      }
    }
  }

  protected void display(Counter c, PrintWriter pw) {
    List cats = new ArrayList(c.keySet());
    Collections.sort(cats, c.comparator(false));
    for (Iterator it = cats.iterator(); it.hasNext(); ) {
      Object ob = it.next();
      pw.println(ob + " " + c.getCount(ob));
    }
  }

  public void display(boolean verbose, PrintWriter pw) {
    pw.println("Most freqently underproposed categories:");
    display(under, pw);
    pw.println("Most freqently overproposed categories:");
    display(over, pw);
  }

  public CatErrorEval(String str) {
    this.str = str;
  }

} // end class CatErrorEval


class CollinsPuncTransformer implements TreeTransformer {

  private TreebankLanguagePack tlp;

  boolean isPunc(Tree t) {
    if (t.isPreTerminal()) {
      String s = t.label().value();
      if (tlp.isEvalBIgnoredPunctuationTag(s)) {
        return true;
      }
    }
    return false;
  }

  LinkedList preTerms(Tree t) {
    LinkedList l = new LinkedList();
    preTermHelper(t, l);
    return l;
  }

  void preTermHelper(Tree t, List l) {
    if (t.isLeaf()) {
      return;
    }
    if (t.isPreTerminal()) {
      l.add(t);
      return;
    }
    Tree[] children = t.children();
    for (int i = 0; i < children.length; i++) {
      preTermHelper(children[i], l);
    }
  }

  Tree transformRoot(Tree tree, TreeFactory tf) {
    // XXXX fix this: use tlp and don't assume 1 daughter of ROOT!
    // leave the root intact
    // if (tlp.isStartSymbol(tlp.basicCategory(tree.label().value())))
    if (tree.label().toString().startsWith("ROOT")) {
      return tf.newTreeNode(tree.label(), Collections.singletonList(transformNode(tree.children()[0], tf)));
    }
    return transformNode(tree, tf);
  }

  Tree transformNode(Tree tree, TreeFactory tf) {
    if (tree.isLeaf()) {
      return tf.newLeaf(tree.label());
    }
    if (tree.isPreTerminal()) {
      return tf.newTreeNode(tree.label(), Collections.singletonList(tf.newLeaf(tree.children()[0].label())));
    }
    List children = tree.getChildrenAsList();
    LinkedList newChildren = new LinkedList();
    // promote lower punctuation
    for (Iterator i = children.iterator(); i.hasNext();) {
      Tree child = (Tree) i.next();
      LinkedList preTerms = preTerms(child);
      while (!preTerms.isEmpty() && isPunc((Tree) preTerms.getFirst())) {
        newChildren.add(preTerms.getFirst());
        preTerms.removeFirst();
      }
      Tree newChild = transformNode(child, tf);
      LinkedList temp = new LinkedList();
      if (newChild.children().length > 0) {
        newChildren.add(newChild);
      }
      while (!preTerms.isEmpty() && isPunc((Tree) preTerms.getLast())) {
        temp.addFirst(preTerms.getLast());
        preTerms.removeLast();
      }
      newChildren.addAll(temp);
    }
    // remove local punctuation
    while (!newChildren.isEmpty() && isPunc((Tree) newChildren.getFirst())) {
      newChildren.removeFirst();
    }
    while (!newChildren.isEmpty() && isPunc((Tree) newChildren.getLast())) {
      newChildren.removeLast();
    }
    return tf.newTreeNode(tree.label(), newChildren);
  }


//   public Tree transformTree(Tree tree) {
//     //System.out.println("PUNCTUATION TRANSFORM:");
//     //tree.pennPrint();
//     //System.out.println("BECOMES:");
//     //transformRoot(tree, tf).pennPrint();
//     return transformRoot(tree, tf);
//   }

   public Tree transformTree(Tree tree) {
     return transformRoot(tree,tree.treeFactory());
   }

  public CollinsPuncTransformer(TreebankLanguagePack tlp) {
    this.tlp = tlp;
  }
}


/**
 * Does detransformations to a parsed sentence to map it back to the
 * standard treebank form for output or evaluation.
 * This version has Penn-Treebank-English-specific details, but can probably
 * be used without harm on other treebanks.
 * Returns labels to their basic category, removes punctuation (with
 * respect to a gold tree), deletes the boundary symbol, changes PRT
 * labels to ADVP.
 */
class TreeCollinizer implements TreeTransformer {

  private TreebankLanguagePack tlp;
  private final boolean deletePunct;
  private final boolean fixCollinsBaseNP;

  public TreeCollinizer(TreebankLanguagePack tlp) {
    this(tlp, true, false);
  }

  public TreeCollinizer(TreebankLanguagePack tlp, boolean deletePunct,
                        boolean fixCollinsBaseNP) {
    this.tlp = tlp;
    this.deletePunct = deletePunct;
    this.fixCollinsBaseNP = fixCollinsBaseNP;
  }

  protected TreeFactory tf = new LabeledScoredTreeFactory();

  public Tree transformTree(Tree tree) {
    String s = tree.value();
    if (tree.isLeaf()) {
      return tf.newLeaf(s);  // make it a StringLabel
    }
    s = tlp.basicCategory(s);
    if (deletePunct) {
      // this is broken as it's not the right thing to do when there
      // is any tag ambiguity -- and there is for ' (POS/'').  Sentences
      // can then have more or less words.  It's also unnecessary for EVALB,
      // since it ignores punctuation anyway (though in the same wrong way)
      if (tree.isPreTerminal() && tlp.isEvalBIgnoredPunctuationTag(s)) {
        return null;
      }
    }
    // remove the extra NPs inserted in the collinsBaseNP option
    if (fixCollinsBaseNP && s.equals("NP")) {
      Tree[] kids = tree.children();
      if (kids.length == 1 && tlp.basicCategory(kids[0].value()).equals("NP")) {
        return transformTree(kids[0]);
      }
    }
    // XXXX below isn't correct if root isn't unary.
    if (tlp.isStartSymbol(s)) {
      // NB: This deletes the boundary symbol, which is in the tree!
      return transformTree(tree.children()[0]);
    }
    // Magerman erased this decision, and everyone else has followed like sheep
    if (s.equals("PRT")) {
      s = "ADVP";
    }
    List children = new ArrayList();
    for (int cNum = 0; cNum < tree.children().length; cNum++) {
      Tree child = tree.children()[cNum];
      Tree newChild = transformTree(child);
      if (newChild != null) {
        children.add(newChild);
      }
    }
    if (children.size() == 0) {
      return null;
    }
    return tf.newTreeNode(s, children);
  }

} // end class TreeColliniser


/**
 * Returns labels to their basic category, removes punctuation (with
 * respect to a gold tree), deletes the boundary symbol, changes PRT
 * labels to ADVP.
 */
class BoundaryRemover implements TreeTransformer {

  public BoundaryRemover() {
  }

  public Tree transformTree(Tree tree) {
    Tree last = tree.lastChild();
    if (last.label().value().equals(Lexicon.BOUNDARY_TAG)) {
      List childList = tree.getChildrenAsList();
      List lastGoneList = childList.subList(0, childList.size() - 1);
      return tree.treeFactory().newTreeNode(tree.label(), lastGoneList);
    } else {
      return tree;
    }
  }

} // end class BoundaryRemover


class NodePruner {
  private ExhaustivePCFGParser parser;
  private TreeTransformer debinarizer;

  List prune(List treeList, Label label, int start, int end) {
    // get reference tree
    if (treeList.size() == 1) {
      return treeList;
    }
    Tree testTree = ((Tree) treeList.get(0)).treeFactory().newTreeNode(label, treeList);
    int goal = Numberer.getGlobalNumberer("states").number(label.value());
    Tree tempTree = parser.extractBestParse(goal, start, end);
    parser.restoreUnaries(tempTree);
    Tree pcfgTree = debinarizer.transformTree(tempTree);
    Set pcfgConstituents = pcfgTree.constituents(new LabeledScoredConstituentFactory());
    // delete child labels that are not in reference but do not cross reference
    List prunedChildren = new ArrayList();
    int childStart = 0;
    for (int c = 0; c < testTree.children().length; c++) {
      Tree child = testTree.children()[c];
      boolean isExtra = true;
      int childEnd = childStart + child.yield().size();
      Constituent childConstituent = new LabeledScoredConstituent(childStart, childEnd, child.label(), 0);
      if (pcfgConstituents.contains(childConstituent)) {
        isExtra = false;
      }
      if (childConstituent.crosses(pcfgConstituents)) {
        isExtra = false;
      }
      if (child.isLeaf() || child.children()[0].isLeaf()) {
        isExtra = false;
      }
      if (pcfgTree.yield().size() != testTree.yield().size()) {
        isExtra = false;
      }
      if (!label.value().startsWith("NP^NP")) {
        isExtra = false;
      }
      if (isExtra) {
        System.err.println("Pruning: " + child.label() + " from " + (childStart + start) + " to " + (childEnd + start));
        System.err.println("Was: " + testTree + " vs " + pcfgTree);
        prunedChildren.addAll(child.getChildrenAsList());
      } else {
        prunedChildren.add(child);
      }
      childStart = childEnd;
    }
    return prunedChildren;
  }

  private List helper(List treeList, int start) {
    List newTreeList = new ArrayList(treeList.size());
    for (Iterator i = treeList.iterator(); i.hasNext();) {
      Tree tree = (Tree) i.next();
      int end = start + tree.yield().size();
      newTreeList.add(prune(tree, start));
      start = end;
    }
    return newTreeList;
  }

  public Tree prune(Tree tree) {
    return prune(tree, 0);
  }

  Tree prune(Tree tree, int start) {
    if (tree.isLeaf() || tree.children()[0].isLeaf()) {
      return tree;
    }
    // check each node's children for deletion
    List children = helper(tree.getChildrenAsList(), start);
    children = prune(children, tree.label(), start, start + tree.yield().size());
    return tree.treeFactory().newTreeNode(tree.label(), children);
  }

  public NodePruner(ExhaustivePCFGParser parser, TreeTransformer debinarizer) {
    this.parser = parser;
    this.debinarizer = debinarizer;
  }
}

class EvalB {

  private static PrintWriter goldWriter, testWriter;

  static void initEVALBfiles(TreebankLangParserParams tlpParams) {
    try {
      goldWriter = tlpParams.pw(new FileOutputStream("parses.gld"));
      testWriter = tlpParams.pw(new FileOutputStream("parses.tst"));
    } catch (IOException e) {
      System.exit(0);
    }
  }

  static void closeEVALBfiles() {
    goldWriter.close();
    testWriter.close();
  }

  static void writeEVALBline(Tree gold, Tree test) {
    if (false) {
      if (gold.yield().length() <= 1) {
        goldWriter.println("(ROOT (X X))");
        testWriter.println("(ROOT (X X))");
        return;
      }
    }
    goldWriter.println(gold.toString());
    testWriter.println(test.toString());
    System.err.println("Wrote EVALB lines.");
  }

}

/* some documentation for Roger's convenience
 * {pcfg,dep,combo}{PE,DE,TE} are precision/dep/tagging evals for the models

 * parser is the PCFG parser
 * dparser is the dependency parser
 * bparser is the combining parser

 * during testing:
 * tree is the test tree
 * tree2 is the best PCFG parse (debinarized)
 * tree3 is the dependency parse
 * tree4 is the combo parse
 */

public class FactoredParser {

  public static void main(String[] args) {

    TreebankLangParserParams tlpParams = new EnglishTreebankParserParams();
    Options op = new Options(tlpParams);
    // tlpParams may be changed to something else later, so don't use it till
    // after options are parsed.

    System.out.println("Currently " + new Date());
    System.out.print("Invoked with arguments:");
    for (int i = 0; i < args.length; i++) {
      System.out.print(" " + args[i]);
    }
    System.out.println();

    String path = "/u/nlp/stuff/corpora/Treebank3/parsed/mrg/wsj";
    int trainLow = 200, trainHigh = 2199, testLow = 2200, testHigh = 2219;
    String serializeFile = null;

    int i = 0;
    while (i < args.length && args[i].startsWith("-")) {
      if (args[i].equalsIgnoreCase("-path") && (i + 1 < args.length)) {
        path = args[i + 1];
        i += 2;
      } else if (args[i].equalsIgnoreCase("-train") && (i + 2 < args.length)) {
        trainLow = Integer.parseInt(args[i + 1]);
        trainHigh = Integer.parseInt(args[i + 2]);
        i += 3;
      } else if (args[i].equalsIgnoreCase("-test") && (i + 2 < args.length)) {
        testLow = Integer.parseInt(args[i + 1]);
        testHigh = Integer.parseInt(args[i + 2]);
        i += 3;
      } else if (args[i].equalsIgnoreCase("-serialize") && (i + 1 < args.length)) {
        serializeFile = args[i + 1];
        i += 2;
      } else if (args[i].equalsIgnoreCase("-tLPP") && (i + 1 < args.length)) {
        try {
          tlpParams = (TreebankLangParserParams) Class.forName(args[i + 1]).newInstance();
          op.tlpParams = tlpParams;
        } catch (ClassNotFoundException e) {
          System.err.println("Class not found: " + args[i + 1]);
        } catch (InstantiationException e) {
          System.err.println("Couldn't instantiate: " + args[i + 1] + ": " + e.toString());
        } catch (IllegalAccessException e) {
          System.err.println("illegal access" + e);
        }
        i += 2;
      } else if (args[i].equals("-encoding")) {
        // sets encoding for TreebankLangParserParams
        tlpParams.setInputEncoding(args[i + 1]);
        tlpParams.setOutputEncoding(args[i + 1]);
        i += 2;
      } else {
        int j = setOptionFlag(op, args, i);
        if (j == i) {
          j = tlpParams.setOptionFlag(args, i);
        }
        if (j == i) {
          System.err.println("Unknown option ignored: " + args[i]);
          j++;
        }
        i = j;
      }
    }
    // System.out.println(tlpParams.getClass());
    TreebankLanguagePack tlp = tlpParams.treebankLanguagePack();

    Train.sisterSplitters = new HashSet(Arrays.asList(tlpParams.sisterSplitters()));
    //    BinarizerFactory.TreeAnnotator.setTreebankLang(tlpParams);
    PrintWriter pw = tlpParams.pw();

    Test.display();
    Train.display();
    op.display();
    tlpParams.display();

    // setup tree transforms
    Treebank trainTreebank = tlpParams.memoryTreebank();
    MemoryTreebank testTreebank = tlpParams.testMemoryTreebank();
    // Treebank blippTreebank = ((EnglishTreebankParserParams) tlpParams).diskTreebank();
    // String blippPath = "/afs/ir.stanford.edu/data/linguistic-data/BLLIP-WSJ/";
    // blippTreebank.loadPath(blippPath, "", true);

    Timing.startTime();
    System.err.print("Reading trees...");
    testTreebank.loadPath(path, new NumberRangeFileFilter(testLow, testHigh, true));
    if (Test.increasingLength) {
      Collections.sort(testTreebank, new TreeLengthComparator());
    }

    trainTreebank.loadPath(path, new NumberRangeFileFilter(trainLow, trainHigh, true));
    Timing.tick("done.");
    System.err.print("Binarizing trees...");
    TreeAnnotatorAndBinarizer binarizer = null;
    if (!Train.leftToRight) {
      binarizer = new TreeAnnotatorAndBinarizer(tlpParams, op.forceCNF, !Train.outsideFactor(), true);
    } else {
      binarizer = new TreeAnnotatorAndBinarizer(tlpParams.headFinder(), new LeftHeadFinder(), tlpParams, op.forceCNF, !Train.outsideFactor(), true);
    }
    CollinsPuncTransformer collinsPuncTransformer = null;
    if (Train.collinsPunc) {
      collinsPuncTransformer = new CollinsPuncTransformer(tlp);
    }
    TreeTransformer debinarizer = new Debinarizer(tlp, op.forceCNF);
    List binaryTrainTrees = new ArrayList();

    if (Train.selectiveSplit) {
      Train.splitters = ParentAnnotationStats.getSplitCategories(trainTreebank, Train.tagSelectiveSplit, 0, Train.selectiveSplitCutOff, Train.tagSelectiveSplitCutOff, tlpParams.treebankLanguagePack());
      if (Train.deleteSplitters != null) {
        List<String> deleted = new ArrayList<String>();
        for (String del : Train.deleteSplitters) {
          String baseDel = tlp.basicCategory(del);
          boolean checkBasic = del.equals(baseDel);
          for (Iterator it = Train.splitters.iterator(); it.hasNext(); ) {
            String elem = (String) it.next();
            String baseElem = tlp.basicCategory(elem);
            boolean delStr = checkBasic && baseElem.equals(baseDel) ||
              elem.equals(del);
            if (delStr) {
              it.remove();
              deleted.add(elem);
            }
          }
        }
        System.err.println("Removed from vertical splitters: " + deleted);
      }
    }
    if (Train.selectivePostSplit) {
      TreeTransformer myTransformer = new TreeAnnotator(tlpParams.headFinder(), tlpParams);
      Treebank annotatedTB = trainTreebank.transform(myTransformer);
      Train.postSplitters = ParentAnnotationStats.getSplitCategories(annotatedTB, true, 0, Train.selectivePostSplitCutOff, Train.tagSelectivePostSplitCutOff, tlpParams.treebankLanguagePack());
    }

    if (Train.hSelSplit) {
      binarizer.setDoSelectiveSplit(false);
      for (Iterator tI = trainTreebank.iterator(); tI.hasNext();) {
        Tree tree = (Tree) tI.next();
        if (Train.collinsPunc) {
          tree = collinsPuncTransformer.transformTree(tree);
        }
        //tree.pennPrint(tlpParams.pw());
        tree = binarizer.transformTree(tree);
        //binaryTrainTrees.add(tree);
      }
      binarizer.setDoSelectiveSplit(true);
    }
    for (Iterator tI = trainTreebank.iterator(); tI.hasNext();) {
      Tree tree = (Tree) tI.next();
      if (Train.collinsPunc) {
        tree = collinsPuncTransformer.transformTree(tree);
      }
      tree = binarizer.transformTree(tree);
      binaryTrainTrees.add(tree);
    }
    if (Test.verbose) {
      binarizer.dumpStats();
    }

    List<Tree> binaryTestTrees = new ArrayList<Tree>();
    for (Iterator tI = testTreebank.iterator(); tI.hasNext();) {
      Tree tree = (Tree) tI.next();
      if (Train.collinsPunc) {
        tree = collinsPuncTransformer.transformTree(tree);
      }
      tree = binarizer.transformTree(tree);
      binaryTestTrees.add(tree);
    }
    Timing.tick("done.");  // binarization
    BinaryGrammar bg = null;
    UnaryGrammar ug = null;
    DependencyGrammar dg = null;
    // DependencyGrammar dgBLIPP = null;
    Lexicon lex = null;
    // extract grammars
    Extractor bgExtractor = new BinaryGrammarExtractor();
    //Extractor bgExtractor = new SmoothedBinaryGrammarExtractor();//new BinaryGrammarExtractor();
    // Extractor lexExtractor = new LexiconExtractor();

    //Extractor dgExtractor = new DependencyMemGrammarExtractor();

    Extractor dgExtractor = new MLEDependencyGrammarExtractor(op);
    if (op.doPCFG) {
      System.err.print("Extracting PCFG...");
      Pair bgug = null;
      if (Train.cheatPCFG) {
        List allTrees = new ArrayList(binaryTrainTrees);
        allTrees.addAll(binaryTestTrees);
        bgug = (Pair) bgExtractor.extract(allTrees);
      } else {
        bgug = (Pair) bgExtractor.extract(binaryTrainTrees);
      }
      bg = (BinaryGrammar) bgug.second;
      bg.splitRules();
      ug = (UnaryGrammar) bgug.first;
      ug.purgeRules();
      Timing.tick("done.");
    }
    System.err.print("Extracting Lexicon...");
    lex = tlpParams.lex(op.lexOptions);
    lex.train(binaryTrainTrees);
    Timing.tick("done.");

    if (op.doDep) {
      System.err.print("Extracting Dependencies...");
      binaryTrainTrees.clear();
      // dgBLIPP = (DependencyGrammar) dgExtractor.extract(new ConcatenationIterator(trainTreebank.iterator(),blippTreebank.iterator()),new TransformTreeDependency(tlpParams,true));

      DependencyGrammar dg1 = (DependencyGrammar) dgExtractor.extract(trainTreebank.iterator(), new TransformTreeDependency(tlpParams, true));
      //dgBLIPP=(DependencyGrammar)dgExtractor.extract(blippTreebank.iterator(),new TransformTreeDependency(tlpParams));

      //dg = (DependencyGrammar) dgExtractor.extract(new ConcatenationIterator(trainTreebank.iterator(),blippTreebank.iterator()),new TransformTreeDependency(tlpParams));
      // dg=new DependencyGrammarCombination(dg1,dgBLIPP,2);
      //dg = (DependencyGrammar) dgExtractor.extract(binaryTrainTrees); //uses information whether the words are known or not, discards unknown words
      Timing.tick("done.");
      //System.out.print("Extracting Unknown Word Model...");
      //UnknownWordModel uwm = (UnknownWordModel)uwmExtractor.extract(binaryTrainTrees);
      //Timing.tick("done.");
      System.out.print("Tuning Dependency Model...");
      dg.tune(binaryTestTrees);
      //System.out.println("TUNE DEPS: "+tuneDeps);
      Timing.tick("done.");
    }

    BinaryGrammar boundBG = bg;
    UnaryGrammar boundUG = ug;

    GrammarProjection gp = new NullGrammarProjection(bg, ug);

    // serialization
    if (serializeFile != null) {
      System.err.print("Serializing parser...");
      LexicalizedParser.saveParserDataToSerialized(new ParserData(lex, bg, ug, dg, Numberer.getNumberers(), op), serializeFile);
      Timing.tick("done.");
    }

    // test: pcfg-parse and output

    ExhaustivePCFGParser parser = null;
    if (op.doPCFG) {
      parser = new ExhaustivePCFGParser(boundBG, boundUG, lex, op);
    }


    ExhaustiveDependencyParser dparser = (op.doDep ? new ExhaustiveDependencyParser(dg, lex, op) : null);

    Scorer scorer = (op.doPCFG ? new TwinScorer(new ProjectionScorer(parser, gp), dparser) : null);
    //Scorer scorer = parser;
    BiLexPCFGParser bparser = null;
    if (op.doPCFG && op.doDep) {
      bparser = (Test.useN5) ? new N5BiLexPCFGParser(scorer, parser, dparser, bg, ug, dg, lex, op, gp) : new BiLexPCFGParser(scorer, parser, dparser, bg, ug, dg, lex, op, gp);
    }

    LabeledConstituentEval pcfgPE = new LabeledConstituentEval("pcfg  PE", tlp);
    LabeledConstituentEval comboPE = new LabeledConstituentEval("combo PE", tlp);
    CBEval pcfgCB = new CBEval("pcfg  CB", tlp);

    TaggingEval pcfgTE = new TaggingEval("pcfg  TE");
    TaggingEval comboTE = new TaggingEval("combo TE");
    TaggingEval pcfgTEnoPunct = new TaggingEval("pcfg nopunct TE");
    TaggingEval comboTEnoPunct = new TaggingEval("combo nopunct TE");
    TaggingEval depTE = new TaggingEval("depnd TE");

    DependencyEval depDE = new DependencyEval("depnd DE", tlp.punctuationWordAcceptFilter());
    DependencyEval comboDE = new DependencyEval("combo DE", tlp.punctuationWordAcceptFilter());

    if (Test.evalb) {
      EvalB.initEVALBfiles(tlpParams);
    }

    // int[] countByLength = new int[Test.maxLength+1];

    // use a reflection ruse, so one can run this without needing the tagger
    //edu.stanford.nlp.process.SentenceTagger tagger = (Test.preTag ? new edu.stanford.nlp.process.SentenceTagger("/u/nlp/data/tagger.params/wsj0-21.holder") : null);
    SentenceProcessor tagger = null;
    if (Test.preTag) {
      try {
        Class[] argsClass = new Class[]{String.class};
        Object[] arguments = new Object[]{"/u/nlp/data/tagger.params/wsj0-21.holder"};
        tagger = (SentenceProcessor) Class.forName("edu.stanford.nlp.process.SentenceTagger").getConstructor(argsClass).newInstance(arguments);
      } catch (Exception e) {
        System.err.println(e);
        System.err.println("Warning: No pretagging of sentences will be done.");
      }
    }

    for (int tNum = 0, ttSize = testTreebank.size(); tNum < ttSize; tNum++) {
      Tree tree = testTreebank.get(tNum);
      int testTreeLen = tree.yield().size();
      if (testTreeLen > Test.maxLength) {
        continue;
      }
      Tree binaryTree = binaryTestTrees.get(tNum);
      // countByLength[testTreeLen]++;
      System.out.println("-------------------------------------");
      System.out.println("Number: " + (tNum + 1));
      System.out.println("Length: " + testTreeLen);

      //tree.pennPrint(pw);
      // System.out.println("XXXX The binary tree is");
      // binaryTree.pennPrint(pw);
      //System.out.println("Here are the tags in the lexicon:");
      //System.out.println(lex.showTags());
      //System.out.println("Here's the tagnumberer:");
      //System.out.println(Numberer.getGlobalNumberer("tags").toString());

      long timeMil1 = System.currentTimeMillis();
      Timing.tick("Starting parse.");
      if (op.doPCFG) {
        //System.err.println(Test.forceTags);
        if (Test.forceTags) {
          if (tagger != null) {
            //System.out.println("Using a tagger to set tags");
            //System.out.println("Tagged sentence as: " + tagger.processSentence(cutLast(wordify(binaryTree.yield()))).toString(false));
            parser.parse(addLast(tagger.processSentence(cutLast(wordify(binaryTree.yield())))));
          } else {
            //System.out.println("Forcing tags to match input.");
            parser.parse(cleanTags(binaryTree.taggedYield(), tlp));
          }
        } else {
          // System.out.println("XXXX Parsing " + binaryTree.yield());
          parser.parse(binaryTree.yield());
        }
        //Timing.tick("Done with pcfg phase.");
      }
      if (op.doDep) {
        dparser.parse(binaryTree.yield());
        //Timing.tick("Done with dependency phase.");
      }
      boolean bothPassed = false;
      if (op.doPCFG && op.doDep) {
        bothPassed = bparser.parse(binaryTree.yield());
        //Timing.tick("Done with combination phase.");
      }
      long timeMil2 = System.currentTimeMillis();
      long elapsed = timeMil2 - timeMil1;
      System.err.println("Time: " + ((int) (elapsed / 100)) / 10.00 + " sec.");
      //System.out.println("PCFG Best Parse:");
      Tree tree2b = null;
      Tree tree2 = null;
      //System.out.println("Got full best parse...");
      if (op.doPCFG) {
        tree2b = parser.getBestParse();
        tree2 = debinarizer.transformTree(tree2b);
      }
      //System.out.println("Debinarized parse...");
      //tree2.pennPrint();
      //System.out.println("DepG Best Parse:");
      Tree tree3 = null;
      Tree tree3db = null;
      if (op.doDep) {
        tree3 = dparser.getBestParse();
        // was: but wrong Tree tree3db = debinarizer.transformTree(tree2);
        tree3db = debinarizer.transformTree(tree3);
        tree3.pennPrint(pw);
      }
      //tree.pennPrint();
      //((Tree)binaryTrainTrees.get(tNum)).pennPrint();
      //System.out.println("Combo Best Parse:");
      Tree tree4 = null;
      if (op.doPCFG && op.doDep) {
        try {
          tree4 = bparser.getBestParse();
          if (tree4 == null) {
            tree4 = tree2b;
          }
        } catch (NullPointerException e) {
          System.err.println("Blocked, using PCFG parse!");
          tree4 = tree2b;
        }
      }
      if (op.doPCFG && !bothPassed) {
        tree4 = tree2b;
      }
      //tree4.pennPrint();
      if (op.doDep) {
        depDE.evaluate(tree3, binaryTree, pw);
        depTE.evaluate(tree3db, tree, pw);
      }
      TreeTransformer tc = tlpParams.collinizer();
      TreeTransformer tcEvalb = tlpParams.collinizerEvalb();
      Tree tree4b = null;
      if (op.doPCFG) {
        // System.out.println("XXXX Best PCFG was: ");
        // tree2.pennPrint();
        // System.out.println("XXXX Transformed best PCFG is: ");
        // tc.transformTree(tree2).pennPrint();
        //System.out.println("True Best Parse:");
        //tree.pennPrint();
        //tc.transformTree(tree).pennPrint();
        pcfgPE.evaluate(tc.transformTree(tree2), tc.transformTree(tree), pw);
        pcfgCB.evaluate(tc.transformTree(tree2), tc.transformTree(tree), pw);
        if (op.doDep) {
          comboDE.evaluate((bothPassed ? tree4 : tree3), binaryTree, pw);
          tree4b = tree4;
          tree4 = debinarizer.transformTree(tree4);
          if (op.nodePrune) {
            NodePruner np = new NodePruner(parser, debinarizer);
            tree4 = np.prune(tree4);
          }
          //tree4.pennPrint();
          comboPE.evaluate(tc.transformTree(tree4), tc.transformTree(tree), pw);
        }
        //pcfgTE.evaluate(tree2, tree);
        pcfgTE.evaluate(tcEvalb.transformTree(tree2), tcEvalb.transformTree(tree), pw);
        pcfgTEnoPunct.evaluate(tc.transformTree(tree2), tc.transformTree(tree), pw);

        if (op.doDep) {
          comboTE.evaluate(tcEvalb.transformTree(tree4), tcEvalb.transformTree(tree), pw);
          comboTEnoPunct.evaluate(tc.transformTree(tree4), tc.transformTree(tree), pw);
        }
        System.out.println("PCFG only: " + parser.scoreBinarizedTree(tree2b, 0));

        //tc.transformTree(tree2).pennPrint();
        tree2.pennPrint(pw);

        if (op.doDep) {
          System.out.println("Combo: " + parser.scoreBinarizedTree(tree4b, 0));
          // tc.transformTree(tree4).pennPrint(pw);
          tree4.pennPrint(pw);
        }
        System.out.println("Correct:" + parser.scoreBinarizedTree(binaryTree, 0));
        /*
        if (parser.scoreBinarizedTree(tree2b,true) < parser.scoreBinarizedTree(binaryTree,true)) {
          System.out.println("SCORE INVERSION");
          parser.validateBinarizedTree(binaryTree,0);
        }
        */
        tree.pennPrint(pw);
      } // end if doPCFG

      if (Test.evalb) {
        if (op.doPCFG && op.doDep) {
          EvalB.writeEVALBline(tcEvalb.transformTree(tree), tcEvalb.transformTree(tree4));
        } else if (op.doPCFG) {
          EvalB.writeEVALBline(tcEvalb.transformTree(tree), tcEvalb.transformTree(tree2));
        } else if (op.doDep) {
          EvalB.writeEVALBline(tcEvalb.transformTree(tree), tcEvalb.transformTree(tree3db));
        }
      }
    } // end for each tree in test treebank

    if (Test.evalb) {
      EvalB.closeEVALBfiles();
    }

    // Test.display();
    if (op.doPCFG) {
      System.out.print("ParseEval for PCFG: ");
      pcfgPE.display(false, pw);
      System.out.println("Grammar size: " + Numberer.getGlobalNumberer("states").total());
      pcfgCB.display(false, pw);
      if (op.doDep) {
        System.out.print("ParseEval for Combo: ");
        comboPE.display(false, pw);
      }
      System.out.print("TagEval for PCFG: ");
      pcfgTE.display(false, pw);
      System.out.print("TagEval for PCFG excluding punct.: ");
      pcfgTEnoPunct.display(false, pw);
      if (op.doDep) {
        System.out.print("TagEval for Combo: ");
        comboTE.display(false, pw);
        System.out.print("TagEval for Combo excluding punct.: ");
        comboTEnoPunct.display(false, pw);
      }
    }
    if (op.doDep) {
      System.out.print("TagEval for Dependency: ");
      depTE.display(false, pw);
      System.out.print("DepEval for Dependency: ");
      depDE.display(false, pw);
    }
    if (op.doPCFG && op.doDep) {
      System.out.print("DepEval for Combo: ");
      comboDE.display(false, pw);
    }
    // pcfgPE.printGoodBad();
  }


  /** Set an option based on flags inside an Options object.
   * Some options:
   * -printTT print the training trees in raw, annotated, and annotated+binarized form.  Useful for debugging and other miscellany.
   * -printAnnotated <filename> use only in conjunction with -printTT.  Redirects printing of annotated training trees to <filename>.
   *
   * @param op An Options argument which is updated according to the option processed.
   * @param args An array of options arguments, command-line style
   * @param i The index in args to start at when processing an option
   * @return The index in args of the position after the last element used in processing
   *     this option, or the value i unchanged if a valid option couldn't be processed
   *     starting at position i.
   */
  static int setOptionFlag(Options op, String[] args, int i) {
    if (args[i].equalsIgnoreCase("-PCFG")) {
      op.doDep = false;
      op.doPCFG = true;
      i++;
    } else if (args[i].equalsIgnoreCase("-dep")) {
      op.doDep = true;
      op.doPCFG = false;
      i++;
    } else if (args[i].equalsIgnoreCase("-factored")) {
      op.doDep = true;
      op.doPCFG = true;
      i++;
    } else if (args[i].equalsIgnoreCase("-noRecoveryTagging")) {
      Test.noRecoveryTagging = true;
      i++;
    } else if (args[i].equalsIgnoreCase("-maxLength") && (i + 1 < args.length)) {
      int length = Integer.parseInt(args[i + 1]);
      Test.maxLength = length;
      i += 2;
    } else if (args[i].equalsIgnoreCase("-MAX_ITEMS") && (i + 1 < args.length)) {
      int items = Integer.parseInt(args[i + 1]);
      Test.MAX_ITEMS = items;
      i += 2;
    } else if (args[i].equalsIgnoreCase("-iterativeCKY")) {
      Test.iterativeCKY = true;
    } else if (args[i].equalsIgnoreCase("-vMarkov") && (i + 1 < args.length)) {
      int order = Integer.parseInt(args[i + 1]);
      if (order <= 1) {
        Train.PA = false;
        Train.gPA = false;
      } else if (order == 2) {
        Train.PA = true;
        Train.gPA = false;
      } else if (order >= 3) {
        Train.PA = true;
        Train.gPA = true;
      }
      i += 2;
    } else if (args[i].equalsIgnoreCase("-vSelSplitCutOff") && (i + 1 < args.length)) {
      Train.selectiveSplitCutOff = Double.parseDouble(args[i + 1]);
      Train.selectiveSplit = Train.selectiveSplitCutOff > 0.0;
      i += 2;
    } else if (args[i].equalsIgnoreCase("-vSelPostSplitCutOff") && (i + 1 < args.length)) {
      Train.selectivePostSplitCutOff = Double.parseDouble(args[i + 1]);
      Train.selectivePostSplit = Train.selectivePostSplitCutOff > 0.0;
      i += 2;
    } else if (args[i].equalsIgnoreCase("-deleteSplitters") && (i+1 < args.length)) {
      String[] toDel = args[i+1].split(" *, *");
      Train.deleteSplitters = new HashSet<String>();
      for (int idx = 0; idx < toDel.length; idx++) {
        Train.deleteSplitters.add(toDel[idx]);
      }
      i += 2;
    } else if (args[i].equalsIgnoreCase("-postSplitWithBaseCategory")) {
      Train.postSplitWithBaseCategory = true;
      i += 1;
    } else if (args[i].equalsIgnoreCase("-vPostMarkov") && (i + 1 < args.length)) {
      int order = Integer.parseInt(args[i + 1]);
      if (order <= 1) {
        Train.postPA = false;
        Train.postGPA = false;
      } else if (order == 2) {
        Train.postPA = true;
        Train.postGPA = false;
      } else if (order >= 3) {
        Train.postPA = true;
        Train.postGPA = true;
      }
      i += 2;
    } else if (args[i].equalsIgnoreCase("-hMarkov") && (i + 1 < args.length)) {
      int order = Integer.parseInt(args[i + 1]);
      if (order >= 0) {
        Train.markovOrder = order;
        Train.markovFactor = true;
      } else {
        Train.markovFactor = false;
      }
      i += 2;
    } else if (args[i].equalsIgnoreCase("-depWeight") && (i + 1 < args.length)) {
      double depWeight = Double.parseDouble(args[i + 1]);
      Test.depWeight = depWeight;
      i += 2;
    } else if (args[i].equalsIgnoreCase("-smoothTagsThresh") && (i + 1 < args.length)) {
      int thresh = Integer.parseInt(args[i + 1]);
      op.lexOptions.smoothInUnknownsThreshold = thresh;
      i += 2;
    } else if (args[i].equalsIgnoreCase("-unseenSmooth") && (i + 1 < args.length)) {
      Test.unseenSmooth = Double.parseDouble(args[i + 1]);
      i += 2;
    } else if (args[i].equalsIgnoreCase("-fractionBeforeUnseenCounting") && (i + 1 < args.length)) {
      Train.fractionBeforeUnseenCounting = Double.parseDouble(args[i + 1]);
      i += 2;
    } else if (args[i].equalsIgnoreCase("-hSelSplitThresh") && (i + 1 < args.length)) {
      int thresh = Integer.parseInt(args[i + 1]);
      Train.HSEL_CUT = thresh;
      i += 2;
    } else if (args[i].equalsIgnoreCase("-hSelSplit")) {
      Train.hSelSplit = true;
      i += 1;
    } else if (args[i].equalsIgnoreCase("-tagPA")) {
      Train.tagPA = true;
      i += 1;
    } else if (args[i].equalsIgnoreCase("-tagSelSplit")) {
      Train.tagSelectiveSplit = true;
      i += 1;
    } else if (args[i].equalsIgnoreCase("-tagSelSplitCutOff") && (i + 1 < args.length)) {
      Train.tagSelectiveSplitCutOff = Double.parseDouble(args[i + 1]);
      i += 2;
    } else if (args[i].equalsIgnoreCase("-tagSelPostSplit")) {
      Train.tagSelectivePostSplit = true;
      i += 1;
    } else if (args[i].equalsIgnoreCase("-tagSelPostSplitCutOff") && (i + 1 < args.length)) {
      Train.tagSelectivePostSplitCutOff = Double.parseDouble(args[i + 1]);
      i += 2;
    } else if (args[i].equalsIgnoreCase("-noTagSplit")) {
      Train.noTagSplit = true;
      i += 1;
    } else if (args[i].equalsIgnoreCase("-uwm") && (i + 1 < args.length)) {
      int order = Integer.parseInt(args[i + 1]);
      if (order < 0) {
        order = 0;
      } else if (order > 5) {
        order = 5;
      }
      op.lexOptions.useUnknownWordSignatures = order;
      i += 2;
    } else if (args[i].equalsIgnoreCase("-openClassThreshold") && (i + 1 < args.length)) {
      int thresh = Integer.parseInt(args[i + 1]);
      Train.openClassTypesThreshold = thresh;
      i += 2;
    } else if (args[i].equalsIgnoreCase("-leaveItAll")) {
      Train.leaveItAll = true;
      i += 1;
    } else if (args[i].equalsIgnoreCase("-unary")) {
      Train.markUnary = true;
      i += 1;
    } else if (args[i].equalsIgnoreCase("-unary2")) {
      Train.markUnary2 = true;
      i += 1;
    } else if (args[i].equalsIgnoreCase("-unaryTags")) {
      Train.markUnaryTags = true;
      i += 1;
    } else if (args[i].equalsIgnoreCase("-mutate")) {
      op.lexOptions.smartMutation = true;
      i += 1;
    } else if (args[i].equalsIgnoreCase("-rightRec")) {
      Train.rightRec = true;
      i += 1;
    } else if (args[i].equalsIgnoreCase("-noRightRec")) {
      Train.rightRec = false;
      i += 1;
    } else if (args[i].equalsIgnoreCase("-preTag")) {
      Test.preTag = true;
      i += 1;
    } else if (args[i].equalsIgnoreCase("-forceTags")) {
      Test.forceTags = true;
      i += 1;
    } else if (args[i].equalsIgnoreCase("-scTags")) {
      op.dcTags = false;
      i += 1;
    } else if (args[i].equalsIgnoreCase("-dcTags")) {
      op.dcTags = true;
      i += 1;
    } else if (args[i].equalsIgnoreCase("-evalb")) {
      Test.evalb = true;
      i += 1;
    } else if (args[i].equalsIgnoreCase("-v") || args[i].equalsIgnoreCase("-verbose")) {
      Test.verbose = true;
      i += 1;
    } else if (args[i].equalsIgnoreCase("-writeOutputFiles")) {
      Test.writeOutputFiles = true;
      i += 1;
    } else if (args[i].equalsIgnoreCase("-printAllBestParses")) {
      Test.printAllBestParses = true;
      i += 1;
    } else if (args[i].equalsIgnoreCase("-outputTreeFormat") || args[i].equalsIgnoreCase("-outputFormat")) {
      Test.outputFormat = args[i + 1];
      i += 2;
    } else if (args[i].equalsIgnoreCase("-outputTreeFormatOptions") || args[i].equalsIgnoreCase("-outputFormatOptions")) {
      Test.outputFormatOptions = args[i + 1];
      i += 2;
    } else if (args[i].equalsIgnoreCase("-flexiTag")) {
      op.flexiTag = true;
      i += 1;
    } else if (args[i].equalsIgnoreCase("-lexiTag")) {
      op.flexiTag = false;
      i += 1;
    } else if (args[i].equalsIgnoreCase("-compactGrammar")) {
      Train.compactGrammar = Integer.parseInt(args[i + 1]);
      i += 2;
    } else if (args[i].equalsIgnoreCase("-markFinalStates")) {
      Train.markFinalStates = args[i + 1].equalsIgnoreCase("true");
      i += 2;
    } else if (args[i].equalsIgnoreCase("-leftToRight")) {
      Train.leftToRight = args[i + 1].equals("true");
      i += 2;
    } else if (args[i].equalsIgnoreCase("-cnf")) {
      op.forceCNF = true;
      i += 1;
    } else if (args[i].equalsIgnoreCase("-nodePrune") && i+1 < args.length) {
      op.nodePrune = args[i+1].equalsIgnoreCase("true");
      i += 2;
    } else if (args[i].equalsIgnoreCase("-acl03pcfg")) {
      op.doDep = false;
      op.doPCFG = true;
      // op.lexOptions.smoothInUnknownsThreshold = 30;
      Train.markUnary = true;
      Train.PA = true;
      Train.gPA = false;
      Train.tagPA = true;
      Train.tagSelectiveSplit = false;
      Train.rightRec = true;
      Train.selectiveSplit = true;
      Train.selectiveSplitCutOff = 400.0;
      Train.markovFactor = true;
      Train.markovOrder = 2;
      Train.hSelSplit = true;
      op.lexOptions.useUnknownWordSignatures = 2;
      op.flexiTag = true;
      // DAN: Tag double-counting is BAD for PCFG-only parsing
      op.dcTags = false;
      // don't increment i so it gets language specific stuff as well
    } else if (args[i].equalsIgnoreCase("-goodPCFG")) {
      op.doDep = false;
      op.doPCFG = true;
      // op.lexOptions.smoothInUnknownsThreshold = 30;
      Train.markUnary = true;
      Train.PA = true;
      Train.gPA = false;
      Train.tagPA = true;
      Train.tagSelectiveSplit = false;
      Train.rightRec = true;
      Train.selectiveSplit = true;
      Train.selectiveSplitCutOff = 400.0;
      Train.markovFactor = true;
      Train.markovOrder = 2;
      Train.hSelSplit = true;
      op.lexOptions.useUnknownWordSignatures = 2;
      op.flexiTag = true;
      // DAN: Tag double-counting is BAD for PCFG-only parsing
      op.dcTags = false;
      // don't increment i so it gets language specific stuff as well
    } else if (args[i].equalsIgnoreCase("-linguisticPCFG")) {
      op.doDep = false;
      op.doPCFG = true;
      // op.lexOptions.smoothInUnknownsThreshold = 30;
      Train.markUnary = true;
      Train.PA = true;
      Train.gPA = false;
      Train.tagPA = true;        // on at the moment, but iffy
      Train.tagSelectiveSplit = false;
      Train.rightRec = false;    // not for linguistic
      Train.selectiveSplit = true;
      Train.selectiveSplitCutOff = 400.0;
      Train.markovFactor = true;
      Train.markovOrder = 2;
      Train.hSelSplit = true;
      op.lexOptions.useUnknownWordSignatures = 5;   // different from acl03pcfg
      op.flexiTag = false;       // different from acl03pcfg
      // DAN: Tag double-counting is BAD for PCFG-only parsing
      op.dcTags = false;
      // don't increment i so it gets language specific stuff as well
    } else if (args[i].equalsIgnoreCase("-ijcai03")) {
      op.doDep = true;
      op.doPCFG = true;
      Train.markUnary = false;
      Train.PA = true;
      Train.gPA = false;
      Train.tagPA = false;
      Train.tagSelectiveSplit = false;
      Train.rightRec = false;
      Train.selectiveSplit = true;
      Train.selectiveSplitCutOff = 300.0;
      Train.markovFactor = true;
      Train.markovOrder = 2;
      Train.hSelSplit = true;
      Train.compactGrammar = 0; /// cdm: May 2005 compacting bad for factored?
      op.lexOptions.useUnknownWordSignatures = 2;
      op.flexiTag = false;
      op.dcTags = true;
      // op.nodePrune = true;  // cdm: May 2005: this doesn't help
      // don't increment i so it gets language specific stuff as well
    } else if (args[i].equalsIgnoreCase("-goodFactored")) {
      op.doDep = true;
      op.doPCFG = true;
      Train.markUnary = false;
      Train.PA = true;
      Train.gPA = false;
      Train.tagPA = false;
      Train.tagSelectiveSplit = false;
      Train.rightRec = false;
      Train.selectiveSplit = true;
      Train.selectiveSplitCutOff = 300.0;
      Train.markovFactor = true;
      Train.markovOrder = 2;
      Train.hSelSplit = true;
      Train.compactGrammar = 0; /// cdm: May 2005 compacting bad for factored?
      op.lexOptions.useUnknownWordSignatures = 5;  // different from ijcai03
      op.flexiTag = false;
      op.dcTags = true;
      // op.nodePrune = true;  // cdm: May 2005: this doesn't help
      // don't increment i so it gets language specific stuff as well
    } else if (args[i].equalsIgnoreCase("-chinesePCFG")) {
      op.doDep = false;
      op.doPCFG = true;
      // Single counting tag->word rewrite is also much better for Chinese PCFG
      // Bracketing F1 is up about 2% and tag accuracy by about 1% (exact by 6%)
      op.dcTags = false;
    } else if (args[i].equalsIgnoreCase("-printTT")) {
      Train.printTreeTransformations = true;
      i++;
    } else if (args[i].equalsIgnoreCase("-printAnnotated") && (i + 1 < args.length)) {
      try {
        Train.printAnnotatedPW = op.tlpParams.pw(new FileOutputStream(args[i + 1]));
      } catch (IOException ioe) {
        Train.printAnnotatedPW = null;
      }
      i += 2;
    } else if (args[i].equalsIgnoreCase("-printBinarized") && (i + 1 < args.length)) {
      try {
        Train.printBinarizedPW = op.tlpParams.pw(new FileOutputStream(args[i + 1]));
      } catch (IOException ioe) {
        Train.printBinarizedPW = null;
      }
      i += 2;
    } else if (args[i].equalsIgnoreCase("-printStates")) {
      Train.printStates = true;
      i++;
    }
    return i;
  }


  private static List cleanTags(List twList, TreebankLanguagePack tlp) {
    List l = new ArrayList();
    for (int i = 0; i < twList.size(); i++) {
      TaggedWord tw = (TaggedWord) twList.get(i);
      TaggedWord tw2 = new TaggedWord(tw.word(), tlp.basicCategory(tw.tag()));
      l.add(tw2);
    }
    return l;
  }

  private static Sentence wordify(List wList) {
    Sentence s = new Sentence();
    for (int i = 0; i < wList.size(); i++) {
      s.add(new Word(wList.get(i).toString()));
    }
    return s;
  }

  private static Sentence cutLast(Sentence s) {
    return new Sentence(s.subList(0, s.size() - 1));
  }

  private static Sentence addLast(Sentence s) {
    Sentence s2 = new Sentence(s);
    //s2.add(new StringLabel(Lexicon.BOUNDARY));
    s2.add(new Word(Lexicon.BOUNDARY));
    return s2;
  }

  /**
   * Not an instantiable class
   */
  private FactoredParser() {
  }

}
