package edu.stanford.nlp.parser.lexparser;

import edu.stanford.nlp.fsm.TransducerGraph;
import edu.stanford.nlp.stats.Counter;
import edu.stanford.nlp.stats.Distribution;
import edu.stanford.nlp.util.Numberer;
import edu.stanford.nlp.util.Pair;

import java.io.*;
import java.util.*;

/**
 * @author Teg Grenager (grenager@cs.stanford.edu)
 */
public abstract class GrammarCompactor {

  // so that the grammar remembers its graphs after compacting them
  Set compactedGraphs;

  public static final Object RAW_COUNTS = new Object();
  public static final Object NORMALIZED_LOG_PROBABILITIES = new Object();

  public Object outputType = RAW_COUNTS; // default value

  protected Numberer stateNumberer;
  protected Numberer newStateNumberer;
  protected String stateSpace;

  String rawBaseDir = "raw";
  String compactedBaseDir = "compacted";
  boolean writeToFile = false;
  protected Distribution inputPrior;
  private static Object END = "END";
  private static Object EPSILON = "EPSILON";
  protected boolean verbose = false;


  protected abstract TransducerGraph doCompaction(TransducerGraph graph, List trainPaths, List testPaths);

  public Pair compactGrammar(Pair grammar) {
    return compactGrammar(grammar, new HashMap(), new HashMap());
  }

  /**
   * Compacts the grammar specified by the Pair.
   *
   * @param grammar       a Pair of grammars, ordered UnaryGrammar BinaryGrammar.
   * @param allTrainPaths a Map from String passive constituents to Lists of paths
   * @param allTestPaths  a Map from String passive constituents to Lists of paths
   * @return a Pair of grammars, ordered UnaryGrammar BinaryGrammar.
   */
  public Pair compactGrammar(Pair grammar, Map allTrainPaths, Map allTestPaths) {
    inputPrior = computeInputPrior(allTrainPaths); // computed once for the whole grammar
    BinaryGrammar bg = (BinaryGrammar) grammar.second;
    stateSpace = bg.stateSpace();
    stateNumberer = Numberer.getGlobalNumberer(stateSpace);
    List trainPaths, testPaths;
    Set unaryRules = new HashSet();
    Set binaryRules = new HashSet();
    Map graphs = convertGrammarToGraphs(grammar, unaryRules, binaryRules);
    compactedGraphs = new HashSet();
    if (verbose) {
      System.out.println("There are " + graphs.size() + " categories to compact.");
    }
    int i = 0;
    for (Iterator graphIter = graphs.entrySet().iterator(); graphIter.hasNext();) {
      Map.Entry entry = (Map.Entry) graphIter.next();
      String cat = (String) entry.getKey();
      TransducerGraph graph = (TransducerGraph) entry.getValue();
      if (verbose) {
        System.out.println("About to compact grammar for " + cat + " with numNodes=" + graph.getNodes().size());
      }
      trainPaths = (List) allTrainPaths.remove(cat);// to save memory
      if (trainPaths == null) {
        trainPaths = new ArrayList();
      }
      testPaths = (List) allTestPaths.remove(cat);// to save memory
      if (testPaths == null) {
        testPaths = new ArrayList();
      }
      TransducerGraph compactedGraph = doCompaction(graph, trainPaths, testPaths);
      i++;
      if (verbose) {
        System.out.println(i + ". Compacted grammar for " + cat + " from " + graph.getArcs().size() + " arcs to " + compactedGraph.getArcs().size() + " arcs.");
      }
      graphIter.remove(); // to save memory, remove the last thing
      compactedGraphs.add(compactedGraph);
    }
    Pair result = convertGraphsToGrammar(compactedGraphs, unaryRules, binaryRules);
    return result;
  }

  protected Distribution computeInputPrior(Map trainPathMap) {
    Counter result = new Counter();
    for (Iterator catI = trainPathMap.values().iterator(); catI.hasNext();) {
      List pathList = (List) catI.next();
      for (Iterator pathI = pathList.iterator(); pathI.hasNext();) {
        List path = (List) pathI.next();
        for (Iterator inputI = path.iterator(); inputI.hasNext();) {
          Object input = inputI.next();
          result.incrementCount(input);
        }
      }
    }
    return Distribution.laplaceSmoothedDistribution(result, result.size() * 2, 0.5);
  }

  private double smartNegate(double output) {
    if (outputType == NORMALIZED_LOG_PROBABILITIES) {
      return -output;
    }
    return output;
  }

  public static boolean writeFile(TransducerGraph graph, String dir, String name) {
    try {
      File baseDir = new File(dir);
      if (baseDir.exists()) {
        if (!baseDir.isDirectory()) {
          return false;
        }
      } else {
        if (!baseDir.mkdirs()) {
          return false;
        }
      }
      File file = new File(baseDir, name + ".dot");
      PrintWriter w;
      try {
        w = new PrintWriter(new FileWriter(file));
        String dotString = graph.asDOTString();
        w.print(dotString);
        w.flush();
        w.close();
      } catch (FileNotFoundException e) {
        System.err.println("Failed to open file in writeToDOTfile: " + file);
        return false;
      } catch (IOException e) {
        System.err.println("Failed to open file in writeToDOTfile: " + file);
        return false;
      }
      return true;
    } catch (Exception e) {
      e.printStackTrace();
      return false;
    }
  }

  /**
   *
   */
  public Map convertGrammarToGraphs(Pair grammar, Set unaryRules, Set binaryRules) {
    int numRules = 0;
    UnaryGrammar ug = (UnaryGrammar) grammar.first;
    BinaryGrammar bg = (BinaryGrammar) grammar.second;
    Map graphs = new HashMap();
    // go through the BinaryGrammar and add everything
    for (Iterator ruleIter = bg.iterator(); ruleIter.hasNext();) {
      BinaryRule rule = (BinaryRule) ruleIter.next();
      numRules++;
      boolean wasAdded = addOneBinaryRule(rule, graphs);
      if (!wasAdded)
      // add it for later, since we don't make graphs for these
      {
        binaryRules.add(rule);
      }
    }
    // now we need to use the UnaryGrammar to
    // add start and end Arcs to the graphs
    for (Iterator ruleIter = ug.iterator(); ruleIter.hasNext();) {
      UnaryRule rule = (UnaryRule) ruleIter.next();
      numRules++;
      boolean wasAdded = addOneUnaryRule(rule, graphs);
      if (!wasAdded)
      // add it for later, since we don't make graphs for these
      {
        unaryRules.add(rule);
      }
    }
    if (verbose) {
      System.out.println("Number of raw rules: " + numRules);
      System.out.println("Number of raw states: " + stateNumberer.total());
    }
    return graphs;
  }

  protected TransducerGraph getGraphFromMap(Map m, Object o) {
    TransducerGraph graph = (TransducerGraph) m.get(o);
    if (graph == null) {
      graph = new TransducerGraph();
      graph.setEndNode(o);
      m.put(o, graph);
    }
    return graph;
  }

  protected String getTopCategoryOfSyntheticState(String s) {
    if (s.charAt(0) != '@') {
      return null;
    }
    int bar = s.indexOf('|');
    if (bar < 0) {
      throw new RuntimeException("Grammar format error. Expected bar in state name: " + s);
    }
    String topcat = s.substring(1, bar);
    return topcat;
  }

  protected boolean addOneUnaryRule(UnaryRule rule, Map graphs) {
    String parentString = (String) stateNumberer.object(rule.parent);
    String childString = (String) stateNumberer.object(rule.child);
    if (isSyntheticState(parentString)) {
      String topcat = getTopCategoryOfSyntheticState(parentString);
      TransducerGraph graph = getGraphFromMap(graphs, topcat);
      Double output = new Double(smartNegate(rule.score()));
      graph.addArc(graph.getStartNode(), parentString, childString, output);
      return true;
    } else if (isSyntheticState(childString)) {
      // need to add Arc from synthetic state to endState
      TransducerGraph graph = getGraphFromMap(graphs, parentString);
      Double output = new Double(smartNegate(rule.score()));
      graph.addArc(childString, parentString, END, output); // parentString should the the same as endState
      graph.setEndNode(parentString);
      return true;
    } else {
      return false;
    }
  }

  protected boolean addOneBinaryRule(BinaryRule rule, Map graphs) {
    // parent has to be synthetic in BinaryRule
    String parentString = (String) stateNumberer.object(rule.parent);
    String leftString = (String) stateNumberer.object(rule.leftChild);
    String rightString = (String) stateNumberer.object(rule.rightChild);
    String source, target, input;
    String bracket = null;
    if (Train.markFinalStates) {
      bracket = parentString.substring(parentString.length() - 1, parentString.length());
    }
    // the below test is not necessary with left to right grammars
    if (isSyntheticState(leftString)) {
      source = leftString;
      input = rightString + (bracket == null ? ">" : bracket);
    } else if (isSyntheticState(rightString)) {
      source = rightString;
      input = leftString + (bracket == null ? "<" : bracket);
    } else {
      // we don't know what to do with this rule
      return false;
    }
    target = parentString;
    Double output = new Double(smartNegate(rule.score())); // makes it a real  0 <= k <= infty
    String topcat = getTopCategoryOfSyntheticState(source);
    if (topcat == null) {
      throw new RuntimeException("can't have null topcat");
    }
    TransducerGraph graph = getGraphFromMap(graphs, topcat);
    graph.addArc(source, target, input, output);
    return true;
  }

  protected boolean isSyntheticState(String state) {
    return state.charAt(0) == '@';
  }


  /**
   * @param graphs      a Map from String categories to TransducerGraph objects
   * @param unaryRules  is a Set of UnaryRule objects that we need to add
   * @param binaryRules is a Set of BinaryRule objects that we need to add
   * @return a new Pair of UnaryGrammar, BinaryGrammar
   */
  public Pair convertGraphsToGrammar(Set graphs, Set unaryRules, Set binaryRules) {
    // first go through all the existing rules and number them with new numberer
    newStateNumberer = new Numberer();
    for (Iterator urIter = unaryRules.iterator(); urIter.hasNext();) {
      UnaryRule rule = (UnaryRule) urIter.next();
      Object parent = stateNumberer.object(rule.parent);
      rule.parent = newStateNumberer.number(parent);
      Object child = stateNumberer.object(rule.child);
      rule.child = newStateNumberer.number(child);
    }
    for (Iterator urIter = binaryRules.iterator(); urIter.hasNext();) {
      BinaryRule rule = (BinaryRule) urIter.next();
      Object parent = stateNumberer.object(rule.parent);
      rule.parent = newStateNumberer.number(parent);
      Object leftChild = stateNumberer.object(rule.leftChild);
      rule.leftChild = newStateNumberer.number(leftChild);
      Object rightChild = stateNumberer.object(rule.rightChild);
      rule.rightChild = newStateNumberer.number(rightChild);
    }


    // put the new (smaller) numberer in place of the old one
    Map numbs = Numberer.getNumberers();
    numbs.put(stateSpace, newStateNumberer);

    // now go through the graphs and add the rules
    for (Iterator graphIter = graphs.iterator(); graphIter.hasNext();) {
      TransducerGraph graph = (TransducerGraph) graphIter.next();
      Object startNode = graph.getStartNode();
      for (Iterator arcIter = graph.getArcs().iterator(); arcIter.hasNext();) {
        TransducerGraph.Arc arc = (TransducerGraph.Arc) arcIter.next();
        Object source = arc.getSourceNode();
        Object target = arc.getTargetNode();
        Object input = arc.getInput();
        String inputString = input.toString();
        double output = ((Double) arc.getOutput()).doubleValue();
        if (source.equals(startNode)) {
          // make a UnaryRule
          UnaryRule ur = new UnaryRule(newStateNumberer.number(target), newStateNumberer.number(inputString), smartNegate(output));
          unaryRules.add(ur);
        } else if (inputString.equals(END) || inputString.equals(EPSILON)) {
          // make a UnaryRule
          UnaryRule ur = new UnaryRule(newStateNumberer.number(target), newStateNumberer.number(source), smartNegate(output));
          unaryRules.add(ur);
        } else {
          // make a BinaryRule
          // figure out whether the input was generated on the left or right
          int length = inputString.length();
          char leftOrRight = inputString.charAt(length - 1);
          inputString = inputString.substring(0, length - 1);
          BinaryRule br = null;
          if (leftOrRight == '<' || leftOrRight == '[') {
            br = new BinaryRule(newStateNumberer.number(target), newStateNumberer.number(inputString), newStateNumberer.number(source), smartNegate(output));
          } else if (leftOrRight == '>' || leftOrRight == ']') {
            br = new BinaryRule(newStateNumberer.number(target), newStateNumberer.number(source), newStateNumberer.number(inputString), smartNegate(output));
          } else {
            throw new RuntimeException("Arc input is in unexpected format: " + arc);
          }
          binaryRules.add(br);
        }
      }
    }
    // by now, the unaryRules and binaryRules Sets have old untouched and new rules with scores
    Counter symbolCounter = new Counter();
    if (outputType == RAW_COUNTS) {
      // now we take the sets of rules and turn them into grammars
      // the scores of the rules we are given are actually counts
      // so we count parent symbol occurences
      for (Iterator ruleIter = unaryRules.iterator(); ruleIter.hasNext();) {
        UnaryRule rule = (UnaryRule) ruleIter.next();
        symbolCounter.incrementCount(newStateNumberer.object(rule.parent), rule.score);
      }
      for (Iterator ruleIter = binaryRules.iterator(); ruleIter.hasNext();) {
        BinaryRule rule = (BinaryRule) ruleIter.next();
        symbolCounter.incrementCount(newStateNumberer.object(rule.parent), rule.score);
      }
    }
    // now we put the rules in the grammars
    int numStates = newStateNumberer.total();     // this should be smaller than last one
    int numRules = 0;
    UnaryGrammar ug = new UnaryGrammar(numStates);
    BinaryGrammar bg = new BinaryGrammar(numStates);
    for (Iterator ruleIter = unaryRules.iterator(); ruleIter.hasNext();) {
      UnaryRule rule = (UnaryRule) ruleIter.next();
      if (outputType == RAW_COUNTS) {
        double count = symbolCounter.getCount(newStateNumberer.object(rule.parent));
        rule.score = Math.log(rule.score / count);
      }
      ug.addRule(rule);
      numRules++;
    }
    for (Iterator ruleIter = binaryRules.iterator(); ruleIter.hasNext();) {
      BinaryRule rule = (BinaryRule) ruleIter.next();
      if (outputType == RAW_COUNTS) {
        double count = symbolCounter.getCount(newStateNumberer.object(rule.parent));
        rule.score = Math.log((rule.score - Train.ruleDiscount) / count);
      }
      bg.addRule(rule);
      numRules++;
    }
    if (verbose) {
      System.out.println("Number of minimized rules: " + numRules);
      System.out.println("Number of minimized states: " + newStateNumberer.total());
    }

    ug.purgeRules();
    bg.splitRules();
    return new Pair(ug, bg);
  }


}
