package edu.stanford.nlp.parser.lexparser;

import edu.stanford.nlp.ling.TaggedWord;
import edu.stanford.nlp.util.Numberer;
import edu.stanford.nlp.util.StringUtils;

import java.io.Serializable;

public class IntTaggedWord implements Serializable, Comparable {

  private static Numberer wordNumberer;
  private static Numberer tagNumberer;

  Numberer wordNumberer() {
    if (wordNumberer == null) {
      wordNumberer = Numberer.getGlobalNumberer("words");
    }
    return wordNumberer;
  }

  Numberer tagNumberer() {
    if (tagNumberer == null) {
      tagNumberer = Numberer.getGlobalNumberer("tags");
    }
    return tagNumberer;
  }

  static void resetNumberers() {
    wordNumberer = null;
    tagNumberer = null;
  }



  public int word;
  public short tag;

  public static final String ANY = ".*.";
  public static final String STOP = "STOP";

  public int tag() {
    return tag;
  }

  public int word() {
    return word;
  }

  public String wordString() {
    String wordStr;
    if (word >= 0) {
      wordStr = (String) wordNumberer().object(word);
    } else if (word == -1) {
      wordStr = ANY;
    } else {
      wordStr = STOP;
    }
    return wordStr;
  }

  public String tagString() {
    String tagStr;
    if (tag >= 0) {
      tagStr = (String) tagNumberer().object(tag);
    } else if (tag == -1) {
      tagStr = ANY;
    } else {
      tagStr = STOP;
    }
    return tagStr;
  }

  public int hashCode() {
    return word ^ (tag << 16);
  }

  public boolean equals(Object o) {
    if (this == o) {
      return true;
    } else if (o instanceof IntTaggedWord) {
      IntTaggedWord i = (IntTaggedWord) o;
      return (word == i.word && tag == i.tag);
    } else {
      return false;
    }
  }

  public int compareTo(Object obj) {
    IntTaggedWord that = (IntTaggedWord) obj;
    if (tag != that.tag) {
      return tag - that.tag;
    } else {
      return word - that.word;
    }
  }

  static final char[] charsToEscape = new char[]{'\"'};


  public String toLexicalEntry() {
    String wordStr = wordString();
    String tagStr = tagString();
    return "\"" + StringUtils.escapeString(tagStr, charsToEscape, '\\') + "\" -> \"" + StringUtils.escapeString(wordStr, charsToEscape, '\\') + "\"";
  }

  public String toString() {
    return wordString() + "/" + tagString();
  }

  public String toString(String arg) {
    if (arg.equals("verbose")) {
      return wordString() + "[" + word + "]/" + tagString() + "[" + tag + "]";
    } else {
      return toString();
    }
  }

  public IntTaggedWord(int word, int tag) {
    this.word = word;
    this.tag = (short) tag;
  }

  public TaggedWord toTaggedWord() {
    String wordStr = wordString();
    String tagStr = tagString();
    return new TaggedWord(wordStr, tagStr);
  }

  /**
   * Creates an IntTaggedWord given by the String representation
   * of the form <word>|<tag>
   */
  public IntTaggedWord(String s, char splitChar) {
    // awkward, calls s.indexOf(splitChar) twice
    this(extractWord(s, splitChar), extractTag(s, splitChar));
    //    System.out.println("s: " + s);
    //    System.out.println("tagNumberer: " + tagNumberer);
    //    System.out.println("word: " + word);
    //    System.out.println("tag: " + tag);
  }

  private static String extractWord(String s, char splitChar) {
    int n = s.lastIndexOf(splitChar);
    String result = s.substring(0, n);
    //    System.out.println("extracted word: " + result);
    return result;
  }

  private static String extractTag(String s, char splitChar) {
    int n = s.lastIndexOf(splitChar);
    String result = s.substring(n + 1);
    //    System.out.println("extracted tag: " + result);
    return result;
  }

  /**
   * Creates an IntTaggedWord given by the tagString and wordString
   */
  public IntTaggedWord(String wordString, String tagString) {
    int word, tag;
    if (wordString.equals(ANY)) {
      word = -1;
    } else if (wordString.equals(STOP)) {
      word = -2;
    } else {
      word = wordNumberer().number(wordString);
    }
    if (tagString.equals(ANY)) {
      tag = -1;
    } else if (wordString.equals(STOP)) {
      tag = -2;
    } else {
      tag = tagNumberer().number(tagString);
    }
    this.word = word;
    this.tag = (short) tag;
  }

  private static final long serialVersionUID = 1L;

} // end class IntTaggedWord
