// StanfordLexicalizedParser -- a probabilistic lexicalized NL CFG parser
// Copyright (c) 2002, 2003, 2004, 2005 The Board of Trustees of 
// The Leland Stanford Junior University. All Rights Reserved.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
// For more information, bug reports, fixes, contact:
//    Christopher Manning
//    Dept of Computer Science, Gates 4A
//    Stanford CA 94305-9040
//    USA
//    parser-support@lists.stanford.edu
//    http://nlp.stanford.edu/downloads/lex-parser.shtml

package edu.stanford.nlp.parser.lexparser;

import edu.stanford.nlp.fsm.ExactGrammarCompactor;
import edu.stanford.nlp.io.NumberRangeFileFilter;
import edu.stanford.nlp.io.NumberRangesFileFilter;
import edu.stanford.nlp.ling.HasWord;
import edu.stanford.nlp.ling.Sentence;
import edu.stanford.nlp.objectbank.TokenizerFactory;
import edu.stanford.nlp.parser.ViterbiParser;
import edu.stanford.nlp.process.DocumentPreprocessor;
import edu.stanford.nlp.process.Function;
import edu.stanford.nlp.process.WhitespaceTokenizer;
import edu.stanford.nlp.trees.*;
import edu.stanford.nlp.util.Numberer;
import edu.stanford.nlp.util.Pair;
import edu.stanford.nlp.util.Timing;

import java.io.*;
import java.net.URL;
import java.net.URLConnection;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.*;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;

// Miscellaneous documentation by Roger (please correct any errors in this documentation!)
//
// I believe that a lexicalized parser is always constructed by creating a ParserData object and then calling
// makeParsers().
//

/**
 * This class provides the top-level API and command-line interface to a set
 * of reasonably good treebank-trained parsers.  The name reflects the main
 * factored parsing model, which provides a lexicalized PCFG parser 
 * implemented as a product
 * model of a plain PCFG parser and a lexicalized dependency parser.  
 * But you can also run either component parser alone.  In particular, it
 * is often useful to do unlexicalized PCFG parsing by using just that 
 * component parser. 
 * <p>
 * See the package documentation for more details and examples of use.
 * See the main method documentation for details of invoking the parser.
 * <p>
 * Note that training requires a lot of memory to run.  Try -mx1500m.
 *
 * @author Dan Klein (original version)
 * @author Christopher Manning (better features, ParserParams, serialization)
 * @author Roger Levy (internationalization)
 * @author Teg Grenager (grammar compaction, tokenization, etc.)
 * @author Galen Andrew (considerable refactoring)
 */
public class LexicalizedParser implements ViterbiParser, Function {

  // this should not really be static but for the moment it is, because I don't want
  // to track down all the places that i'd have to thread it down into -- Roger
  static boolean basicCategoryTagsInDependencyGrammar = false;

  private boolean fallbackToPCFG = false;

  /** The PCFG parser. */
  protected ExhaustivePCFGParser pparser;
  /** The dependency parser. */
  protected ExhaustiveDependencyParser dparser;
  /** The factored parser that combines the dependency and PCFG parsers. */
  protected BiLexPCFGParser bparser;
  protected TreeTransformer debinarizer;
  private TreeTransformer subcategoryStripper;

  private transient ParserData pd;

  public Options getOp() {
    return op;
  }

  private Options op;

  private static final String SERIALIZED_PARSER_PROPERTY = "edu.stanford.nlp.SerializedLexicalizedParser";
  private static final String DEFAULT_PARSER_LOC = "/u/nlp/data/lexparser/englishPCFG.ser.gz";


  /**
   * Converts a Sentence/List/String into a Tree.  If it can't be parsed,
   * it is made into a trivial tree in which each word is attached to a 
   * dummy tag ("X") and then to a start nonterminal (also "X").
   *
   * @param in The input Sentence/List/String
   * @return A Tree that is the parse tree for the sentence.  If the parser
   *         fails, a new Tree is synthesized which attaches all words to the
   *         root.
   * @throws IllegalArgumentException If argument isn't a List or String
   */
  public Object apply(Object in) {
    List lst;
    if (in instanceof String) {
      DocumentPreprocessor dp = new DocumentPreprocessor(op.tlpParams.treebankLanguagePack().getTokenizerFactory());
      lst = dp.getWordsFromString((String) in);
    } else if (in instanceof List) {
      lst = (List) in;
    } else {
      throw new IllegalArgumentException("Can only parse Sentence/List/String");
    }

    try {
      if (parse(lst)) {
        return getBestParse();
      }
    } catch (Exception e) {
      e.printStackTrace();
    }
    // if can't parse or exception, fall through
    TreeFactory lstf = new LabeledScoredTreeFactory();
    List lst2 = new Sentence();
    for (Iterator it = lst.iterator(); it.hasNext();) {
      String s = it.next().toString();
      Tree t = lstf.newLeaf(s);
      Tree t2 = lstf.newTreeNode("X", Collections.singletonList(t));
      lst2.add(t2);
    }
    return lstf.newTreeNode("X", lst2);
  }


  /**
   * Parse a Sentence.
   *
   * @return true iff it could be parsed
   */
  public boolean parse(Sentence sentence) {
    return parse((List) sentence);
  }


  /**
   * Parse a Sentence.
   * This hasn't yet been implemented.  At present the goal is ignored.
   *
   * @return true iff it could be parsed
   */
  public boolean parse(Sentence sentence, String goal) {
    return parse((List) sentence);
  }


  /**
   * Return a TreePrint for formatting parsed output trees.
   */
  public TreePrint getTreePrint() {
    return Test.treePrint(op.tlpParams);
  }

  private static boolean parseSucceeded = false;

  /**
   * Tokenize and parse a sentence.
   *
   * @param sentence
   * @return true iff it could be parsed
   */

  public boolean parse(String sentence) {
    DocumentPreprocessor dp = new DocumentPreprocessor(op.tlpParams.treebankLanguagePack().getTokenizerFactory());

    return parse(dp.getWordsFromString(sentence));
  }

  /**
   * Parse a sentence represented as a List.
   *
   * @param sentence The sentence to parse
   * @return true Iff the sentence was accepted by the grammar
   * @throws UnsupportedOperationException If the Sentence is too long or
   *                                       of zero length or the parse
   *                                       otherwise fails for resource reasons
   */
  public boolean parse(List sentence) {
    int length = sentence.size();
    if (length == 0) {
      throw new UnsupportedOperationException("Can't parse a zero-length sentence!");
    }
    if (length > Test.maxLength) { 
      throw new UnsupportedOperationException("Sentence too long: length " + length);
    }
    TreePrint treePrint = getTreePrint();
    PrintWriter pwOut = op.tlpParams.pw();
    parseSucceeded = false;
    List sentenceB = new ArrayList(sentence);
    sentenceB.add(Lexicon.BOUNDARY);
    if (op.doPCFG) {
      if (!pparser.parse(sentenceB)) {
        return parseSucceeded;
      }
      if (Test.verbose) {
        System.out.println("PParser output");
        treePrint.printTree(debinarizer.transformTree(pparser.getBestParse()), pwOut);
      }
    }
    if (op.doDep) {
      if (!dparser.parse(sentenceB)) {
        return parseSucceeded;
      }
      if (Test.verbose) {
        System.out.println("DParser output");
        treePrint.printTree(dparser.getBestParse(), pwOut);
      }
    }
    if (op.doPCFG && op.doDep) {
      if (!bparser.parse(sentenceB)) {
        return parseSucceeded;
      } else {
        parseSucceeded = true;
      }
    }
    return true;
  }

  /**
   * Parse a lattice with PCFG parser.
   *
   * @param lr a lattice to parse
   * @return Whether the lattice could be parsed by the grammar
   */
  public boolean parse(LatticeReader lr) {
    TreePrint treePrint = getTreePrint();
    PrintWriter pwOut = op.tlpParams.pw();
    parseSucceeded = false;
    if (lr.getNumStates() > Test.maxLength + 1) {  // + 1 for boundary symbol
      throw new UnsupportedOperationException("Lattice too big: " + lr.getNumStates());
    }
    if (op.doPCFG) {
      if (!pparser.parse(lr)) {
        return parseSucceeded;
      }
      if (Test.verbose) {
        System.out.println("PParser output");
        treePrint.printTree(debinarizer.transformTree(pparser.getBestParse()), pwOut);
      }
    }
    return true;
  }


  /**
   * Return the best parse of the sentence most recently parsed.
   * This will be from the factored parser, if it was used and it succeeeded
   * else from the PCFG if it was used and succeed, else from the dependency
   * parser.
   *
   * @return The best tree
   * @throws NoSuchElementException If no previously successfully parsed
   *                                sentence
   */
  public Tree getBestParse() {
    if (bparser != null && parseSucceeded) {
      Tree binaryTree = bparser.getBestParse();
      Tree tree = debinarizer.transformTree(binaryTree);
      if (op.nodePrune) {
        NodePruner np = new NodePruner(pparser, debinarizer);
        tree = np.prune(tree);
      }
      return subcategoryStripper.transformTree(tree);
    } else if (pparser != null && pparser.hasParse() && fallbackToPCFG) {
      return getBestPCFGParse();
    } else if (dparser != null && dparser.hasParse()) { // && fallbackaToDG
      return getBestDependencyParse(true);
    } else {
      throw new NoSuchElementException();
    }
  }

  public Tree getBestPCFGParse() {
    return getBestPCFGParse(true);
  }

  public Tree getBestPCFGParse(boolean stripSubcategories) {
    if (pparser == null) {
      return null;
    }
    Tree binaryTree = pparser.getBestParse();
    if (binaryTree == null) {
      return null;
    }
    Tree t = debinarizer.transformTree(binaryTree);
    if (stripSubcategories) {
      t = subcategoryStripper.transformTree(t);
    }
    return t;
  }

  public double getPCFGScore() {
    return getPCFGScore(pparser.goalStr);
  }

  public double getPCFGScore(String goalStr) {
    if (!pparser.stateNumberer.hasSeen(goalStr)) {
      return Double.NEGATIVE_INFINITY;
    }
    pparser.setGoalString(goalStr);
    return pparser.getLogScore();
  }

  public Tree getBestDependencyParse() {
    return getBestDependencyParse(false);
  }

  public Tree getBestDependencyParse(boolean debinarize) {
    Tree t = dparser != null ? dparser.getBestParse() : null;
    if (debinarize && t != null) {
      t = debinarizer.transformTree(t);
    }
    return t;
  }


  /**
   * Construct a new LexicalizedParser object from a previously serialized
   * grammar read from a property
   * <code>edu.stanford.nlp.SerializedLexicalizedParser</code>,
   * or a default file location.
   */
  public LexicalizedParser() {
    this(new Options());
  }


  /**
   * Construct a new LexicalizedParser object from a previously serialized
   * grammar read from a System property
   * <code>edu.stanford.nlp.SerializedLexicalizedParser</code>,
   * or a default file location
   * (<code>/u/nlp/data/lexparser/englishPCFG.ser.gz</code>).
   *
   * @param op Options to the parser.  These get overwritten by the
   *           Options read from the serialized parser; I think the only
   *           thing determined by them is the encoding of the grammar
   *           iff it is a text grammar
   */
  public LexicalizedParser(Options op) {
    this.op = op;
    String source = System.getProperty(SERIALIZED_PARSER_PROPERTY);
    if (source == null) {
      source = DEFAULT_PARSER_LOC;
    }
    pd = getParserDataFromFile(source, op);
    this.op = pd.pt;
    makeParsers();
  }


  /**
   * Construct a new LexicalizedParser.  This loads a grammar that
   * was previously assembled and stored.
   *
   * @throws IllegalArgumentException If parser data cannot be loaded
   */
  public LexicalizedParser(String parserFileOrUrl, Options op) {
    this.op = op;
    //    System.err.print("Loading parser from file " + parserFileOrUrl);
    pd = getParserDataFromFile(parserFileOrUrl, op);
    this.op = pd.pt; // in case a serialized options was read in
    makeParsers();
  }

  public LexicalizedParser(String parserFileOrUrl) {
    this(parserFileOrUrl, new Options());
  }

  /**
   * Set the maximum length of a sentence that the parser will be willing
   * to parse.  Sentences longer than this will not be parsed (an Exception
   * will be thrown).
   *
   * @param maxLength
   */
  public void setMaxLength(int maxLength) {
    Test.maxLength = maxLength;
  }

  public static ParserData getParserDataFromFile(String parserFileOrUrl, Options op) {
    ParserData pd = getParserDataFromSerializedFile(parserFileOrUrl);
    if (pd == null) {
      pd = getParserDataFromTextFile(parserFileOrUrl, op);
    }
    return pd;
  }

  /**
   * Construct a new LexicalizedParser.  This loads a grammar that
   * was previously assembled and stored.
   *
   * @throws IllegalArgumentException If parser data cannot be loaded
   */
  public LexicalizedParser(String parserFileOrUrl, boolean isTextGrammar, Options op) {
    this.op = op;
    if (isTextGrammar) {
      pd = getParserDataFromTextFile(parserFileOrUrl, op);
    } else {
      pd = getParserDataFromSerializedFile(parserFileOrUrl);
      this.op = pd.pt;
    }
    makeParsers();
  }

  /**
   * Construct a new LexicalizedParser object from a previously assembled
   * grammar.
   *
   * @param pd A <code>ParserData</code> object (not <code>null</code>)
   */
  public LexicalizedParser(ParserData pd) {
    this.pd = pd;
    makeParsers();
  }


  /**
   * Construct a new LexicalizedParser object from a previously assembled
   * grammar read from an InputStream.
   * One (ParserData) object is read from the stream.  It isn't closed.
   *
   * @param in The ObjectInputStream
   */
  public LexicalizedParser(ObjectInputStream in) throws Exception {
    this((ParserData) in.readObject());
  }

  /**
   * Construct a new LexicalizedParser.
   *
   * @param trainTreebank a treebank to train from
   */
  public LexicalizedParser(Treebank trainTreebank, GrammarCompactor compactor, Options op) {
    this(trainTreebank, compactor, op, null);
  }

  public LexicalizedParser(String treebankPath, FileFilter filt, Options op) {
    this(makeTreebank(treebankPath, op, filt), op);
  }

  private static Treebank makeTreebank(String treebankPath, Options op, FileFilter filt) {
    System.err.println("Training a parser from treebank dir: " + treebankPath);
    Treebank trainTreebank = op.tlpParams.memoryTreebank();
    System.err.print("Reading trees...");
    if (filt == null) {
      trainTreebank.loadPath(treebankPath);
    } else {
      trainTreebank.loadPath(treebankPath, filt);
    }

    Timing.tick("done [read " + trainTreebank.size() + " trees].");
    return trainTreebank;
  }

  private static DiskTreebank makeSecondaryTreebank(String treebankPath, Options op, FileFilter filt) {
    System.err.println("Additionally using secondary disk treebank: " + treebankPath);
    DiskTreebank trainTreebank = op.tlpParams.diskTreebank();
    if (filt == null) {
      trainTreebank.loadPath(treebankPath);
    } else {
      trainTreebank.loadPath(treebankPath, filt);
    }
    return trainTreebank;
  }


  /**
   * Construct a new LexicalizedParser.
   *
   * @param trainTreebank a treebank to train from
   * @param tuneTreebank  a treebank to tune free params on (may be null)
   */
  public LexicalizedParser(Treebank trainTreebank, GrammarCompactor compactor, Options op, Treebank tuneTreebank) {
    this.op = op;
    pd = getParserDataFromTreebank(trainTreebank, compactor, tuneTreebank);
    makeParsers();
  }

  public LexicalizedParser(Treebank trainTreebank, DiskTreebank secondaryTrainTreebank, double weight, GrammarCompactor compactor, Options op) {
    this.op = op;
    pd = getParserDataFromTreebank(trainTreebank, secondaryTrainTreebank, weight, compactor);
    makeParsers();
  }

  public LexicalizedParser(Treebank trainTreebank, Options op) {
    this(trainTreebank, null, op);
  }

  public ParserData parserData() {
    return pd;
  }

  public Lexicon getLexicon() {
    return pd.lex;
  }

  static void saveParserDataToSerialized(ParserData pd, String filename) {
    try {
      System.err.print("Writing parser in serialized format to file " + filename + " ");
      ObjectOutputStream out;
      if (filename.endsWith(".gz")) {
        // it's faster to do the buffering _outside_ the gzipping as here
        out = new ObjectOutputStream(new BufferedOutputStream(new GZIPOutputStream(new FileOutputStream(filename))));
      } else {
        out = new ObjectOutputStream(new BufferedOutputStream(new FileOutputStream(filename)));
      }
      out.writeObject(pd);
      out.close();
      System.err.println("done.");
    } catch (IOException ioe) {
      ioe.printStackTrace();
    }
  }

  static void saveParserDataToText(ParserData pd, String filename) {
    try {
      System.err.print("Writing parser in text grammar format to file " + filename);
      OutputStream os;
      if (filename.endsWith(".gz")) {
        // it's faster to do the buffering _outside_ the gzipping as here
        os = new BufferedOutputStream(new GZIPOutputStream(new FileOutputStream(filename)));
      } else {
        os = new BufferedOutputStream(new FileOutputStream(filename));
      }
      PrintWriter out = new PrintWriter(os);
      String prefix = "BEGIN ";
      out.println(prefix + "OPTIONS");
      if (pd.pt != null) {
        pd.pt.writeData(out);
      }
      out.println();
      System.err.print(".");
      out.println(prefix + "LEXICON");
      if (pd.lex != null) {
        pd.lex.writeData(out);
      }
      out.println();
      System.err.print(".");
      out.println(prefix + "UNARY_GRAMMAR");
      if (pd.ug != null) {
        pd.ug.writeData(out);
      }
      out.println();
      System.err.print(".");
      out.println(prefix + "BINARY_GRAMMAR");
      if (pd.bg != null) {
        pd.bg.writeData(out);
      }
      out.println();
      System.err.print(".");
      out.println(prefix + "DEPENDENCY_GRAMMAR");
      if (pd.dg != null) {
        pd.dg.writeData(out);
      }
      out.println();
      System.err.print(".");
      out.flush();
      out.close();
      System.err.println("done.");
    } catch (IOException e) {
      System.err.println("Trouble saving parser data to ASCII format.");
      e.printStackTrace();
    }
  }


  protected static ParserData getParserDataFromTextFile(String textFileOrUrl, Options op) {
    try {
      System.err.print("Loading parser from text file " + textFileOrUrl + " ");
      InputStream is;
      BufferedReader in;
      if (textFileOrUrl.startsWith("http://")) {
        URL u = new URL(textFileOrUrl);
        URLConnection uc = u.openConnection();
        is = uc.getInputStream();
      } else {
        is = new FileInputStream(textFileOrUrl);
      }
      if (textFileOrUrl.endsWith(".gz")) {
        in = new BufferedReader(new InputStreamReader(new GZIPInputStream(is)));
      } else {
        in = new BufferedReader(new InputStreamReader(is));
      }
      Timing.startTime();
      String line;
      line = in.readLine();
      if (!line.startsWith("BEGIN")) {
        throw new RuntimeException("Didn't expect " + line);
      }
      op.readData(in);
      System.err.print(".");
      line = in.readLine();
      if (!line.startsWith("BEGIN")) {
        throw new RuntimeException("Didn't expect " + line);
      }
      Lexicon lex = op.tlpParams.lex(op.lexOptions);
      lex.readData(in);
      System.err.print(".");
      line = in.readLine();
      if (!line.startsWith("BEGIN")) {
        throw new RuntimeException("Didn't expect " + line);
      }
      UnaryGrammar ug = new UnaryGrammar(op.numStates);
      ug.readData(in);
      System.err.print(".");
      line = in.readLine();
      if (!line.startsWith("BEGIN")) {
        throw new RuntimeException("Didn't expect " + line);
      }
      BinaryGrammar bg = new BinaryGrammar(op.numStates);
      bg.readData(in);
      System.err.print(".");
      line = in.readLine();
      if (!line.startsWith("BEGIN")) {
        throw new RuntimeException("Didn't expect " + line);
      }
      DependencyGrammar dg = new MLEDependencyGrammar(op.tlpParams, op.directional, op.distance, op.coarseDistance);
      dg.readData(in);
      System.err.print(".");
      Map numbs = Numberer.getNumberers();
      in.close();
      System.err.println(" done.");
      return new ParserData(lex, bg, ug, dg, numbs, op);
    } catch (IOException e) {
      e.printStackTrace();
    }
    return null;
  }


  protected static ParserData getParserDataFromSerializedFile(String serializedFileOrUrl) {
    try {
      System.err.print("Loading parser from serialized file " + serializedFileOrUrl + " ...");
      ObjectInputStream in;
      InputStream is;
      if (serializedFileOrUrl.startsWith("http://")) {
        URL u = new URL(serializedFileOrUrl);
        URLConnection uc = u.openConnection();
        is = uc.getInputStream();
      } else {
        is = new FileInputStream(serializedFileOrUrl);
      }
      if (serializedFileOrUrl.endsWith(".gz")) {
        // it's faster to do the buffering _outside_ the gzipping as here
        in = new ObjectInputStream(new BufferedInputStream(new GZIPInputStream(is)));
      } else {
        in = new ObjectInputStream(new BufferedInputStream(is));
      }
      ParserData pd = (ParserData) in.readObject();
      //      Numberer.setNumberers(pd.numbs); // will happen later in makeParsers()
      in.close();
      System.err.println(" done.");
      return pd;
    } catch (InvalidClassException ice) {
      // For this, it's not a good idea to continue and try it as a text file!
      System.err.println();   // as in middle of line from above message
      ice.printStackTrace();
      System.exit(2);
    } catch (FileNotFoundException fnfe) {
      // For this, it's not a good idea to continue and try it as a text file!
      System.err.println();   // as in middle of line from above message
      fnfe.printStackTrace();
      System.exit(2);
    } catch (StreamCorruptedException sce) {
      // suppress error message, on the assumption that we've really got
      // a text grammar, and that'll be tried next
      System.err.println();
    } catch (Exception e) {
      System.err.println();   // as in middle of line from above message
      e.printStackTrace();
    }
    return null;
  }


  private static void printOptions(boolean train, Options op) {
    op.display();
    if (train) {
      Train.display();
    } else {
      Test.display();
    }
    op.tlpParams.display();
  }

  protected final ParserData getParserDataFromTreebank(Treebank trainTreebank, GrammarCompactor compactor, Treebank tuneTreebank) {
    System.err.println("Currently " + new Date());
    printOptions(true, op);
    Timing.startTime();
    // setup tree transforms
    TreebankLangParserParams tlpParams = op.tlpParams;
    TreebankLanguagePack tlp = tlpParams.treebankLanguagePack();

    if (Test.verbose) {
      System.out.print("Training ");
      System.out.println(trainTreebank.textualSummary());
    }

    System.err.print("Binarizing trees...");
    TreeAnnotatorAndBinarizer binarizer;
    if (!Train.leftToRight) {
      binarizer = new TreeAnnotatorAndBinarizer(tlpParams, op.forceCNF, !Train.outsideFactor(), true);
    } else {
      binarizer = new TreeAnnotatorAndBinarizer(tlpParams.headFinder(), new LeftHeadFinder(), tlpParams, op.forceCNF, !Train.outsideFactor(), true);
    }
    CollinsPuncTransformer collinsPuncTransformer = null;
    if (Train.collinsPunc) {
      collinsPuncTransformer = new CollinsPuncTransformer(tlp);
    }
    List<Tree> binaryTrainTrees = new ArrayList<Tree>();
    List<Tree> binaryTuneTrees = new ArrayList<Tree>();

    if (Train.selectiveSplit) {
      Train.splitters = ParentAnnotationStats.getSplitCategories(trainTreebank, Train.tagSelectiveSplit, 0, Train.selectiveSplitCutOff, Train.tagSelectiveSplitCutOff, tlp);
      if (Train.deleteSplitters != null) {
        List<String> deleted = new ArrayList<String>();
        for (String del : Train.deleteSplitters) {
          String baseDel = tlp.basicCategory(del);
          boolean checkBasic = del.equals(baseDel);
          for (Iterator it = Train.splitters.iterator(); it.hasNext(); ) {
            String elem = (String) it.next();
            String baseElem = tlp.basicCategory(elem);
            boolean delStr = checkBasic && baseElem.equals(baseDel) ||
              elem.equals(del);
            if (delStr) {
              it.remove();
              deleted.add(elem);
            }
          }
        }
        if (Test.verbose) {
          System.err.println("Removed from vertical splitters: " + deleted);
        }
      }
      if (Test.verbose) {
        List list = new ArrayList(Train.splitters);
        Collections.sort(list);
        System.err.println("Parent split categories: " + list);
      }
    }
    if (Train.selectivePostSplit) {
      TreeTransformer myTransformer = new TreeAnnotator(tlpParams.headFinder(), tlpParams);
      Treebank annotatedTB = trainTreebank.transform(myTransformer);
      Train.postSplitters = ParentAnnotationStats.getSplitCategories(annotatedTB, true, 0, Train.selectivePostSplitCutOff, Train.tagSelectivePostSplitCutOff, tlp);
      if (Test.verbose) {
        System.err.println("Parent post annotation split categories: " + Train.postSplitters);
      }
    }
    if (Train.hSelSplit) {
      // We run through all the trees once just to gather counts for hSelSplit!
      binarizer.setDoSelectiveSplit(false);
      for (Tree tree : trainTreebank) {
        if (Train.collinsPunc) {
          tree = collinsPuncTransformer.transformTree(tree);
        }
        binarizer.transformTree(tree);  
      }
      binarizer.setDoSelectiveSplit(true);
    }
    for (Tree tree : trainTreebank) {
      if (Train.collinsPunc) {
        tree = collinsPuncTransformer.transformTree(tree);
      }
      tree = binarizer.transformTree(tree);
      if (tree.yield().size() <= trainLengthLimit) {
        binaryTrainTrees.add(tree);
      }
    }

    if (tuneTreebank != null) {
      for (Iterator tI = tuneTreebank.iterator(); tI.hasNext();) {
        Tree tree = (Tree) tI.next();
        if (Train.collinsPunc) {
          tree = collinsPuncTransformer.transformTree(tree);
        }
        tree = binarizer.transformTree(tree);
        if (tree.yield().size() > trainLengthLimit) {
          continue;
        }
        binaryTuneTrees.add(tree);
      }
    }

    Timing.tick("done.");
    if (Test.verbose) {
      binarizer.dumpStats();
    }
    // extract grammars
    //Extractor bgExtractor = new SmoothedBinaryGrammarExtractor();
    Extractor bgExtractor = new BinaryGrammarExtractor();
    // Extractor lexExtractor = new LexiconExtractor();
    Extractor dgExtractor = op.tlpParams.dependencyGrammarExtractor(op);
    //TreeExtractor uwmExtractor = new UnknownWordModelExtractor(binaryTrainTrees.size());
    System.err.print("Extracting PCFG...");
    Pair bgug = (Pair) bgExtractor.extract(binaryTrainTrees);
    Timing.tick("done.");
    if (compactor != null) {
      System.err.print("Compacting grammar...");
      bgug = compactor.compactGrammar(bgug);
      Timing.tick("done.");
    }
    System.err.print("Compiling grammar...");
    BinaryGrammar bg = (BinaryGrammar) bgug.second;
    bg.splitRules();
    UnaryGrammar ug = (UnaryGrammar) bgug.first;
    ug.purgeRules();
    Timing.tick("done");
    System.err.print("Extracting Lexicon...");
    Lexicon lex = op.tlpParams.lex(op.lexOptions);
    lex.train(binaryTrainTrees);
    Timing.tick("done.");

    DependencyGrammar dg = null;
    if (op.doDep) {
      System.err.print("Extracting Dependencies...");
      dg = (DependencyGrammar) dgExtractor.extract(binaryTrainTrees);
      //      ((ChineseSimWordAvgDepGrammar)dg).setLex(lex);
      Timing.tick("done.");
      //System.out.println(dg);
      //System.err.print("Extracting Unknown Word Model...");
      //UnknownWordModel uwm = (UnknownWordModel)uwmExtractor.extract(binaryTrainTrees);
      //Timing.tick("done.");
      if (tuneTreebank != null) {
        System.err.print("Tuning Dependency Model...");
        dg.tune(binaryTuneTrees);
        Timing.tick("done.");
      }
    }
    Map numbs = Numberer.getNumberers();

    int numStates = Numberer.getGlobalNumberer("states").total(); // TEG debug
    System.err.println("numStates: " + numStates); // TEG debug

    System.err.println("Done training parser.");
    return new ParserData(lex, bg, ug, dg, numbs, op);
  }

  /**
   * A method for training from two different treebanks, the second of which is presumed
   * to be orders of magnitude larger.
   * <p/>
   * Trees are not read into memory but processed as they are read from disk.
   * <p/>
   * A weight (typically < 1) can be put on the second treebank.
   */
  protected final ParserData getParserDataFromTreebank(Treebank trainTreebank, DiskTreebank secondaryTrainTreebank, double weight, GrammarCompactor compactor) {
    System.out.println("Currently " + new Date());
    printOptions(true, op);
    Timing.startTime();

    // setup tree transforms
    TreebankLangParserParams tlpParams = op.tlpParams;
    TreebankLanguagePack tlp = tlpParams.treebankLanguagePack();
    if (Test.verbose) {
      System.out.print("Training ");
      System.out.println(trainTreebank.textualSummary());
    }

    CompositeTreeTransformer trainTransformer = new CompositeTreeTransformer();

    if (Train.collinsPunc) {
      CollinsPuncTransformer collinsPuncTransformer = new CollinsPuncTransformer(tlp);
      trainTransformer.addTransformer(collinsPuncTransformer);
    }

    TreeAnnotatorAndBinarizer binarizer;
    if (!Train.leftToRight) {
      binarizer = new TreeAnnotatorAndBinarizer(tlpParams, op.forceCNF, !Train.outsideFactor(), true);
    } else {
      binarizer = new TreeAnnotatorAndBinarizer(tlpParams.headFinder(), new LeftHeadFinder(), tlpParams, op.forceCNF, !Train.outsideFactor(), true);
    }
    trainTransformer.addTransformer(binarizer);

    CompositeTreebank wholeTreebank = new CompositeTreebank(trainTreebank, secondaryTrainTreebank);
    if (Train.selectiveSplit) {
      Train.splitters = ParentAnnotationStats.getSplitCategories(wholeTreebank, Train.tagSelectiveSplit, 0, Train.selectiveSplitCutOff, Train.tagSelectiveSplitCutOff, tlp);
      if (Train.deleteSplitters != null) {
        List<String> deleted = new ArrayList<String>();
        for (String del : Train.deleteSplitters) {
          String baseDel = tlp.basicCategory(del);
          boolean checkBasic = del.equals(baseDel);
          for (Iterator it = Train.splitters.iterator(); it.hasNext(); ) {
            String elem = (String) it.next();
            String baseElem = tlp.basicCategory(elem);
            boolean delStr = checkBasic && baseElem.equals(baseDel) ||
              elem.equals(del);
            if (delStr) {
              it.remove();
              deleted.add(elem);
            }
          }
        }
        if (Test.verbose) {
          System.err.println("Removed from vertical splitters: " + deleted);
        }
      }
      if (Test.verbose) {
        System.err.println("Parent split categories: " + Train.splitters);
      }
    }
    if (Train.selectivePostSplit) {
      TreeTransformer annotator = new TreeAnnotator(tlpParams.headFinder(), tlpParams);
      wholeTreebank.transformOnRead(annotator);
      Train.postSplitters = ParentAnnotationStats.getSplitCategories(wholeTreebank, true, 0, Train.selectivePostSplitCutOff, Train.tagSelectivePostSplitCutOff, tlp);
      if (Test.verbose) {
        System.err.println("Parent post annotation split categories: " + Train.postSplitters);
      }
    }
    if (Train.hSelSplit) {
      binarizer.setDoSelectiveSplit(false);
      for (Tree tree : wholeTreebank) {
        trainTransformer.transformTree(tree);
      }
      binarizer.setDoSelectiveSplit(true);
    }

    trainTreebank.transform(trainTransformer);
    secondaryTrainTreebank.transformOnRead(trainTransformer);

    // extract grammars
    //Extractor bgExtractor = new SmoothedBinaryGrammarExtractor();
    BinaryGrammarExtractor bgExtractor = new BinaryGrammarExtractor();
    // Extractor lexExtractor = new LexiconExtractor();
    MLEDependencyGrammarExtractor dgExtractor = new MLEDependencyGrammarExtractor(op);
    //TreeExtractor uwmExtractor = new UnknownWordModelExtractor(binaryTrainTrees.size());
    System.out.print("Extracting PCFG...");
    Pair bgug = (Pair) bgExtractor.extract(trainTreebank, secondaryTrainTreebank, weight);
    Timing.tick("done.");
    if (compactor != null) {
      System.out.print("Compacting grammar...");
      bgug = compactor.compactGrammar(bgug);
      Timing.tick("done.");
    }
    System.out.print("Compiling grammar...");
    BinaryGrammar bg = (BinaryGrammar) bgug.second;
    bg.splitRules();
    UnaryGrammar ug = (UnaryGrammar) bgug.first;
    ug.purgeRules();
    Timing.tick("done");
    System.out.print("Extracting Lexicon...");
    Lexicon lex = op.tlpParams.lex(op.lexOptions);
    //    lex.train(wholeTreebank);
    lex.train(trainTreebank);
    //    ((BaseLexicon)lex).train(secondaryTrainTreebank, weight);
    Timing.tick("done.");

    DependencyGrammar dg = null;
    if (op.doDep) {
      System.out.print("Extracting Dependencies...");
      dg = (DependencyGrammar) dgExtractor.extract(trainTreebank, secondaryTrainTreebank, weight);
      Timing.tick("done.");
    }
    Map numbs = Numberer.getNumberers();

    int numStates = Numberer.getGlobalNumberer("states").total(); // TEG debug
    System.out.println("numStates: " + numStates); // TEG debug

    System.err.println("Done training parser.");
    return new ParserData(lex, bg, ug, dg, numbs, op);
  }


  protected final void makeParsers() {
    if (pd == null) {
      throw new IllegalArgumentException("Error loading parser data: pd null");
    }
    Numberer.setNumberers(pd.numbs);
    BinaryGrammar bg = pd.bg;
    bg.splitRules();
    UnaryGrammar ug = pd.ug;
    Lexicon lex = pd.lex;
    DependencyGrammar dg = pd.dg;
    op = pd.pt;
    if (op.doPCFG) {
      if (Test.iterativeCKY) {
        pparser = new IterativeCKYPCFGParser(bg, ug, lex, op);
      } else {
        pparser = new ExhaustivePCFGParser(bg, ug, lex, op);
      }
    }
    if (op.doDep) {
      dparser = new ExhaustiveDependencyParser(dg, lex, op);

      Scorer scorer = new TwinScorer(pparser, dparser);
      //Scorer scorer = parser;
      if (op.doPCFG) {
        bparser = new BiLexPCFGParser(scorer, pparser, dparser, bg, ug, dg, lex, op);
      }
    }
    fallbackToPCFG = true;

    debinarizer = new Debinarizer(op.langpack(), op.forceCNF);
    subcategoryStripper = op.tlpParams.subcategoryStripper();
  }


  /**
   * returns the input sentence for the parser
   */
  static Sentence getInputSentence(Tree t) {
    if (Test.forceTags) {
      if (Test.preTag) // doesn't yet support
      {
        throw new RuntimeException("Sorry -- haven't implemented support for external tagger yet.");
      } else {
        return t.taggedYield();
      }
    } else {
      return t.yield();
    }
  }

  // helper function
  private static int numSubArgs(String[] args, int index) {
    int i = index;
    while (i + 1 < args.length && args[i + 1].charAt(0) != '-') {
      i++;
    }
    return i - index;
  }


  /** Test the parser on a treebank. Parses will be written to stdout, and
   *  various other information will be written to stderr and stdout, 
   *  particularly if <code>Test.verbose</code> is true.
   *
   *  @param testTreebank The treebank to parse
   *  @return The labeled precision/recall F<sub>1</sub> (EVALB measure) 
   *          of the parser on the treebank.
   */
  public double testOnTreebank(Treebank testTreebank) {
    System.err.println("Testing on treebank");
    Timing.startTime();
    if (Test.verbose) {
      System.out.print("Testing ");
      System.out.println(testTreebank.textualSummary());
    }
    TreePrint treePrint = getTreePrint();
    TreebankLangParserParams tlpParams = op.tlpParams;
    TreebankLanguagePack tlp = op.langpack();
    if (Test.evalb) {
      EvalB.initEVALBfiles(tlpParams);
    }
    PrintWriter pwOut = tlpParams.pw();
    PrintWriter pwErr = tlpParams.pw(System.err);
    TreeTransformer tc = tlpParams.collinizer();
    TreeTransformer br = new BoundaryRemover();
    LabeledConstituentEval pcfgPE = new LabeledConstituentEval("pcfg LP/LR", tlp);
    DependencyEval depDE = new DependencyEval("depend DE", tlp.punctuationWordAcceptFilter());
    LabeledConstituentEval comboPE = new LabeledConstituentEval("factor LP/LR", tlp);
    AbstractEval ruleErrorEval = null;
    AbstractEval catErrorEval = null;
    if (Test.verbose) {
      ruleErrorEval = new RuleErrorEval("Rule under/over");
      catErrorEval = new CatErrorEval("Category under/over");
    }
    TaggingEval comboTE = new TaggingEval("Tag", pd.lex);
    // no annotation
    TreeAnnotatorAndBinarizer binarizerOnly;
    if (!Train.leftToRight) {
      binarizerOnly = new TreeAnnotatorAndBinarizer(tlpParams, op.forceCNF, false, false);
    } else {
      binarizerOnly = new TreeAnnotatorAndBinarizer(tlpParams.headFinder(), new LeftHeadFinder(), tlpParams, op.forceCNF, false, false);
    }
    boolean saidMemMessage = false;
    Timing timer = new Timing();
    for (Tree goldTree : testTreebank) {
      Sentence s = getInputSentence(goldTree);
      timer.start();
      pwErr.println("Parsing [len. " + s.length() + "]: " + s);
      Tree tree = null;
      try {
        if ( ! parse(s)) {
          pwErr.print("Sentence couldn't be parsed by grammar.");
          if (pparser != null && pparser.hasParse() && fallbackToPCFG) {
            pwErr.println("... falling back to PCFG parse.");
            tree = getBestPCFGParse();
          } else {
            pwErr.println();
          }
        } else {
          tree = getBestParse();
        }
        //combo parse goes to System.out
        if (Test.verbose) {
          pwOut.println("ComboParser best");
          Tree ot = tree;
          if (!tlpParams.treebankLanguagePack().isStartSymbol(ot.value())) {
            ot = ot.treeFactory().newTreeNode(tlpParams.treebankLanguagePack().startSymbol(), Collections.singletonList(ot));
          }
          treePrint.printTree(ot, pwOut);
        } else {
          treePrint.printTree(tree, pwOut);
        }
        if (Test.printAllBestParses) {
          List parses = pparser.getBestParses();
          int sz = parses.size();
          if (sz > 1) {
            pwOut.println("There were " + sz + " best PCFG parses.");
            Tree transGoldTree = tc.transformTree(goldTree);
            for (Iterator itb = parses.iterator(); itb.hasNext();) {
              Tree tb = (Tree) itb.next();
              Tree tbd = debinarizer.transformTree(tb);
              tbd = subcategoryStripper.transformTree(tbd);
              Tree tbtr = tc.transformTree(tbd);
              pwOut.println("Tree size = " + tbtr.size() + "; depth = " + tbtr.depth());
              tbtr.pennPrint(pwOut);
              pcfgPE.evaluate(tbtr, transGoldTree, pwOut);
            }
          }
        }
        // if (pparser != null)
        //   pwErr.println("Correct parse\tScore: " + pparser.scoreBinarizedTree(binarizer.transformTree(goldTree),true));
        // else
        if (Test.verbose) {
          pwOut.println("Correct parse");
          treePrint.printTree(goldTree, pwOut);
        }
      } catch (OutOfMemoryError e) {
        if (Test.maxLength != -0xDEADBEEF) {
          // this means they explicitly asked for a length they cannot handle.  Throw exception.
          pwErr.println("NOT ENOUGH MEMORY TO PARSE SENTENCES OF LENGTH " + Test.maxLength);
          throw e;
        } else {
          if (!saidMemMessage) {
            pwErr.println();
            pwErr.println("*************************");
            pwErr.println("***     WARNING!!     ***");
            pwErr.println("*************************");
            pwErr.println();
            pwErr.println("OUT OF MEMORY!!!  YOU SHOULD GIVE THE JVM MORE MEMORY OR SET THE MAXIMUM SENTENCE SIZE WITH -maxLength");
            saidMemMessage = true;
          }
          if (pparser.hasParse() && fallbackToPCFG) {
            pwErr.println("Sentence too long for dependency parser.  Falling back to PCFG parse...");
            tree = getBestPCFGParse();
          } else {
            pwErr.println("Sentence too long.  Skipping...");
          }
          pwErr.println();
        }
      } catch (UnsupportedOperationException uoe) {
        pwErr.println("Sentence too long");
      }
      timer.report("Parsing Sentence");
      if (tree != null) {
        Tree transGoldTree = tc.transformTree(goldTree);
        if (transGoldTree == null) {
          continue;
        }
        Tree treePCFG = getBestPCFGParse();
        if (treePCFG != null) {
          pcfgPE.evaluate(tc.transformTree(treePCFG), transGoldTree, pwErr);
          if (Test.verbose) {
            ruleErrorEval.evaluate(tc.transformTree(treePCFG), transGoldTree, pwErr);
            catErrorEval.evaluate(tc.transformTree(treePCFG), transGoldTree, pwErr);
          }
        }
        Tree treeDep = getBestDependencyParse();
        if (treeDep != null) {
          depDE.evaluate(treeDep, binarizerOnly.transformTree(goldTree), pwErr);
        }
        comboPE.evaluate(tc.transformTree(tree), transGoldTree, pwErr);
        comboTE.evaluate(tree, br.transformTree(goldTree), pwErr);
        if (Test.evalb) {
          EvalB.writeEVALBline(goldTree, tree);
        }
      }
      pwErr.println();
    } // for tree iterator
    Timing.tick("Done testing on treebank");
    if (saidMemMessage) {
      pwErr.println();
      pwErr.println("*************************");
      pwErr.println("***     WARNING!!     ***");
      pwErr.println("*************************");
      pwErr.println();
      pwErr.println("THERE WAS NOT ENOUGH MEMORY TO USE THE DEPENDENCY PARSER ON ALL SENTENCES!!!");
      pwErr.println("YOU SHOULD GIVE THE JVM MORE MEMORY OR SET THE MAXIMUM SENTENCE SIZE WITH -maxLength");
      pwErr.println();
    }
    if (Test.evalb) {
      EvalB.closeEVALBfiles();
    }
    if (Test.verbose) {
      ruleErrorEval.display(true, pwErr);
      catErrorEval.display(true, pwErr);
    }
    return comboPE.getEvalbF1();
  }


  /**
   * This will set options to the parser, in a way exactly equivalent to
   * passing in the same sequence of command-line arguments.  This is a useful
   * convenience method when building a parser programmatically. The options
   * passed in should 
   * be specified like command-line arguments, including with an initial
   * minus sign.
   * <p/>
   * <i>Notes:</i> This can be used to set parsing-time flags for a
   * serialized parser.  You can also still change things serialized in 
   * Options, 
   * but this will probably degrade parsing performance.  The vast majority
   * of command line flags can be passed to this method, but you cannot 
   * pass in options that specify the treebank or grammar to be loaded, 
   * the grammar to be written, trees or files to be parsed or details of 
   * their encoding, nor the TreebankLangParserParams (<code>-tLPP</code>) 
   * to use. The TreebankLangParserParams should be set up on construction of
   * a LexicalizedParser, by constructing an Options that uses the required
   * TreebankLangParserParams, and passing that to a LexicalizedParser 
   * constructor.  Note that despite
   * this method being an instance method, many flags are actually set as
   * static class variables.
   *
   * @param flags Arguments to the parser, for example,
   *              {"-outputFormat", "typedDependencies", "-maxLength", "70"}
   * @throws IllegalArgumentException If an unknown flag is passed in
   */
  public void setOptionFlags(String[] flags) {
    for (int i = 0; i < flags.length;) {
      int j = FactoredParser.setOptionFlag(op, flags, i);
      if (j == i) {
        j = op.tlpParams.setOptionFlag(flags, i);
        if (j == i) {
          throw new IllegalArgumentException("Unknown option: " + flags[i]);
        }
      }
      i = j;
    }
  }

  private static void printArgs(String[] args, PrintStream ps) {
    ps.print("LexicalizedParser invoked with arguments:");
    for (int i = 0; i < args.length; i++) {
      ps.print(" " + args[i]);
    }
    ps.println();
  }

  static int trainLengthLimit = 100000;

  /**
   * A main program for using the parser with various options.
   * This program can be used for building and serializing
   * a parser from treebank data, for parsing sentences from a file
   * or URL using a serialized or text grammar parser,
   * and (mainly for parser quality testing)
   * for training and testing a parser on a treebank all in one go. <p>
   * Sample Usages: <br><code>
   * java -mx1500m edu.stanford.nlp.parser.lexparser.LexicalizedParser [-v]
   * -train trainFilesPath fileRange -saveToSerializedFile
   * serializedGrammarFilename</code><p>
   * <code> java -mx1500m edu.stanford.nlp.parser.lexparser.LexicalizedParser
   * [-v] -train trainFilesPath fileRange
   * -testTreebank testFilePath fileRange</code><p>
   * <code>java -mx512m edu.stanford.nlp.parser.lexparser.LexicalizedParser
   * [-v] serializedGrammarPath filename+</code><p>
   * <code>java -mx512m edu.stanford.nlp.parser.lexparser.LexicalizedParser
   * [-v] -loadFromSerializedFile serializedGrammarPath 
   * -testTreebank testFilePath fileRange
   * </code><p>
   * If the <code>serializedGrammarPath</code> ends in <code>.gz</code>,
   * then the grammar is written and read as a compressed file (GZip).
   * If the <code>serializedGrammarPath</code> is a URL, starting with
   * <code>http://</code>, then the parser is read from the URL.
   * A fileRange specifies a numeric value that must be included within a 
   * filename for it to be used in training or testing (this works well with
   * most current treebanks).  It can be specified like a range of pages to be
   * printed, for instance as <code>200-2199</code> or
   * <code>1-300,500-725,9000</code> or just as <code>1</code>.
   * The parser can write a grammar as either a serialized Java object file
   * or in a text format, specified with the following alternate
   * usages:<p>
   * <code>java edu.stanford.nlp.parser.lexparser.LexicalizedParser [-v] -train
   * trainFilesPath [fileRange] [-saveToSerializedFile grammarPath]
   * [-saveToTextFile grammarPath]</code><p>
   * If no files are supplied to parse, then a hardwired sentence
   * is parsed. <p>
   *
   * In the same position as the verbose flag (<code>-v</code>), many other
   * options can be specified.  The most useful to an end user are:
   * <UL>
   * <LI><code>-tLPP class</code> Specify a different
   * TreebankLangParserParams, for when using a different language or
   * treebank (the default is English Penn Treebank)</LI>
   * <LI><code>-encoding charset</code> Specify the character encoding of the
   * input files.  This will override the value in the 
   * <code>TreebankLangParserParams</code>, provided this option appears
   * <i>after</i> any <code>-tLPP</code> </LI>
   * <LI><code>-tokenized</code> Says that the input is already separated
   * into whitespace-delimited tokens.  Unless you also specify
   * <code>-escaper</code>, these tokens <i>must</i> all be correctly 
   * tokenized tokens of the appropriate treebank for the parser to work 
   * well (for instance, if using the Penn English Treebank, you must have
   * coded "(" as "-LRB-", "3/4" as "3\/4", etc.)</LI>
   * <li><code>-escaper class</code> specify a class of type
   * {@link Function}&lt;List&lt;HasWord&gt;,List&lt;HasWord&gt;&gt; to do
   * customized escaping of tokenized text.  This class will be run over the
   * tokenized text and can fix the representation of tokens. For instance,
   * it could change "(" to "-LRB-" for the Penn English Treebank.  A 
   * provided escaper that does such things for the Penn English Treebank is
   * <code>edu.stanford.nlp.process.PTBEscapingProcessor</code>
   * <LI><code>-tokenizerFactory class</code> Specifies a
   * TokenizerFactory class to be used for tokenization</LI>
   * <LI><code>-sentences token </code> Specifies a token that marks sentence
   * boundaries.  All other tokens will be interpreted literally, and must be
   * tokens returned by the tokenizer.  A value of
   * "newline" causes sentence breaking on newlines.</LI>
   * <LI><code>-parseInside element</code> Specifies that parsing should only
   * be done for tokens inside the indicated XML-style
   * elements (done as simple pattern matching, rather than XML parsing).
   * For example, if this is specified as <code>sentence</code>, then
   * the text inside the <code>sentence</code> element
   * would be parsed  as a sentence.  Sentences cannot span elements.
   * (Using "-parseInside s" gives you support for the input format of
   * Charniak's parser.)
   * <LI><code>-tagSeparator char</code> Specifies to look for tags on words
   * following the word and separated from it by a special character 
   * <code>char</code>.  For instance, many tagged corpora have the
   * representation "house/NN" and you would use <code>-tagSeparator /</code>.
   * Notes: This option requires that the input be pretokenized.
   * The separator has to be only a single character, and there is no
   * escaping mechanism. However, splitting is done on the <i>last</i>
   * instance of the character in the token, so that cases like
   * "3\/4/CD" are handled correctly.  The parser will in all normal
   * circumstances use the tag you provide, but will override it in the
   * case of very common words in cases where the tag that you provide
   * is not one that it regards as a possible tagging for the word.
   * The parser supports a format where only some of the words in a sentence
   * have a tag (if you are calling the parser programmatically, you indicate
   * them by having them implement the <code>HasTag</code> interface).
   * You can do this at the command-line by only having tags after some words,
   * but you are limited by the fact that there is no way to escape the
   * tagSeparator character.</LI>
   * <LI><code>-maxLength leng</code> Specify the longest sentence that
   * will be parsed (and hence indirectly the amount of memory
   * needed for the parser). If this is not specified, the parser will 
   * try to dynamically grow its parse chart when long sentence are 
   * encountered, but may run out of memory trying to do so.</LI>
   * <LI><code>-outputFormat styles</code> Choose the style(s) of output
   * sentences: <code>penn</code> for prettyprinting as in the Penn
   * treebank files, or <code>oneline</code> for printing sentences one
   * per line, <code>words</code>, <code>wordsAndTags</code>,
   * <code>dependencies</code>, <code>typedDependencies</code>,
   * or <code>typedDependenciesCollapsed</code>.
   * Multiple options may be specified as a comma-separated
   * list.  See TreePrint class for further documentation.</LI>
   * <LI><code>-outputFormatOptions</code> Provide options that control the
   * behavior of various <code>-outputFormat</code> choices, such as
   * <code>lexicalize</code>, <code>stem</code>, <code>markHeadNodes</code>,
   * or <code>xml</code>.
   * Options are specified as a comma-separated list.</LI>
   * </UL>
   * See also the package documentation for more details and examples of use. 
   *
   * @param args Command line arguments, as above
   */
  public static void main(String[] args) {
    boolean train = false;
    boolean saveToSerializedFile = false;
    boolean saveToTextFile = false;
    String serializedInputFileOrUrl = null;
    String textInputFileOrUrl = null;
    String serializedOutputFileOrUrl = null;
    String textOutputFileOrUrl = null;
    String treebankPath = null;
    Treebank testTreebank = null;
    Treebank tuneTreebank = null;
    String testPath = null;
    FileFilter testFilter = null;
    String tunePath = null;
    FileFilter tuneFilter = null;
    FileFilter trainFilter = null;
    String secondaryTreebankPath = null;
    double secondaryTreebankWeight = 1.0;
    FileFilter secondaryTrainFilter = null;

    // variables needed to process the files to be parsed
    TokenizerFactory tokenizerFactory = null;
    DocumentPreprocessor documentPreprocessor = new DocumentPreprocessor();
    boolean tokenized = false; // whether or not the input file has already been tokenized
    Function<List<HasWord>, List<HasWord>> escaper = null;
    int tagDelimiter = -1;
    String sentenceDelimiter = null;
    boolean fromXML = false;
    int argIndex = 0;
    if (args.length < 1) {
      System.err.println("usage: java edu.stanford.nlp.parser.lexparser." + "LexicalizedParser parserFileOrUrl filename*");
      System.exit(1);
    }

    Options op = new Options();
    // while loop through option arguments
    while (argIndex < args.length && args[argIndex].charAt(0) == '-') {
      if (args[argIndex].equalsIgnoreCase("-train")) {
        train = true;
        int numSubArgs = numSubArgs(args, argIndex);
        argIndex++;
        if (numSubArgs > 1) {
          treebankPath = args[argIndex];
          argIndex++;
        } else {
          throw new RuntimeException("Error: -train option must have treebankPath as first argument.");
        }
        if (numSubArgs == 2) {
          trainFilter = new NumberRangesFileFilter(args[argIndex++], true);
        } else if (numSubArgs >= 3) {
          try {
            int low = Integer.parseInt(args[argIndex]);
            int high = Integer.parseInt(args[argIndex + 1]);
            trainFilter = new NumberRangeFileFilter(low, high, true);
            argIndex += 2;
          } catch (NumberFormatException e) {
            // maybe it's a ranges expression?
            trainFilter = new NumberRangesFileFilter(args[argIndex], true);
            argIndex++;
          }
        }
      } else if (args[argIndex].equalsIgnoreCase("-train2")) {
        // train = true;     // cdm july 2005: should require -train for this
        int numSubArgs = numSubArgs(args, argIndex);
        argIndex++;
        if (numSubArgs < 3) {
          throw new RuntimeException("Error: -train2 <treebankPath> <ranges> <weight>.");
        }
        secondaryTreebankPath = args[argIndex++];
        secondaryTrainFilter = new NumberRangesFileFilter(args[argIndex++], true);
        secondaryTreebankWeight = Double.parseDouble(args[argIndex++]);
      } else if (args[argIndex].equalsIgnoreCase("-tLPP") && (argIndex + 1 < args.length)) {
        try {
          op.tlpParams = (TreebankLangParserParams) Class.forName(args[argIndex + 1]).newInstance();
        } catch (ClassNotFoundException e) {
          System.err.println("Class not found: " + args[argIndex + 1]);
        } catch (InstantiationException e) {
          System.err.println("Couldn't instantiate: " + args[argIndex + 1] + ": " + e.toString());
        } catch (IllegalAccessException e) {
          System.err.println("Illegal access" + e);
        }
        argIndex += 2;
      } else if (args[argIndex].equalsIgnoreCase("-encoding")) { // sets encoding for TreebankLangParserParams
        String encoding = args[argIndex + 1];
        op.tlpParams.setInputEncoding(encoding);
        op.tlpParams.setOutputEncoding(encoding);
        argIndex += 2;
      } else if (args[argIndex].equalsIgnoreCase("-useBasicCategoryTagsInDependencyGrammar")) {
        LexicalizedParser.basicCategoryTagsInDependencyGrammar = true;
        argIndex++;
      } else if (args[argIndex].equalsIgnoreCase("-tokenized")) {
        tokenized = true;
        argIndex += 1;
      } else if (args[argIndex].equalsIgnoreCase("-escaper")) {
        try {
          escaper = (Function<List<HasWord>, List<HasWord>>) Class.forName(args[argIndex + 1]).newInstance();
        } catch (Exception e) {
          System.err.println("Couldn't instantiate escaper " + args[argIndex + 1] + ": " + e);
        }
        argIndex += 2;
      } else if (args[argIndex].equalsIgnoreCase("-tokenizerFactory")) {
        try {
          tokenizerFactory = (TokenizerFactory) Class.forName(args[argIndex + 1]).newInstance();
        } catch (Exception e) {
          System.err.println("Couldn't instantiate TokenizerFactory " + args[argIndex + 1]);
        }
        argIndex += 2;
      } else if (args[argIndex].equalsIgnoreCase("-sentences")) {
        sentenceDelimiter = args[argIndex + 1];
        if (sentenceDelimiter.equalsIgnoreCase("newline")) {
          sentenceDelimiter = "\n";
        }
        argIndex += 2;
      } else if (args[argIndex].equalsIgnoreCase("-parseInside")) {
        sentenceDelimiter = args[argIndex + 1];
        fromXML = true;
        argIndex += 2;
      } else if (args[argIndex].equalsIgnoreCase("-tagSeparator")) {
        tagDelimiter = args[argIndex + 1].charAt(0);
        argIndex += 2;
      } else if (args[argIndex].equalsIgnoreCase("-loadFromSerializedFile")) {
        // load the parser from a binary serialized file
        // the next argument must be the path to the parser file
        serializedInputFileOrUrl = args[argIndex + 1];
        argIndex += 2;
      } else if (args[argIndex].equalsIgnoreCase("-loadFromTextFile")) {
        // load the parser from declarative text file
        // the next argument must be the path to the parser file
        textInputFileOrUrl = args[argIndex + 1];
        argIndex += 2;
      } else if (args[argIndex].equalsIgnoreCase("-saveToSerializedFile")) {
        saveToSerializedFile = true;
        if (numSubArgs(args, argIndex) < 1) {
          System.err.println("Missing path: -saveToSerialized filename");
        } else {
          serializedOutputFileOrUrl = args[argIndex + 1];
        }
        argIndex += 2;
      } else if (args[argIndex].equalsIgnoreCase("-saveToTextFile")) {
        // save the parser to declarative text file
        saveToTextFile = true;
        textOutputFileOrUrl = args[argIndex + 1];
        argIndex += 2;
      } else if (args[argIndex].equalsIgnoreCase("-trainLength")) {
        // train on only short sentences
        trainLengthLimit = Integer.parseInt(args[argIndex + 1]);
        argIndex += 2;
      } else if (args[argIndex].equalsIgnoreCase("-lengthNormalization")) {
        Test.lengthNormalization = true;
        argIndex++;
      } else if (args[argIndex].equalsIgnoreCase("-treebank") ||
                 args[argIndex].equalsIgnoreCase("-testTreebank")) {
        // the next argument is the treebank path and range for testing
        int numSubArgs = numSubArgs(args, argIndex);
        argIndex++;
        if (numSubArgs == 1) {
          testFilter = new NumberRangesFileFilter(args[argIndex++], true);
        } else if (numSubArgs > 1) {
          testPath = args[argIndex++];
          if (numSubArgs == 2) {
            testFilter = new NumberRangesFileFilter(args[argIndex++], true);
          } else if (numSubArgs >= 3) {
            try {
              int low = Integer.parseInt(args[argIndex]);
              int high = Integer.parseInt(args[argIndex + 1]);
              testFilter = new NumberRangeFileFilter(low, high, true);
              argIndex += 2;
            } catch (NumberFormatException e) {
              // maybe it's a ranges expression?
              testFilter = new NumberRangesFileFilter(args[argIndex++], true);
            }
          }
        }
      } else if (args[argIndex].equalsIgnoreCase("-tune")) {
        // the next argument is the treebank path and range for tuning
        int numSubArgs = numSubArgs(args, argIndex);
        argIndex++;
        if (numSubArgs == 1) {
          tuneFilter = new NumberRangesFileFilter(args[argIndex++], true);
        } else if (numSubArgs > 1) {
          tunePath = args[argIndex++];
          if (numSubArgs == 2) {
            tuneFilter = new NumberRangesFileFilter(args[argIndex++], true);
          } else if (numSubArgs >= 3) {
            try {
              int low = Integer.parseInt(args[argIndex]);
              int high = Integer.parseInt(args[argIndex + 1]);
              tuneFilter = new NumberRangeFileFilter(low, high, true);
              argIndex += 2;
            } catch (NumberFormatException e) {
              // maybe it's a ranges expression?
              tuneFilter = new NumberRangesFileFilter(args[argIndex++], true);
            }
          }
        }
      } else {
        int j = FactoredParser.setOptionFlag(op, args, argIndex);
        if (j == argIndex) {
          j = op.tlpParams.setOptionFlag(args, argIndex);
        }
        if (j == argIndex) {
          System.err.println("Unknown option ignored: " + args[argIndex]);
          j++;
        }
        argIndex = j;
      }
    } // end while loop through arguments

    // all other arguments are order dependent and
    // are processed in order below

    if (tuneFilter != null) {
      if (tunePath == null) {
        if (treebankPath == null) {
          throw new RuntimeException("No tune treebank path specified...");
        } else {
          System.err.println("No tune treebank path specified.  Using train path: \"" + treebankPath + "\"");
          tunePath = treebankPath;
        }
      }
      tuneTreebank = op.tlpParams.testMemoryTreebank();
      tuneTreebank.loadPath(tunePath, tuneFilter);
    }

    LexicalizedParser lp = null;
    if (!train && Test.verbose) {
      System.err.println("Currently " + new Date());
      printArgs(args, System.err);
    }
    if (train) {
      printArgs(args, System.err);
      // so we train a parser using the treebank
      GrammarCompactor compactor = null;
      if (Train.compactGrammar() == 3) {
        compactor = new ExactGrammarCompactor(false, false);
      }
      Treebank trainTreebank = makeTreebank(treebankPath, op, trainFilter);
      if (secondaryTreebankPath != null) {
        System.err.println("Additionally training using secondary treebank dir: " + secondaryTreebankPath);
        DiskTreebank secondaryTrainTreebank = makeSecondaryTreebank(secondaryTreebankPath, op, secondaryTrainFilter);
        lp = new LexicalizedParser(trainTreebank, secondaryTrainTreebank, secondaryTreebankWeight, compactor, op);
      } else {
        lp = new LexicalizedParser(trainTreebank, compactor, op, tuneTreebank);
      }
    } else if (textInputFileOrUrl != null) {
      // so we load the parser from a text grammar file
      lp = new LexicalizedParser(textInputFileOrUrl, true, op);
    } else {
      // so we load a serialized parser
      if (serializedInputFileOrUrl == null && argIndex < args.length) {
        // the next argument must be the path to the serialized parser
        serializedInputFileOrUrl = args[argIndex];
        argIndex++;
      }
      if (serializedInputFileOrUrl == null) {
        System.err.println("No grammar specified, exiting...");
        System.exit(0);
      }
      try {
        lp = new LexicalizedParser(serializedInputFileOrUrl, op);
        op = lp.op;
      } catch (IllegalArgumentException e) {
        System.err.println("Error loading parser, exiting...");
        System.exit(0);
      }
    }

    // the following has to go after reading parser to make sure
    // op and tlpParams are the same for train and test
    TreePrint treePrint = Test.treePrint(op.tlpParams);

    if (testFilter != null || testPath != null) {
      if (testPath == null) {
        if (treebankPath == null) {
          throw new RuntimeException("No test treebank path specified...");
        } else {
          System.err.println("No test treebank path specified.  Using train path: \"" + treebankPath + "\"");
          testPath = treebankPath;
        }
      }
      testTreebank = op.tlpParams.testMemoryTreebank();
      testTreebank.loadPath(testPath, testFilter);
    }

    Train.sisterSplitters = new HashSet(Arrays.asList(op.tlpParams.sisterSplitters()));

    // at this point we should be sure that op.tlpParams is
    // set appropriately (from command line, or from grammar file),
    // and will never change again.  -- Roger

    if (Test.verbose) {
      System.err.println("ParserPack is " + op.tlpParams.getClass().getName());
      System.err.println("Lexicon is " + lp.pd.lex.getClass().getName());
    }

    PrintWriter pwOut = op.tlpParams.pw();
    PrintWriter pwErr = op.tlpParams.pw(System.err);

    // Now what do we do with the parser we've made
    if (saveToTextFile) {
      // save the parser to textGrammar format
      if (textOutputFileOrUrl != null) {
        saveParserDataToText(lp.pd, textOutputFileOrUrl);
      } else {
        System.err.println("Usage: must specify a text grammar output path");
      }
    }
    if (saveToSerializedFile) {
      if (serializedOutputFileOrUrl != null) {
        saveParserDataToSerialized(lp.pd, serializedOutputFileOrUrl);
      } else if (textOutputFileOrUrl == null && testTreebank == null) {
        // no saving/parsing request has been specified
        System.err.println("usage: " + "java edu.stanford.nlp.parser.lexparser.LexicalizedParser " + "-train trainFilesPath [fileRange] -saveToSerializedFile serializedParserFilename");
      }
    }
    if (Test.verbose) {
      printOptions(false, op);
    }
    if (testTreebank != null) {
      // test parser on treebank
      lp.testOnTreebank(testTreebank);
    } else if (argIndex >= args.length) {
      // no more arguments, so we just parse our own test sentence
      if (lp.parse(op.tlpParams.defaultTestSentence())) {
        treePrint.printTree(lp.getBestParse(), pwOut);
      } else {
        pwErr.println("Error. Can't parse test sentence: " + 
              lp.parse(op.tlpParams.defaultTestSentence()));
      }
    } else {
      // We parse filenames given by the remaining arguments
      int numWords = 0;
      Timing timer = new Timing();
      // set the tokenizer
      if (tokenized) {
        tokenizerFactory = WhitespaceTokenizer.factory();
      }
      TreebankLanguagePack tlp = op.tlpParams.treebankLanguagePack();
      if (tokenizerFactory == null) {
        tokenizerFactory = tlp.getTokenizerFactory();
      }
      documentPreprocessor.setTokenizerFactory(tokenizerFactory);
      documentPreprocessor.setSentenceFinalPuncWords(tlp.sentenceFinalPunctuationWords());
      documentPreprocessor.setEncoding(op.tlpParams.getInputEncoding());
      timer.start();
      for (int i = argIndex; i < args.length; i++) {
        String filename = args[i];
        try {
          List document; // initialized immediately below
          if (fromXML) {
            document = documentPreprocessor.getSentencesFromXML(filename, sentenceDelimiter,true);
          } else {
            document = documentPreprocessor.getSentencesFromText(filename, escaper, sentenceDelimiter, tagDelimiter);
          }
          System.err.println("Parsing file: " + filename + " with " + document.size() + " sentences.");
          PrintWriter pwo = pwOut;
          if (Test.writeOutputFiles) {
            try {
              pwo = op.tlpParams.pw(new FileOutputStream(filename + ".stp"));
            } catch (IOException ioe) {
              ioe.printStackTrace();
            }
          }
          int num = 0;
          treePrint.printHeader(pwo, tlp.getEncoding());
          for (Iterator it = document.iterator(); it.hasNext();) {
            num++;
            List sentence = (List) it.next();
            int len = sentence.size();
            numWords += len;
            pwErr.println("Parsing [sent. " + num + " len. " + len + "]: " + sentence);
            try {
              lp.parse(sentence);
              // System.out.println("Score: " + lp.pparser.bestScore);
              treePrint.printTree(lp.getBestParse(), Integer.toString(num), pwo);
            } catch (UnsupportedOperationException uoe) {
              pwErr.println("Sentence too long (or zero words).");
              numWords -= len;
            }
          }
          treePrint.printFooter(pwo);
          if (Test.writeOutputFiles) {
            pwo.close();
          }
        } catch (IOException e) {
          pwErr.println("Couldn't find file: " + filename);
        }

      } // end for each file
      long millis = timer.stop();
      double wordspersec = numWords / (((double) millis) / 1000);
      NumberFormat nf = new DecimalFormat("0.00"); // easier way!
      pwErr.println("Parsed " + numWords + " words at " + nf.format(wordspersec) + " words per second.");

    }
  } // end main

} // end class LexicalizedParser
