package edu.stanford.nlp.parser.lexparser;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.Serializable;
import java.io.Writer;
import java.util.Collection;
import java.util.Iterator;

/**
 * An interface for lexicons interfacing to lexparser.
 *
 * @author Galen Andrew
 */
public interface Lexicon extends Serializable {

  String UNKNOWN_WORD = "UNK";  // if UNK were a word, counts would merge
  String BOUNDARY = ".$.";      // boundary word -- assumed not a real word
  String BOUNDARY_TAG = "*";    // boundary tag -- assumed not a real tag

  /**
   * Checks whether a word is in the lexicon.
   *
   * @param word The word as an int
   * @return Whether the word is in the lexicon
   */
  boolean isKnown(int word);

  /**
   * Checks whether a word is in the lexicon.
   *
   * @param word The word as a String
   * @return Whether the word is in the lexicon
   */
  boolean isKnown(String word);

  /**
   * Get an iterator over all rules (pairs of (word, POS)) for this word.
   *
   *  @param word The word, represented as an integer in Numberer
   *  @param loc  The position of the word in the sentence (counting from 0). 
   *                <i>Implementation note: The BaseLexicon class doesn't
   *                actually make use of this position information.</i>
   *  @return An Iterator over a List ofIntTaggedWords, which pair the word 
   *                with possible taggings as integer pairs.  (Each can be
   *                thought of as a <code>tag -&gt; word<code> rule.)
   */
  Iterator ruleIteratorByWord(int word, int loc);

  /**
   * Trains this lexicon on the Collection of trees.
   */
  void train(Collection trees);

  /**
   * Get the score of this word with this tag (as an IntTaggedWord) at this 
   * loc.
   * (Presumably an estimate of P(word | tag).)
   *
   * @param iTW An IntTaggedWord pairing a word and POS tag
   * @param loc The position in the sentence.  <i>In the default implementation
   *               this is used only for unknown words to change their
   *               probability distribution when sentence initial
   * @return A double valued score, usually - log P(word|tag)
   */
  double score(IntTaggedWord iTW, int loc);

  /**
   * Write the lexicon in human-readable format to the Writer.
   * (An optional operation.)
   *
   * @param w The writer to output to
   * @throws IOException
   */
  public void writeData(Writer w) throws IOException;

  /**
   * Read the lexicon from the BufferedReader in the format written by 
   * writeData.
   * (An optional operation.)
   *
   * @param in The BufferedReader to read from
   * @throws IOException
   */
  public void readData(BufferedReader in) throws IOException;

}
