package edu.stanford.nlp.parser.lexparser;

import edu.stanford.nlp.ling.HasTag;
import edu.stanford.nlp.ling.HasWord;
import edu.stanford.nlp.trees.Tree;
import edu.stanford.nlp.stats.Counter;
import edu.stanford.nlp.util.StringUtils;
import edu.stanford.nlp.stats.Counter;

import java.io.*;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;

public class MLEDependencyGrammar extends AbstractDependencyGrammar {

  protected int numWords;

  protected Counter argCounter;
  protected Counter stopCounter;

  public double smooth_aT_hTWd = 16.0;
  public double smooth_aTW_hTWd = 16.0;
  public double smooth_stop = 4.0;
  public double interp = 0.6;
//  public double distanceDecay = 0.0;

  public MLEDependencyGrammar(TreebankLangParserParams tlpParams, boolean directional, boolean distance, boolean coarseDistance) {
    // This use of the LexicalizedParser static variable is an UGLY TEMPORARY HACK!!!
    this(LexicalizedParser.basicCategoryTagsInDependencyGrammar ? (TagProjection) new BasicCategoryTagProjection(tlpParams.treebankLanguagePack()) : new TestTagProjection(), tlpParams, directional, distance, coarseDistance);

  }

  public MLEDependencyGrammar(TagProjection tagProjection, TreebankLangParserParams tlpParams, boolean directional, boolean useDistance, boolean useCoarseDistance) {
    super(tlpParams.treebankLanguagePack(), tagProjection, directional, useDistance, useCoarseDistance);
    argCounter = new Counter();
    stopCounter = new Counter();
    numWords = wordNumberer().total();
    double[] smoothParams = tlpParams.MLEDependencyGrammarSmoothingParams();
    smooth_aT_hTWd = smoothParams[0];
    smooth_aTW_hTWd = smoothParams[1];
    smooth_stop = smoothParams[2];
    interp = smoothParams[3];
  }

  public String toString() {
    NumberFormat nf = NumberFormat.getNumberInstance();
    nf.setMaximumFractionDigits(2);
    StringBuffer sb = new StringBuffer(2000);
    String cl = getClass().getName();
    sb.append(cl.substring(cl.lastIndexOf(".") + 1) + "[tagbins=" + numTagBins + ",words=" + numWords + "; head -> arg\n");
//    for (Iterator dI = coreDependencies.keySet().iterator(); dI.hasNext();) {
//      Dependency d = (Dependency) dI.next();
//      double count = coreDependencies.getCount(d);
//      sb.append(d + " count " + nf.format(count));
//      if (dI.hasNext()) {
//        sb.append(",");
//      }
//      sb.append("\n");
//    }
    sb.append("]");
    return sb.toString();
  }

  public boolean pruneTW(IntTaggedWord argTW) {
    String[] punctTags = tlp.punctuationTags();
    for (int i = 0; i < punctTags.length; i++) {
      if (argTW.tag == tagNumberer().number(punctTags[i])) {
        return true;
      }
    }
    return false;
  }

  static class EndHead {
    public int end;
    public int head;
  }

  static transient EndHead tempEndHead = new EndHead();

  /** Adds dependencies to list depList.  These are in terms of the original
   *  tag set not the reduced (projected) tag set.
   */
  protected static EndHead treeToDependencyHelper(Tree tree, List<Dependency> depList, int loc) {
    //       try {
    // 	PrintWriter pw = new PrintWriter(new OutputStreamWriter(System.out,"GB18030"),true);
    // 	tree.pennPrint(pw);
    //       }
    //       catch (UnsupportedEncodingException e) {}

    if (tree.isLeaf() || tree.isPreTerminal()) {
      tempEndHead.head = loc;
      tempEndHead.end = loc + 1;
      return tempEndHead;
    }
    Tree[] kids = tree.children();
    if (kids.length == 1) {
      return treeToDependencyHelper(kids[0], depList, loc);
    }
    tempEndHead = treeToDependencyHelper(kids[0], depList, loc);
    int lHead = tempEndHead.head;
    int split = tempEndHead.end;
    tempEndHead = treeToDependencyHelper(kids[1], depList, tempEndHead.end);
    int end = tempEndHead.end;
    int rHead = tempEndHead.head;
    String hTag = ((HasTag) tree.label()).tag();
    String lTag = ((HasTag) kids[0].label()).tag();
    String rTag = ((HasTag) kids[1].label()).tag();
    String hWord = ((HasWord) tree.label()).word();
    String lWord = ((HasWord) kids[0].label()).word();
    String rWord = ((HasWord) kids[1].label()).word();
    boolean leftHeaded = hWord.equals(lWord);
    String aTag = (leftHeaded ? rTag : lTag);
    String aWord = (leftHeaded ? rWord : lWord);
    int hT = tagNumberer().number(hTag);
    int aT = tagNumberer().number(aTag);
    int hW = (wordNumberer().hasSeen(hWord) ? wordNumberer().number(hWord) : wordNumberer().number(Lexicon.UNKNOWN_WORD));
    int aW = (wordNumberer().hasSeen(aWord) ? wordNumberer().number(aWord) : wordNumberer().number(Lexicon.UNKNOWN_WORD));
    int head = (leftHeaded ? lHead : rHead);
    int arg = (leftHeaded ? rHead : lHead);
    Dependency dependency = new Dependency(hW, hT, aW, aT, leftHeaded, (leftHeaded ? split - head - 1 : head - split));
    depList.add(dependency);
    Dependency stopL = new Dependency(aW, aT, -2, -2, false, (leftHeaded ? arg - split : arg - loc));
    depList.add(stopL);
    Dependency stopR = new Dependency(aW, aT, -2, -2, true, (leftHeaded ? end - arg - 1 : split - arg - 1));
    depList.add(stopR);
    //System.out.println("Adding: "+dependency+" at "+tree.label());
    tempEndHead.head = head;
    return tempEndHead;
  }


  public void dumpSizes() {
//    System.out.println("core dep " + coreDependencies.size());
    System.out.println("arg counter " + argCounter.size());
    System.out.println("stop counter " + stopCounter.size());
  }

  /** Returns the List of dependencies for a Tree.  These are in terms of
   *  the original tag set not the reduced (projected) tag set.
   */
  public static List<Dependency> treeToDependencyList(Tree tree) {
    List<Dependency> depList = new ArrayList<Dependency>();
    treeToDependencyHelper(tree, depList, 0);
    if (false) {
      System.out.println("----------------------------");
      tree.pennPrint();
      System.out.println(depList);
    }
    return depList;
  }

  protected double scoreAll(Collection deps) {
    double totalScore = 0.0;
    for (Iterator<Dependency> iter = deps.iterator(); iter.hasNext(); ) {
      Dependency d = iter.next();
      //if (d.head.word == Numberer.number("words","via") ||
      //          d.arg.word == Numberer.number("words","via"))
      //System.out.println(d+" at "+score(d));
      double score = score(d);
      if (score > Double.NEGATIVE_INFINITY) {
        totalScore += score;
      }
    }
    return totalScore;
  }

  public void tune(Collection<Tree> trees) {
    List<Dependency> deps = new ArrayList<Dependency>();
    for (Tree tree : trees) {
      deps.addAll(treeToDependencyList(tree));
    }

    double bestScore = Double.NEGATIVE_INFINITY;
    double bestSmooth_stop = 0.0;
    double bestSmooth_aTW_hTWd = 0.0;
    double bestSmooth_aT_hTWd = 0.0;
    double bestInterp = 0.0;

    System.err.println("Tuning smooth_stop...");
    for (smooth_stop = 1.0/100.0; smooth_stop < 100.0; smooth_stop *= 1.25) {
      double totalScore = 0.0;
      for (Dependency dep : deps) {
        if (!rootTW(dep.head)) {
          double stopProb = getStopProb(dep);
          if (!dep.arg.equals(stopTW)) {
            stopProb = 1.0 - stopProb;
          }
          if (stopProb > 0.0) {
            totalScore += Math.log(stopProb);
          }
        }
      }
      if (totalScore > bestScore) {
        bestScore = totalScore;
        bestSmooth_stop = smooth_stop;
      }
    }
    smooth_stop = bestSmooth_stop;
    System.err.println("Tuning selected smooth_stop: " + smooth_stop);

    for (Iterator iter = deps.iterator(); iter.hasNext();) {
      Dependency dep = (Dependency) iter.next();
      if (dep.arg.equals(stopTW)) {
        iter.remove();
      }
    }

    System.err.print("Tuning other parameters...");
    bestScore = Double.NEGATIVE_INFINITY;
    for (smooth_aTW_hTWd = 0.5; smooth_aTW_hTWd < 100.0; smooth_aTW_hTWd *= 1.25) {
      System.err.print(".");
      for (smooth_aT_hTWd = 0.5; smooth_aT_hTWd < 100.0; smooth_aT_hTWd *= 1.25) {
        for (interp = 0.02; interp < 1.0; interp += 0.02) {
          double totalScore = 0.0;
          for (Dependency dep : deps) {
            double score = score(dep);
            if (score > Double.NEGATIVE_INFINITY) {
              totalScore += score;
            }
          }
          if (totalScore > bestScore) {
            bestScore = totalScore;
            bestInterp = interp;
            bestSmooth_aTW_hTWd = smooth_aTW_hTWd;
            bestSmooth_aT_hTWd = smooth_aT_hTWd;
            //        System.err.println("Current best interp: " + interp + " with score " + totalScore);
          }
        }
      }
    }

    smooth_aTW_hTWd = bestSmooth_aTW_hTWd;
    smooth_aT_hTWd = bestSmooth_aT_hTWd;
    interp = bestInterp;

    System.out.println("\nTuning selected smooth_aTW_hTWd: " + smooth_aTW_hTWd + " smooth_aT_hTWd: " + smooth_aT_hTWd + " interp: " + interp);
  }

  /** Add this dependency with the given count to the grammar.
   *  This is a dependency represented in the full tag space.
   */
  public void addRule(Dependency dependency, double count) {
    if (!directional) {
      dependency.leftHeaded = false;
    }
//    coreDependencies.incrementCount(dependency, count);
    /*new Dependency(dependency.head.word,
                                        dependency.head.tag,
                                        dependency.arg.word,
                                        dependency.arg.tag,
                                        dependency.leftHeaded,
                                        dependency.distance), count);
    */
    expandDependency(dependency, count);
  }

  protected transient List<IntTaggedWord> tagITWList = null;//new ArrayList();

  private IntTaggedWord getCachedITW(short tag) {
    if (tagITWList == null) {
      tagITWList = new ArrayList<IntTaggedWord>(numTagBins + 2);
      for (int i=0; i<numTagBins + 2; i++) {
        tagITWList.add(i, null);
      }
    }
    IntTaggedWord headT = tagITWList.get(tag + 2);
    if (headT == null) {
      headT = new IntTaggedWord(-1, tagBin(tag));
      tagITWList.set(tag + 2, headT);
    }
    return headT;
  }

  protected void expandDependency(Dependency dependency, double count) {
    //if (Test.prunePunc && pruneTW(dependency.arg))
    //  return;
    if (dependency.head == null) {
      return;
    }
    if (dependency.arg == null) {
      return;
    }

    if (dependency.arg.word != -2) {
      expandArg(dependency, valenceBin(dependency.distance), count);
    }
    expandStop(dependency, distanceBin(dependency.distance), count, true);
  }

  private void expandArg(Dependency dependency, short valBinDist, double count) {
    IntTaggedWord headT = getCachedITW(dependency.head.tag);
    IntTaggedWord argT = getCachedITW(dependency.arg.tag);
    IntTaggedWord head = new IntTaggedWord(dependency.head.word, tagBin(dependency.head.tag));//dependency.head;
    IntTaggedWord arg = new IntTaggedWord(dependency.arg.word, tagBin(dependency.arg.tag));//dependency.arg;
    boolean leftHeaded = dependency.leftHeaded;

    argCounter.incrementCount(intern(head, arg, leftHeaded, valBinDist), count);
    argCounter.incrementCount(intern(headT, arg, leftHeaded, valBinDist), count);
    argCounter.incrementCount(intern(head, argT, leftHeaded, valBinDist), count);
    argCounter.incrementCount(intern(headT, argT, leftHeaded, valBinDist), count);

    argCounter.incrementCount(intern(head, wildTW, leftHeaded, valBinDist), count);
    argCounter.incrementCount(intern(headT, wildTW, leftHeaded, valBinDist), count);

    argCounter.incrementCount(intern(wildTW, arg, false, -1), count);
    argCounter.incrementCount(intern(wildTW, argT, false, -1), count);
  }

  private void expandStop(Dependency dependency, short distBinDist, double count, boolean wildForStop) {
    IntTaggedWord headT = getCachedITW(dependency.head.tag);
    IntTaggedWord head = new IntTaggedWord(dependency.head.word, tagBin(dependency.head.tag));//dependency.head;
    IntTaggedWord arg = new IntTaggedWord(dependency.arg.word, tagBin(dependency.arg.tag));//dependency.arg;
    boolean leftHeaded = dependency.leftHeaded;

    if (arg.word == -2) {
      stopCounter.incrementCount(intern(head, arg, leftHeaded, distBinDist), count);
      stopCounter.incrementCount(intern(headT, arg, leftHeaded, distBinDist), count);
    }
    if (wildForStop || arg.word != -2) {
      stopCounter.incrementCount(intern(head, wildTW, leftHeaded, distBinDist), count);
      stopCounter.incrementCount(intern(headT, wildTW, leftHeaded, distBinDist), count);
    }
  }

  public double countHistory(Dependency dependency) {
    short hTBackup = dependency.head.tag;
    IntTaggedWord aTWBackup = dependency.arg;
    short dist = dependency.distance;
    dependency.head.tag = (short) tagBin(dependency.head.tag);
    dependency.distance = valenceBin(dist);

    dependency.arg = wildTW;


    double s = argCounter.getCount(dependency);

    dependency.head.tag = hTBackup;
    dependency.arg = aTWBackup;
    dependency.distance = dist;

    return s;
  }

  public double scoreTB(Dependency dependency) {
    return Test.depWeight * Math.log(probTB(dependency));
  }

  private boolean verbose = false;

  protected static final double MIN_PROBABILITY = Math.exp(-100.0);

  /** Calculate the probability of a dependency as a real probability between
   *  0 and 1 inclusive.  
   *  @param dependency The dependency for which the probability is to be
   *       calculated.   The tags in this dependency are in the reduced
   *       TagProjection space.
   *  @return The probability of the dependency
   */
  protected double probTB(Dependency dependency) {
    if ( ! directional) {
      dependency.leftHeaded = false;
    }
    if (verbose) {
      // System.out.println("tagNumberer: " + tagNumberer());
      System.out.println("Generating " + dependency);
    }

    short distance = dependency.distance;
    boolean leftHeaded = dependency.leftHeaded;
    int hW = dependency.head.word;
    int aW = dependency.arg.word;
    IntTaggedWord aTW = dependency.arg;
    IntTaggedWord hTW = dependency.head;

    double pb_stop_hTWds = getStopProb(dependency);

    boolean isRoot = rootTW(dependency.head);
    if (dependency.arg.word == -2) {
      // did we generate stop?
      if (isRoot) {
        return 0.0;
      }
      return pb_stop_hTWds;
    }

    double pb_go_hTWds = 1.0 - pb_stop_hTWds;

    if (isRoot) {
      pb_go_hTWds = 1.0;
    }

    // generate the argument

    dependency.distance = valenceBin(distance);

    // KEY:
    // c_     count of
    // p_     MLE prob of
    // pb_    MAP prob of
    // a      arg
    // h      head
    // T      tag
    // W      word
    // d      direction
    // ds     distance

    double c_aTW_hTWd = argCounter.getCount(dependency);
    dependency.arg.word = -1;
    double c_aT_hTWd = argCounter.getCount(dependency);
    dependency.arg.word = aW;
    dependency.arg = wildTW;
    double c_hTWd = argCounter.getCount(dependency);
    dependency.arg = aTW;

    dependency.head.word = -1;
    double c_aTW_hTd = argCounter.getCount(dependency);
    dependency.arg.word = -1;
    double c_aT_hTd = argCounter.getCount(dependency);
    dependency.arg.word = aW;
    dependency.arg = wildTW;
    double c_hTd = argCounter.getCount(dependency);
    dependency.arg = aTW;
    dependency.head.word = hW;

    dependency.head = wildTW;
    dependency.leftHeaded = false;
    dependency.distance = -1;
    double c_aTW = argCounter.getCount(dependency);
    dependency.arg.word = -1;
    double c_aT = argCounter.getCount(dependency);
    dependency.arg.word = aW;
    dependency.leftHeaded = leftHeaded;
    dependency.head = hTW;

    dependency.distance = distance;

    // do the magic
    double p_aTW_hTd = (c_hTd > 0.0 ? c_aTW_hTd / c_hTd : 0.0);
    double p_aT_hTd = (c_hTd > 0.0 ? c_aT_hTd / c_hTd : 0.0);
    double p_aTW_aT = (c_aTW > 0.0 ? c_aTW / c_aT : 1.0);

    double pb_aTW_hTWd = (c_aTW_hTWd + smooth_aTW_hTWd * p_aTW_hTd) / (c_hTWd + smooth_aTW_hTWd);
    double pb_aT_hTWd = (c_aT_hTWd + smooth_aT_hTWd * p_aT_hTd) / (c_hTWd + smooth_aT_hTWd);

    double score = (interp * pb_aTW_hTWd + (1.0 - interp) * p_aTW_aT * pb_aT_hTWd) * pb_go_hTWds;

    if (verbose) {
      NumberFormat nf = NumberFormat.getNumberInstance();
      nf.setMaximumFractionDigits(2);
      System.out.println("  c_aTW_hTWd: " + c_aTW_hTWd + "; c_aT_hTWd: " + c_aT_hTWd + "; c_hTWd: " + c_hTWd);
      System.out.println("  c_aTW_hTd: " + c_aTW_hTd + "; c_aT_hTd: " + c_aT_hTd + "; c_hTd: " + c_hTd);
      System.out.println("  Generated with pb_go_hTWds: " + nf.format(pb_go_hTWds) + " pb_aTW_hTWd: " + nf.format(pb_aTW_hTWd) + " p_aTW_aT: " + nf.format(p_aTW_aT) + " pb_aT_hTWd: " + nf.format(pb_aT_hTWd));
      System.out.println("  NoDist score: " + score);
    }

    if (Test.prunePunc && pruneTW(aTW)) {
      return 1.0;
    }

    if (Double.isNaN(score)) {
      score = 0.0;
    }

    //if (Test.rightBonus && ! dependency.leftHeaded)
    //  score -= 0.2;

    if (score < MIN_PROBABILITY) {
      score = 0.0;
    }

    return score;
  }

  /** Return the probability (as a real number between 0 and 1) of stopping
   *  rather than generating another argument at this position.
   *  @param dependency The dependency used as the basis for stopping on.
   *     Tags are assumed in the TagProjection space
   */
  protected double getStopProb(Dependency dependency) {
    int hW = dependency.head.word;
    IntTaggedWord aTW = dependency.arg;
    short distance = dependency.distance;

    dependency.distance = distanceBin(distance);
    dependency.arg = stopTW;
    double c_stop_hTWds = stopCounter.getCount(dependency);
    dependency.head.word = -1;
    double c_stop_hTds = stopCounter.getCount(dependency);
    dependency.head.word = hW;
    dependency.arg = wildTW;
    double c_hTWds = stopCounter.getCount(dependency);
    dependency.head.word = -1;
    double c_hTds = stopCounter.getCount(dependency);

    dependency.head.word = hW;
    dependency.arg = aTW;
    dependency.distance = distance;

    double p_stop_hTds = (c_hTds > 0.0 ? c_stop_hTds / c_hTds : 1.0);

    double pb_stop_hTWds = (c_stop_hTWds + smooth_stop * p_stop_hTds) / (c_hTWds + smooth_stop);

    if (verbose) {
      System.out.println("  c_stop_hTWds: " + c_stop_hTWds + "; c_hTWds: " + c_hTWds + "; c_stop_hTds: " + c_stop_hTds + "; c_hTds: " + c_hTds);
      System.out.println("  Generate STOP prob: " + pb_stop_hTWds);
    }
    return pb_stop_hTWds;
  }

  private void readObject(ObjectInputStream stream) throws IOException, ClassNotFoundException {
    stream.defaultReadObject();
//    System.err.println("Before decompression:");
//    System.err.println("arg size: " + argCounter.size() + "  total: " + argCounter.totalCount());
//    System.err.println("stop size: " + stopCounter.size() + "  total: " + stopCounter.totalCount());

    Counter compressedArgC = argCounter;
    argCounter = new Counter();
    Counter compressedStopC = stopCounter;
    stopCounter = new Counter();
    for (Iterator dI = compressedArgC.keySet().iterator(); dI.hasNext();) {
      Dependency d = (Dependency) dI.next();
      double count = compressedArgC.getCount(d);
      expandArg(d, d.distance, count);
    }

    for (Iterator dI = compressedStopC.keySet().iterator(); dI.hasNext();) {
      Dependency d = (Dependency) dI.next();
      double count = compressedStopC.getCount(d);
      expandStop(d, d.distance, count, false);
    }

//    System.err.println("After decompression:");
//    System.err.println("arg size: " + argCounter.size() + "  total: " + argCounter.totalCount());
//    System.err.println("stop size: " + stopCounter.size() + "  total: " + stopCounter.totalCount());

    expandDependencyMap = null;
  }

  private void writeObject(ObjectOutputStream stream) throws IOException {
//    System.err.println("\nBefore compression:");
//    System.err.println("arg size: " + argCounter.size() + "  total: " + argCounter.totalCount());
//    System.err.println("stop size: " + stopCounter.size() + "  total: " + stopCounter.totalCount());

    Counter fullArgCounter = argCounter;
    argCounter = new Counter();
    for (Iterator iterator = fullArgCounter.keySet().iterator(); iterator.hasNext();) {
      Dependency dependency = (Dependency) iterator.next();
      if (dependency.head != wildTW && dependency.arg != wildTW &&
              dependency.head.word != -1 && dependency.arg.word != -1) {
        argCounter.incrementCount(dependency, fullArgCounter.getCount(dependency));
      }
    }

    Counter fullStopCounter = stopCounter;
    stopCounter = new Counter();
    for (Iterator iterator = fullStopCounter.keySet().iterator(); iterator.hasNext();) {
      Dependency dependency = (Dependency) iterator.next();
      if (dependency.head.word != -1) {
        stopCounter.incrementCount(dependency, fullStopCounter.getCount(dependency));
      }
    }

//    System.err.println("After compression:");
//    System.err.println("arg size: " + argCounter.size() + "  total: " + argCounter.totalCount());
//    System.err.println("stop size: " + stopCounter.size() + "  total: " + stopCounter.totalCount());

    stream.defaultWriteObject();

    argCounter = fullArgCounter;
    stopCounter = fullStopCounter;
  }

  /**
   * Populates data in this DependencyGrammar from the character stream
   * given by the Reader r.
   */
  public void readData(BufferedReader in) throws IOException {
    final String LEFT = "left";
    int lineNum = 1;
    // all lines have one rule per line
    boolean doingStop = false;
    Dependency tempDependency = new Dependency(-2, -2, -2, -2, false, 0);

    for (String line = in.readLine(); line != null && line.length() > 0; line = in.readLine()) {
      try {
        if (line.equals("BEGIN_STOP")) {
          doingStop = true;
          continue;
        }
        String[] fields = StringUtils.splitOnCharWithQuoting(line, ' ', '\"', '\\'); // split on spaces, quote with doublequote, and escape with backslash
        //        System.out.println("fields:\n" + fields[0] + "\n" + fields[1] + "\n" + fields[2] + "\n" + fields[3] + "\n" + fields[4] + "\n" + fields[5]);

        tempDependency.leftHeaded = fields[3].equals(LEFT);
        short distance = (short)Integer.parseInt(fields[4]);
        tempDependency.head = new IntTaggedWord(fields[0], '/');
        tempDependency.arg = new IntTaggedWord(fields[2], '/');

        double count = Double.parseDouble(fields[5]);
        if (doingStop) {
          expandStop(tempDependency, distance, count, false);
        } else {
          expandArg(tempDependency, distance, count);
        }
      } catch (Exception e) {
        e.printStackTrace();
        throw new IOException("Error on line " + lineNum + ": " + line);
      }
      //      System.out.println("read line " + lineNum + ": " + line);
      lineNum++;
    }
  }

  /**
   * Writes out data from this Object to the Writer w.
   */
  public void writeData(PrintWriter out) throws IOException {
    // all lines have one rule per line

    for (Iterator iterator = argCounter.keySet().iterator(); iterator.hasNext();) {
      Dependency dependency = (Dependency) iterator.next();
      if (dependency.head != wildTW && dependency.arg != wildTW &&
              dependency.head.word != -1 && dependency.arg.word != -1) {
        double count = argCounter.getCount(dependency);
        out.println(dependency + " " + count);
      }
    }

    out.println("BEGIN_STOP");

    for (Iterator iterator = stopCounter.keySet().iterator(); iterator.hasNext();) {
      Dependency dependency = (Dependency) iterator.next();
      if (dependency.head.word != -1) {
        double count = stopCounter.getCount(dependency);
        out.println(dependency + " " + count);
      }
    }

    out.flush();
  }

  private static final long serialVersionUID = 1L;

} // end class DependencyGrammar

