package edu.stanford.nlp.parser.lexparser;

import edu.stanford.nlp.trees.TreebankLanguagePack;
import edu.stanford.nlp.util.Numberer;

import java.io.*;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

/**
 * Options to the parser which MUST be the SAME at both training and testing (parsing)
 * time in order for the parser to work properly.
 */
public class Options implements Serializable {

  public Options() {
    this(new EnglishTreebankParserParams());
  }

  public Options(TreebankLangParserParams tlpParams) {
    this.tlpParams = tlpParams;
  }

  public int numWords = -1;
  public int numTags = -1;
  public int numStates = -1;

  public static class LexOptions implements Serializable {
    /**
     * Use suffix and capitalization information for unknowns.
     * 0 means a single unknown token.  1 uses suffix, and capitalization.
     * 2 uses a variant (richer) form of signature.  Good.
     * Use this one.  Using the richer signatures in versions 3 or 4 seems
     * to have very marginal or no positive value.
     * 3 uses a richer form of signature that mimics the NER word type
     * patterns.  4 is a variant of 2.  5 is another with more English
     * specific morphology (good for English unknowns!)
     */
    public int useUnknownWordSignatures = 0;

    /**
     * Words more common than this are tagged with MLE P(t|w). Default 100. The
     * smoothing is sufficiently slight that changing this has little effect.
     */
    public int smoothInUnknownsThreshold = 100;

    /**
     * Smarter smoothing for rare words.
     */
    public boolean smartMutation = false;
  };

  public LexOptions lexOptions = new LexOptions();

  /**
   * The treebank-specific parser parameters  to use.
   */
  public TreebankLangParserParams tlpParams;

  /**
   * returns the treebank language pack for the treebank the parser
   * is trained on.
   */
  public TreebankLanguagePack langpack() {
    return tlpParams.treebankLanguagePack();
  }


  /**
   * Forces parsing with strictly CNF grammar -- unary chains are converted
   * to XP&YP symbols and back
   */
  public boolean forceCNF = false;

  /**
   * Do a PCFG parse of the sentence.  If both variables are on,
   * also do a combined parse of the sentence.
   */
  public boolean doPCFG = true;

  /**
   * Do a dependency parse of the sentence.
   */
  public boolean doDep = true;

  /**
   * if true, any child can be the head (seems rather bad!)
   */
  public boolean freeDependencies = false;

  /**
   * Whether dependency grammar considers left/right direction. Good.
   */
  public boolean directional = true;
  public boolean genStop = true;

  /**
   * Use distance bins in the dependency calculations
   */
  public boolean distance = true;
  /**
   * Use coarser distance (4 bins) in dependency calculations
   */
  public boolean coarseDistance = false;

  /* Setting this option to true means that the parser generates ALL tags in
   * the lexicon for every word (some may still be scored at -Inf, but lots
   * of tags that it's never been seen with will have positive probability).
   * If false, it only generates seen tags, and if the word was of low
   * enough frequency in the training data, other tags deemed as open class.
   */
  public boolean flexiTag = false;

  /**
   * "double count" tags rewrites as word in PCFG and Dep parser.  Good for
   * combined parsing only (it used to not kick in for PCFG parsing).  This
   * option is only used at Test time, but it is now in Options, so the
   * correct choice for a grammar is recorded by a serialized parser.
   */
  public boolean dcTags = true;

  /**
   * If true, inside the factored parser, remove any node from the final
   * chosen tree which improves the PCFG score. This was added as the
   * dependency factor tends to encourage 'deep' trees.
   */
  public boolean nodePrune = false;


  public void display() {
    try {
      System.err.println("Options parameters:");
      writeData(new PrintWriter(System.err));
      if (Train.printStates) {
        dumpNumberer(Numberer.getGlobalNumberer("states"), "states", 
                     tlpParams.pw(System.err));
      }
    } catch (IOException e) {
      e.printStackTrace();
    }
  }

  public void writeData(Writer w) throws IOException {
    numWords = Numberer.getGlobalNumberer("words").total();
    numTags = Numberer.getGlobalNumberer("tags").total();
    numStates = Numberer.getGlobalNumberer("states").total();
    PrintWriter out = new PrintWriter(w);
    StringBuilder sb = new StringBuilder();
    sb.append("numWords " + numWords + "\n");
    sb.append("numTags " + numTags + "\n");
    sb.append("numStates " + numStates + "\n");
    sb.append("useUnknownWordSignatures " + lexOptions.useUnknownWordSignatures + "\n");
    sb.append("smoothInUnknownsThreshold " + lexOptions.smoothInUnknownsThreshold + "\n");
    sb.append("parserParams " + tlpParams.getClass().getName() + "\n");
    sb.append("smartMutation " + lexOptions.smartMutation + "\n");
    sb.append("forceCNF " + forceCNF + "\n");
    sb.append("doPCFG " + doPCFG + "\n");
    sb.append("doDep " + doDep + "\n");
    sb.append("freeDependencies " + freeDependencies + "\n");
    sb.append("directional " + directional + "\n");
    sb.append("genStop " + genStop + "\n");
    sb.append("distance " + distance + "\n");
    sb.append("coarseDistance " + coarseDistance + "\n");
    sb.append("flexiTag " + flexiTag + "\n");
    sb.append("dcTags " + dcTags + "\n");
    sb.append("nPrune " + nodePrune + "\n");
    out.print(sb.toString());
    out.flush();
  }


  private void dumpNumberer(Numberer num, String name, PrintWriter pw) {
    pw.println("### Sorted contents of " + name);
    List lis = new ArrayList(num.objects());
    Collections.sort(lis);
    for (Iterator it = lis.iterator(); it.hasNext();) {
      Object obj = it.next();
      pw.println(obj);
    }
    pw.println("### End sorted contents of " + name);
    pw.flush();
  }


  /**
   * Populates data in this Options from the character stream
   * given by the Reader r.
   */
  public void readData(BufferedReader in) throws IOException {
    String line, value;
    final String TRUE = "true";
    line = in.readLine();
    value = line.substring(line.indexOf(' ') + 1);
    numWords = Integer.parseInt(value);
    line = in.readLine();
    value = line.substring(line.indexOf(' ') + 1);
    numTags = Integer.parseInt(value);
    line = in.readLine();
    value = line.substring(line.indexOf(' ') + 1);
    numStates = Integer.parseInt(value);
    line = in.readLine();
    value = line.substring(line.indexOf(' ') + 1);
    lexOptions.useUnknownWordSignatures = Integer.parseInt(value);
    line = in.readLine();
    value = line.substring(line.indexOf(' ') + 1);
    lexOptions.smoothInUnknownsThreshold = Integer.parseInt(value);
    line = in.readLine();
    value = line.substring(line.indexOf(' ') + 1);
    try {
      tlpParams = (TreebankLangParserParams) Class.forName(value).newInstance();
    } catch (Exception e) {
      e.printStackTrace();
      throw new IOException("Problem instantiating parserParams: " + line);
    }
    line = in.readLine();
    value = line.substring(line.indexOf(' ') + 1);
    lexOptions.smartMutation = value.equals(TRUE);
    line = in.readLine();
    // ensure backwards compatibility
    if (line.matches("^forceCNF.*")) {
      value = line.substring(line.indexOf(' ') + 1);
      forceCNF = value.equals(TRUE);
      line = in.readLine();
    }
    value = line.substring(line.indexOf(' ') + 1);
    doPCFG = value.equals(TRUE);
    line = in.readLine();
    value = line.substring(line.indexOf(' ') + 1);
    doDep = value.equals(TRUE);
    line = in.readLine();
    value = line.substring(line.indexOf(' ') + 1);
    freeDependencies = value.equals(TRUE);
    line = in.readLine();
    value = line.substring(line.indexOf(' ') + 1);
    directional = value.equals(TRUE);
    line = in.readLine();
    value = line.substring(line.indexOf(' ') + 1);
    genStop = value.equals(TRUE);
    line = in.readLine();
    value = line.substring(line.indexOf(' ') + 1);
    distance = value.equals(TRUE);
    line = in.readLine();
    value = line.substring(line.indexOf(' ') + 1);
    coarseDistance = value.equals(TRUE);
    line = in.readLine();
    value = line.substring(line.indexOf(' ') + 1);
    flexiTag = value.equals(TRUE);
    line = in.readLine();
    value = line.substring(line.indexOf(' ') + 1);
    dcTags = value.equals(TRUE);
    line = in.readLine();
    if ( ! line.matches("^nPrune.*")) {
      throw new RuntimeException("Expected nPrune, found: " + line);
    }
    value = line.substring(line.indexOf(' ') + 1);
    nodePrune = value.equals(TRUE);
    line = in.readLine(); // get rid of last line
    if (line.length() != 0) {
      throw new RuntimeException("Expected blank line, found: " + line);
    }
  }

  private static final long serialVersionUID = 3L;

} // end class Options

