package edu.stanford.nlp.parser.lexparser;

import edu.stanford.nlp.trees.TreePrint;
import edu.stanford.nlp.trees.TreebankLanguagePack;

import java.util.List;
import java.util.regex.Pattern;

/**
 * Options to the parser which affect performance only at testing (parsing)
 * time.
 */
public class Test {

  /**
   * If false, then failure of the PCFG factor to parse a sentence
   * will trigger parse recovery mode, in which more flexible tagging
   * is allowed. False by default. Use option -noRecoveryTagging to set
   * to true.
   */
  public static boolean noRecoveryTagging = false;

  /* If true, then  failure of the PCFG factor to parse a sentence
   * will trigger parse recovery mode.
   */
  public static boolean doRecovery = true;

  /**
   * If true, the n^4 "speed-up" is not used.
   */
  public static boolean useN5 = false;

  /** If true, use faster iterative deepening CKY algorithm. */
  public static boolean iterativeCKY = false;

  /**
   * The maximum sentence length (including punctuation, etc.) to parse.
   */
  public static int maxLength = -0xDEADBEEF;
  // initial value is -0xDEADBEEF (actually positive because of 2s complement)

  /**
   * The maximum number of edges and hooks combined that the factored parser 
   * will build before giving up.  This number should probably be relative to
   * the sentence length parsed. In general, though, if the parser cannot parse
   * a sentence after this much work then there is no good parse consistent
   * between the PCFG and Dependency parsers.  (Normally, depending on other 
   * flags), the parser will then just return the best PCFG parse.)
   */
  public static int MAX_ITEMS = 200000;

  /**
   *  The amount of smoothing put in (as an m-estimate) for unknown words.
   *  If negative, set by the code in the lexicon class.
   */
  public static double unseenSmooth = -1.0;

  /**
   * Parse trees in test treebank in order of increasing length.
   */
  public static boolean increasingLength = false;

  /**
   * Tag the sentences first, then parse given those (coarse) tags.
   */
  public static boolean preTag = false;

  /**
   * Parse using only tags given from correct answer or the POS tagger
   */
  public static boolean forceTags = preTag;


  /**
   * Write EvalB-readable output files.
   */
  public static boolean evalb = false;

  /**
   * Print a lot of extra output as you parse.
   */
  public static boolean verbose = false; // Don't change this; set with -v

  public static final boolean exhaustiveTest = false;
  public static final boolean pcfgThreshold = false;
  public static final double pcfgThresholdValue = -2.0;

  /**
   * Print out all best PCFG parses.
   */
  public static boolean printAllBestParses = false;

  /**
   * Weighting on dependency log probs
   */
  public static double depWeight = 1.0;
  public static boolean prunePunc = false;


  /**
   * Determines format of output trees: choose among penn, oneline
   */
  public static String outputFormat = "penn";
  public static String outputFormatOptions = "";


  /** If true, write files parsed to a new file with the same name except
   *  for an added ".stp" extension.
   */
  public static boolean writeOutputFiles;

  /**
   * The largest span to consider for word-hood.  Used for parsing unsegmented
   * Chinese text and parsing lattices.  Keep it at 1 unless you know what
   * you're doing.
   */
  public static int maxSpanForTags = 1;

  /**
   * Turns on normalizing scores for sentence length.  Makes no difference
   * (except decreased efficiency) unless maxSpanForTags is greater than one.
   * Works only for PCFG (so far).
   */
  public static boolean lengthNormalization = false;

  /**
   * When you want to force the parse to parse a particular subsequence
   * into a particular state.  Parses will only be made where there is
   * a constituent over the given span which matches (as regular expression)
   * the state Pattern given.
   */
  public static List<Constraint> constraints = null;

  public static class Constraint {
    public int start;
    public int end;
    public Pattern state;
  }

  /**
   * Determines method for print trees on output.
   */
  public static TreePrint treePrint(TreebankLangParserParams tlpParams) {
    TreebankLanguagePack tlp = tlpParams.treebankLanguagePack();
    return new TreePrint(outputFormat, outputFormatOptions, tlp, tlpParams.headFinder());
  }


  public static void display() {
    String str = "Test parameters maxLength=" + maxLength + " preTag=" + preTag + " outputFormat=" + outputFormat + " outputFormatOptions=" + outputFormatOptions + " printAllBestParses=" + printAllBestParses;
    System.err.println(str);
  }

}
