package edu.stanford.nlp.parser.lexparser;

import edu.stanford.nlp.util.Numberer;

import java.io.*;
import java.util.*;

/**
 * Maintains efficient indexing of binary grammar rules.
 *
 * @author Dan Klein
 */
public class UnaryGrammar implements Serializable, Iterable<UnaryRule> {

  /** The number of UnaryRule items in the UnaryGrammar. */
  private int numStates = -1;
  /** The name of the Numberer state space for unary rules. */
  private String stateSpace;

  private transient List[] rulesWithParent = null;
  private transient List[] rulesWithChild = null;

  private transient List[] closedRulesWithParent = null;
  private transient List[] closedRulesWithChild = null;

  private transient UnaryRule[][] closedRulesWithP = null;
  private transient UnaryRule[][] closedRulesWithC = null;

  protected Map coreRules = null;
  private transient Map bestRulesUnderMax = null;
  private transient Map backTrace = null;

  public int numClosedRules() {
    return bestRulesUnderMax.keySet().size();
  }

  public Iterator closedRuleIterator() {
    return bestRulesUnderMax.keySet().iterator();
  }

  public int numRules() {
    return coreRules.keySet().size();
  }

  public Iterator iterator() {
    return ruleIterator();
  }

  public Iterator ruleIterator() {
    return coreRules.keySet().iterator();
  }

  public List rules() {
    return new ArrayList(coreRules.keySet());
  }

  public void purgeRules() {
    Map bR = new HashMap();
    for (Iterator i = bestRulesUnderMax.keySet().iterator(); i.hasNext();) {
      UnaryRule ur = (UnaryRule) i.next();
      if (ur.parent != ur.child) {
        bR.put(ur, ur);
      }
    }
    bestRulesUnderMax = bR;
  }

  public List getBestPath(int parent, int child) {
    List path = new ArrayList();
    UnaryRule tempR = new UnaryRule();
    tempR.parent = parent;
    tempR.child = child;
    //System.out.println("Building path...");
    int loc = parent;
    while (loc != child) {
      path.add(new Integer(loc));
      //System.out.println("Path is "+path);
      tempR.parent = loc;
      Integer nextInt = (Integer) backTrace.get(tempR);
      if (nextInt == null) {
        loc = child;
      } else {
        loc = nextInt.intValue();
      }
      //System.out.println(Numberer.getGlobalNumberer("states").object(parent)+"->"+Numberer.getGlobalNumberer("states").object(child)+" went via "+Numberer.getGlobalNumberer("states").object(loc));
      if (path.size() > 10) {
        System.exit(0);
      }
    }
    path.add(new Integer(child));
    return path;
  }

  private void closeRulesUnderMax(UnaryRule ur) {
    for (int i = 0; i < closedRulesWithChild[ur.parent].size(); i++) {
      UnaryRule pr = (UnaryRule) closedRulesWithChild[ur.parent].get(i);
      for (int j = 0; j < closedRulesWithParent[ur.child].size(); j++) {
        UnaryRule cr = (UnaryRule) closedRulesWithParent[ur.child].get(j);
        UnaryRule resultR = new UnaryRule();
        resultR.parent = pr.parent;
        resultR.child = cr.child;
        resultR.score = pr.score + cr.score + ur.score;
        if (relaxRule(resultR)) {
          if (resultR.parent != ur.parent) {
            backTrace.put(resultR, new Integer(ur.parent));
          } else {
            backTrace.put(resultR, new Integer(ur.child));
          }
        }
      }
    }
  }

  private boolean relaxRule(UnaryRule ur) {
    UnaryRule bestR = (UnaryRule) bestRulesUnderMax.get(ur);
    if (bestR == null) {
      bestRulesUnderMax.put(ur, ur);
      closedRulesWithParent[ur.parent].add(ur);
      closedRulesWithChild[ur.child].add(ur);
      return true;
    } else {
      if (bestR.score < ur.score) {
        bestR.score = ur.score;
        return true;
      }
      return false;
    }
  }

  public double scoreRule(UnaryRule ur) {
    UnaryRule bestR = (UnaryRule) bestRulesUnderMax.get(ur);
    return (bestR != null ? bestR.score : Double.NEGATIVE_INFINITY);
  }

  public void addRule(UnaryRule ur) {
    // add rules' closure
    closeRulesUnderMax(ur);
    coreRules.put(ur, ur);
    rulesWithParent[ur.parent].add(ur);
    rulesWithChild[ur.child].add(ur);
  }
  //public Iterator closedRuleIterator() {
  //  return bestRulesUnderMax.keySet().iterator();
  //}

  void makeCRArrays() {
    int numStates = closedRulesWithParent.length;
    closedRulesWithP = new UnaryRule[numStates][];
    closedRulesWithC = new UnaryRule[numStates][];
    for (int i = 0; i < numStates; i++) {
      closedRulesWithP[i] = (UnaryRule[]) closedRulesWithParent[i].toArray(new UnaryRule[0]);
      closedRulesWithC[i] = (UnaryRule[]) closedRulesWithChild[i].toArray(new UnaryRule[0]);
    }
  }

  public UnaryRule[] closedRulesByParent(int state) {
    if (closedRulesWithP == null) {
      makeCRArrays();
    }
    if (state >= closedRulesWithP.length) {
      return new UnaryRule[0];
    }
    return closedRulesWithP[state];
  }

  public UnaryRule[] closedRulesByChild(int state) {
    if (closedRulesWithC == null) {
      makeCRArrays();
    }
    if (state >= closedRulesWithC.length) {
      return new UnaryRule[0];
    }
    return closedRulesWithC[state];
  }

  public Iterator closedRuleIteratorByParent(int state) {
    if (state >= closedRulesWithParent.length) {
      return Collections.EMPTY_LIST.iterator();
    }
    return closedRulesWithParent[state].iterator();
  }

  public Iterator closedRuleIteratorByChild(int state) {
    if (state >= closedRulesWithChild.length) {
      return Collections.EMPTY_LIST.iterator();
    }
    return closedRulesWithChild[state].iterator();
  }

  public Iterator ruleIteratorByParent(int state) {
    if (state >= rulesWithParent.length) {
      return Collections.EMPTY_LIST.iterator();
    }
    return rulesWithParent[state].iterator();
  }

  public Iterator ruleIteratorByChild(int state) {
    if (state >= rulesWithChild.length) {
      return Collections.EMPTY_LIST.iterator();
    }
    return rulesWithChild[state].iterator();
  }

  public List rulesByParent(int state) {
    if (state >= rulesWithParent.length) {
      return Collections.EMPTY_LIST;
    }
    return rulesWithParent[state];
  }

  public List rulesByChild(int state) {
    if (state >= rulesWithChild.length) {
      return Collections.EMPTY_LIST;
    }
    return rulesWithChild[state];
  }

  public List[] rulesWithParent() {
    return rulesWithParent;
  }

  private void readObject(ObjectInputStream stream) throws IOException, ClassNotFoundException {
    stream.defaultReadObject();
    Set allRules = new HashSet(coreRules.keySet());
    init();
    for (Iterator i = allRules.iterator(); i.hasNext();) {
      UnaryRule ur = (UnaryRule) i.next();
      addRule(ur);
    }
    purgeRules();
  }


  void init() {
    coreRules = new HashMap();
    rulesWithParent = new List[numStates];
    rulesWithChild = new List[numStates];
    closedRulesWithParent = new List[numStates];
    closedRulesWithChild = new List[numStates];
    bestRulesUnderMax = new HashMap();
    backTrace = new HashMap();
    for (int s = 0; s < numStates; s++) {
      rulesWithParent[s] = new ArrayList();
      rulesWithChild[s] = new ArrayList();
      closedRulesWithParent[s] = new ArrayList();
      closedRulesWithChild[s] = new ArrayList();
      UnaryRule selfR = new UnaryRule();
      selfR.parent = s;
      selfR.child = s;
      selfR.score = 0;
      relaxRule(selfR);
    }
  }

  public UnaryGrammar(int numStates) {
    this(numStates, "states");
  }

  public UnaryGrammar(int numStates, String stateSpace) {
    this.numStates = numStates;
    this.stateSpace = stateSpace;
    init();
  }

  /**
   * Populates data in this UnaryGrammar from the character stream
   * given by the Reader r.
   */
  public void readData(BufferedReader in) throws IOException {
    String line;
    int lineNum = 1;
    Numberer n = Numberer.getGlobalNumberer(stateSpace);
    // all lines have one rule per line
    line = in.readLine();
    while (line != null && line.length() > 0) {
      try {
        addRule(new UnaryRule(line, n));
      } catch (Exception e) {
        throw new IOException("Error on line " + lineNum);
      }
      lineNum++;
      line = in.readLine();
    }
    purgeRules();
  }

  /**
   * Writes out data from this Object to the Writer w.
   */
  public void writeData(Writer w) throws IOException {
    PrintWriter out = new PrintWriter(w);
    // all lines have one rule per line
    for (Iterator rI = ruleIterator(); rI.hasNext();) {
      out.println(((UnaryRule) rI.next()).toString());
    }
    out.flush();
  }

  private static final long serialVersionUID = 1L;

} // end class UnaryGrammar
