package edu.stanford.nlp.process;

import edu.stanford.nlp.ling.HasWord;
import edu.stanford.nlp.ling.Sentence;
import edu.stanford.nlp.ling.Word;
import edu.stanford.nlp.objectbank.TokenizerFactory;
import edu.stanford.nlp.objectbank.XMLBeginEndIterator;
import edu.stanford.nlp.util.StringUtils;
import edu.stanford.nlp.web.HTMLParser;

import java.io.*;
import java.net.URL;
import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Fully customizable preprocessor for XML, HTML, and plain text documents.
 * Can take any of a number of input formats and return a {@link List} of tokenized strings.
 *
 * @author Chris Cox
 * @author Jenny Finkel
 */

public class DocumentPreprocessor {


  private static final boolean DEBUG = false;

  private TokenizerFactory tokenizerFactory;
  private String encoding;
  private String[] sentenceFinalPuncWords;

  /*
   *  SET TOKENIZER METHODS
   */
  public DocumentPreprocessor(TokenizerFactory tokenizerFactory) {
    this.tokenizerFactory = tokenizerFactory;
  }

  /**
   * Constructs a preprocessor using the default tokenzier: {@link PTBTokenizer}.
   */
  public DocumentPreprocessor() {
    this.tokenizerFactory = PTBTokenizer.factory();
  }

  /**
   * Set the character encoding.
   *
   * @param encoding
   */
  public void setEncoding(String encoding) {
    this.encoding = encoding;
  }


  public void setSentenceFinalPuncWords(String[] sentenceFinalPuncWords) {
    this.sentenceFinalPuncWords = sentenceFinalPuncWords;
  }

  /**
   * Sets the factory from which to produce a {@link Tokenizer}.  The default is
   * {@link PTBTokenizer}.
   *
   * @param newTokenizerFactory
   */
  public void setTokenizerFactory(TokenizerFactory newTokenizerFactory) {
    tokenizerFactory = newTokenizerFactory;
  }

  public void usePTBTokenizer() {
    tokenizerFactory = PTBTokenizer.factory();
  }

  /**
   * Use tokenizers which tokenize on whitespace.
   */
  public void useWhitespaceTokenizer() {
    tokenizerFactory = WhitespaceTokenizer.factory();
  }

  //
  // METHODS TO PROCESS PLAIN TEXT
  //

  /**
   * Reads the file into a single list of words.
   *
   * @param fileOrURL the path of a text file or URL
   * @return a list of objects of type Word representing words
   * @throws IOException
   */
  public List getWordsFromText(String fileOrURL) throws IOException {
    return getWordsFromText(fileOrURLToReader(fileOrURL));
  }

  /**
   * @param input a Reader of text
   * @return a List of objects of type Word representing words
   */
  public List getWordsFromText(Reader input) {
    Tokenizer tokenizer = tokenizerFactory.getTokenizer(new BufferedReader(input));
    return tokenizer.tokenize();
  }

  /**
   * Reads the file and outputs a list of sentences.
   *
   * @param fileOrURL the path of a text file or URL
   * @return a list of objects of type {@link edu.stanford.nlp.ling.Sentence}
   * @throws IOException
   */
  public List getSentencesFromText(String fileOrURL) throws IOException {
    return getSentencesFromText(fileOrURLToReader(fileOrURL));
  }

  public List getSentencesFromText(String fileOrURL, boolean doPTBEscaping, String sentenceDelimiter, int tagDelimiter) throws IOException {
    return getSentencesFromText(fileOrURLToReader(fileOrURL), doPTBEscaping, sentenceDelimiter, tagDelimiter);
  }

  /**
   * @param input a Reader of text
   * @return a List of objects of type Sentence
   */
  public List getSentencesFromText(Reader input) {
    return getSentencesFromText(input, false, null, -1);
  }

  /**
   * Produce a list of sentences from text.
   *
   * @param input             the path to the filename or URL
   * @param escaper           a {@link Function} that takes a List of HasWords and returns an escaped version of those words.
   *                          Passing in <code>null</code> here means that no escaping is done.
   * @param sentenceDelimiter If null, means that sentences are not segmented already, and should be using default sentence delimiters
   *                          if non-null, means that sentences have already been segmented, and are delimited with this token.
   * @param tagDelimiter
   * @return a List of objects of type List representing sentences
   */
  public List getSentencesFromText(String input, Function<List<HasWord>, List<HasWord>> escaper, String sentenceDelimiter, int tagDelimiter) throws IOException {
    return getSentencesFromText(fileOrURLToReader(input), escaper, sentenceDelimiter, tagDelimiter);
  }


  /**
   * Produce a list of sentences from a reader.
   *
   * @param input             the input
   * @param escaper           a {@link Function} that takes a List of HasWords and returns an escaped version of those words.
   *                          Passing in <code>null</code> here means that no escaping is done.
   * @param sentenceDelimiter If null, means that sentences are not segmented already, and should be using default sentence delimiters
   *                          if non-null, means that sentences have already been segmented, and are delimited with this token.
   * @param tagDelimiter
   * @return a List of objects of type List representing sentences
   */

  public List getSentencesFromText(Reader input, Function<List<HasWord>, List<HasWord>> escaper, String sentenceDelimiter, int tagDelimiter) {
    if (escaper == null) {
      escaper = new NullEscaper();
    }
    ListEscaper listEscaper = new ListEscaper(escaper);

    if (tokenizerFactory instanceof WhitespaceTokenizer.WhitespaceTokenizerFactory) {
      if (sentenceDelimiter == null) {
        // tokenized but sentences are not delimited
        Tokenizer tokenizer = new WhitespaceTokenizer(input, false);
        List words = tokenizer.tokenize();
        if (tagDelimiter >= 0) {
          // split off tags first
          WordToTaggedWordProcessor wttwp = new WordToTaggedWordProcessor((char) tagDelimiter);
          words = wttwp.process(words);
        }
        words = escaper.apply(words);

        WordToSentenceProcessor sp;
        if (sentenceFinalPuncWords != null) {
          sp = new WordToSentenceProcessor(new HashSet(Arrays.asList(sentenceFinalPuncWords)));
        } else {
          sp = new WordToSentenceProcessor();
        }

        List list = sp.process(words);
        return list;
      } else {
        // tokenized, and sentences are delimited
        Tokenizer tokenizer = new WhitespaceTokenizer(input, sentenceDelimiter.equals("\n"));
        List words = tokenizer.tokenize();
        List sentences = splitListsOnToken(words, sentenceDelimiter);
        if (tagDelimiter >= 0) {
          sentences = tagSplitSentences(sentences, tagDelimiter);
        }
        sentences = listEscaper.apply(sentences);
        return sentences;
      }
    } else {
      if (tagDelimiter >= 0) {
        throw new RuntimeException("Can't read tags from untokenized document.");
      }
      // no tags
      if (sentenceDelimiter == null) {
        // the simple case: not tokenized, not delimited, not tagged
        if (DEBUG) {
          System.out.println("doing simple case");
        }
        Tokenizer tokenizer = tokenizerFactory.getTokenizer(new BufferedReader(input));
        List words = tokenizer.tokenize();
        words = escaper.apply(words);

        WordToSentenceProcessor sp;
        if (sentenceFinalPuncWords != null) {
          sp = new WordToSentenceProcessor(new HashSet(Arrays.asList(sentenceFinalPuncWords)));
        } else {
          sp = new WordToSentenceProcessor();
        }

        return sp.process(words);
      } else {
        // not tokenized, but sentence delimited, so we must tokenize twice
        // first tokenization to look for the delimiter
        Tokenizer tokenizer = new WhitespaceTokenizer(input, true);
        List tokens = tokenizer.tokenize();
        List sentences = glueSentences(splitListsOnToken(tokens, sentenceDelimiter));
        // second tokenization done per sentence
        return tokenizeSentences(sentences);
      }
    }

  }


  /**
   * Gets a list of words from a string.
   *
   * @param input string
   * @return a List of objects of type Word representing words
   */
  public List getWordsFromString(String input) {
    Tokenizer tokenizer = tokenizerFactory.getTokenizer(new BufferedReader(new StringReader(input)));
    return tokenizer.tokenize();
  }


  //
  // METHODS TO PROCESS XML
  //
  /**
   * Returns a list of sentences contained in an XML file, occuring between the begin and end of a selected tag.
   * It escapes sentences using a {@link WordToSentenceProcessor}.
   * By default, it does PTBEscaping as well.
   *
   * @param fileOrURL
   * @param splitOnTag the tag which denotes text boundaries
   * @return
   * @throws IOException
   */
  public List getSentencesFromXML(String fileOrURL, String splitOnTag) throws IOException {
    return getSentencesFromXML(fileOrURL, splitOnTag, true);
  }

  /**
   * Returns a list of sentences contained in an XML file, occuring between the begin and end of a selected tag.
   * It escapes sentences using a {@link WordToSentenceProcessor}.
   *
   * @param fileOrURL
   * @param splitOnTag    the tag which denotes text boundaries
   * @param doPTBEscaping whether to escape PTB tokens using a {@link PTBEscapingProcessor}
   * @return
   * @throws IOException
   */
  public List getSentencesFromXML(String fileOrURL, String splitOnTag, boolean doPTBEscaping) throws IOException {
    return getSentencesFromXML(fileOrURLToReader(fileOrURL), splitOnTag, doPTBEscaping);
  }

  /**
   * Returns a list of sentences contained in an XML file, occuring between the begin and end of a selected tag.
   *
   * @param input
   * @param splitOnTag    the tag which denotes text boundaries
   * @param doPTBEscaping whether to escape PTB tokens using a {@link PTBEscapingProcessor}
   * @return
   */
  public List getSentencesFromXML(Reader input, String splitOnTag, boolean doPTBEscaping) {


    XMLBeginEndIterator xmlIter = new XMLBeginEndIterator(input, splitOnTag);
    List<Sentence> l = new ArrayList<Sentence>();

    while (xmlIter.hasNext()) {
      //get the next string delimited by split tags
      String s = (String) xmlIter.next();
      //TODO:Create more options here
      //get this string as a list of sentences with appropriate options
      List<Sentence> section = this.getSentencesFromText(new BufferedReader(new StringReader(s)), doPTBEscaping, null, -1);
      //place these sentences on the master list
      for (Sentence individual : section) {
        l.add(individual);
      }
    }

    return l;
  }


  //
  // METHODS TO PROCESS HTML
  //

  public List getWordsFromHTML(String fileOrURL) throws IOException {
    return getWordsFromHTML(fileOrURLToReader(fileOrURL));
  }

  public List getWordsFromHTML(Reader input) {
    HTMLParser parser = new HTMLParser();
    String s;
    try {
      s = parser.parse(input);
      return getWordsFromText(new StringReader(s));
    } catch (IOException e) {
      System.err.println("IOException" + e.getMessage());
    }
    return null;
  }

  public List getSentencesFromHTML(String fileOrURL) throws IOException {
    return getSentencesFromHTML(fileOrURLToReader(fileOrURL));
  }

  public List getSentencesFromHTML(Reader input) {
    HTMLParser parser = new HTMLParser();
    String s;
    try {
      s = parser.parse(input);
      return getSentencesFromText(new StringReader(s));
    } catch (IOException e) {
      System.err.println("IOException" + e.getMessage());
    }
    return null;
  }


  //
  // PRIVATE
  //


  private List getSentencesFromText(Reader fileOrURL, boolean doPTBEscaping, String sentenceDelimiter, int tagDelimiter) {
    Function<List<HasWord>, List<HasWord>> escaper = null;
    if (doPTBEscaping) {
      escaper = new PTBEscapingProcessor();
    } else {
      escaper = new NullEscaper();
    }
    return getSentencesFromText(fileOrURL, escaper, sentenceDelimiter, tagDelimiter);
  }


  private static class NullEscaper implements Function<List<HasWord>, List<HasWord>> {
    public List<HasWord> apply(List<HasWord> hasWords) {
      return hasWords;
    }
  }

  private static class ListEscaper implements Function<List<List<HasWord>>, List<List<HasWord>>> {

    public ListEscaper(Function<List<HasWord>, List<HasWord>> f) {
      this.f = f;
    }

    Function<List<HasWord>, List<HasWord>> f;

    public List<List<HasWord>> apply(List<List<HasWord>> lists) {
      List<List<HasWord>> result = new ArrayList<List<HasWord>>(lists.size());
      for (List<HasWord> l : lists) {
        result.add(f.apply(l));
      }
      return result;
    }
  }


  /**
   * @return a List of Lists of tokens. delimiter tokens are removed.
   */
  private List splitListsOnToken(List tokens, String sentenceDelimiter) {
    List result = new ArrayList();
    List sentence = new ArrayList();
    for (int i = 0; i < tokens.size(); i++) {
      Word word = (Word) tokens.get(i);
      if (word.word().equals(sentenceDelimiter)) {
        // don't add this token
        // finish this sentence and add it
        result.add(sentence);
        sentence = new ArrayList();
      } else {
        sentence.add(word);
      }
    }
    if (!sentence.isEmpty()) {
      result.add(sentence);
    }
    return result;
  }

  private List glueSentences(List sentences) {
    List result = new ArrayList();
    for (int i = 0; i < sentences.size(); i++) {
      List sentence = (List) sentences.get(i);
      result.add(glueSentence(sentence));
    }
    return result;
  }

  private String glueSentence(List sentence) {
    StringBuffer result = new StringBuffer();
    Word word = (Word) sentence.get(0);
    String s = word.word();
    result.append(s);
    for (int i = 1; i < sentence.size(); i++) {
      word = (Word) sentence.get(i);
      s = word.word();
      result.append(" " + s);
    }
    return result.toString();
  }


  private List tokenizeSentences(List sentences) {
    List result = new ArrayList();
    for (int i = 0; i < sentences.size(); i++) {
      String sentence = (String) sentences.get(i);
      Tokenizer tok = tokenizerFactory.getTokenizer(new StringReader(sentence));
      result.add(tok.tokenize());
    }
    return result;
  }

  private List tagSplitSentences(List sentences, int tagDelimiter) {
    List result = new ArrayList();
    WordToTaggedWordProcessor wttwp = new WordToTaggedWordProcessor((char) tagDelimiter);
    for (int i = 0; i < sentences.size(); i++) {
      List sentence = (List) sentences.get(i);
      sentence = wttwp.process(sentence);
      result.add(sentence);
    }
    return result;
  }

  private Pattern urlPattern = Pattern.compile(".*?tp://.*?");

  private Reader fileOrURLToReader(String fileOrURL) throws IOException {
    if (DEBUG) {
      System.out.println("fileOrURLToReader(" + fileOrURL + ")");
    }
    Matcher m = urlPattern.matcher(fileOrURL);
    if (m.matches()) {
      URL url = new URL(fileOrURL);
      return new BufferedReader(new StringReader(StringUtils.slurpURL(url)));
    } else {
      if (encoding == null) {
        return new FileReader(fileOrURL);
      } else {
        return new BufferedReader(new InputStreamReader(new FileInputStream(fileOrURL), encoding));
      }
    }
  }

  private List tokenizeListOfStrings(List stringList) {
    List docList = new ArrayList();
    Reader r;
    Tokenizer tok;
    for (int i = 0; i < stringList.size(); i++) {
      r = new BufferedReader(new StringReader((String) stringList.get(i)));
      tok = tokenizerFactory.getTokenizer(r);
      docList.add(tok.tokenize());
    }
    return docList;
  }

  static final int plain = 0;
  static final int xml = 1;
  static final int html = 2;


  /** This provides a simple test method for DocumentPreprocessor.
   *  Usage:
   *  DocumentPreprocessor -file filename [-xml tag|-html] [-noSplitSentence]
   */
  public static void main(String[] args) throws IOException {
    if (args.length == 0) {
      System.err.println("usage: DocumentPreprocessor -file filename [-xml tag|-html] [-noSplitSentence]");
      return;
    }

    boolean splitSentences = true;
    String xmlTag = null;
    DocumentPreprocessor docPreprocessor = new DocumentPreprocessor();

    int fileType = 0;

    String file = null;

    for (int i = 0; i < args.length; i++) {
      if (args[i].equals("-file")) {
        file = args[++i];
      } else if (args[i].equals("-xml")) {
        fileType = xml;
        xmlTag = args[++i];
      } else if (args[i].equals("-html")) {
        fileType = html;
      } else if (args[i].equals("-noSplitSentence")) {
        splitSentences = false;
      }
    }

    List docs = null;

    switch (fileType) {
      case (plain):
        if (splitSentences) {
          docs = docPreprocessor.getSentencesFromText(file);
        } else {
          docs = docPreprocessor.getWordsFromText(file);
        }
        break;
      case (xml):
        docs = docPreprocessor.getSentencesFromXML(file, xmlTag);
        break;
      case (html):
        if (splitSentences) {
          docs = docPreprocessor.getSentencesFromHTML(file);
        } else {
          docs = docPreprocessor.getWordsFromHTML(file);
        }
        break;
    }

    System.err.println(docs.size());
    for (Iterator iter = docs.iterator(); iter.hasNext();) {
      List l = (List) iter.next();
      System.err.println(l);
    }

  }
}
