package edu.stanford.nlp.process;

import edu.stanford.nlp.ling.Word;
import edu.stanford.nlp.objectbank.TokenizerFactory;

import java.io.FileReader;
import java.io.IOException;
import java.io.Reader;
import java.io.StreamTokenizer;
import java.util.Iterator;

/**
 * Simple Tokenizer implementation that tokenizes on whitespace.
 * This implementation returns Word objects. It has a parameter for whether
 * to make EOL a token.  If it is, it is return as a Word with String value
 * "\n".
 *
 * @author Joseph Smarr (jsmarr@stanford.edu)
 * @author Teg Grenager (grenager@stanford.edu)
 */
public class WhitespaceTokenizer extends AbstractTokenizer {

  /**
   * A factory which vends SimpleTokenizers.
   *
   * @author Christopher Manning
   */
  static class WhitespaceTokenizerFactory implements TokenizerFactory {

    private final boolean eolIsSignificant;

    public WhitespaceTokenizerFactory() {
      this(false);
    }

    public WhitespaceTokenizerFactory(boolean eolIsSignificant) {
      this.eolIsSignificant = eolIsSignificant;
    }

    public Iterator getIterator(Reader r) {
      return getTokenizer(r);
    }

    public Tokenizer getTokenizer(Reader r) {
      return new WhitespaceTokenizer(r, eolIsSignificant);
    }

  } // end class WhitespaceTokenizerFactory

  private StreamTokenizer st;
  private final boolean eolIsSignificant;

  /**
   * Internally fetches the next token.
   *
   * @return the next token in the token stream, or null if none exists.
   */
  protected Object getNext() {
    Object token = null;
    if (st == null) {
      return null;
    }
    try {
      st.nextToken();
      if (st.ttype == StreamTokenizer.TT_EOF) {
        return null;
      } else if (st.ttype == StreamTokenizer.TT_EOL) {
        if (eolIsSignificant) {
          token = new Word("\n");
        } else {
          return null;
        }
      } else {
        token = new Word(st.sval);
      }
    } catch (IOException e) {
      // do nothing, return null
    }
    return token;
  }

  /**
   * Constructs a new WhitespaceTokenizer with the Reader r as its source.
   */
  public WhitespaceTokenizer(Reader r) {
    this(r, false);
  }


  /**
   * Constructs a new WhitespaceTokenizer with the Reader r as its source.
   */
  public WhitespaceTokenizer(Reader r, boolean eolIsSignificant) {
    this.eolIsSignificant = eolIsSignificant;
    // The conditional below is currently needed in LexicalizedParser, since 
    // it passes in a null arg while doing type-checking for sentence escaping
    // but StreamTokenizer barfs on that.  But maybe shouldn't be here.
    if (r != null) {
      setSource(r);
    }
  }

  /**
   * Sets the source of this Tokenizer to be the Reader r.
   */
  private void setSource(Reader r) {
    st = new StreamTokenizer(r);
    st.resetSyntax();
    st.whitespaceChars(0, 0x20);
    st.wordChars(0x21, 0x7E);
    st.wordChars(0xA1, 0xFF); // perhaps okay for iso-8859-1 too
    st.eolIsSignificant(eolIsSignificant);
  }

  public static TokenizerFactory factory() {
    return new WhitespaceTokenizerFactory(false);
  }

  public static TokenizerFactory factory(boolean eolIsSignificant) {
    return new WhitespaceTokenizerFactory(eolIsSignificant);
  }

  // for debugging
  private String type(int n) {
    if (n == StreamTokenizer.TT_EOF) {
      return "EOF";
    }
    if (n == StreamTokenizer.TT_EOL) {
      return "EOL";
    }
    if (n == StreamTokenizer.TT_NUMBER) {
      return "NUMBER";
    }
    if (n == StreamTokenizer.TT_WORD) {
      return "WORD";
    }
    return "nothing";
  }

  /**
   * Reads a file from the argument and prints its tokens one per line.
   * This is mainly as a testing aid, but it can also be quite useful
   * standalone to turn a corpus into a one token per line file of tokens.
   * <p/>
   * Usage: <code>java edu.stanford.nlp.process.WhitespaceTokenizer filename
   * </code>
   *
   * @param args Command line arguments
   */
  public static void main(String[] args) throws IOException {
    if (args.length < 1) {
      System.err.println("usage: java edu.stanford.nlp.process." + "WhitespaceTokenizer [-cr] filename");
      return;
    }
    Tokenizer tokenizer = new WhitespaceTokenizer(new FileReader(args[args.length - 1]), args[0].equals("-cr"));
    while (tokenizer.hasNext()) {
      System.out.println(tokenizer.next());
    }
  }

} // end class WhitespaceTokenizer

