package edu.stanford.nlp.trees;

import edu.stanford.nlp.ling.*;
import edu.stanford.nlp.process.Morphology;
import edu.stanford.nlp.util.Pair;
import edu.stanford.nlp.util.ArrayUtils;
import static java.lang.System.err;
import java.util.ArrayList;
import java.util.List;
import java.util.Set;
import java.util.HashSet;
import java.io.PrintWriter;

/**
 * Finds WordNet collocations in parse trees.
 *
 * @author Chris Cox
 */

public class CollocationFinder {
  private static boolean DEBUG = false;
  private Tree qTree;
  private HeadFinder hf;
  private ArrayList<Collocation> collocationCollector;
  private WordNetConnection wnConnect;

  /**
   * Construct a new <code>CollocationFinder</code> over the <code>Tree</code> t.
   * The default {@link HeadFinder} is a {@link CollinsHeadFinder}.
   * @param t parse tree
   * @param w wordnet connection
   */
  public CollocationFinder(Tree t,WordNetConnection w) {
    this(t,w,new CollinsHeadFinder());
  }
  /**
   * Construct a new <code>CollocationFinder</code> over the <code>Tree</code> t.
   * @param t parse tree
   * @param w w
   * @param hf {@link HeadFinder} to use
   */
  public CollocationFinder(Tree t,WordNetConnection w, HeadFinder hf){
    this.wnConnect=w;
    this.qTree=t;
    this.collocationCollector = new ArrayList<Collocation>();
    this.hf=hf;
    this.getCollocationsList();
  }

  /**
   * Returns the "collocations included" parse tree.
   *
   * @return the mangled tree which applies collocations found in this object.
   */
  public Tree getMangledTree(){
    return getMangledTree(qTree);
  }

  private Tree getMangledTree(Tree t) {
    Collocation matchingColl = null;
    for(Tree child : t.children()){
      child=getMangledTree(child);
    }
    for (Collocation c : collocationCollector) {
      //if there are multiple collocations with the same parent node,
      //this will take the longer one.
      if (t.equals(c.parentNode)) {
        matchingColl = c;
        if (DEBUG) {
          err.println("Founding matching collocation for tree\n");
          t.pennPrint();
          err.println("head label: " + c.headLabel);
          err.println("collocation string: " + c.collocationString);
          err.println("Constituents: "+ c.indicesOfConstituentChildren);
        }
      }
    }
    if (matchingColl == null) {
      return t;
    } else {
      Tree[] allChildren = t.children();
      //get the earliest child in the collocation and store it as first child.
      //delete the rest.
      StringBuffer mutatedString = new StringBuffer(16000);
      for (int i : matchingColl.indicesOfConstituentChildren) {
        String strToAppend = mergeLeavesIntoCollocatedString(allChildren[i]);
        mutatedString.append(strToAppend);
        mutatedString.append("_"); 
      }
      //Starting with the latest constituent, delete all the "pruned" children.
      for (int index = matchingColl.indicesOfConstituentChildren.size() - 1; index > 0; index--) {
        int thisConstituent = matchingColl.indicesOfConstituentChildren.get(index);
        allChildren = (Tree[]) ArrayUtils.removeAt(allChildren, thisConstituent);
      }
      //name for the leaf string of our new collocation
      String newNodeString = mutatedString.substring(0, mutatedString.length() - 1);

      if (DEBUG) {
        err.println("Merged leaves into: " + newNodeString);
      }

      int firstChildIndex = matchingColl.indicesOfConstituentChildren.get(0);
      //now we mutate the earliest constitutent
      Tree newCollocationChild = allChildren[firstChildIndex].localTree();
      assert(newCollocationChild.children().length==1);
      newCollocationChild.setValue(matchingColl.headLabel.value());
      newCollocationChild.firstChild().setValue(newNodeString);

      allChildren[firstChildIndex] = newCollocationChild;
      t.setChildren(allChildren);
      return t;
    }
  }

/**
 * Traverses the parse tree to find WordNet collocations.
 *
 * @return a {@link List} of lists of the indices of collocation constituents.
 */
  private ArrayList<Pair<Integer,Integer>> getCollocationsList() {
    getCollocationsList(qTree);
    return null;
  }

  /**
   * Prints the collocations found in this <code>Tree</code> as strings.
   * Each is followed by its boundary constituent indices in the original tree.
   * <br>Example: <code> throw_up (2,3) </code>
   *   <br>       <code> came_up_with (7,9) </code>
   */
  public void PrintCollocationStrings(PrintWriter pw){
    ArrayList<String> strs = new ArrayList<String>();
    for(Collocation c: collocationCollector){
      String cs = c.collocationString;
      pw.println(cs+" ("+(c.span.first()+1)+","+(c.span.second()+1)+")");
    }
  }

  /**
   * This method does the work of traversing the tree and writing collocations to the
   * CollocationCollector.
   * @param t
   */
  private void getCollocationsList(Tree t){
    int leftMostLeaf = Trees.leftEdge(t,qTree);
    if(t.isPreTerminal()) return;
    List<Tree> children = t.getChildrenAsList();
    if(children.size()==0) return;
    //TODO: fix determineHead
    // - in phrases like "World Trade Organization 's" the head of the parent NP is "POS".
    // - this is problematic for the collocationFinder which assigns this head
    // as the POS for the collocation "World_Trade_Organization"!
    Label headLabel= hf.determineHead(t).label();
    StringBuffer testString = null;
    Integer leftSistersBuffer=0;//measures the length of sisters in words when reading
    for(int i =0; i < children.size();i++){
      ArrayList<Integer> childConstituents = new ArrayList<Integer>();
      childConstituents.add(i);
      Tree subtree = children.get(i);
      Integer currWindowLength=0; //measures the length in words of the current collocation.
      getCollocationsList(subtree); //recursive call to get colls in subtrees.
      testString=new StringBuffer(16000);
      testString.append(treeAsStemmedCollocation(subtree));
      testString.append("_");
      Integer thisSubtreeLength= subtree.yield().size();
      currWindowLength+=thisSubtreeLength;
      
      //for each subtree i, we iteratively append word yields of succeeding sister
      //subtrees j and check their wordnet entries.  if they exist we write them to
      //the global collocationCollectior pair by the indices of the leftmost and 
      //rightmost words in the collocation.

      for(int j = i+1; j < children.size();j++){
        Tree sisterNode = children.get(j);
        childConstituents.add(j);
        testString.append(treeAsStemmedCollocation(sisterNode));
        currWindowLength+=sisterNode.yield().size();
        if(DEBUG){
       //   err.println("Testing string w/ reported indices:" + testString.toString()
         //             + " (" +(leftMostLeaf+leftSistersBuffer)+","+(leftMostLeaf+leftSistersBuffer+currWindowLength-1)+")");
        }
        //ignore collocations beginning with "the"
        if(wordNetContains(testString)){
          Pair <Integer, Integer> c = new Pair<Integer,Integer>(leftMostLeaf+leftSistersBuffer,leftMostLeaf+leftSistersBuffer+currWindowLength-1);

          Collocation col = new Collocation(c,t,(ArrayList<Integer>)childConstituents.clone(),testString.toString(),headLabel);
          collocationCollector.add(col);
          if(DEBUG){
            err.println("Found collocation in wordnet: "+ testString.toString());
          }
        }
        testString.append("_");
      }
      leftSistersBuffer+=thisSubtreeLength;

    }
    return;
  };

  private static String treeAsStemmedCollocation(Tree t){
    ArrayList<WordTag> list= getStemmedWordTagsFromTree(t);
   // err.println(list.size());
    StringBuffer s=new StringBuffer(16000);
    WordTag firstWord = list.remove(0);
    s.append(firstWord.word());

    for(WordTag wt : list){
      s.append("_");
      s.append(wt.word());

    }
    //err.println("Expressing this as:"+s.toString());
    return s.toString();
  }
  private static String mergeLeavesIntoCollocatedString(Tree t){
    StringBuffer sb = new StringBuffer(16000);
    Sentence sent = t.taggedYield();
    for(int i =0;i<sent.size();i++){
      sb.append((sent.getHasWord(i).word())+"_");
    }
    return sb.substring(0,sb.length() -1);
  }

  private static String mergeLeavesIntoCollocatedString(Tree[] trees){
    StringBuffer sb = new StringBuffer(16000);
    for(Tree t: trees){
      Sentence sent = t.taggedYield();
      for(int i =0;i<sent.size();i++){
        sb.append((sent.getHasWord(i).word())+"_");
      }
    }
    return sb.substring(0,sb.length() -1);
  }
  /**
   *
   * @param t a tree
   * @return the WordTags corresponding to the leaves of the tree,
   * stemmed according to their POS tags in the tree.
   */
  private static ArrayList<WordTag> getStemmedWordTagsFromTree(Tree t) {
    ArrayList<WordTag> stemmedWordTags = new ArrayList<WordTag>();
    Sentence s = t.taggedYield();
    for(int i = 0; i < s.size(); i++) {
      TaggedWord w = (TaggedWord) s.getHasWord(i);
      WordTag wt = Morphology.stemStatic(w.word(), w.tag());
      stemmedWordTags.add(wt);
    }
    return stemmedWordTags;
  }

  private boolean wordNetContains(StringBuffer s) {
    return wnConnect.wordNetContains(s);
  }

/**
 * Holds information for one collocation.
 */
  private class Collocation{

    Pair<Integer,Integer> span;
    Tree parentNode;
    Label headLabel;
    ArrayList<Integer> indicesOfConstituentChildren;
    String collocationString;

    private Collocation(Pair<Integer,Integer> span, Tree parentNode,ArrayList<Integer>
                        indicesOfConstituentChildren,String collocationString,Label headLabel){
      this.span=span;
      this.parentNode = parentNode;
      this.collocationString=collocationString;
      this.indicesOfConstituentChildren=indicesOfConstituentChildren;
      this.headLabel=headLabel;
    }

  }

}
