package edu.stanford.nlp.trees;

import edu.stanford.nlp.io.NumberRangesFileFilter;
import edu.stanford.nlp.util.FilePathProcessor;
import edu.stanford.nlp.util.FileProcessor;
import edu.stanford.nlp.util.Timing;

import java.io.File;
import java.io.FileFilter;
import java.io.Reader;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

/**
 * A <code>DiskTreebank</code> is a <code>Collection</code> of 
 * <codeTree</code>s.
 * A <code>DiskTreebank</code> object stores merely the information to
 * get at a corpus of trees that is stored on disk.  Access is usually
 * via apply()'ing a TreeVisitor to each Tree in the Treebank or by using
 * an iterator() to get an iteration over the Trees.
 *
 * @author Christopher Manning
 */
public final class DiskTreebank extends Treebank {

  private static final boolean PRINT_FILENAMES = false;
  private ArrayList filePaths = new ArrayList();
  private ArrayList fileFilters = new ArrayList();

  /**
   * Maintains as a class variable the <code>File</code> from which
   * trees are currently being read.
   */
  private File currentFile = null;

  /**
   * If this is true, the system will retry opening files a few times
   * before concluding that there is really a problem. This seems to
   * be necessary with NFS on Linux boxes -- at least the DB ones.
   */
  private static final boolean BROKEN_NFS = true;


  /**
   * Create a new DiskTreebank.
   * The trees are made with a <code>LabeledScoredTreeReaderFactory</code>.
   * <p/>
   * <i>Compatibility note: Until Sep 2004, this used to create a Treebank
   * with a SimpleTreeReaderFactory, but this was changed as the old
   * default wasn't very useful, especially to naive users.</i>
   */
  public DiskTreebank() {
    this(new LabeledScoredTreeReaderFactory());
  }

  /**
   * Create a new tree bank, set the encoding for file access.
   *
   * @param encoding The charset encoding to use for treebank file decoding
   */
  public DiskTreebank(String encoding) {
    this(new LabeledScoredTreeReaderFactory(), encoding);
  }

  /**
   * Create a new DiskTreebank.
   *
   * @param trf the factory class to be called to create a new
   *            <code>TreeReader</code>
   */
  public DiskTreebank(TreeReaderFactory trf) {
    super(trf);
  }

  /**
   * Create a new DiskTreebank.
   *
   * @param trf      the factory class to be called to create a new
   *                 <code>TreeReader</code>
   * @param encoding The charset encoding to use for treebank file decoding
   */
  public DiskTreebank(TreeReaderFactory trf, String encoding) {
    super(trf, encoding);
  }

  /**
   * Create a new Treebank.
   * The trees are made with a <code>LabeledScoredTreeReaderFactory</code>.
   * <p/>
   * <i>Compatibility note: Until Sep 2004, this used to create a Treebank
   * with a SimpleTreeReaderFactory, but this was changed as the old
   * default wasn't very useful, especially to naive users.</i>
   *
   * @param initialCapacity The initial size of the underlying Collection.
   *                        For a <code>DiskTreebank</code>, this parameter is ignored.
   */
  public DiskTreebank(int initialCapacity) {
    this(initialCapacity, new LabeledScoredTreeReaderFactory());
  }

  /**
   * Create a new Treebank.
   *
   * @param initialCapacity The initial size of the underlying Collection,
   *                        For a <code>DiskTreebank</code>, this parameter is ignored.
   * @param trf             the factory class to be called to create a new
   *                        <code>TreeReader</code>
   */
  public DiskTreebank(int initialCapacity, TreeReaderFactory trf) {
    this(trf);
  }

  private TreeTransformer readTransformer;
  /**
   * Transform all trees with this TreeTransformer as they are read.
   * @param tt
   */
  public void transformOnRead(TreeTransformer tt) {
    readTransformer = tt;
  }

  /**
   * Empty a <code>Treebank</code>.
   */
  public void clear() {
    filePaths.clear();
    fileFilters.clear();
  }

  /**
   * Load trees from given directory.  This version just records
   * the paths to be processed, and actually processes them at apply time.
   *
   * @param path file or directory to load from
   * @param filt a FilenameFilter of files to load
   */
  public void loadPath(File path, FileFilter filt) {
    filePaths.add(path);
    fileFilters.add(filt);
  }

  /**
   * Applies the TreeVisitor to to all trees in the Treebank.
   *
   * @param tp A class that can process trees.
   */
  public void apply(final TreeVisitor tp) {
    for (Tree t : this) {
      tp.visitTree(t);
    }
  }

  /**
   * Return the <code>File</code> from which trees are currently being
   * read by <code>apply()</code>, and pased to a
   * <code>TreePprocessor</code>.  This is useful if one wants to map
   * the original file and
   * directory structure over to a set of modified trees.
   *
   * @return the file that trees are currently being read from, or
   *         <code>null</code> if no file is currently open
   */
  public File getCurrentFile() {
    return currentFile;
  }

  private class DiskTreebankIterator implements Iterator<Tree> {

    private int fileUpto = -1; // before starting on index array 0
    private int treeUpto; // = 0
    private List files;
    private MemoryTreebank currentFileTrees;
    private boolean hasNext;

    private DiskTreebankIterator() {
      files = new ArrayList();
      // get the list of files in the Treebank via a new
      // FilePathProcessor()
      FileProcessor dtifp = new FileProcessor() {
        public void processFile(File file) {
          files.add(file.toString());
        };
      };
      int numPaths = filePaths.size();
      for (int i = 0; i < numPaths; i++) {
        FilePathProcessor.processPath((File) filePaths.get(i), (FileFilter) fileFilters.get(i), dtifp);
      }
      currentFileTrees = new MemoryTreebank(treeReaderFactory(), encoding());
      // we're now all setup to read a new file on the priming call
      // first time treeUpto = currentFileTrees.size() = 0
      hasNext = primeNextFile();
    }

    private boolean primeNextFile() {
      while (fileUpto < files.size()) {
        if (treeUpto < currentFileTrees.size()) {
          return true;
        } else {
          // load the next file
          currentFileTrees.clear();
          fileUpto++;
          treeUpto = 0;
          if (fileUpto < files.size()) {
            String fname = (String) files.get(fileUpto);
            // maybe print file name to stdout to get some feedback
            if (PRINT_FILENAMES) {
              System.err.println(fname);
            }
            currentFileTrees.loadPath(fname);
          }
        }
      }
      // there's nothing left;
      return false;
    }

    /**
     * Returns true if the iteration has more elements.
     */
    public boolean hasNext() {
      return hasNext;
    }

    /**
     * Returns the next element in the iteration.
     */
    public Tree next() {
      Tree ret = currentFileTrees.get(treeUpto++);
      hasNext = primeNextFile();
      if (readTransformer != null) {
        ret = readTransformer.transformTree(ret);
      }
      return ret;
    }

    /**
     * Not supported
     */
    public void remove() {
      throw new UnsupportedOperationException();
    }

  }

  /**
   * Return an Iterator over Trees in the Treebank.  This is implemented
   * by building per-file MemoryTreebanks for the files in the
   * DiskTreebank.  As such, it isn't as efficient as using
   * <code>apply()</code>.
   */
  public Iterator<Tree> iterator() {
    return new DiskTreebankIterator();
  }

  /**
   * Loads treebank and prints it.
   * All files below the designated <code>filePath</code> within the given 
   * number range if any are loaded.  You can normalize the trees or not 
   * (English-specific) and print trees one per line up to a certain length
   * (for EVALB).
   * <p>
   * Usage: <code>
   * java edu.stanford.nlp.trees.DiskTreebank [-maxLength n|-normalize] filePath numberRanges
   * </code>
   *
   * @param args Array of command-line arguments
   */
  public static void main(String[] args) {
    int i = 0;
    int maxLength = -1;
    boolean normalized = false;
    while (i < args.length && args[i].startsWith("-")) {
      if (args[i].equals("-maxLength") && i + 1 < args.length) {
        maxLength = Integer.parseInt(args[i+1]);
        i += 2;
      } else if (args[i].equals("-normalized")) {
        normalized = true;
        i += 1;
      } else {
        System.err.println("Unknown option: " + args[i]);
        i++;
      }
    }
    Treebank treebank;
    if (normalized) {
      treebank = new DiskTreebank();
    } else {
      treebank = new DiskTreebank(new TreeReaderFactory() {
          public TreeReader newTreeReader(Reader in) {
            return new PennTreeReader(in, new LabeledScoredTreeFactory());
          }
        });
    }

    if (i + 1 < args.length ) {
      treebank.loadPath(args[i], new NumberRangesFileFilter(args[i+1], true));
    } else {
      treebank.loadPath(args[i]);
    }

    if (maxLength >= 0) {
      for (Tree t: treebank) {
        if (t.yield().length() <= maxLength) {
          System.out.println(t);
        }
      }
    } else {
      System.out.println();
      Timing.startTime();
      int num = 0;
      for (Tree t : treebank) {
        num += t.yield().length();
      }
      Timing.endTime("traversing corpus, counting words with iterator");

      treebank.apply(new TreeVisitor() {
          int num = 0;
          public void visitTree(final Tree t) {
            num += t.yield().length();
          }
        });
      System.out.println();
      Timing.endTime("traversing corpus, counting words with TreeVisitor");

      System.out.println();
      Timing.startTime();
      System.out.println("This treebank contains " + treebank.size() + " trees.");
      Timing.endTime("size of corpus");
    }
  }

}
