package edu.stanford.nlp.trees;

import java.util.Arrays;
import java.util.Collections;
import java.util.List;

import static edu.stanford.nlp.trees.GrammaticalRelation.*;


/**
 * <code>EnglishGrammaticalRelations</code> is a
 * set of {@link GrammaticalRelation} objects for the English language.<p>
 *
 * This set of English grammatical relations is not intended to be
 * exhaustive or immutable.  It's just where we're at now.<p>
 *
 * @see GrammaticalStructure
 * @see GrammaticalRelation
 * @see EnglishGrammaticalStructure

 * @author Bill MacCartney
 * @author Galen Andrew (refactoring English-specific stuff)
 */

public class EnglishGrammaticalRelations {

  public static List<GrammaticalRelation> values() {
    return Collections.unmodifiableList(Arrays.asList(values));
  }

  /**
   * The "predicate" grammatical relation.  The predicate of a
   * clause is the main VP of that clause; the predicate of a
   * subject is the predicate of the clause to which the subject
   * belongs.<p>
   * <p/>
   * Example: "Reagan died" &rarr; <code>pred</code>(Reagan, died)
   */
  public static final GrammaticalRelation PREDICATE = new GrammaticalRelation("pred", "predicate", DEPENDENT, "S", new String[]{"S < VP=target"});

  /**
   * The auxiliary grammatical relation.  An auxiliary of a clause is a
   * non-main verb of the clause.<p>
   * <p/>
   * Example: "Reagan has died" &rarr; <code>aux</code>(died, has)
   */
  public static final GrammaticalRelation AUX_MODIFIER = new GrammaticalRelation("aux", "auxiliary", DEPENDENT, "S", new String[]{"S << (VP < VP|ADJP < /^(?:TO|MD|VB.*)$/=target)"});

  /**
   * The "argument" grammatical relation.  An argument of a VP is a
   * subject or complement of that VP; an argument of a clause is
   * an argument of the VP which is the predicate of that
   * clause.<p>
   * <p/>
   * Example: "Clinton defeated Dole" &rarr;
   * <code>arg</code>(defeated, Clinton),
   * <code>arg</code>(defeated, Dole)
   */
  public static final GrammaticalRelation ARGUMENT = new GrammaticalRelation("arg", "argument", DEPENDENT, null, new String[0]);

  /**
   * The "subject" grammatical relation.  The subject of a VP is
   * the noun or clause that performs or experiences the VP; the
   * subject of a clause is the subject of the VP which is the
   * predicate of that clause.<p>
   * <p/>
   * Examples:
   * "Clinton defeated Dole" &rarr; <code>subj</code>(defeated, Clinton),
   * "What she said is untrue" &rarr; <code>subj</code>(is, What she said)
   */
  public static final GrammaticalRelation SUBJECT = new GrammaticalRelation("subj", "subject", ARGUMENT, null, new String[0]);

  /**
   * The "nominal subject" grammatical relation.  A nominal subject is
   * a subject which is an noun phrase.<p>
   * <p/>
   * Example: "Clinton defeated Dole" &rarr;
   * <code>nsubj</code>(defeated, Clinton),
   */
  public static final GrammaticalRelation NOMINAL_SUBJECT = new GrammaticalRelation("nsubj", "nominal subject", SUBJECT, "S", new String[]{"S < (/^NP$/=target $++ VP)"});

  /**
   * The "nominal passive subject" grammatical relation.  A nominal passive
   * subject is a subject of a passive which is an noun phrase.<p>
   * <p/>
   * Example: "Dole was defeated by Clinton" &rarr;
   * <code>nsubjpass</code>(defeated, Dole)
   */
  public static final GrammaticalRelation NOMINAL_PASSIVE_SUBJECT = new GrammaticalRelation("nsubjpass", "nominal passive subject", NOMINAL_SUBJECT, "S", new String[]{"S < /^NP$/=target < (VP|SQ < (VP < VBN|VBD) < (/^(VB|AUX)/ < /be|was|is|are|were|been|being|'s|'re|'m|am|Been|Being|WAS|IS|get|got|getting|gets|Get|gotten|become|became|felt|feels|feel|seems|seem|seemed|remains|remained|remain/))"});

  /**
   * The "clausal subject" grammatical relation.  A clausal subject is
   * a subject which is a clause.<p>
   * <p/>
   * Example:
   * "What she said is untrue" &rarr; <code>csubj</code>(is, What she said)
   */
  public static final GrammaticalRelation CLAUSAL_SUBJECT = new GrammaticalRelation("csubj", "clausal subject", SUBJECT, "S", new String[]{"S < (SBAR=target $++ (VP !$-- NP))"});

  /**
   * The "complement" grammatical relation.  A complement of a VP
   * is any object (direct or indirect) of that VP, or a clause or
   * adjectival phrase which functions like an object; a complement
   * of a clause is an complement of the VP which is the predicate
   * of that clause.<p>
   * <p/>
   * Examples:
   * "She gave me a raise" &rarr;
   * <code>comp</code>(gave, me),
   * <code>comp</code>(gave, a raise);
   * "I like to swim" &rarr;
   * <code>comp</code>(like, to swim)
   */
  public static final GrammaticalRelation COMPLEMENT = new GrammaticalRelation("comp", "complement", ARGUMENT, null, new String[0]);

  /**
   * The "object" grammatical relation.  An object of a VP
   * is any direct object or indirect object of that VP; an object
   * of a clause is an object of the VP which is the predicate
   * of that clause.<p>
   * <p/>
   * Examples:
   * "She gave me a raise" &rarr;
   * <code>obj</code>(gave, me),
   * <code>obj</code>(gave, raise)
   */
  public static final GrammaticalRelation OBJECT = new GrammaticalRelation("obj", "object", COMPLEMENT, null, new String[0]);

  /**
   * The "direct object" grammatical relation.  The direct object
   * of a VP is the noun phrase which is the (accusative) object of
   * the verb; the direct object of a clause is the direct object of the VP
   * which is the predicate of that clause.<p>
   * <p/>
   * Example:
   * "She gave me a raise" &rarr;
   * <code>dobj</code>(gave, raise)
   */
  public static final GrammaticalRelation DIRECT_OBJECT = new GrammaticalRelation("dobj", "direct object", OBJECT, "VP", new String[]{"VP <+(VP) (/^NP$/ . /^NP$/=target)", "VP <+(VP) (NP < (/^NP$/ . /^NP$/=target))", "VP <+(VP) /^NP$/=target"});

  /**
   * The "indirect object" grammatical relation.  The indirect
   * object of a VP is the noun phrase which is the (dative) object
   * of the verb; the indirect object of a clause is the indirect
   * object of the VP which is the predicate of that clause.<p>
   * <p/>
   * Example:
   * "She gave me a raise" &rarr;
   * <code>iobj</code>(gave, me)
   */
  public static final GrammaticalRelation INDIRECT_OBJECT = new GrammaticalRelation("iobj", "indirect object", OBJECT, "S", new String[]{"S < (VP <+(VP) (/^NP$/=target . /^NP$/))", "S < (VP <+(VP) (NP=target < (/^NP$/ . /^NP$/)))"});

  /**
   * The "clausal complement" grammatical relation.  A clausal
   * complement of a VP is a clause which functions like an object
   * of the verb; a clausal complement of a clause is the clausal
   * complement of the VP which is the predicate of that clause.<p>
   * <p/>
   * Example:
   * "I like to swim" &rarr;
   * <code>ccomp</code>(like, swim)
   */
  public static final GrammaticalRelation CLAUSAL_COMPLEMENT = new GrammaticalRelation("ccomp", "clausal complement", COMPLEMENT, "S", new String[]{"S <+(VP) (VP < (S=target < VP !$-- NP))", // allow SBAR here too.  but leads to some false positives
                                                                                                                                                    // only if no IN, or else done in collapsing
                                                                                                                                                    "S <+(VP) (VP < (SBAR=target !<# IN < (S < VP)))"});

  /**
   * The "adjectival complement" grammatical relation.  An
   * adjectival complement of a VP is a adjectival phrase which
   * functions like an object of the verb; an adjectival complement
   * of a clause is the adjectival complement of the VP which is
   * the predicate of that clause.<p>
   * <p/>
   * Example:
   * "She looks very beautiful" &rarr;
   * <code>acomp</code>(looks, very beautiful)
   */
  public static final GrammaticalRelation ADJECTIVAL_COMPLEMENT = new GrammaticalRelation("acomp", "adjectival complement", COMPLEMENT, "S", new String[]{"S <+(VP) (VP < (ADJP=target !$-- NP))"});

  /**
   * The "modifier" grammatical relation.  A modifier of a VP is
   * any constituent that serves to modify the meaning of the VP
   * (but is not an {@link ARGUMENT <code>ARGUMENT</code>} of that
   * VP); a modifier of a clause is an modifier of the VP which is
   * the predicate of that clause.<p>
   * <p/>
   * Examples:
   * "I swam in the pool last night" &rarr;
   * <code>mod</code>(swam, in the pool),
   * <code>mod</code>(swam, last night)
   */
  public static final GrammaticalRelation MODIFIER = new GrammaticalRelation("mod", "modifier", DEPENDENT, null, new String[0]);

  /**
   * The "temporal modifier" grammatical relation.  A temporal
   * modifier of a VP is any constituent that serves to modify the
   * meaning of the VP by specifying a time; a temporal modifier of a
   * clause is an temporal modifier of the VP which is the
   * predicate of that clause.<p>
   * <p/>
   * Examples:
   * "I swam in the pool last night" &rarr;
   * <code>tmod</code>(swam, last night)
   */
  public static final GrammaticalRelation TEMPORAL_MODIFIER = new GrammaticalRelation("tmod", "temporal modifier", MODIFIER, "S", new String[]{"S <+(VP) /^NP-TMP$/=target", "S < (/^NP-TMP$/=target $++ (NP $++ VP))"});

  /**
   * The "adjectival modifier" grammatical relation.  An adjectival
   * modifier of an NP is any adjectival phrase that serves to modify
   * the meaning of the NP.<p>
   * <p/>
   * Examples:
   * "Sam eats red meat" &rarr;
   * <code>amod</code>(meat, red)
   */
  public static final GrammaticalRelation ADJECTIVAL_MODIFIER = 
    new GrammaticalRelation("amod", 
                            "adjectival modifier", 
                            MODIFIER, 
                            "^NP(?:-TMP)?$", 
                            new String[]{"/^NP(?:-TMP)?$/ < ADJP|JJ|JJR|JJS|VBN|VBG|VBD=target"});

  /**
   * The "numeric modifier" grammatical relation.  A numeric
   * modifier of an NP is any number phrase that serves to modify
   * the meaning of the NP.<p>
   * <p/>
   * Examples:
   * "Sam eats 3 sheep" &rarr;
   * <code>num</code>(sheep, 3)
   */
  public static final GrammaticalRelation NUMERIC_MODIFIER = new GrammaticalRelation("num", "numeric modifier", MODIFIER, "^NP(-TMP)?$", new String[]{"/^NP(-TMP)?$/ < CD|QP=target"});

  /**
   * The "noun compound modifier" grammatical relation.  A noun compound
   * modifier of an NP is any noun that serves to modify the head noun.
   * Note that this has all nouns modify the rightmost a la Penn headship
   * rules.  There is no intelligent noun compound analysis. <p>
   * <p/>
   * Example:
   * "Oil price futures"
   * <code>nn</code>(futures, oil)
   * <code>nn</code>(futures, price)
   */
  public static final GrammaticalRelation NOUN_COMPOUND_MODIFIER = new GrammaticalRelation("nn", "nn modifier", MODIFIER, "^NP(-TMP)?$", new String[]{"/^NP(-TMP)?$/ < (NN|NNS|NNP|NNPS=target .. NN|NNS|NNP|NNPS)"});

  /**
   * The "appositional modifier" grammatical relation.  An appositional
   * modifier of an NP is an NP that serves to modify
   * the meaning of the NP.<p>
   * <p/>
   * Examples:
   * "Sam, my brother, eats red meat" &rarr;
   * <code>appos</code>(Sam, brother)
   */
  public static final GrammaticalRelation APPOSITIONAL_MODIFIER = new GrammaticalRelation("appos", "appositional modifier", MODIFIER, "^NP(-TMP)?$", new String[]{"/^NP(?:-TMP)?$/ < (NP=target $- /^,$/ $-- NP !$ CC)"});

  /**
   * The "abbreviation appositional modifier" grammatical relation.
   * An abbreviation modifier of an NP is an NP that serves to abbreviate
   * the NP.<p>
   * <p/>
   * Examples:
   * "The Australian Broadcasting Corporation (ABC)" &rarr;
   * <code>abbrev</code>(Corporation, ABC)
   */
  public static final GrammaticalRelation ABBREVIATION_MODIFIER = new GrammaticalRelation("abbrev", "abbreviation modifier", APPOSITIONAL_MODIFIER, "^NP(?:-TMP)?$", new String[]{"/^NP(?:-TMP)?$/ < (PRN < (NP=target < NNP $- /^-LRB-$/ $+ /^-RRB-$/))"});

  /**
   * The "participial modifier" grammatical relation.  A participial
   * modifier of an NP is a VP that serves to modify
   * the meaning of the NP.<p>
   * <p/>
   * Examples:
   * "truffles picked during the spring are tasty" &rarr;
   * <code>partmod</code>(truffles, picked)
   */
  public static final GrammaticalRelation PARTICIPIAL_MODIFIER = new GrammaticalRelation("partmod", "participial modifier", MODIFIER, "^NP(?:-TMP)?$", new String[]{"/^NP(?:-TMP)?$/ < (VP=target $-- NP)"});

  /**
   * The "infinitval modifier" grammatical relation.  A participial
   * modifier of an NP is an S/VP that serves to modify
   * the meaning of the NP.<p>
   * <p/>
   * Examples:
   * "points to establish are ..." &rarr;
   * <code>infmod</code>(points, establish)
   */
  public static final GrammaticalRelation INFINITIVAL_MODIFIER = new GrammaticalRelation("infmod", "infinitival modifier", MODIFIER, "^NP(-TMP)?$", new String[]{"/^NP(?:-[A-Z]+)?$/ < (S=target < (VP < TO) $-- /^NP|NNP?S?$/)"});

  /**
   * The "adverbial modifier" grammatical relation.  An adverbial
   * modifier of a word is an RB or ADVP that serves to modify
   * the meaning of the word.<p>
   * <p/>
   * Examples:
   * "genetically modified food" &rarr;
   * <code>advmod</code>(modified, genetically)
   */
  public static final GrammaticalRelation ADVERBIAL_MODIFIER = new GrammaticalRelation("advmod", "adverbial modifier", MODIFIER, "VP|ADJP|S|SBAR|SINV|SQ", new String[]{"VP|ADJP|S|SBAR|SINV|SQ < ADVP=target", "VP|ADJP < RB=target"});

  /**
   * The "determiner" grammatical relation.  We're treating these as
   * a sort of degenerate NP modifier, for now.
   */
  public static final GrammaticalRelation DETERMINER = new GrammaticalRelation("det", "determiner", MODIFIER, "^NP(-TMP)?$", new String[]{"/^NP(-TMP)?$/ < DT|PDT=target"});

  /**
   * The "prepositional modifier" grammatical relation.  A prepositional
   * modifier of an NP is any prepositional phrase that serves to modify
   * the meaning of the NP.<p>
   * <p/>
   * Examples:
   * "I saw a cat in a hat" &rarr;
   * <code>prep</code>(cat, in)
   */
  public static final GrammaticalRelation PREPOSITIONAL_MODIFIER = 
    new GrammaticalRelation("prep", 
                            "prepositional modifier", 
                            MODIFIER, 
                            "^NP(?:-TMP)?$", 
                            new String[]{"/^NP(?:-TMP)?$/ < PP=target"});

  /**
   * The "semantic dependent" grammatical relation has been
   * introduced as a supertype for the controlling subject relation.
   */
  public static final GrammaticalRelation SEMANTIC_DEPENDENT = new GrammaticalRelation("sdep", "semantic dependent", null, null, new String[0]);

  /**
   * The "controlling subject" grammatical relation.<p>
   * <p/>
   * Examples:
   * "Tom likes to eat fish" &rarr;
   * <code>xsubj</code>(eat, Tom)
   */
  public static final GrammaticalRelation CONTROLLING_SUBJECT = new GrammaticalRelation("xsubj", "controlling subject", SEMANTIC_DEPENDENT, null, new String[0]);

  private static final GrammaticalRelation[] values
          = new GrammaticalRelation[] { GOVERNOR, DEPENDENT, PREDICATE, AUX_MODIFIER, ARGUMENT, SUBJECT, NOMINAL_SUBJECT, NOMINAL_PASSIVE_SUBJECT, CLAUSAL_SUBJECT, COMPLEMENT, OBJECT, DIRECT_OBJECT, INDIRECT_OBJECT, CLAUSAL_COMPLEMENT, ADJECTIVAL_COMPLEMENT, MODIFIER, TEMPORAL_MODIFIER, ADJECTIVAL_MODIFIER, NUMERIC_MODIFIER, NOUN_COMPOUND_MODIFIER, APPOSITIONAL_MODIFIER, ABBREVIATION_MODIFIER, PARTICIPIAL_MODIFIER, INFINITIVAL_MODIFIER, ADVERBIAL_MODIFIER, DETERMINER, PREPOSITIONAL_MODIFIER, SEMANTIC_DEPENDENT, CONTROLLING_SUBJECT };
  
  public static void main(String[] args) {
    System.out.println(DEPENDENT.toPrettyString());
  }

}
