package edu.stanford.nlp.trees;

import edu.stanford.nlp.util.Filter;

import java.io.*;
import java.util.*;

import static edu.stanford.nlp.trees.EnglishGrammaticalRelations.*;
import static edu.stanford.nlp.trees.GrammaticalRelation.*;

/**
 * A GrammaticalStructure for English.
 *
 * <b>Caveat emptor!</b> This is a work in progress.
 * Nothing in here should be relied upon to function perfectly.
 * Feedback welcome.
 * @author Bill MacCartney
 */
public class EnglishGrammaticalStructure extends GrammaticalStructure {

  private static final HeadFinder shf = new SemanticHeadFinder();
  private static final Filter puncFilter = new PennTreebankLanguagePack().punctuationWordRejectFilter();
  protected TypedDependencyList moreTypedDependencies = null;

  /**
   * Construct a new <code>GrammaticalStructure</code> from an
   * existing parse tree.  The new <code>GrammaticalStructure</code>
   * has the same tree structure and label values as the given tree
   * (but no shared storage).  As part of construction, the parse tree
   * is analyzed using definitions from {@link GrammaticalRelation
   * <code>GrammaticalRelation</code>} to populate the new
   * <code>GrammaticalStructure</code> with as many labeled
   * grammatical relations as it can.
   */
  public EnglishGrammaticalStructure(Tree t) {
    super(t, EnglishGrammaticalRelations.values(), shf, puncFilter);
    fixupForConjunctions(typedDependencies);
    moreTypedDependencies = addSubjectsOfInfinitivalComplements(root);
    fixupForConjunctions(moreTypedDependencies);
  }

  /**
   * Get "extra" dependencies.  Currently, the only "extra"
   *  implemented is the controlling subject relation.  For example,
   *  in the sentence "Tom likes to eat fish", requesting extras will
   *  cause "xsubj(eat, Tom)" to be added to the typed dependency
   *  list.
   *
   */
  protected TypedDependencyList moreTypedDependencies() {
    return moreTypedDependencies;
  }

  private void fixupForConjunctions(TypedDependencyList list) {
    for (TypedDependency td : list) {
      if (td.reln() == DEPENDENT && (td.dep().value().equals("and") || td.dep().value().equals("or"))) {
        td.setReln("cc");
        TreeGraphNode gov = td.gov();
        for (TypedDependency td1 : list) {
          if (td1.reln() == DEPENDENT && td1.gov() == gov) {
            td1.setReln("conj");
          }
        }
      }
    }
  }

  private TypedDependencyList addSubjectsOfInfinitivalComplements(TreeGraphNode t) {
    TypedDependencyList tdl = new TypedDependencyList();
    if (t.numChildren() > 0) {          // don't do leaves
      TreeGraphNode ccomp = getNodeInRelation(t, CLAUSAL_COMPLEMENT);
      if (ccomp != null && getSubject(ccomp) == null) {
        TreeGraphNode subj = getSubject(t);
        if (subj != null) {
          ccomp.addArc(CONTROLLING_SUBJECT, subj);
          TreeGraphNode ccompHead = ccomp.headWordNode();
          TreeGraphNode subjHead = subj.headWordNode();
          if (ccompHead != null && subjHead != null) {
            tdl.add(new TypedDependency(CONTROLLING_SUBJECT, ccompHead, subjHead));
          }
        }
      }
      // now recurse into children
      for (Tree kid : t.children()) {
        tdl.addAll(addSubjectsOfInfinitivalComplements((TreeGraphNode) kid));
      }
    }
    return tdl;
  }

  /**
   * Tries to return a node representing the {@link
   * GrammaticalRelation#SUBJECT <code>SUBJECT</code>} (whether
   * nominal or clausal) of the given node <code>t</code>.
   * Probably, node <code>t</code> should represent a clause or
   * verb phrase.
   *
   * @param t a node in this <code>GrammaticalStructure</code>
   * @return a node which is the subject of node
   *         <code>t</code>, or else <code>null</code>
   */
  public TreeGraphNode getSubject(TreeGraphNode t) {
    TreeGraphNode subj = getNodeInRelation(t, NOMINAL_SUBJECT);
    if (subj == null) {
      return getNodeInRelation(t, CLAUSAL_SUBJECT);
    } else {
      return subj;
    }
  }

  /** Destructively modifies this <code>TypedDepedencyList</code> by
   *  collapsing two types of transitive pairs of dependencies:
   *  <dl>
   *  <dt>prepositional dependencies</dt>
   *  <dd>
   *    <code>dep(cat, in)</code> and
   *    <code>dep(in, hat)</code> are collapsed to
   *    <code>in(cat, hat)</code>
   *  </dd>
   *  <dt>possessive dependencies</dt>
   *  <dd>
   *    <code>dep(revenge, 's)</code> and
   *    <code>dep('s, Montezuma)</code> are collapsed to
   *    <code>poss(revenge, Montezuma)</code>
   *  </dd>
   *  </dl>
   */
  protected void collapseDependencies(TypedDependencyList list) {
//    System.out.println("Before conj:");
//    System.out.println(list);
    collapseConj(list);
//    System.out.println("Before PrepAndPoss:");
//    System.out.println(list);
    collapsePrepAndPoss(list);
//    System.out.println("Before multiword:");
//    System.out.println(list);
    collapseMultiwordPreps(list);
//    System.out.println("After all collapse:");
//    System.out.println(list);
  }

  private void collapseConj(TypedDependencyList list) {
    // find typed deps of form cc(gov, dep)
    for (TypedDependency td : list) {
      if (td.reln().equals("cc")) {
        TreeGraphNode gov = td.gov();
        Object conj = td.dep().value();
        // find other deps of that gov having reln "conj"
        for (TypedDependency td1 : list) {
          if (td1.reln().equals("conj") && td1.gov() == gov) {
            // and change "conj" to the actual (lexical) conjunction
            td1.setReln(conj);
          }
        }
      }
    }
    // now remove typed dependencies with reln "cc"
    TypedDependencyList newTypedDeps = new TypedDependencyList();
    for (TypedDependency td: list) {
      if (!td.reln().equals("cc")) {
        newTypedDeps.add(td);
      }
    }
    list.clear();
    list.addAll(newTypedDeps);
  }

  private void collapsePrepAndPoss(TypedDependencyList list) {

    // Man oh man, how gnarly is the logic of this method.

    TypedDependencyList newTypedDeps = new TypedDependencyList();

    // Construct a map from tree nodes to the set of typed
    // dependencies in which the node appears as governor.
    Map<TreeGraphNode, Set<TypedDependency>> map =
            new HashMap<TreeGraphNode, Set<TypedDependency>>();
    for (TypedDependency typedDep: list) {
      if (!map.containsKey(typedDep.gov())) {
        map.put(typedDep.gov(), new HashSet<TypedDependency>());
      }
      map.get(typedDep.gov()).add(typedDep);
    }
    //System.err.println("here's the map: " + map);

    for (TypedDependency td1: list) {
      if (!td1.reln().equals("kill")) {
        TreeGraphNode td1Dep = td1.dep();
        String td1DepPOS = td1Dep.parent().value();
        //System.err.println("td1 & td1DepPOS: " + td1 + "&" + td1DepPOS);
        // find all other typedDeps having our dep as gov
        Set<TypedDependency> possibles = map.get(td1Dep);
        if (possibles != null) {

          //          System.out.println("Possibles: " + possibles);
          // look for the "second half"
          for (TypedDependency td2 : possibles) {
            TreeGraphNode td2Dep = td2.dep();
            String td2DepPOS = td2Dep.parent().value();
            if (td2 != null
                && (td1.reln() == DEPENDENT 
                    || td1.reln() == PREPOSITIONAL_MODIFIER)
                && td2.reln() == DEPENDENT
                && (td1DepPOS.equals("IN")
                    || td1DepPOS.equals("TO")
                    || td1DepPOS.equals("POS"))
                && (!td2DepPOS.equals("RB"))) {
              // OK, we have a pair td1, td2 to collapse to td3
              TypedDependency td3 =
                      new TypedDependency((td1DepPOS.equals("POS") ?
                      "poss" : // for possessives, use "poss"
                      td1Dep.value()), // for prepositions, use the preposition
                              td1.gov(),
                              td2.dep());
              //System.err.println("adding: " + td3);
              newTypedDeps.add(td3);
              td1.setReln("kill");        // remember these are "used up"
              td2.setReln("kill");        // remember these are "used up"
            }
          }

          // Now we need to see if there any TDs that will be "orphaned"
          // by this collapse.  Example: if we have:
          //   dep(drew, on)
          //   dep(on, book)
          //   dep(on, right)
          // the first two will be collapsed to on(drew, book), but then
          // the third one will be orphaned, since its governor no
          // longer appears.  So, change its governor to 'drew'.
          if (td1.reln().equals("kill")) {
            for (TypedDependency td2 : possibles) {
              if (!td2.reln().equals("kill")) {
                //System.err.println("td1 & td2: " + td1 + " & " + td2);
                td2.setGov(td1.gov());
              }
            }
          }
        }
      }
    }

    // now copy remaining unkilled TDs from here to new
    for (TypedDependency td: list) {
      if (!td.reln().equals("kill")) {
        newTypedDeps.add(td);
      }
    }

    list.clear();                            // forget all (esp. killed) TDs
    list.addAll(newTypedDeps);
  } // end collapsePrepAndPoss()

  // used by collapseMultiwordPreps()
  private static String[][] MULTIWORD_PREPS = {{"according", "to"},
                                               {"based", "on"},
                                               {"depending", "on"},
                                               {"compared", "with"},
                                               {"compared", "to"},
                                               {"followed", "by"},
                                               {"along", "with"}};

  private void collapseMultiwordPreps(TypedDependencyList list) {
    TypedDependencyList newTypedDeps = new TypedDependencyList();

    for (String[] mwp : MULTIWORD_PREPS) {

      // find typed deps of form dep(gov, mwp[0])
      for (TypedDependency td1 : list) {
        if (//td1.reln() == DEPENDENT &&
                td1.dep().value().equals(mwp[0])) {
          // find other deps of that dep having reln mwp[1]
          for (TypedDependency td2 : list) {
            if (td2.gov() == td1.dep() && td2.reln().equals(mwp[1])) {
              td1.setReln("kill");
              td2.setReln("kill");
              newTypedDeps.add(new TypedDependency(mwp[0] + "_" + mwp[1],
                      td1.gov(),
                      td2.dep()));
            }
          }
        }
      }
    }

    // now remove typed dependencies with reln "kill"
    for (TypedDependency td: list) {
      if (!td.reln().equals("kill")) {
        newTypedDeps.add(td);
      }
    }
    list.clear();
    list.addAll(newTypedDeps);
  }

  /**
   * Just for testing.
   * Usage: java GrammaticalStructure [treeFile]
   */
  public static void main(String[] args) {

    // System.out.print("GrammaticalRelations under DEPENDENT:");
    // System.out.println(DEPENDENT.toPrettyString());

    Treebank tb = new MemoryTreebank();

    if (args.length == 0) {
      try {
        TreeReader tr = new PennTreeReader(new StringReader("(S (NP (NNP Sam)) (VP (VBD died) (NP-TMP (NN today))))"), new LabeledScoredTreeFactory());
        tb.add(tr.readTree());
      } catch (Exception e) {
        System.err.println("Horrible error: " + e);
        e.printStackTrace();
      }
    } else {
      try {
        TreeReader tr = new PennTreeReader(new BufferedReader(new InputStreamReader(new FileInputStream(args[0]))), new LabeledScoredTreeFactory());
        Tree t;
        while ((t = tr.readTree()) != null) {
          tb.add(t);
        }
      } catch (IOException e) {
        throw new RuntimeException("File problem: " + e);
      }
    }

    for (Iterator it = tb.iterator(); it.hasNext();) {

      System.out.println("==================================================");
      Tree t = (Tree) it.next();
      GrammaticalStructure gs = new EnglishGrammaticalStructure(t);

      t.pennPrint();

      //System.out.println("----------------------------");
      //TreeGraph tg = new TreeGraph(t);
      //System.out.println(tg);

      //System.out.println("----------------------------");
      //System.out.println(gs);

      System.out.println("----------------------------");
      gs.typedDependencies().print();

      System.out.println("----------------------------");
      gs.typedDependenciesCollapsed(true).print();
      
      //gs.printTypedDependencies("xml");
    }
  }

}
