package edu.stanford.nlp.trees;

import edu.stanford.nlp.trees.tregex.TregexMatcher;
import edu.stanford.nlp.trees.tregex.TregexPattern;

import java.util.*;
import java.util.regex.Pattern;


/**
 * <code>GrammaticalRelation</code> is used to define a
 * standardized, hierarchical set of grammatical relations,
 * together with patterns for identifying them in
 * parse trees.<p>
 *
 * Each <code>GrammaticalRelation</code> has:
 * <ul>
 *   <li>a <code>String</code> short name, which should be a lowercase
 *       abbreviation of some kind,</li>
 *   <li>a <code>String</code> long name, which should be descriptive,</li>
 *   <li>a parent in the <code>GrammaticalRelation</code> hierarchy,</li>
 *   <li>a {@link java.util.regex.Pattern <code>Pattern</code>} called
 *   <code>sourcePattern</code> which matches nodes from which
 *   this <code>GrammaticalRelation</code> could hold, and</li>
 *   <li>a list of zero or more {@link TregexPattern
 *   <code>TregexPattern</code>s} called <code>targetPatterns</code>,
 *   which describe the local tree structure which must hold between
 *   the source node and a target node for the
 *   <code>GrammaticalRelation</code> to apply.</li>
 * </ul>
 *
 * The <code>targetPatterns</code> associated
 * with a <code>GrammaticalRelation</code> are designed as follows.
 * In order to recognize a grammatical relation X holding between
 * nodes A and B in a parse tree, we want to associate with
 * <code>GrammaticalRelation</code> X a {@link TregexPattern
 * <code>TregexPattern</code>} such that:
 * <ul>
 *   <li>the root of the pattern matches A, and</li>
 *   <li>the pattern includes a special node label, "target", which matches B.</li>
 * </ul>
 * For example, for the grammatical relation <code>PREDICATE</code>
 * which holds between a clause and its primary verb phrase, we might
 * want to use the pattern <code>"S < VP=target"</code>, in which the
 * root will match a clause and the node labeled <code>"target"</code>
 * will match the verb phrase.<p>
 *
 * For a given grammatical relation, the method {@link
 * GrammaticalRelation#getRelatedNodes <code>getRelatedNodes()</code>}
 * takes a <code>Tree</code> node as an argument and attempts to
 * return other nodes which have this grammatical relation to the
 * argument node.  By default, this method operates as follows: it
 * steps through the patterns in the pattern list, trying to match
 * each pattern against the argument node, until it finds some
 * matches.  If a pattern matches, all matching nodes (that is, each
 * node which corresponds to node label "target" in some match) are
 * returned as a list; otherwise the next pattern is tried.<p>
 * 
 * For some grammatical relations, we need more sophisticated logic to
 * identify related nodes.  In such cases, {@link
 * GrammaticalRelation#getRelatedNodes <code>getRelatedNodes()</code>}
 * can be overridden on a per-relation basis using anonymous subclassing.<p>
 *
 * @see GrammaticalStructure
 * @see EnglishGrammaticalRelations
 * @see EnglishGrammaticalStructure
 *
 * @author Bill MacCartney
 * @author Galen Andrew (refactoring English-specific stuff)
 */
public class GrammaticalRelation {

  /**
   * The "governor" grammatical relation, which is the inverse of "dependent".<p>
   * <p/>
   * Example: "the red car" &rarr; <code>gov</code>(red, car)
   */
  public static final GrammaticalRelation GOVERNOR = new GrammaticalRelation("gov", "governor", null, null, new String[0]);

  /**
   * The "dependent" grammatical relation, which is the inverse of "governor".<p>
   * <p/>
   * Example: "the red car" &rarr; <code>dep</code>(car, red)
   */
  public static final GrammaticalRelation DEPENDENT = new GrammaticalRelation("dep", "dependent", null, null, new String[0]);

  private final String shortName;
  private final String longName;
  private GrammaticalRelation parent;
  private List<GrammaticalRelation> children = new ArrayList<GrammaticalRelation>();
  // a regexp for node values at which this relation can hold
  private Pattern sourcePattern = null;
  private List<TregexPattern> targetPatterns = new ArrayList<TregexPattern>();

  public GrammaticalRelation(String shortName,
                      String longName,
                      GrammaticalRelation parent,
                      String sourcePattern,
                      String[] targetPatterns) {
    this.shortName = shortName;
    this.longName = longName;
    this.parent = parent;
    if (parent != null) {
      parent.addChild(this);
    }
    if (sourcePattern != null) {
      try {
        this.sourcePattern = Pattern.compile(sourcePattern);
      } catch (java.util.regex.PatternSyntaxException e) {
        throw new RuntimeException("Bad pattern: " + sourcePattern);
      }
    }
    for (String pattern : targetPatterns) {
      TregexPattern p = null;
      try {
        p = TregexPattern.compile(pattern);
      } catch (edu.stanford.nlp.trees.tregex.ParseException pe) {
        throw new RuntimeException("Bad pattern: " + pattern);
      }
      this.targetPatterns.add(p);
    }
  }

  private void addChild(GrammaticalRelation child) {
    children.add(child);
  }

  /** Given a <code>Tree</code> node <code>t</code>, attempts to
   *  return a list of nodes to which node <code>t</code> has this
   *  grammatical relation.
   */
  public Collection<Tree> getRelatedNodes(Tree t) {
    Set<Tree> nodeList = new LinkedHashSet<Tree>();
    Iterator<TregexPattern> it = targetPatterns.iterator();
    while (it.hasNext()) {   // cdm: I deleted: && nodeList.isEmpty()
      TregexPattern p = it.next();
      TregexMatcher m = p.matcher(t);
      while (m.find()) {
        if (m.getMatch() == t) {
          nodeList.add(m.getNode("target"));
          //System.err.println("found " + this + "(" + t + ", " + m.getNode("target") + ")");
        }
      }
    }
    return nodeList;
  }
    
  /** Returns <code>true</code> iff the value of <code>Tree</code>
   *  node <code>t</code> matches the <code>sourcePattern</code> for
   *  this <code>GrammaticalRelation</code>, indicating that this
   *  <code>GrammaticalRelation</code> is one that could hold between
   *  <code>Tree</code> node <code>t</code> and some other node.
   */
  public boolean isApplicable(Tree t) {
    return (sourcePattern != null) &&
      sourcePattern.matcher(t.value()).matches();
  }
    
  public boolean isAncestor(GrammaticalRelation gr) {
    while (gr != null) {
      if (this == gr) { return true; }
      gr = gr.parent;
    }
    return false;
  }

  /**
   * Returns short name (abbreviation) for this
   * <code>GrammaticalRelation</code>.
   */
  public String toString() {
    return shortName;
  }

  /**
   * Returns a <code>String</code> representation of this
   * <code>GrammaticalRelation</code> and the hierarchy below
   * it, with one node per line, indented according to level.
   *
   * @return <code>String</code> representation of this
   *         <code>GrammaticalRelation</code>
   */
  public String toPrettyString() {
    return toPrettyString(0);
  }

  /**
   * Returns a <code>String</code> representation of this
   * <code>GrammaticalRelation</code> and the hierarchy below
   * it, with one node per line, indented according to
   * <code>indentLevel</code>.
   *
   * @param indentLevel how many levels to indent (0 for root node)
   * @return <code>String</code> representation of this
   *         <code>GrammaticalRelation</code>
   */
  private String toPrettyString(int indentLevel) {
    StringBuffer buf = new StringBuffer("\n");
    for (int i = 0; i < indentLevel; i++) {
      buf.append("  ");
    }
    buf.append(shortName + ": " + targetPatterns);
    for (GrammaticalRelation child : children) {
      buf.append(" " + child.toPrettyString(indentLevel + 1));
    }
    return buf.toString();
  }

}
