package edu.stanford.nlp.trees;

import edu.stanford.nlp.ling.Label;
import edu.stanford.nlp.util.Filter;

import java.io.Serializable;
import java.util.*;

import static edu.stanford.nlp.trees.GrammaticalRelation.*;

/**
 * A <code>GrammaticalStructure</code> is a {@link TreeGraph
 * <code>TreeGraph</code>} (that is, a tree with additional labeled
 * arcs between nodes) for representing the grammatical relations in a
 * parse tree.  A new <code>GrammaticalStructure</code> is constructed
 * from an existing parse tree with the help of {@link
 * GrammaticalRelation <code>GrammaticalRelation</code>}, which
 * defines a hierarchy of grammatical relations, along with
 * patterns for identifying them in parse trees.  The constructor for
 * <code>GrammaticalStructure</code> uses these definitions to
 * populate the new <code>GrammaticalStructure</code> with as many
 * labeled grammatical relations as it can.  Once constructed, the new
 * <code>GrammaticalStructure</code> can be printed in various
 * formats, or interrogated using the interface methods in this
 * class.<p>
 *
 * @author Bill MacCartney
 * @author Galen Andrew (refactoring English-specific stuff)
 * @see EnglishGrammaticalRelations
 * @see GrammaticalRelation
 * @see EnglishGrammaticalStructure
 *      <p/>
 *      <b>Caveat emptor!</b> This is a work in progress.
 *      Nothing in here should be relied upon to function perfectly.
 *      Feedback welcome.
 */

public abstract class GrammaticalStructure extends TreeGraph implements Serializable {
  protected Set<Dependency> dependencies = null;
  protected TypedDependencyList typedDependencies = null;

  /**
   * Create a new GrammaticalStructure, analyzing the parse tree and
   * populate the GrammaticalStructure with as many labeled
   * grammatical relation arcs as possible.
   *
   * @param relations  a set of GrammaticalRelations to consider
   * @param t          a Tree to analyze
   * @param hf         a HeadFinder for analysis
   * @param puncFilter a Filter to reject punctuation
   */
  public GrammaticalStructure(Tree t, Collection<GrammaticalRelation> relations, HeadFinder hf, Filter puncFilter) {
    super(t);

    // add head word and tag to phrase nodes
    root.percolateHeads(hf);

    // add dependencies, using heads
    dependencies = root.dependencies(new NoPunctFilter(puncFilter));
    for (Dependency p : dependencies) {
      TreeGraphNode gov = (TreeGraphNode) p.governor();
      TreeGraphNode dep = (TreeGraphNode) p.dependent();
      dep.addArc(GOVERNOR, gov);
    }

    // analyze the root (and its descendants, recursively)
    analyzeNode(root, relations);

    // add typed dependencies
    typedDependencies = new TypedDependencyList(this);
  }

  public String toString() {
    StringBuffer sb = new StringBuffer(super.toString());
    //    sb.append("Dependencies:");
    //    sb.append("\n" + dependencies);
    //    sb.append("Typed Dependencies:");
    //    sb.append("\n" + typedDependencies);
    //    sb.append("More Typed Dependencies:");
    //    sb.append("\n" + moreTypedDependencies());
    return sb.toString();
  }

  private void analyzeNode(TreeGraphNode t, Collection<GrammaticalRelation> relations) {
    if (t.numChildren() > 0) {          // don't do leaves
      TreeGraphNode tHigh = t.highestNodeWithSameHead();
      for (GrammaticalRelation egr : relations) {
        if (egr.isApplicable(t)) {
          for (Tree u : egr.getRelatedNodes(t)) {
            tHigh.addArc(egr, (TreeGraphNode) u);
          }
        }
      }
      // now recurse into children
      for (Tree kid : t.children()) {
        analyzeNode((TreeGraphNode) kid, relations);
      }
    }
  }

  private static class NoPunctFilter implements Filter {

    private Filter npf;

    NoPunctFilter(Filter f) {
      this.npf = f;
    }

    public boolean accept(Object o) {
      if (o == null || !(o instanceof Dependency)) {
        return false;
      }
      Object s = ((Dependency) o).dependent();
      if (s == null || !(s instanceof TreeGraphNode)) {
        return false;
      }
      Object l = ((TreeGraphNode) s).label();
      if (l == null || !(l instanceof Label)) {
        return false;
      }
      return npf.accept(((Label) l).value());
    }

  } // end static class NoPunctFilter

  /**
   * Returns the set of (governor, dependent) dependencies in this
   * <code>GrammaticalStructure</code>.
   */
  public Set<Dependency> dependencies() {
    return dependencies;
  }

  /**
   * Tries to return a <code>Set</code> of leaf (terminal) nodes
   * which are the {@link GrammaticalRelation#DEPENDENT
   * <code>DEPENDENT</code>}s of the given node <code>t</code>.
   * Probably, <code>t</code> should be a leaf node as well.
   *
   * @param t a leaf node in this <code>GrammaticalStructure</code>
   * @return a <code>Set</code> of nodes which are dependents of
   *         node <code>t</code>, or else <code>null</code>
   */
  public Set getDependents(TreeGraphNode t) {
    Set deps = new HashSet();
    Set nodes = root.subTrees();
    for (Iterator it = nodes.iterator(); it.hasNext();) {
      TreeGraphNode node = (TreeGraphNode) it.next();
      TreeGraphNode gov = getNodeInRelation(node, GOVERNOR);
      if (gov != null && gov == t) {
        deps.add(node);
      }
    }
    return deps;
  }

  /**
   * Tries to return a leaf (terminal) node which is the {@link
   * GrammaticalRelation#GOVERNOR
   * <code>GOVERNOR</code>} of the given node <code>t</code>.
   * Probably, <code>t</code> should be a leaf node as well.
   *
   * @param t a leaf node in this <code>GrammaticalStructure</code>
   * @return a node which is the governor for node
   *            <code>t</code>, or else <code>null</code>
   */
  public TreeGraphNode getGovernor(TreeGraphNode t) {
    return t.followArcToNode(GOVERNOR);
  }

  public TreeGraphNode getNodeInRelation(TreeGraphNode t, GrammaticalRelation r) {
    return t.followArcToNode(r);
  }

  /**
   * Get GrammaticalRelation between gov and dep, and null if gov  is not the
   * governor of dep
   */
  public GrammaticalRelation getGrammaticalRelation(int govIndex, int depIndex) {
    TreeGraphNode gov = getNodeByIndex(govIndex);
    TreeGraphNode dep = getNodeByIndex(depIndex);
    return getGrammaticalRelation(gov, dep);
  }

  /**
   * Get GrammaticalRelation between gov and dep, and null if gov is not the
   * governor of dep
   */
  public GrammaticalRelation getGrammaticalRelation(TreeGraphNode gov, TreeGraphNode dep) {
    GrammaticalRelation reln = DEPENDENT;
    TreeGraphNode govH = gov.highestNodeWithSameHead();
    TreeGraphNode depH = dep.highestNodeWithSameHead();

    //    System.out.println("govH " + govH);
    //    System.out.println("depH " + depH);

    Set arcLabels = govH.arcLabelsToNode(depH);
    //    System.out.println("arcLabels: " + arcLabels);

    for (Object arcLabel : arcLabels) {

      if (arcLabel != null && arcLabel instanceof GrammaticalRelation) {
        GrammaticalRelation reln2 = (GrammaticalRelation) arcLabel;
        if (reln.isAncestor(reln2)) {
          reln = reln2;
        }
      }
    }
    return reln;
  }

  /**
   * Returns the typed dependencies of this grammatical structure.
   */
  public TypedDependencyList typedDependencies() {
    return typedDependencies(false);
  }

  /**
   * Returns the typed dependencies of this grammatical structure.
   * <p/>
   * If the boolean argument is true, the list of typed dependencies
   * returned may include "extras".
   */
  public TypedDependencyList typedDependencies(boolean includeExtras) {
    if (includeExtras) {
      TypedDependencyList tdl = new TypedDependencyList();
      tdl.addAll(typedDependencies);
      tdl.addAll(moreTypedDependencies());
      return tdl;
    }
    return typedDependencies;
  }

  /**
   * Get the typed dependencies after collapsing them.
   *
   * @return collapsed dependencies
   */
  public TypedDependencyList typedDependenciesCollapsed() {
    return typedDependenciesCollapsed(false);
  }

  /**
   * Get the typed dependencies after collapsing them.
   * <p/>
   * If the boolean argument is true, the list of typed dependencies
   * returned may include "extras".
   *
   * @return collapsed dependencies
   */
  public TypedDependencyList typedDependenciesCollapsed(boolean includeExtras) {
    TypedDependencyList tdl = typedDependencies(includeExtras);
    collapseDependencies(tdl);
    return tdl;
  }

  /**
   * Destructively modify the <code>TypedDependencyList</code> to collapse
   * language-dependent transitive dependencies.
   * <p/>
   * Default is no-op; to be over-ridden in subclasses.
   *
   * @param list
   */
  protected void collapseDependencies(TypedDependencyList list) {

  }

  /**
   * Get "extra" dependencies.
   * <p/>
   * Default is an empty list; to be overridden in subclasses.
   */
  protected TypedDependencyList moreTypedDependencies() {
    return new TypedDependencyList();
  }

  /**
   * Returns the dependency path as a list of labels, from node to root, it is assumed that
   * that root is an ancestor of node
   *
   * @param nodeIndex
   * @param rootIndex
   * @return a list of dependency labels
   */
  public List getDependencyPath(int nodeIndex, int rootIndex) {
    TreeGraphNode node = getNodeByIndex(nodeIndex);
    TreeGraphNode root = getNodeByIndex(rootIndex);
    return getDependencyPath(node, root);
  }

  /**
   * Returns the dependency path as a list of labels, from node to root, it is assumed that
   * that root is an ancestor of node
   *
   * @param node
   * @param root
   * @return a list of dependency labels
   */
  public List getDependencyPath(TreeGraphNode node, TreeGraphNode root) {
    List path = new ArrayList();
    while (true) {
      TreeGraphNode gov = node.followArcToNode(GOVERNOR);
      Set relations = node.arcLabelsToNode(gov);
      StringBuffer sb = new StringBuffer();
      for (Object arcLabel : relations) {
        //if (!arcLabel.equals(GOVERNOR))
        sb.append((sb.toString().length() == 0 ? "" : "+") + arcLabel.toString());
      }
      path.add(sb.toString());
      if (gov.equals(root)) {
        break;
      }
      node = gov;
    }
    return path;
  }
}
