package edu.stanford.nlp.trees;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;

/**
 * A general factory for {@link GrammaticalStructure} objects.
 *
 * Uses reflection to create objects of any subclass of {@link GrammaticalStructure}.
 *
 * @author Galen Andrew
 */
public class GrammaticalStructureFactory {
  final Constructor con;

  /**
   * Make a new GrammaticalStructureFactory that dispenses {@link GrammaticalStructure}
   * objects of the given specific type
   * @param name the fully qualified name of a subclass of GrammaticalStructure
   */
  public GrammaticalStructureFactory(String name) {
    Class c, t;
    try {
      c = Class.forName(name);
      if (!Class.forName("edu.stanford.nlp.trees.GrammaticalStructure").isAssignableFrom(c)) {
        throw new ClassNotFoundException();
      }
      t = Class.forName("edu.stanford.nlp.trees.Tree");
    } catch (ClassNotFoundException e) {
      throw new RuntimeException("Class " + name + " does not exist or does not extend edu.stanford.nlp.trees.GrammaticalStructure.");
    }

    try {
      con = c.getConstructor(t);
    } catch (NoSuchMethodException e) {
      // this is impossible, given above exceptions
      throw new RuntimeException();
    }
  }

  /**
   * Vend a new {@link GrammaticalStructure} based on the given {@link Tree}.
   *
   * @param t the tree to analyze
   * @return a GrammaticalStructure based on the tree
   */
  public GrammaticalStructure newGrammaticalStructure(Tree t) {
    try {
      return (GrammaticalStructure)con.newInstance(t);
    } catch (InstantiationException e) {
      throw new RuntimeException("Cannot instantiate " + con.getDeclaringClass().getName());
    } catch (IllegalAccessException e) {
      throw new RuntimeException(con.getDeclaringClass().getName() + "(Tree t) does not have public access");
    } catch (InvocationTargetException e) {
      throw new RuntimeException(e);
    }
  }
}
