package edu.stanford.nlp.trees;

import edu.stanford.nlp.process.Tokenizer;
import edu.stanford.nlp.objectbank.TokenizerFactory;

import java.io.*;
import java.util.ArrayList;
import java.util.List;


/**
 * A <code>PennTreeReader</code> is a <code>TreeReader</code> that
 * reads in Penn Treebank-style files.  Example usage:
 * <br>
 * <code> TreeReader tr = new PennTreeReader(new BufferedReader(new
 * InputStreamReader(new FileInputStream(file),"UTF-8")),
 * myTreeFactory); </code>
 *
 * @author Christopher Manning
 * @author Roger Levy
 * @version 2003/01
 */
public class PennTreeReader implements TreeReader {

  private Reader in;
  private Tokenizer st;
  private TreeNormalizer tn;
  private TreeFactory tf;

  private static final boolean DEBUG = false;


  /**
   * Read parse trees from a <code>Reader</code>.
   * For the defaulted arguments, you get a
   * <code>SimpleTreeFactory</code>, no <code>TreeNormalizer</code>, and
   * a <code>PennTreebankTokenizer</code>.
   *
   * @param in The <code>Reader</code>
   */
  public PennTreeReader(Reader in) {
    this(in, new SimpleTreeFactory());
  }


  /**
   * Read parse trees from a <code>Reader</code>.
   *
   * @param in the Reader
   * @param tf TreeFactory -- factory to create some kind of Tree
   */
  public PennTreeReader(Reader in, TreeFactory tf) {
    this(in, tf, null, new PennTreebankTokenizer(in));
  }

  /**
   * Read parse trees from a <code>Reader</code>.
   *
   * @param in The Reader
   * @param st The Tokenizer
   */
  public PennTreeReader(Reader in, Tokenizer st) {
    this(in, new SimpleTreeFactory(), null, st);
  }


  /**
   * Read parse trees from a Reader.
   *
   * @param in Reader
   * @param tf TreeFactory -- factory to create some kind of Tree
   * @param tn the method of normalizing trees
   */
  public PennTreeReader(Reader in, TreeFactory tf, TreeNormalizer tn) {
    this(in, tf, tn, new PennTreebankTokenizer(in));
  }


  /**
   * Read parse trees from a Reader.
   *
   * @param in Reader
   * @param tf TreeFactory -- factory to create some kind of Tree
   * @param tn the method of normalizing trees
   * @param st Tokenizer that divides up Reader
   */
  public PennTreeReader(Reader in, TreeFactory tf, TreeNormalizer tn, Tokenizer st) {
    this.in = in;
    this.tf = tf;
    this.tn = tn;
    this.st = st;
    // check for whacked out headers still present in Brown corpus in Treebank 3
    String first = (String) st.peek();
    if (first != null && first.startsWith("*x*x*x")) {
      if (DEBUG) {
        System.err.println("PennTreeReader: skipping past whacked out header.");
      }
      int foundCount = 0;
      while (foundCount < 4 && st.hasNext()) {
        first = (String) st.next();
        if (first != null && first.startsWith("*x*x*x")) {
          foundCount++;
        }
      }
    }

    if (DEBUG) {
      System.err.println("Built PennTreeReader from " + in.getClass().getName() + " " + ((tf == null) ? "no tf" : tf.getClass().getName()) + " " + ((tn == null) ? "no tn" : tn.getClass().getName()) + " " + ((st == null) ? "no st" : st.getClass().getName()));
    }
  }


  /**
   * Reads a single tree in standard Penn Treebank format,
   * with or without an additional set of parens around it (an unnamed
   * ROOT node).  If the token stream ends before the current tree is complete, a
   * {@link java.util.NoSuchElementException} will get thrown from
   * deep within the innards of this method.
   *
   * @return A single tree, or <code>null</code> at end of token stream.
   */
  public Tree readTree() throws IOException {
    if (!st.hasNext()) {
      return null;
    }
    Tree tr = readTree((String) st.next());
    if (tr == null || tn == null) {
      return tr;
    } else {
      return tn.normalizeWholeTree(tr, tf);
    }
  }


  private Tree readTree(String token) throws IOException {
    if (DEBUG) {
      System.out.println("readTree() next token " + token);
    }
    String name;
    // a paren starts new tree, a string is a leaf symbol,
    // o.w. IO exception
    if (token == null) {
      return null;
    } else if (token.equals("(")) {
      // looks at next
      name = (String) st.peek();
      if (DEBUG) {
        System.out.println("  peeked is " + name);
      }
      // checks if it's a normal string and returns it as the label
      if (name.equals("(") || name.equals(")")) {
        name = null;
      } else {
        // get it for real
        name = (String) st.next();
      }
      if (tn != null) {
        name = tn.normalizeNonterminal(name); // we used to .intern();
      }
      return (tf.newTreeNode(name, readTrees()));
    } else {
      if (tn != null) {
        name = tn.normalizeTerminal(token); // was: .intern();
      } else {
        name = token;
      }
      return tf.newLeaf(name);
    }
  }


  /**
   * Parse sequence of trees, followed by a single right paren.
   */
  private List readTrees() throws IOException {
    // allocate array list for temporarily storing trees
    ArrayList parseTrees = new ArrayList();
    // until a paren closes all subtrees, keep reading trees
    String nextToken = null;
    while (st.hasNext()) {
      nextToken = (String) st.next();
      if (nextToken.equals(")")) {
        break;
      }
      parseTrees.add(readTree(nextToken));
    }
    if (!nextToken.equals(")")) {
      throw(new IOException("Expecting right paren found eof"));
    }
    return parseTrees;
  }


  /**
   * Close the Reader behind this <code>TreeReader</code>.
   */
  public void close() throws IOException {
    in.close();
  }

  public static TokenizerFactory<Tree> tokenizerFactory(final TreeFactory tf, final TreeNormalizer tn, final Tokenizer stringTokenizer) {
    return new TreeTokenizerFactory(new TreeReaderFactory() {
      public TreeReader newTreeReader(Reader in) {
        return new PennTreeReader(in,tf,tn,stringTokenizer);
      }
    });
  }
  /*
  private static class TreeTokenizerFactory implements TokenizerFactory<Tree> {
    TreeFactory tf;
    TreeNormalizer tn;
    Tokenizer t;

    public TreeTokenizerFactory(TreeFactory tf, TreeNormalizer tn, Tokenizer t) {
      this.tf = tf;
      this.tn = tn;
      this.t = t;
    }

    public Tokenizer<Tree> getTokenizer(final Reader r) {
      return new AbstractTokenizer<Tree>() {
        PennTreeReader tr = new PennTreeReader(r,tf,tn,t);
        public Tree getNext() {
          try {
            return tr.readTree();
          }
          catch(IOException e) {
            System.err.println("Error in reading tree.");
            return null;
          }
        }
      };
    }

    public Iterator<Tree> getIterator(Reader r) {
      return getTokenizer(r);
    }
  }
  */


  /**
   * Loads treebank data from first argument and prints it.
   *
   * @param args Array of command-line arguments: specifies a filename
   */
  public static void main(String[] args) {
    try {
      TreeFactory tf = new LabeledScoredTreeFactory();
      Reader r = new BufferedReader(new InputStreamReader(new FileInputStream(args[0]), "UTF-8"));
      TreeReader tr = new PennTreeReader(r, tf);
      Tree t = tr.readTree();
      while (t != null) {
        System.out.println(t);
        System.out.println();
        t = tr.readTree();
      }
      r.close();
    } catch (IOException ioe) {
      ioe.printStackTrace();
    }
  }

}
