package edu.stanford.nlp.trees;

import edu.stanford.nlp.ling.HasTag;
import edu.stanford.nlp.ling.HasWord;
import edu.stanford.nlp.ling.Label;

import java.util.Arrays;
import java.util.HashSet;


/**
 * Implements a 'semantic head' variant of the the HeadFinder found
 * in Michael Collins' 1999 thesis.
 * This version chooses the semantic head verb rather than the verb form
 * for cases with verbs.  Should remember auxiliaries to differentiate past
 * and passive, though.
 * Except: we've added a head rule for NX (returns rightmost), whereas
 * no rule for the head of NX is found in any of the versions of
 * Collins' head table that we have (did he perhaps use the NP rules
 * for NX?
 *
 * @author John Rappaport
 */
public class SemanticHeadFinder extends ModCollinsHeadFinder {

  private static final boolean DEBUG = false;

  private HashSet<String> verbalAuxiliaries;
  private HashSet<String> verbalTags;

  public SemanticHeadFinder() {
    this(new PennTreebankLanguagePack());
  }

  public SemanticHeadFinder(TreebankLanguagePack tlp) {
    super(tlp);

    verbalAuxiliaries = new HashSet<String>();
    verbalAuxiliaries.addAll(Arrays.asList(new String[]{"will", "wo", "shall", "may", "might", "should", "would", "can", "could", "ca", "must", "has", "have", "had", "having", "am", "are", "is", "was", "were", "be", "being", "been", "get", "gets", "getting", "got", "gotten", "do", "does", "did", "to", "'m", "'re", "'s", "'ve", "'d", "'ll"}));

    verbalTags = new HashSet<String>();
    // include Charniak tags so can do BLLIP right
    verbalTags.addAll(Arrays.asList(new String[]{"TO", "MD", "VB", "VBD", "VBP", "VBZ", "VBG", "VBN", "AUX", "AUXG"}));
  }

  /**
   * Determine which daughter of the current parse tree is the
   * head.  It assumes that the daughters alread have had their
   * heads determined.  Uses special rule for VPheads
   *
   * @param t The parse tree to examine the daughters of.
   *          This is assumed to never be a leaf
   * @return The parse tree that is the head
   */
  protected Tree determineNonTrivialHead(Tree t) {
    String motherCat = tlp.basicCategory(t.label().value());

    // do VPs with auxiliary as special case
    if (motherCat.equals("VP")) {
      Tree[] kids = t.children();
      // try to find if there is an auxiliary verb

      if (DEBUG) {
        System.err.println("Semantic head finder: at VP");
        System.err.println("Class is " + t.getClass().getName());
        t.pennPrint(System.err);
        System.err.println("hasVerbalAuxiliary = " + hasVerbalAuxiliary(kids));
      }
      if (hasVerbalAuxiliary(kids)) {
        // String[] how = new String[] {"left", "VP", "ADJP", "NP"});
        // Including NP etc seems okay for copular sentences but is
        // problematic for other auxiliaries, like 'he has an answer'
        // But maybe doing ADJP is fine!
        String[] how = new String[]{"left", "VP", "ADJP"};
        Tree pti = traverseLocate(kids, how, false);
        if (DEBUG) {
          System.err.println("Determined head is: " + pti);
        }
        if (pti != null) {
          return pti;
        } else {
          // System.err.println("------");
          // System.err.println("SemanticHeadFinder failed to reassign head for");
          // t.pennPrint(System.err);
          // System.err.println("------");
        }
      }
    }

    return super.determineNonTrivialHead(t);
  }


  private boolean hasVerbalAuxiliary(Tree[] kids) {

    for (int i = 0; i < kids.length; i++) {

      Label kidLabel = kids[i].label();

      String cat = tlp.basicCategory(kidLabel.value());
      String word = null;
      if (kidLabel instanceof HasWord) {
        word = ((HasWord) kidLabel).word();
      }
      if (word == null) {
        Label htl = kids[i].headTerminal(this).label();
        if (htl instanceof HasWord) {
          word = ((HasWord) htl).word();
        }
        if (word == null) {
          word = htl.value();
        }
      }

      String tag = null;
      if (kidLabel instanceof HasTag) {
        tag = ((HasTag) kidLabel).tag();
      }
      if (tag == null) {
        tag = kids[i].headPreTerminal(this).value();
      }
      if (DEBUG) {
        System.err.println("Checking " + kids[i].value() + " head is " +
                           word + "/" + tag);
      }
      // got to not match on to/TO if in PP!
      if (( ! "PP".equals(cat)) && verbalTags.contains(tag) && verbalAuxiliaries.contains(word)) {
        return true;
      }
    }

    return false;
  }

  private static final long serialVersionUID = 5721799188009249808L;

}
