package edu.stanford.nlp.trees;

import edu.stanford.nlp.ling.*;
import edu.stanford.nlp.parser.lexparser.ChineseTreebankParserParams;
import edu.stanford.nlp.parser.lexparser.TreebankLangParserParams;
import edu.stanford.nlp.process.Function;
import edu.stanford.nlp.trees.international.pennchinese.ChineseEnglishWordMap;
import edu.stanford.nlp.util.Filter;
import edu.stanford.nlp.util.StringUtils;

import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.io.Reader;
import java.util.*;


/**
 * A class for customizing the print method(s) for a
 * <code>edu.stanford.nlp.trees.Tree</code> as the output of the
 * parser.  This now supports printing in multiple ways and altering
 * behavior via properties.
 * <p>
 * <i>Note that turning on markHeadNodes is incompatible with (i.e., breaks)
 * typedDependencies output. This should be fixed at some point.
 *
 * @author Roger Levy
 * @author Christopher Manning
 * @author Galen Andrew
 */
public class TreePrint {

  public static final String rootLabelOnlyFormat = "rootSymbolOnly";

  /* The legal output tree formats. */
  public static final String[] outputTreeFormats = {
    "penn",
    "oneline",
    rootLabelOnlyFormat,
    "words",
    "wordsAndTags",
    "dependencies",
    "typedDependencies",
    "typedDependenciesCollapsed",
    "latexTree",
    "collocations" 
  };

  public static final String headMark = "=H";

  private Properties formats;
  private Properties options;

  private boolean markHeadNodes = false;
  private boolean lexicalize = false;
  private boolean stem = false;
  private boolean transChinese = false;

  private HeadFinder hf;
  private TreebankLanguagePack tlp;
  private WordStemmer stemmer;
  private Filter punctuationTagRejectFilter;
  private GrammaticalStructureFactory gsf;

  private PrintWriter pw = new PrintWriter(System.out, true);

  /** The anglocentric constructor */
  public TreePrint(String formats) {
    this(formats, "", new PennTreebankLanguagePack());
  }

  public TreePrint(String formats, TreebankLanguagePack tlp) {
    this(formats, "", tlp);
  }

  public TreePrint(String formats, String options, TreebankLanguagePack tlp) {
    this(formats, options, tlp, new CollinsHeadFinder());
  }

  /**
   * Make a TreePrint instance.
   *
   * @param formatString A comma separated list of ways to print each Tree.
   *                Known formats are: oneline, penn, latexTree, wordsAndTags, dependencies,
   *                typedDependencies, collocations.  All of them print a blank line after
   *                the output except for oneline.
   * @param optionsString Options that additionally specify how trees are to
   *                be printed (for instance, whether stemming should be done).
   *                Known options are: stem, lexicalize, markHeadNodes, xml,
   *                removeTopBracket.
   * @param tlp     The TreebankLanguagePack used to do things like delete
   *                or ignore punctuation in output
   * @param hf      The HeadFinder used in printing output
   */
  public TreePrint(String formatString, String optionsString, TreebankLanguagePack tlp, HeadFinder hf) {
    formats = StringUtils.stringToProperties(formatString);
    options = StringUtils.stringToProperties(optionsString);
    Set es = this.formats.keySet();
    List okOutputs = Arrays.asList(outputTreeFormats);
    for (Iterator it = es.iterator(); it.hasNext();) {
      String format = (String) it.next();
      if (!okOutputs.contains(format)) {
        throw new RuntimeException("Error: output tree format " + format + " not supported");
      }
    }

    setHeadFinder(hf);
    this.tlp = tlp;
    punctuationTagRejectFilter = tlp.punctuationTagRejectFilter();
    stem = propertyToBoolean(this.options, "stem");
    if (stem) {
      stemmer = new WordStemmer();
    }
    if (formats.containsKey("typedDependenciesCollapsed") || formats.containsKey("typedDependencies")) {
      gsf = tlp.grammaticalStructureFactory();
    }

    lexicalize = propertyToBoolean(this.options, "lexicalize");
    markHeadNodes = propertyToBoolean(this.options, "markHeadNodes");
    transChinese = propertyToBoolean(this.options, "transChinese");
  }


  private static boolean propertyToBoolean(Properties prop, String key) {
    String val = prop.getProperty(key);
    if (val == null) {
      return false;
    } else if (val.trim().equals("")) {
      return true;
    } else if (val.trim().equalsIgnoreCase("true")) {
      return true;
    } else if (val.trim().equalsIgnoreCase("false")) {
      return false;
    } else {
      throw new RuntimeException("Property " + key + " should have a boolean value but it's " + val);
    }
  }


  /**
   * Prints the tree to the default PrintWriter.
   */
  public void printTree(Tree t) {
    printTree(t, pw);
  }

  /**
   * Prints the tree, with an empty ID.
   */
  public void printTree(final Tree t, PrintWriter pw) {
    printTree(t, "", pw);
  }

  /**
   * Prints the tree.
   *
   * @param id A name for this sentence.
   */
  public void printTree(final Tree t, final String id, final PrintWriter pw) {
    Tree outputTree = t;

    if (propertyToBoolean(options, "xml")) {
      Sentence sentUnstemmed = outputTree.yield();
      pw.print("<s");
      if (id != null && !"".equals(id)) {
        pw.print(" id=\"" + id + "\"");
      }
      pw.println(">");
      if (formats.containsKey("words")) {
        pw.println("  <words>");
        int i = 1;
        for (Iterator it = sentUnstemmed.iterator(); it.hasNext();) {
          Label w = (Label) it.next();
          pw.println("    <word ind=\"" + i + "\">" + w + "</word>");
          i++;
        }
        pw.println("  </words>");
      }
    } else {
      if (formats.containsKey("words")) {
        pw.println(outputTree.yield().toString(false));
        pw.println();
      }
    }

    if (propertyToBoolean(options, "removeTopBracket")) {
      String s = outputTree.label().value();
      if (tlp.isStartSymbol(s) && outputTree.isUnaryRewrite()) {
        outputTree = outputTree.firstChild();
      } else {
        // It's not quite clear what to do if the tree isn't unary at the top
        // but we then don't strip the ROOT symbol, since that seems closer
        // than losing part of the tree altogether....
        System.err.println("TreePrint: can't remove top bracket: not unary");
      }
    }
    if (stem) {
      stemmer.visitTree(outputTree);
    }
    if (lexicalize) {
      Function a = TreeFunctions.getLabeledTreeToCategoryWordTagTreeFunction();
      outputTree = (Tree) a.apply(outputTree);
      outputTree.percolateHeads(hf);
    }

    if (markHeadNodes) {
      outputTree = markHeadNodes(outputTree);
    }
    if (formats.containsKey("collocations")) {
      /* Reflection to check that a Wordnet connection exists.  Otherwise
       * we print an error and exit.
       */
      try {
        Class cl = Class.forName("edu.stanford.nlp.trees.WordNetInstance");
        WordNetConnection wn = (WordNetConnection) cl.newInstance();
        CollocationFinder cf = new CollocationFinder(outputTree,wn);
        outputTree = cf.getMangledTree();
      } catch (Exception e) {
        e.printStackTrace();
        System.err.println("Couldn't open WordNet Connection.  Aborting collocation detection.");
      }


    }

    if (!lexicalize) { // delexicalize the output tree
      Function a = TreeFunctions.getLabeledTreeToStringLabeledTreeFunction();
      outputTree = (Tree) a.apply(outputTree);
    }
    if (transChinese) {
      TreeTransformer tt = new TreeTransformer() {
        public Tree transformTree(Tree t) {
          t = t.deepCopy();
          for (Iterator it = t.iterator(); it.hasNext();) {
            Tree subtree = (Tree) it.next();
            if (subtree.isLeaf()) {
              Label oldLabel = subtree.label();
              String translation = ChineseEnglishWordMap.getStaticMap().getFirstTranslation(oldLabel.value());
              if (translation == null) translation = "[UNK]";
              Label newLabel = new StringLabel(oldLabel.value()+":"+translation);
              subtree.setLabel(newLabel);
            }
          }
          return t;
        }
      };
      outputTree = tt.transformTree(outputTree);
    }

    if (propertyToBoolean(options, "xml")) {
      if (formats.containsKey("wordsAndTags")) {
        Sentence sent = outputTree.taggedYield();
        pw.println("  <words pos=\"true\">");
        int i = 1;
        for (Iterator it = sent.iterator(); it.hasNext();) {
          TaggedWord tw = (TaggedWord) it.next();
          pw.println("    <word ind=\"" + i + "\" pos=\"" + tw.tag() + "\">" + tw.word() + "</word>");
          i++;
        }
        pw.println("  </words>");
      }
      if (formats.containsKey("penn")) {
        pw.println("  <tree style=\"penn\">");
        outputTree.pennPrint(pw);
        pw.println("  </tree>");
      }
      if (formats.containsKey("latexTree")) {
        pw.println(".[");
        outputTree.indentedListPrint(pw,false);
        pw.println(".]");
      }
      if (formats.containsKey("dependencies")) {
        Set depsSet = outputTree.taggedDependencies(new Filter() {
          public boolean accept(Object ob) {
            if (!(ob instanceof Dependency)) {
              return false;
            }
            Dependency d = (Dependency) ob;
	    if ( ! (d.dependent() instanceof HasTag)) {
	      return false;
	    }
	    String tag = ((HasTag) d.dependent()).tag();
            return punctuationTagRejectFilter.accept(tag);
          }
        }, hf);
        pw.println("  <dependencies style=\"untyped\">");
        for (Iterator it = depsSet.iterator(); it.hasNext();) {
          Dependency d = (Dependency) it.next();
          pw.println("    " + d.toString("xml"));
        }
        pw.println("  </dependencies>");
      }
      if (formats.containsKey("typedDependencies")) {
        GrammaticalStructure gs = gsf.newGrammaticalStructure(outputTree);
        gs.typedDependencies().print("xml", pw);
      }
      if (formats.containsKey("typedDependenciesCollapsed")) {
        GrammaticalStructure gs = gsf.newGrammaticalStructure(outputTree);
        gs.typedDependenciesCollapsed().print("xml", pw);
      }
      pw.println("</s>");
      pw.println();
    } else {
      if (formats.containsKey("wordsAndTags")) {
        pw.println(outputTree.taggedYield().toString(false));
        pw.println();
      }
      if (formats.containsKey("oneline")) {
        pw.println(outputTree.toString());
      }
      if (formats.containsKey("penn")) {
        outputTree.pennPrint(pw);
        pw.println();
      }
      if (formats.containsKey(rootLabelOnlyFormat)) {
        pw.println(outputTree.label());
      }
      if (formats.containsKey("latexTree")) {
        pw.println(".[");
        outputTree.indentedListPrint(pw,false);
        pw.println(".]");
      }
      if (formats.containsKey("dependencies")) {
        Set depsSet = outputTree.taggedDependencies(new Filter() {
          public boolean accept(Object ob) {
            if (!(ob instanceof Dependency)) {
              return false;
            }
            Dependency d = (Dependency) ob;
	    if ( ! (d.dependent() instanceof HasTag)) {
	      return false;
	    }
	    String tag = ((HasTag) d.dependent()).tag();
            return punctuationTagRejectFilter.accept(tag);
          }
        }, hf);
        pw.println(depsSet);
        pw.println();
      }
      if (formats.containsKey("typedDependencies")) {
        // pw.println(new GrammaticalStructure(outputTree).typedDependencies());
        GrammaticalStructure gs = gsf.newGrammaticalStructure(outputTree);
//        pw.print(gs);
        gs.typedDependencies().print(pw);
        pw.println();
      }
      if (formats.containsKey("typedDependenciesCollapsed")) {
        GrammaticalStructure gs = gsf.newGrammaticalStructure(outputTree);
//        pw.print(gs);
        gs.typedDependenciesCollapsed().print(pw);
      }

    }
    }


  public void printHeader(PrintWriter pw, String charset) {
    if (propertyToBoolean(options, "xml")) {
      pw.println("<?xml version=\"1.0\" encoding=\"" + charset + "\"?>");
      pw.println("<corpus>");
    }
  }


  public void printFooter(PrintWriter pw) {
    if (propertyToBoolean(options, "xml")) {
      pw.println("</corpus>");
    }
  }


  /**
   * Sets whether or not to stem the Trees using Morphology.
   */
  public void setStem(boolean stem) {
    this.stem = stem;
    if (stem) {
      stemmer = new WordStemmer();
    } else {
      stemmer = null;
    }
  }

  /**
   * Set the headfinder to be used for lexicalizing trees.  Relevant
   * if the output is to be lexicalized.  By default,
   * <code>CollinsHeadFinder</code> is used.
   */
  public void setHeadFinder(HeadFinder hf) {
    this.hf = hf;
  }

  public HeadFinder getHeadFinder() {
    return hf;
  }


  /**
   * Sets the default print writer for printing trees with the instance.
   * Unless this is set, the default is <code>System.out</code>.
   */
  public void setPrintWriter(PrintWriter pw) {
    this.pw = pw;
  }


  public Tree markHeadNodes(Tree t) {
    return markHeadNodes(t, null);
  }

  private Tree markHeadNodes(Tree t, Tree head) {
    if (t.isLeaf()) {
      return t; // don't worry about head-marking leaves
    }
    Label newLabel;
    if (t == head) {
      newLabel = headMark(t.label());
    } else {
      newLabel = t.label();
    }
    Tree newHead = hf.determineHead(t);
    return t.treeFactory().newTreeNode(newLabel, Arrays.asList(headMarkChildren(t, newHead)));
  }

  private Label headMark(Label l) {
    Label l1 = l.labelFactory().newLabel(l);
    l1.setValue(l1.value() + headMark);
    return l1;
  }

  private Tree[] headMarkChildren(Tree t, Tree head) {
    Tree[] kids = t.children();
    Tree[] newKids = new Tree[kids.length];
    for (int i = 0, n = kids.length; i < n; i++) {
      newKids[i] = markHeadNodes(kids[i], head);
    }
    return newKids;
  }

  /**
   * utility code...
   */
  public static void main(String[] args) {
    String format = "penn";
    String options = "";
    Map<String,Integer> flagMap = new HashMap<String,Integer>();
    flagMap.put("-format",1);
    flagMap.put("-options",1);
    Map<String,String[]> argsMap = StringUtils.argsToMap(args,flagMap);
    args = argsMap.get(null);
    if(argsMap.keySet().contains("-format"))
      format = argsMap.get("-format")[0];
    if(argsMap.keySet().contains("-options"))
      options = argsMap.get("-options")[0];
    TreebankLangParserParams tlpp = new ChineseTreebankParserParams();
    TreePrint print = new TreePrint(format, options, tlpp.treebankLanguagePack(), tlpp.headFinder());
    Iterator<Tree> i; // initialized below
    if(args.length > 0) {
      Treebank trees; // initialized below
      if(argsMap.keySet().contains("-arabic"))
        trees = tlpp.diskTreebank();
      else {
        trees = new DiskTreebank(new TreeReaderFactory() {
            public TreeReader newTreeReader(Reader in) {
              return new PennTreeReader(in, new LabeledScoredTreeFactory(new StringLabelFactory()), new TreeNormalizer());
            }
          });
      }
      trees.loadPath(args[0]);
      i = trees.iterator();
    }
    else {
      i = tlpp.treeTokenizerFactory().getIterator(new BufferedReader(new InputStreamReader(System.in)));
    }
    while(i.hasNext()) {
      print.printTree(i.next());
    }
  }

}
