package edu.stanford.nlp.trees;

import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;

/**
 * A <code>TypedDependencyList</code> represents a complete list of
 * typed dependencies for a <code>GrammaticalStructure</code>,
 * with special methods for printing in different formats.
 *
 * @author Bill MacCartney
 * @author Galen Andrew (refactoring English-specific stuff)
 * @see GrammaticalStructure
 * @see GrammaticalRelation
 * @see EnglishGrammaticalStructure
 */
public class TypedDependencyList extends ArrayList<TypedDependency> {

  public void sort() {
    Collections.sort(this, new Comparator() {
      // this idiosyncratic compare method enables sorting by index of
      // dependent, which is very natural for printing
      public int compare(Object o1, Object o2) {
        return ((TypedDependency) o1).dep().index() - ((TypedDependency) o2).dep().index();
      }
    });
  }

  public TypedDependencyList() {
    super();
  }

  /**
   * The constructor builds a list of typed dependencies using
   * information from a <code>GrammaticalStructure</code>.
   */
  public TypedDependencyList(GrammaticalStructure gs) {
    for (Dependency d : gs.dependencies()) {
      TreeGraphNode gov = (TreeGraphNode) d.governor();
      TreeGraphNode dep = (TreeGraphNode) d.dependent();
      //      System.out.println("Gov: " + gov);
      //      System.out.println("Dep: " + dep);
      GrammaticalRelation reln = gs.getGrammaticalRelation(gov, dep);
      //      System.out.println("Reln: " + reln);
      add(new TypedDependency(reln, gov, dep));
    }
  }

  /**
   * Returns a String representation of the result of this set of
   * typed dependencies in a user-specified format.
   * Currently, three formats are supported:
   * <dl>
   * <dt>"plain"</dt>
   * <dd>(Default.)  Formats the dependencies as logical relations,
   * as exemplified by the following:
   * <pre>
   *  nsubj(died-1, Sam-0)
   *  tmod(died-1, today-2)
   *  </pre>
   * </dd>
   * <dt>"readable"</dt>
   * <dd>Formats the dependencies as a table with columns
   * <code>dependent</code>, <code>relation</code>, and
   * <code>governor</code>, as exemplified by the following:
   * <pre>
   *  Sam-0               nsubj               died-1
   *  today-2             tmod                died-1
   *  </pre>
   * </dd>
   * <dt>"xml"</dt>
   * <dd>Formats the dependencies as XML, as exemplified by the following:
   * <pre>
   *  &lt;dependencies&gt;
   *    &lt;dep type="nsubj"&gt;
   *      &lt;governor idx="1"&gt;died&lt;/governor&gt;
   *      &lt;dependent idx="0"&gt;Sam&lt;/dependent&gt;
   *    &lt;/dep&gt;
   *    &lt;dep type="tmod"&gt;
   *      &lt;governor idx="1"&gt;died&lt;/governor&gt;
   *      &lt;dependent idx="2"&gt;today&lt;/dependent&gt;
   *    &lt;/dep&gt;
   *  &lt;/dependencies&gt;
   *  </pre>
   * </dd>
   * </dl>
   *
   * @param format a <code>String</code> specifying the desired format
   * @return a <code>String</code> representation of the typed
   *         dependencies in this <code>GrammaticalStructure</code>
   */
  public String toString(String format) {
    if (format != null && format.equals("xml")) {
      return toXMLString();
    } else if (format != null && format.equals("readable")) {
      return toReadableString();
    } else {
      return toString();
    }
  }

  /**
   * Returns a String representation of this set of typed dependencies
   * as exemplified by the following:
   * <pre>
   *  tmod(died-6, today-9)
   *  nsubj(died-6, Sam-3)
   *  </pre>
   *
   * @return a <code>String</code> representation of this set of
   *         typed dependencies
   */
  public String toString() {
    sort();
    StringBuffer buf = new StringBuffer();
    for (TypedDependency td : this) {
      buf.append(td.toString() + "\n");
    }
    return buf.toString();
  }

  private String toReadableString() {
    sort();
    StringBuffer buf = new StringBuffer();
    buf.append(String.format("%-20s%-20s%-20s%n", "dep", "reln", "gov"));
    buf.append(String.format("%-20s%-20s%-20s%n", "---", "----", "---"));
    for (TypedDependency td : this) {
      buf.append(String.format("%-20s%-20s%-20s%n", td.dep(), td.reln(), td.gov()));
    }
    return buf.toString();
  }

  private String toXMLString() {
    sort();
    StringBuffer buf = new StringBuffer("<dependencies style=\"typed\">\n");
    for (TypedDependency td : this) {
      String reln = td.reln().toString();
      String gov = td.gov().value();
      int govIdx = td.gov().index();
      String dep = td.dep().value();
      int depIdx = td.dep().index();
      buf.append("  <dep type=\"" + reln + "\">\n");
      buf.append("    <governor idx=\"" + govIdx + "\">" + gov + "</governor>\n");
      buf.append("    <dependent idx=\"" + depIdx + "\">" + dep + "</dependent>\n");
      buf.append("  </dep>\n");
    }
    buf.append("</dependencies>");
    return buf.toString();
  }

  /**
   * Prints this set of typed dependencies to <code>stdout</code>.
   */
  public void print() {
    print(new PrintWriter(System.out, true));
  }

  /**
   * Prints this set of typed dependencies to the specified
   * <code>PrintWriter</code>.
   */
  public void print(PrintWriter pw) {
    pw.println(toString());
  }

  /**
   * Prints this set of typed dependencies to <code>stdout</code>
   * in the specified format.
   */
  public void print(String format) {
    print(format, new PrintWriter(System.out, true));
  }

  /**
   * Prints this set of typed dependencies to the specified
   * <code>PrintWriter</code> in the specified format.
   */
  public void print(String format, PrintWriter pw) {
    pw.println(toString(format));
  }

}

