// ChineseEnglishWordMap -- a mapping from Chinese to English words
// Copyright (c) 2002, 2003, 2004 The Board of Trustees of
// The Leland Stanford Junior University. All Rights Reserved.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// Map is taken from CEDict Chinese-English Lexicon.  Future versions
// will support multiple Lexicons.
//
// http://www.mandarintools.com/cedict.html
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
// For more information, bug reports, fixes, contact:
//    Christopher Manning
//    Dept of Computer Science, Gates 4A
//    Stanford CA 94305-9040
//    USA
//    java-nlp-support@lists.stanford.edu

package edu.stanford.nlp.trees.international.pennchinese;

import edu.stanford.nlp.trees.TreebankLanguagePack;
import edu.stanford.nlp.util.StringUtils;

import java.io.*;
import java.util.*;
import java.util.HashMap;
import java.util.Map;
import java.util.ArrayList;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * A class for mapping Chinese words to English.  Uses CEDict free Lexicon.
 * @author Galen Andrew
 */
public class ChineseEnglishWordMap {
  private Map<String, Set<String>> map = new HashMap<String, Set<String>>();

  private static ChineseEnglishWordMap staticMap;
  private static final String defaultPath = "cedict_ts.u8";
  private static final String defaultPattern = "[^ ]+ ([^ ]+)[^/]+/(.+)/";
  private static final String defaultDelimiter = "/";
  private static final String defaultCharset = "UTF-8";

  /**
   * A method for getting the one static copy of the map.
   * @return the static copy of ChineseEnglishWordMap
   */
  public static ChineseEnglishWordMap getStaticMap() {
    if (staticMap == null) {
      staticMap = new ChineseEnglishWordMap();
    }
    return staticMap;
  }

  /**
   * Does the word exist in the dictionary?
   */
  public boolean containsKey(String key) {
    key = key.toLowerCase();    
    return map.containsKey(key);
  }

  /**
   *
   * @param key a Chinese word
   * @return the English translation (null if not in dictionary)
   */
  public Set<String> getAllTranslations(String key) {
    key = key.toLowerCase();
    return map.get(key);
  }

  /**
   *
   * @param key a Chinese word
   * @return the English translations as an array (null if not in dictionary)
   */
  public String getFirstTranslation(String key) {
    key = key.toLowerCase();
    Set<String> strings = map.get(key);
    if (strings == null) return null;
    else return strings.iterator().next();
  }

  public void readCEDict(String dictPath) {
    readCEDict(dictPath, defaultPattern, defaultDelimiter, defaultCharset);
  }

  public void readCEDict(String dictPath, String pattern, String delimiter, String charset) {
    BufferedReader infile = null;
    try {
      infile = new BufferedReader(new InputStreamReader(new FileInputStream(dictPath), charset));

      Pattern p = Pattern.compile(pattern);
      for (String line = infile.readLine(); line != null; line = infile.readLine()) {
        Matcher m = p.matcher(line);
        if (m.matches()) {
          String word = (m.group(1)).toLowerCase();
          String transGroup = m.group(2);
          String[] trans = transGroup.split(delimiter);
          if (map.containsKey(word)) {
            Set<String> oldtrans = map.get(word);
            for (String t : trans) {
              if(!oldtrans.contains(t)) {
                oldtrans.add(t);
              }
            }
          } else {
            Set transList = new HashSet(Arrays.asList(trans));
            map.put(word, transList);
          }
        }
      }
    } catch (IOException e) {
      throw new RuntimeException("IOException reading CEDict from file " + dictPath, e);
    }
  }

  /**
   * Make a ChineseEnglishWordMap with the default CEDict path ("cedict_ts.u8")
   */
  public ChineseEnglishWordMap() {
    this(defaultPath);
  }

  /**
   * Make a ChineseEnglishWordMap
   * @param dictPath the path/filename of the CEDict
   */
  public ChineseEnglishWordMap(String dictPath) {
    readCEDict(dictPath);
  }

  public ChineseEnglishWordMap(String dictPath, String pattern, String delimiter, String charset) {
    readCEDict(dictPath, pattern, delimiter, charset);
  }


  private static boolean isDigits(String in) {
    for (int i = 0, len = in.length(); i < len; i++) {
      if ( ! Character.isDigit(in.charAt(i))) {
	return false;
      }
    }
    return true;
  }

  /**
   * return a reversed map of the current map
   */
  public Map<String, Set<String>> getReverseMap() {
    Map<String, Set<String>> rMap = new HashMap<String, Set<String>>();
    for (String k : map.keySet()) {
      Set<String> transList = map.get(k);
      for (String trans : transList) {
        Set<String> entry = rMap.get(trans);
        if (entry == null) {
          Set<String> toAdd = new HashSet<String>();
          toAdd.add(k);
          rMap.put(trans, toAdd);
        } else {
          entry.add(k);
        }
      }
    }
    return rMap;
  }
  
  /**
   * Add all of the mappings from the specified map to the current map.
   */

  public void addMap(Map<String, Set<String>> addM) {
    for (String k : addM.keySet()) {
      Set<String> origList = map.get(k);
      Set<String> addList = addM.get(k);
      if (origList == null) {
        map.put(k, new HashSet(addList));
      } else {
        for (String toAdd : addList) {
          if (!(origList.contains(toAdd))) {
            origList.add(toAdd);
          }
        }
      }
    }
  }
  
  public String toString() {
    return map.toString();
  }

  public int size() {
    return map.size();
  }


  /**
   * The main method reads (segmented, whitespace delimited) words from a file
   * and prints them with their English translation(s).
   *
   * The path and filename of the CEDict Lexicon can be supplied via the
   * "-dictPath" flag; otherwise the default filename "cedict_ts.u8" in the
   * current directory is checked.
   *
   * By default, only the first translation is printed.  If the "-all" flag
   * is given, all translations are printed.
   *
   * The input and output encoding can be specified using the "-encoding" flag.
   * Otherwise UTF-8 is assumed.
   */
  public static void main(String[] args) throws IOException {
    Map<String, Integer> flagsToNumArgs = new HashMap<String, Integer>();
    flagsToNumArgs.put("-dictPath" , 1);
    flagsToNumArgs.put("-encoding" , 1);
    Map<String, String[]> argMap = StringUtils.argsToMap(args, flagsToNumArgs);
    String[] otherArgs = argMap.get(null);
    if (otherArgs.length < 1) {
      System.err.println("usage: ChineseEnglishWordMap [-all] [-dictPath path] [-encoding enc_string] inputFile");
      System.exit(1);
    }
    String filename = otherArgs[0];
    boolean allTranslations = argMap.containsKey("-all");
    String charset = defaultCharset;
    if (argMap.containsKey("-encoding")) {
      charset = argMap.get("-encoding")[0];
    }
    BufferedReader r = new BufferedReader(new InputStreamReader(new FileInputStream(filename), charset));

    TreebankLanguagePack tlp = new ChineseTreebankLanguagePack();
    String[] dpString = argMap.get("-dictPath");
    ChineseEnglishWordMap cewm = (dpString == null) ? new ChineseEnglishWordMap() : new ChineseEnglishWordMap(dpString[0]);
    int totalWords = 0, coveredWords = 0;

    PrintWriter pw = new PrintWriter(new OutputStreamWriter(System.out, charset), true);

    for (String line = r.readLine(); line != null; line = r.readLine()) {
      String[] words = line.split("\\s", 1000);
      for (String word : words) {
        totalWords++;
        if (word.length() == 0) continue;
        pw.print(StringUtils.pad(word + ':', 8));
        if (tlp.isPunctuationWord(word)) {
          totalWords--;
          pw.print(word);
	} else if (isDigits(word)) {
	  pw.print(word + " [NUMBER]");
        } else if (cewm.containsKey(word)) {
          coveredWords++;
          if (allTranslations) {
            List<String> trans = new ArrayList(cewm.getAllTranslations(word));
            for (String s : trans) {
              pw.print((trans.indexOf(s) > 0 ? "|" : "") + s);
            }
          } else {
            pw.print(cewm.getFirstTranslation(word));
          }
        } else {
          pw.print("[UNK]");
        }
	pw.println();
      }
      pw.println();
    }

    System.err.print("Finished translating " + totalWords + " words (");
    System.err.println(coveredWords + " were in dictionary).");
  }
}
