// TregexPattern -- a Tgrep2-style utility for recognizing patterns in trees.
// Copyright (c) 2004,2005 The Board of Trustees of
// The Leland Stanford Junior University. All Rights Reserved.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
// For more information, bug reports, fixes, contact:
//    Christopher Manning
//    Dept of Computer Science, Gates 4A
//    Stanford CA 94305-9040
//    USA
//    http://nlp.stanford.edu/downloads/tregex.shtml

package edu.stanford.nlp.trees.tregex;

import edu.stanford.nlp.ling.StringLabelFactory;
import edu.stanford.nlp.process.Function;
import edu.stanford.nlp.trees.*;
import edu.stanford.nlp.util.StringUtils;
import edu.stanford.nlp.util.Timing;

import java.io.*;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

/**
 * A TregexPattern is a <code>tgrep</code>-type pattern.
 *
 * <p> TregexPattern instances can be matched against instances of the {@link Tree} class.
 * The {@link #main} method can be used to find matching nodes of a treebank from the command line.
 *
 * <p>
 * Currently supported node/node relations and their symbols:
 * <p>
 * <table border = "1">
 * <tr><th>Symbol<th>Meaning
 * <tr><td>A &#60;&#60; B <td>A dominates B
 * <tr><td>A &#62;&#62; B <td>A is dominated by B
 * <tr><td>A &#60; B <td>A immediately dominates B
 * <tr><td>A &#62; B <td>A is immediately dominated by B
 * <tr><td>A &#36; B <td>A is a sister of B (and not equal to B)
 * <tr><td>A .. B <td>A precedes B
 * <tr><td>A . B <td>A immediately precedes B
 * <tr><td>A ,, B <td>A follows B
 * <tr><td>A , B <td>A immediately follows B
 * <tr><td>A &#60;&#60;, B <td>B is a leftmost descendent of A
 * <tr><td>A &#60;&#60;- B <td>B is a rightmost descendent of A
 * <tr><td>A &#62;&#62;, B <td>A is a leftmost descendent of B
 * <tr><td>A &#62;&#62;- B <td>A is a rightmost descendent of B
 * <tr><td>A &#60;, B <td>B is the first child of A
 * <tr><td>A &#62;, B <td>A is the first child of B
 * <tr><td>A &#60;- B <td>B is the last child of A
 * <tr><td>A &#62;- B <td>A is the last child of B
 * <tr><td>A &#60;` B <td>B is the last child of A
 * <tr><td>A &#62;` B <td>A is the last child of B
 * <tr><td>A &#60;i B <td>B is the ith child of A (i > 0)
 * <tr><td>A &#62;i B <td>A is the ith child of B (i > 0)
 * <tr><td>A &#60;-i B <td>B is the ith-to-last child of A (i > 0)
 * <tr><td>A &#62;-i B <td>A is the ith-to-last child of B (i > 0)
 * <tr><td>A &#60;: B <td>B is the only child of A
 * <tr><td>A &#62;: B <td>A is the only child of B
 * <tr><td>A &#60;&#60;: B <td>A dominates B via an unbroken chain (length > 0) of unary local trees.
 * <tr><td>A &#62;&#62;: B <td>A is dominated by B via an unbroken chain (length > 0) of unary local trees.
 * <tr><td>A &#36;++ B <td>A is a left sister of B (same as &#36;.. for context-free trees)
 * <tr><td>A &#36;-- B <td>A is a right sister of B (same as &#36;,, for context-free trees)
 * <tr><td>A &#36;+ B <td>A is the immediate left sister of B (same as &#36;. for context-free trees)
 * <tr><td>A &#36;- B <td>A is the immediate right sister of B (same as &#36;, for context-free trees)
 * <tr><td>A &#36;.. B <td>A is a sister of B and precedes B
 * <tr><td>A &#36;,, B <td>A is a sister of B and follows B
 * <tr><td>A &#36;. B <td>A is a sister of B and immediately precedes B
 * <tr><td>A &#36;, B <td>A is a sister of B and immediately follows B
 * <tr><td>A <+(C) B <td>A dominates B via an unbroken chain of (zero or more) nodes matching description C
 * <tr><td>A >+(C) B <td>A is dominated by B via an unbroken chain of (zero or more) nodes matching description C
 * <tr><td>A &#60;&#60;&#35; B <td>B is a head of phrase A
 * <tr><td>A &#62;&#62;&#35; B <td>A is a head of phrase B
 * <tr><td>A &#60;&#35; B <td>B is the immediate head of phrase A
 * <tr><td>A &#62;&#35; B <td>B is the immediate head of phrase A
 * <tr><td>A == B <td>A and B are the same node
 * </table>
 * <p> Label descriptions can be literal strings, which much match labels
 * exactly, or regular expressions in regular expression bars: /regex/.
 * In order to prevent ambiguity with other Tregex symbols, only standard
 * "identifiers" are allowed as literals, i.e. strings matching [a-zA-z]([a-zA-Z0-9_])*
 * If you want to use other symbols, you can do so by using a regular expression
 * instead of a literal string.
 * A disjunctive list of literal strings can be given separated by '|'.
 * The special string '__' (two underscores) can be used to match any
 * node.  (WARNING!!  Use of the '__' node description may seriously
 * slow down search.)  If a label description is preceeded by '@', the
 * label will match any node whose </em> basicCategory </em> matches the
 * description.  The basicCategory is defined according to a Function
 * mapping Strings to Strings, as provided by
 * {@link edu.stanford.nlp.trees.AbstractTreebankLanguagePack#getBasicCategoryFunction()}.
 * <p> In a chain of relations, all relations are relative to the first node in the chain.
 * For example, <code> (S < VP < NP) </code> means "an S over a VP and also over an NP".
 * If instead what you want is an S above a VP above an NP, you should write
 * "<code>S < (VP < NP)</code>".
 * <p> Nodes can be grouped using parens '(' and ')'
 * as in <code> S < (NP $++ VP) </code> to match an S
 * over an NP, where the NP has a VP as a right sister.
 *
 * <p> Relations can be combined using the '&' and '|' operators.  Thus
 * <code> (NP < NN | < NNS) </code> will match an NP node dominating either
 * an NN or an NNS.  <code> (NP > S & $++ VP) </code> matches an NP that
 * is both under an S and has a VP as a right sister.
 *
 * <p> Relations can be grouped using brackets '[' and ']'.  So the
 * expression <code> NP [< NN | < NNS] & > S </code> matches an NP
 * that dominates either an NN or an NNS and is under an S.  Without
 * brackets, & takes precidence over |, and equivalent operators are
 * left-associative.  Also note that & is the default combining operator if the
 * operator is omitted in a chain of relations, so that <code> (S < VP < NP) </code>
 * is equivalent to <code> (S < VP & < NP) </code>.
 * As another example, <code> (VP < VV | < NP % NP)
 * </code> can be written explicitly as <code> (VP [< VV | [< NP & % NP] ] )
 *
 * <p> Relations can be negated with the '!' operator, in which case the
 * expression will match only if there is no node satisfying the relation.
 * For example <code> (NP !< NNP) </code> matches only NPs not dominating
 * an NNP.  Label descriptions can also be negated with '!': (NP < !NNP|NNS) matches
 * NPs dominating some node that is not an NNP or an NNS.
 *
 * <p> In order to consider only the "basic category" of a tree label,
 * i.e. to ignore functional tags or other annotations on the label,
 * prefix that node's description with the &#64; symbol.  For example
 * <code> (@NP < @/NN.?/) </code>  This can only be used for individual nodes;
 * if you want all nodes to use the basic category, it would be more efficient
 * to use a {@link edu.stanford.nlp.trees.TreeNormalizer} to remove functional
 * tags before passing the tree to the TregexPattern.
 *
 * <p><h3>Naming nodes</h3>
 *
 * <p> Nodes can be given names using '='.  A named node will be stored in a
 * map that maps names to nodes so that if a match is found, the node
 * corresponding to the named node can be extracted from the map.  For
 * example <code> (NP < NNP=name) </code> will match an NP dominating an NNP
 * and after a match is found, the map can be queried with the
 * name to retreived the matched node using {@link TregexMatcher#getNode(Object o)}
 * with (String) argument "name" (<it>not</it> "=name").
 * Note that a {@link ParseException} will be thrown if a named node is used in the
 * scope of a negated relation, because the semantics would be unclear.
 *
 * <p> Named nodes that refer back to previous named nodes need not have a node
 * description -- this is known as "backreferencing".  In this case, the expression
 * will match only if the subsequently named node is equal to the previously named
 * node (in the == sense).
 * For example: <code> (@NP <, (@NP $+ (/,/ $+ (@NP $+ /,/=comma))) <- =comma) </code>
 * matches an NP dominating exactly the sequence NP comma NP comma.  Multiple
 * backreferences are allowed.  If the node w/ no node description does not refer
 * to a previously named node, there will be no error, the expression simply will
 * not match anything.
 *
 * <p> Another way to refer to previously named nodes is with the "link" symbol: '~'.
 * A link is like a backreference, except that instead of having to be *equal to* the
 * referred node, the current node only has to match the label of the referred to node.
 * A link cannot have a node description, i.e. the '~' symbol must immediately follow a
 * relation symbol.
 *
 * <p> Relations can be made optional with the '?' operator.  This way the
 * expression will match even if the optional relation is not satisfied, but
 * if it is satisfied named nodes under it will still be put into the map.
 *
 * <p> The HeadFinder used to determine heads for the head relations, and also
 * the Function mapping from labels to Basic Category tags can be
 * chosen by using a {@link TregexPatternCompiler}.
 *
 * <p><h3>Variable Groups</h3>
 *
 * <p> If you write a node description using a regular expression, you can assign its matching groups to variable names.
 * If more than one node has a group assigned to the same variable name, then matching will only occur when all such groups
 * capture the same string.  This is useful for enforcing coindexation constraints.  The syntax is
 *
 * <blockquote>
 * <code> / < regex-stuff > /#< group-number > % < variable-name > </code>
 * </blockquote>
 *
 * For example, the pattern (designed for Penn Treebank trees)
 *
 * <blockquote>
 * <code> @SBAR < /^WH.*-([0-9]+)$/#1%index << (__=empty < (/^-NONE-/ < /^\*T\*-([0-9]+)$/#1%index)) </code>
 * </blockquote>
 *
 * will match only such that the WH- node under the SBAR is coindexed with the trace node that gets the name <code>empty</code>.
 *
 *
 * <p><h3>Current known bugs/shortcomings:</h3>
 *
 * <ul>
 *
 * <li> Node search currently takes no advantage of limitations
 * imposed by the queried relations.  This reduces the efficiency of
 * the search, quite a bit in some cases.
 * <li> Due to the lack of parent pointers in {@link edu.stanford.nlp.trees.Tree}s, parents are found via
 * exhaustive depth-first search from the root.  This is a serious efficiency bottleneck.
 *
 * </ul>
 *
 * @author Galen Andrew
 * @author Roger Levy (rog@csli.stanford.edu)
 */
public abstract class TregexPattern implements Serializable {
  protected static Function<String, String> currentBasicCatFunction;

  static void setBasicCatFunction(Function<String, String> f) {
    currentBasicCatFunction = f;
  }

  private boolean neg = false;
  private boolean opt = false;
  private String patternString;

  void negate() {
    neg = true;
    if (neg && opt) {
      throw new RuntimeException("Node cannot be both negated and optional.");
    }
  }

  void makeOptional() {
    opt = true;
    if (neg && opt) {
      throw new RuntimeException("Node cannot be both negated and optional.");
    }
  }

  private void prettyPrint(PrintWriter pw, int indent) {
    for (int i = 0; i < indent; i++) {
      pw.print("   ");
    }
    if (neg) {
      pw.print('!');
    }
    if (opt) {
      pw.print('?');
    }
    pw.println(localString());
    for (Iterator iter = getChildren().iterator(); iter.hasNext();) {
      TregexPattern child = (TregexPattern) iter.next();
      child.prettyPrint(pw, indent + 1);
    }
  }

  // package private constructor
  TregexPattern() {
  }

  abstract List getChildren();

  abstract String localString();

  boolean isNegated() {
    return neg;
  }

  boolean isOptional() {
    return opt;
  }

  abstract TregexMatcher matcher(Tree root, Tree tree, Map<Object, Tree> namesToNodes, VariableStrings variableStrings);

  /**
   * Get a {@link TregexMatcher} for this pattern on this tree.
   *
   * @param t a tree to match on
   * @return a TregexMatcher
   */
  public TregexMatcher matcher(Tree t) {
    return matcher(t, t, new HashMap<Object, Tree>(), new VariableStrings());
  }

  /**
   * Creates a pattern from the given string using the default Headfinder and
   * BasicCategoryFunction.  If you want to use a different HeadFinder or
   * BasicCategoryFunction, use a {@link TregexPatternCompiler} object.
   *
   * @param tregex the pattern string
   * @return a TregexPattern for the string.
   * @throws ParseException if the string does not parse
   */
  public static TregexPattern compile(String tregex) throws ParseException {
    return TregexPatternCompiler.defaultCompiler.compile(tregex);
  }

  public String pattern() {
    return patternString;
  }

  public void setPatternString(String patternString) {
    this.patternString = patternString;
  }

  /**
   * @return A single-line string representation of the pattern
   */
  public abstract String toString();

  /**
   * Print a multi-line respresentation
   * of the pattern illustrating it's syntax.
   */
  public void prettyPrint(PrintWriter pw) {
    prettyPrint(pw, 0);
  }

  /**
   * Print a multi-line respresentation
   * of the pattern illustrating it's syntax.
   */
  public void prettyPrint(PrintStream ps) {
    prettyPrint(new PrintWriter(new OutputStreamWriter(ps), true));
  }

  /**
   * Print a multi-line respresentation of the pattern illustrating
   * it's syntax to System.out.
   */
  public void prettyPrint() {
    prettyPrint(System.out);
  }

  /**
   * Prints out all matches of a tree pattern on each tree in the path.
   * Usage: <br><br><code>
   * java edu.stanford.nlp.trees.tregex.TregexPattern [-TCwfosnu] pattern
   * [handle] filepath   </code>
   *
   * <p>
   * Arguments:<br>
   * <ul><li><code>pattern</code>: the tree
   * pattern which optionally names some node =name (for some arbitrary
   * string "name")
   * <li> <code>handle</code> (optional): the named node in <code>pattern</code>
   * <li> <code>filepath</code>: the path to files with trees
   * </ul><p>
   * Options:<br>
   * <ul><li> -T causes all trees to be printed as processed.  Otherwise only matching nodes
   * are printed.
   * <li> -C suppresses printing of matches, so only the
   * number of matches is printed.
   * <li> -w causes the whole of a tree that matches to be printed.
   * <li> -f causes the filename to be printed.
   * <li> -o allows each tree node to be reported only once as the root of a match (by default a node will
   * be printed once for every <em>way</em> the pattern matches).
   * <li> -s causes trees to be printed all on one line (by default they are pretty printed).
   * <li> -n causes the number of the tree in which the match was found to be
   * printed before every match.
   * <li> -u causes only the label of each matching node to be printed.
   * <li> -t causes only the yield of the selected node (or the yield of the whole tree, if the <code>-w</code> option is used) to be printed.
   * <li> -encoding <charset_encoding> option allows specification of character set.
   * </ul>
   */
  public static void main(String[] args) {
    Timing.startTime();

    String treePrintFormats = "";

    String encodingOption = "-encoding";
    String encoding = "UTF-8";
    Map<String,Integer> flagMap = new HashMap<String,Integer>();
    flagMap.put(encodingOption,1);
    Map<String, String[]> argsMap = StringUtils.argsToMap(args, flagMap);
    args = argsMap.get(null);

    if (args.length < 1) {
      System.err.println("Usage: java edu.stanford.nlp.trees.tregex.TregexPattern [-T] [-C] [-w] [-f] [-o] [-n] pattern [handle] [filepath]");
      System.exit(0);
    }

    if (argsMap.containsKey("-T")) {
      TRegexTreeVisitor.printTree = true;
    }
    if (argsMap.containsKey("-C")) {
      TRegexTreeVisitor.printMatches = false;
    }
    if (argsMap.containsKey("-w")) {
      TRegexTreeVisitor.printWholeTree = true;
    }
    if (argsMap.containsKey("-f")) {
      TRegexTreeVisitor.printFilename = true;
    }
    if(argsMap.containsKey(encodingOption)) {
      encoding = argsMap.get(encodingOption)[0];
    }
    if(argsMap.containsKey("-o"))
      TRegexTreeVisitor.oneMatchPerRootNode = true;
    if(argsMap.containsKey("-n"))
      TRegexTreeVisitor.reportTreeNumbers = true;
    if(argsMap.containsKey("-u"))
      treePrintFormats +=  TreePrint.rootLabelOnlyFormat + ",";
    else if(argsMap.containsKey("-s")) // display short form
      treePrintFormats += "oneline,";
    else if(argsMap.containsKey("-t"))
      treePrintFormats += "words,";
    else
      treePrintFormats += "penn,";

    TRegexTreeVisitor.tp = new TreePrint(treePrintFormats, new PennTreebankLanguagePack());

    try {
      //TreePattern p = TreePattern.compile("/^S/ > S=dt $++ '' $-- ``");
      TregexPattern p = TregexPattern.compile(args[0]);
      System.err.println("Pattern string:\n" + p.pattern());
      System.err.println("Parsed representation:");
      p.prettyPrint(System.err);

      String label = null;
      if (args.length == 1) {
        System.err.println("using default tree");
        TreeReader r = new PennTreeReader(new StringReader("(VP (VP (VBZ Try) (NP (NP (DT this) (NN wine)) (CC and) (NP (DT these) (NNS snails)))) (PUNCT .))"), new LabeledScoredTreeFactory(new StringLabelFactory()));
        Tree t = r.readTree();
        treebank = new MemoryTreebank();
        treebank.add(t);
      } else {
        int last = args.length - 1;
        System.err.println("Reading trees from file(s) " + args[last]);
        TreeReaderFactory trf = new TreeMatcher.TRegexTreeReaderFactory();
        treebank = new DiskTreebank(trf,encoding);
        treebank.loadPath(args[last], null, true);
        if (args.length == 3) {
          label = args[1];
        }
      }
      TRegexTreeVisitor vis = new TRegexTreeVisitor(p, label);

      treebank.apply(vis);
      Timing.endTime();
      if(TRegexTreeVisitor.printMatches)
        System.err.println("There were " + vis.numMatches() + " matches in total.");
      else
        System.out.println(vis.numMatches());
    } catch (IOException e) {
      e.printStackTrace();
    } catch (ParseException e) {
      System.err.println("Error parsing expression: " + args[0]);
      System.err.println("Parse exception: " + e.toString());
    }
  }

  static Treebank treebank; // used by main method, must be accessible

  static class TRegexTreeVisitor implements TreeVisitor {

    static boolean printTree = false;
    static boolean printWholeTree = false;
    static boolean printMatches = true;
    static boolean printFilename = false;
    static boolean oneMatchPerRootNode = false;
    static boolean reportTreeNumbers = false;
    
    static TreePrint tp;
    PrintWriter pw = new PrintWriter(System.out,true);

    int treeNumber = 0;

    TregexPattern p;
    String node;
    int numMatches;

    TRegexTreeVisitor(TregexPattern p, String node) {
      this.p = p;
      this.node = node;
    }

    public void visitTree(Tree t) {
      treeNumber++;
      if (printTree) {
        pw.print(treeNumber+":");
        pw.println("Next tree read:");
        tp.printTree(t,pw);
      }
      TregexMatcher match = p.matcher(t);
      Tree lastMatchingRootNode = null;
      while (match.find()) {
        if(oneMatchPerRootNode) {
          if(lastMatchingRootNode == match.getMatch())
            continue;
          else
            lastMatchingRootNode = match.getMatch();
        }
        numMatches++;
        if (printFilename && treebank instanceof DiskTreebank) {
          DiskTreebank dtb = (DiskTreebank) treebank;
          pw.print("# ");
          pw.println(dtb.getCurrentFile());
        }
        if (printMatches) {
          if(reportTreeNumbers) {
            pw.print(treeNumber + ": ");
          }
          if (printTree) {
            pw.println("Found a full match:");
          }
          if (printWholeTree) {
            tp.printTree(t,pw);
          } else if (node != null) {
            if (printTree) {
              pw.println("Here's the node you were interested in:");
            }
            Tree labeledNode = match.getNode(node);
            if (labeledNode == null) {
              System.err.println("ACK!!  There is no matched node \"" + node + "\"!  Did you specify such a label in the pattern?");
            } else {
              tp.printTree(labeledNode,pw);
            }
          } else {
            tp.printTree(match.getMatch(),pw);
          }
          pw.println();
        }
      }
    }

    public int numMatches() {
      return numMatches;
    }

  } // end class TRegexTreeVisitor

}
