package edu.stanford.nlp.util;

import java.io.Serializable;
import java.util.*;

/**
 * Factory for vending Maps.
 *
 * @author Dan Klein (klein@cs.stanford.edu)
 * @author Kayur Patel (kdpatel@cs)
 */
abstract public class MapFactory implements Serializable {

  private static final long serialVersionUID = 4529666940763477360L;

  public static final MapFactory HASH_MAP_FACTORY = new HashMapFactory();

  public static final MapFactory IDENTITY_HASH_MAP_FACTORY = new IdentityHashMapFactory();

  public static final MapFactory WEAK_HASH_MAP_FACTORY = new WeakHashMapFactory();

  public static final MapFactory TREE_MAP_FACTORY = new TreeMapFactory();

  public static final MapFactory ARRAY_MAP_FACTORY = new ArrayMapFactory();

  private static class HashMapFactory extends MapFactory {

    private static final long serialVersionUID = -9222344631596580863L;

    public Map newMap() {
      return new HashMap();
    }

    public <K1, V1> Map<K1, V1> setMap(Map<K1,V1> map) {
      map = new HashMap<K1,V1>();
      return map;
    }

  } // end class HashMapFactory

  private static class IdentityHashMapFactory extends MapFactory {

    private static final long serialVersionUID = -9222344631596580863L;

    public Map newMap() {
      return new IdentityHashMap();
    }

    public <K1, V1> Map<K1, V1> setMap(Map<K1,V1> map) {
      map = new IdentityHashMap<K1,V1>();
      return map;
    }

  } // end class IdentityHashMapFactory

  private static class WeakHashMapFactory extends MapFactory {

    private static final long serialVersionUID = 4790014244304941000L;

    public Map newMap() {
      return new WeakHashMap();
    }

    public <K1, V1> Map<K1, V1> setMap(Map<K1,V1> map) {
      map = new WeakHashMap<K1,V1>();
      return map;
    }

  } // end class WeakHashMapFactory

  private static class TreeMapFactory extends MapFactory {

    private static final long serialVersionUID = -9138736068025818670L;

    public Map newMap() {
      return new TreeMap();
    }   

    public <K1, V1> Map<K1, V1> setMap(Map<K1,V1> map) {
      map = new TreeMap<K1,V1>();
      return map;
    }

  } // end class TreeMapFactory

  private static class ArrayMapFactory extends MapFactory {

    public Map newMap() {
      return new ArrayMap();
    }

    public <K1, V1> Map<K1, V1> setMap(Map<K1, V1> map) {
      return new ArrayMap<K1,V1>();
    }
  }

  abstract public Map newMap();

  /**
   * A method to get a parameterized map out.
   * @param map a type-parameterized {@link Map} argument
   * @return a {@link Map} with type-parameterization identical to that of the argument. 
   */
  abstract public <K1, V1> Map<K1, V1> setMap(Map<K1,V1> map);

}
