package edu.stanford.nlp.util;

import java.util.List;
import java.util.Set;

/**
 * A Set that also represents an ordering of its elements, and responds
 * quickly to add(), changePriority(), removeFirst(), and getFirst() method calls.
 *
 * @author Teg Grenager (greuager@cs.stanford.edu)
 */
public interface PriorityQueue<E> extends Set<E> {

  /**
   * Finds the object with the highest priority, removes it,
   * and returns it.
   *
   * @return the object with highest priority
   */
  public E removeFirst();

  /**
   * Finds the object with the highest priority and returns it, without
   * modifying the queue.
   *
   * @return the object with minimum key
   */
  public E getFirst();

  /**
   * Get the priority of a key.
   *
   * @param key The object to assess
   * @return A key's priority. If the key is not in the queue,
   *         Double.NEGATIVE_INFINITY is returned.
   */
  public double getPriority(Object key);

  /**
   * Convenience method for if you want to pretend relaxPriority doesn't exist,
   * or if you really want add's return conditions.
   *
   * @return <tt>true</tt> if this set did not already contain the specified
   *         element.
   */
  public boolean add(E key, double priority);


  /**
   * Changes a priority, either up or down, adding the key it if it wasn't there already.
   *
   * @param key an <code>Object</code> value
   * @return whether the priority actually changed.
   */
  public boolean changePriority(E key, double priority);

  /**
   * Increases the priority of the Object key to the new priority if the old priority
   * was lower than the new priority. Otherwise, does nothing.
   *
   * @param key
   * @param priority
   * @return
   */
  public boolean relaxPriority(E key, double priority);

  public List<E> toSortedList();

}
