package edu.stanford.nlp.util;

import java.io.IOException;
import java.io.Reader;
import java.io.StringReader;
import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Class XMLUtils
 *
 * @author Teg Grenager
 */
public class XMLUtils {

  /**
   * Block-level HTML tags that are rendered with surrounding line breaks.
   */
  public static final Set breakingTags = new HashSet(Arrays.asList(new String[]{"blockquote", "br", "div", "h1", "h2", "h3", "h4", "h5", "h6", "hr", "li", "ol", "p", "pre", "ul", "tr", "td"}));

  /**
   *
   * @param r the reader to read the XML/HTML from
   * @param mapBack a List of Integers mapping the positions in the result buffer
   * to positions in the original Reader, will be cleared on receipt
   * @return the String containing the resulting text
   */
  public static String stripTags(Reader r, List mapBack, boolean markLineBreaks) {
    if (mapBack!=null) mapBack.clear(); // just in case it has something in it!
    StringBuffer result = new StringBuffer();
    String text;
    String tag;
    int position = 0;
    try {
      do {
        text = XMLUtils.readUntilTag(r); // will do nothing if the next thing is a tag
        if (text.length()>0) {
          // add offsets to the map back
          for (int i=0; i<text.length(); i++) {
            result.append(text.charAt(i));
            if (mapBack!=null) mapBack.add(new Integer(position+i));
          }
          position += text.length();
        }
//        System.out.println(position + " got text: " + text);
        tag = XMLUtils.readTag(r);
        if (tag==null) {
          break;
        }
        if (markLineBreaks && XMLUtils.isBreaking(parseTag(tag))) {
          result.append("\n");
          if (mapBack!=null) mapBack.add(new Integer(-position));
        }
        position += tag.length();
//        System.out.println(position + " got tag: " + tag);
      } while (tag != null);
    } catch (Exception e) {
      System.err.println("Error reading string");
      e.printStackTrace();
    }
    return result.toString();
  }

  public static boolean isBreaking(String tag) {
    return breakingTags.contains(tag);
  }

  public static boolean isBreaking(XMLTag tag) {
    return breakingTags.contains(tag.name);
  }

  /**
   * Reads all text up to next XML tag and returns it as a String.
   *
   * @return the String of the text read, which may be empty.
   */
  public static String readUntilTag(Reader r) throws IOException {
    if (!r.ready()) {
      return "";
    }
    StringBuffer b = new StringBuffer();
    int c = r.read();
    while (c >= 0 && c != '<') {
      b.append((char) c);
      c = r.read();
    }
    return b.toString();
  }

  /**
   * @return the new XMLTag object, or null if couldn't be created
   */
  public static XMLTag readAndParseTag(Reader r) throws Exception {
    String s = readTag(r);
    if (s == null) {
      return null;
    }
    return new XMLTag(s);
  }

  public static String unescapeStringForXML(String s) {
    Pattern p = Pattern.compile("\\&.+?;");
    StringBuffer result = new StringBuffer();
    Matcher m = p.matcher(s);
    int start = -1;
    int end = 0;
    while (m.find()) {
      start = m.start();
      result.append(s.substring(end, start));
      end = m.end();
      result.append(translate(s.substring(start, end)));
    }
    result.append(s.substring(end, s.length()));
    return result.toString();
  }

  private static char translate(String s) {
    if (s.equals("&amp;")) {
      return '&';
    } else if (s.equals("&lt;")) {
      return '<';
    } else if (s.equals("&gt;")) {
      return '>';
    } else if (s.equals("&quot;")) {
      return '\"';
    } else if (s.equals("&apos;")) {
      return '\'';
    } else {
      return '?';
    }
  }

  /**
   * Returns a String in which all of the special characters of XML have been escaped. The resulting String
   * can be used as text in well-formed XML.
   * @param s
   * @return
   */
  public static String escapeStringForXML(String s) {
    StringBuffer result = new StringBuffer();
    for (int i = 0; i < s.length(); i++) {
      int c = s.charAt(i);
      if (c < 0x20 || c>=0x80) {
        // weird chars
        result.append(" ");
      } else {
        switch (c) {
          case 0x26:
            result.append("&amp;");
            break;
          case 0x3c:
            result.append("&lt;");
            break;
          case 0x3e:
            result.append("&gt;");
            break;
          case 0x22:
            result.append("&quot;");
            break;
          case 0x27:
            result.append("&apos;");
            break;
          default:
            result.append((char) c);
        }
      }
    }
    return result.toString();
  }

  public static String escapeTextAroundXMLTags(String s) {
    StringBuffer result = new StringBuffer();
    String text;
    XMLTag tag;
    Reader r = new StringReader(s);
    try {
      do {
        text = readUntilTag(r); // will do nothing if the next thing is a tag
        //      System.out.println("got text: " + text);
        result.append(escapeStringForXML(text));
        tag = readAndParseTag(r);
        //      System.out.println("got tag: " + tag);
        if (tag == null) {
          break;
        }
        result.append(tag.toString());
      } while (tag != null);
    } catch (Exception e) {
      System.err.println("Error reading string");
      e.printStackTrace();
    }
    return result.toString();
  }

  public static class XMLTag {
    public String text;
    public String name;
    public Map attributes;
    public boolean isEndTag;
    public boolean isSingleTag;

    /**
     * Assumes that String contains XML tag.
     *
     * @param tag
     */
    public XMLTag(String tag) throws Exception {
      if (tag == null || tag.length() == 0) {
        throw new Exception("Can't parse tag");
      }
      text = tag;
      int begin = 1;
      if (tag.charAt(1) == '/') {
        begin = 2;
        isEndTag = true;
      } else {
        isEndTag = false;
      }
      int end = tag.length() - 1;
      if (tag.charAt(tag.length() - 2) == '/') {
        end = tag.length() - 2;
        isSingleTag = true;
      } else {
        isSingleTag = false;
      }
      tag = tag.substring(begin, end);
      attributes = new HashMap();
      begin = 0;
      end = tag.indexOf(' ');
      if (end < 0) {
        name = tag;
      } else {
        name = tag.substring(begin, end);
        do {
          begin = end + 1;
          while (tag.charAt(begin) < 0x21) {
            begin++; // get rid of leading whitespace
          }
          end = tag.indexOf('=', begin);
          if (end < 0) {
            break;
          }
          String att = tag.substring(begin, end);
          begin = end + 1;
          String value = null;
          if (tag.length()>begin) {
            if (tag.charAt(begin)=='\"') {
              // get quoted expression
              begin++;
              end = tag.indexOf('\"', begin);
              if (end < 0) {
                break; // this is a problem
              }
              value = tag.substring(begin, end);
              end++;
            } else {
              // get unquoted expression
              end = tag.indexOf(' ', begin);
              if (end < 0) {
                end = tag.length();
              }
              value = tag.substring(begin, end);
            }
          }
          attributes.put(att, value);
        } while (end < tag.length() - 3);
      }
    }

    public String toString() {
      return text;
    }
  }

  /**
   * Reads all text of the XML tag and returns it as a String.
   * Assumes that a '<' character has already been read.
   *
   * @param r
   * @return the String representing the tag, or null if one couldn't be read
   */
  public static String readTag(Reader r) throws IOException {
    if (!r.ready()) {
      return null;
    }
    StringBuffer b = new StringBuffer();
    int c = r.read();
    while (c >= 0) {
      b.append((char) c);
      if (c == '>') {
        break;
      }
      c = r.read();
    }
    if (b.length() == 0) {
      return null;
    }
    return "<" + b.toString();
  }

  public static void main(String[] args) throws Exception {
    String s = StringUtils.slurpFile(args[0]);
    Reader r = new StringReader(s);
    String text = readUntilTag(r);
    String tag = readTag(r);
    while (tag.length() > 0) {
      text = readUntilTag(r);
      tag = readTag(r);
      if (tag.length() == 0) {
        break;
      }
      System.out.println("got tag=" + new XMLTag(tag));
    }
  }

  public static XMLTag parseTag(String tagString) throws Exception {
    if (tagString == null || tagString.length() == 0) {
      return null;
    }
    return new XMLTag(tagString);
  }

}
